/* -*- c -*- ****************************************************************
 *
 *  			Copyright 1995-1999 Jasspa.
 *
 *  System        : MicroEmacs Jasspa Distribution
 *  Object Name   : abbrev.c
 *  Revision      : 2.1
 *  Date          : 12th June 1995
 *  Created By    : Steve Philips
 *  Last Modified : <990706.1155>
 *
 *  Description	
 *        Handle character expansion. The expansion sequences are held in 
 *        abbreviation files (.eaf). The file is loaded into a buffer and 
 *        subsequent expansion requests use the buffer for the matching and
 *        expansion text.
 *
 *        The .abr file is organsied with the competion match string 
 *        followed by white space, followed by the replacement text.
 *        Each entry occupies one line. The replacement text includes 
 *        magic tags for cursor palcement and spacial characters which 
 *        include:-
 *
 *              \n - new line
 *              \t - Tab character
 *              \b - backwards 1 character
 *              \d - delete tab backwards (want previous tab stop)
 *              \p - Position of cursor after expansion.
 *  Notes
 *
 *  History
 *	
 *  $Log$
 *
 ****************************************************************************
 *
 * Copyright (C) 1995 - 1999, JASSPA 
 * The MicroEmacs Jasspa distribution can be copied and distributed freely for
 * any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
 * incorporated into commercial software with the expressed permission of
 * JASSPA.
 * 
 ****************************************************************************/
#include "emain.h"
#include "efunc.h"

#define __ABREVC

#if ABREV


int
setAbbrev(int f, int n, meABREV **abrevPtr)
{
    register meABREV *abrev ;
    register int      status;		/* return status of name query */
    char              buf[MAXBUF]; 	/* name of file to execute */
    
    if(n == 0)
    {
        if((abrev=*abrevPtr) != NULL)
        {
            abrev->loaded = 0 ;
            freeLineLoop(&(abrev->hlp),0) ;
        }
        return TRUE ;
    }
        
    if((status=mlreply("Abbrev file",MLFILECASE,0,buf,MAXBUF)) != TRUE)
        return status ;
    
    if(buf[0] == '\0')
        abrev = NULL ;
    else
    {
        abrev = aheadp ;
        while(abrev != NULL)
        {
            if(!fnamecmp(buf,abrev->fname))
                break ;
            abrev = abrev->next ;
        }
        if((abrev == NULL) &&
           ((abrev = meMalloc(sizeof(meABREV))) != NULL))
        {
            abrev->fname = strdup(buf) ;
            abrev->loaded = 0 ;
            abrev->next = aheadp ;
            aheadp = abrev ;
            abrev->hlp.l_fp = &(abrev->hlp) ;
            abrev->hlp.l_bp = &(abrev->hlp) ;
        }
    }
    *abrevPtr = abrev ;
    
    return TRUE ;
}

int
bufferAbbrev(int f, int n)
{
    return setAbbrev(f,n,&(curbp->abrevFile)) ;
}
int
globalAbbrev(int f, int n)
{
    return setAbbrev(f,n,&globalAbrev) ;
}


static int
doExpandAbbrev(unsigned char *abName, int abLen, meABREV *abrev)
{
    register LINE    *lp, *hlp ;
    
    hlp = &abrev->hlp ;
    if(!abrev->loaded)
    {
        char fname[FILEBUF] ;
        
        if(!fileLookup(abrev->fname,".eaf",meFL_CHECKDOT|meFL_USESRCHPATH,fname) ||
           (ffReadFile(fname,READ_SILENT,NULL,hlp) == ABORT))
            return mlwrite(MWABORT,"[Failed to abbrev file %s]",abrev->fname);
        abrev->loaded = 1 ;
    }
    lp = lforw(hlp) ;
    while(lp != hlp)
    {
        if((lp->l_used > abLen) &&
           !strncmp((char *) lp->l_text,(char *) abName,abLen) &&
           (lp->l_text[abLen] == ' '))
        {
            unsigned char buf[MAXBUF] ;
            curwp->w_doto -= abLen ;
            ldelete(abLen,2) ;
            /* grab token */
            token((char *) lp->l_text+abLen, (char *) buf);
            return stringExec(FALSE,1,buf+1) ;
        }
        lp = lforw(lp) ;
    }
    return FALSE ;
}

int
expandAbbrev(int f, int n)
{
    register int     len, ii ;
    unsigned char    buf[MAXBUF] ;
    
    if(bchange() != TRUE)               /* Check we can change the buffer */
        return ABORT ;
    ii = curwp->w_doto ;
    if(((curbp->abrevFile != NULL) || (globalAbrev != NULL)) &&
       (--ii >= 0) && (!isSpace(curwp->w_dotp->l_text[ii])))
    {
        len = 1 ;
        while((--ii >= 0) && !isSpace(curwp->w_dotp->l_text[ii]))
            len++ ;
        strncpy((char *) buf,(char *) &(curwp->w_dotp->l_text[++ii]),len) ;
        if(((curbp->abrevFile != NULL) && ((ii=doExpandAbbrev(buf,len,curbp->abrevFile)) != FALSE)) ||
           ((globalAbrev != NULL)      && ((ii=doExpandAbbrev(buf,len,globalAbrev)) != FALSE)))
            return ii ;
    }
    /* We used to insert a space if the expansion was not defined
     * there seems to be no good reason to do this. Instead bitch
     * if we are not in a macro and return a false condition. The
     * error is not serious enough to abort */
    mlwrite (MWCLEXEC, "[No abbrev expansion defined]"); 
    return FALSE;
    
}
#endif
