/*****************************************************************************
*
*	Title:		%M%
*
*	Synopsis:	Binding functions.
*
******************************************************************************
*
*	Filename:		%P%
*
*	Author:			Daniel Lawrence
*
*	Creation Date:		11/02/86 12:37		<990714.0825>
*
*	Modification date:	%G% : %U%
*
*	Current rev:		%I%
*
*	Special Comments:	
*
*	Contents Description:	
*
*	This file is for functions having to do with key bindings,
*	descriptions, help commands and startup file.
*
*****************************************************************************
* 
* (C)opyright 1987 by Daniel M. Lawrence
* MicroEMACS 3.8 can be copied and distributed freely for any
* non-commercial purposes. MicroEMACS 3.8 can only be incorporated
* into commercial software with the permission of the current author.
* 
* Modifications to the original file by Jasspa. 
* 
* Copyright (C) 1988 - 1999, JASSPA 
* The MicroEmacs Jasspa distribution can be copied and distributed freely for
* any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
* incorporated into commercial software with the expressed permission of
* JASSPA.
* 
****************************************************************************/

/*---	Include defintions */

#define	__BINDC				/* Define file name */

/*---	Include files */

#ifdef _UNIX
#include <sys/stat.h>
#endif

#include "emain.h"
#include "efunc.h"
#include "esearch.h"
#include "eskeys.h"

/*---	Local variable definitions */
static char helpFileName[] = "me.ehf" ;

unsigned short
getskey(unsigned char **tp)
{
    unsigned char *ss=*tp, cc, dd ;
    unsigned short ii=0 ;
    
    for(;;)
    {
        if(((cc=*ss++) == '\0') ||
           ((dd=*ss++) != '-'))
            break ;
        if     (cc == 'C')
            ii |= ME_CONTROL ;
        else if(cc == 'S')
            ii |= ME_SHIFT ;
        else if(cc == 'A')
            ii |= ME_ALT ;
    }
    if(cc != '\0')
    {
        if(!isSpace(dd))
        {
            unsigned char *s1=ss ;
            int skey ;
            
            while(((dd=*s1) != '\0') && (dd != ' '))
                s1++ ;
            *s1 = '\0' ;
            skey = biChopFindString((char *) ss-2,13,specKeyNames,SKEY_MAX) ;
            *s1 = dd ;
            
            if(skey >= 0)
            {
                if(skey == SKEY_space)
                    ii |= 32 ;
                else
                    ii |= (skey|ME_SPECIAL) ;
                *tp = s1 ;
                return ii ;
            }
        }
        if(ii)
        {
            cc = toUpper(cc) ;
            if((ii == ME_CONTROL) && (cc >= 'A') && (cc <= '_'))
                ii = cc - '@' ;
            else
            {
                cc = toLower(cc) ;
                ii |= cc ;
            }
        }
        else
            ii |= cc ;
    }
    *tp = ss-1 ;
    return ii ;
}

/* get a command key sequence from the keyboard	*/
unsigned short
getckey(int mflag, int silent)
/* mflag	going for a meta sequence? */
/* silent;      print any thing */
{
    register unsigned short cc ;	/* character fetched */
    
    /* check to see if we are executing a command line */
    if (clexec)
    {
        unsigned char *tp;		/* pointer into the token */
        unsigned char tok[MAXBUF];	/* command incoming */
        
        macarg((char *) tok);	/* get the next token */
        
        tp = tok ;
        cc = getskey(&tp) ;
        
        if(!mflag)
        {
            int ii=ME_PREFIX_NUM+1 ;
            
            while(--ii > 0)
                if(cc == prefixc[ii])
                {
                    cc = ii << ME_PREFIX_BIT ;
                    break ;
                }
            if(cc & ME_PREFIX_MASK)
            {
                unsigned short ee ;	/* character fetched */
                unsigned char  dd ;
                while(((dd=*tp) != '\0') && isSpace(dd))
                    tp++ ;
                ee = getskey(&tp) ;
                if((ee >= 'A') && (ee <= 'Z'))
                    /* with a prefix make a letter lower case */
                    ee ^= 0x20 ;
                cc |= ee ;
            }
        }
    }
    /* or the normal way */
    else if (mflag)
        cc = tgetc();
    else
        cc = getkeycmd(FALSE, 0, silent);
    return cc ;
}

int
cmdchar(unsigned short cc, char *d)
{
    register int  doff=0, spec ;
    register char *ss ;
    
    if(cc & ME_ALT)
    {
        d[doff++] = 'A' ;
        d[doff++] = '-' ;
    }        
    if(cc & ME_CONTROL)
    {
        d[doff++] = 'C' ;
        d[doff++] = '-' ;
    }        
    if(cc & ME_SHIFT)
    {
        d[doff++] = 'S' ;
        d[doff++] = '-' ;
    }
    spec = (cc & ME_SPECIAL) ;
    cc &= 0x0ff ;
    
    if(spec)
    {
        register char ch ;
        
        ss = specKeyNames[cc] ;
name_copy:
        while((ch = *ss++) != '\0')
            d[doff++] = ch ;
        return doff ;
    }
    
    if((cc > 0) && (cc <= 0x20))	/* control character */
    {
        if(cc == 0x20)
        {
            ss = specKeyNames[SKEY_space] ;
            goto name_copy ;
        }
        cc |= 0x40 ;
        cc = toLower(cc) ;
        d[doff++] = 'C' ;
        d[doff++] = '-' ;
        d[doff++] = (char) cc ;
    }
    else if((cc == 0) || (cc >= 0x80))
    {
        d[doff++] = '\\';
        d[doff++] = '0';
        d[doff++] = hexdigits[cc >> 4] ;
        d[doff++] = hexdigits[cc & 0x0f] ;
    }
    else if((cc == '\\') || (cc == '"'))
    {
        d[doff++] = '\\';
        d[doff++] = (char) cc ;
    }
    else			/* any other character */
        d[doff++] = (char) cc ;
    
    return doff ;
}

/* change a key command to a string we can print out */
/* c   - 	sequence to translate */
/* seq - 	destination string for sequence */
void
cmdstr(unsigned short cc, register char * seq)
{
    /* apply prefix sequence if needed */
    if(cc & ME_PREFIX_MASK)
    {
        seq += cmdchar(prefixc[(cc&ME_PREFIX_MASK)>>ME_PREFIX_BIT],seq);
        *seq++ = ' ' ;
    }
    seq += cmdchar(cc,seq);
    *seq = 0;	/* terminate the string */
}

int
descKey(int f, int n)	/* describe the command for a certain key */
{
    static   char   notBound[] = "Not Bound" ;
    static   char   dskyPrompt[]="Show binding: " ;
    register unsigned short cc;	/* command character to describe */
    register char  *ptr; 	/* string pointer to scan output strings */
    register int    found;	/* matched command flag */
    unsigned int    arg ;	/* argument */
    char            outseq[40];	/* output buffer for command sequence */
    char            argStr[20];	/* argument string */
    
    /* prompt the user to type us a key to describe */
    mlwrite(MWCURSOR,dskyPrompt);
    
    /* get the command sequence to describe */
    cc = getckey(FALSE,1);	/* get a silent command sequence */
    
    /* change it to something we can print as well */
    cmdstr(cc, outseq);
    
    /* find the right function */
    if ((found = decode_key(cc,&arg)) < 0)
    {
        ptr = notBound ;
        argStr[0] = 0 ;
    }
    else
    {
        ptr = getCommandName(found) ;
        if(arg)
            sprintf(argStr,"%d ",(int) (arg + 0x80000000)) ;
        else
            argStr[0] = 0 ;
    }

    /* output the resultant string */
    mlwrite(MWCURSOR,"%s\"%s\" %s%s",dskyPrompt,outseq,argStr,ptr);
    
    return TRUE ;
}


static int
loadHelpFile(void)
{
    char fname[FILEBUF] ;
    if(!fileLookup(helpFileName,NULL,meFL_CHECKDOT|meFL_USESRCHPATH,fname))
        return mlwrite(MWABORT,"[Help file [%s] is not online]",helpFileName);
    /* and read the stuff in */
    ffReadFile(fname,READ_SILENT,NULL,&helpStore) ;
    return TRUE ;
}

static int
findHelpItem(char *item, int silent)
{
    unsigned char cc ;
    WINDOW *wp ;
    BUFFER *bp ;
    LINE   *lp, *elp ;
    int     sectLen, itemLen, ii ;
    char   *ss,  sect[5] ;
    
    itemLen = strlen(item) ;
    if((item[itemLen-1] == ')') &&
       ((item[(ii=itemLen-3)] == '(') ||
        (item[(ii=itemLen-4)] == '(') ))
    {
        sectLen = itemLen-ii-2 ;
        strcpy(sect,item+ii) ;
        itemLen = ii ;
        item[itemLen] = '\0' ;
    }
    else
    {
        sectLen = 0 ;
        sect[0] = '\0' ;
        sect[1] = '\0' ;
    }
    
    elp = &helpStore ;
try_again:
    lp = lforw(elp) ;
    while(lp != elp)
    {
        if((lp->l_text[0] == '!') &&
           (!sectLen || 
            ((sect[1] == (char) lp->l_text[2]) &&
             (((sectLen == 1) && (lp->l_text[3] == ' ')) || 
              (sect[2] == (char) lp->l_text[3])))))
        {
            if((cc=lp->l_text[1]) == ' ')
            {
                ii = llength(lp) - 4 ;
                if(ii != itemLen)
                    ii = -1 ;
            }
            else
            {
                ii = cc - '0' ;
                if(ii > itemLen)
                    ii = -1 ;
            }
            if((ii > 0) && !strncmp(item,(char *) lp->l_text+4,ii))
                break ;
        }
        lp = lforw(lp) ;
    }
        
    if(lp == elp)
    {
        static unsigned char loaded=0 ;
        meMACRO *mac ;
        if(!loaded)
        {
            if(loadHelpFile() != TRUE)
                return ABORT ;
            loaded = 1 ;
            goto try_again ;
        }
        if((getMacroTypeS(item) == TKCMD) &&
           ((ii = decode_fncname(item,1)) >= CK_MAX) &&
           ((mac = getMacro(ii)) != NULL) &&
           (mac->hlp->l_flag & MACFILE))
        {
            if(mac->fname != NULL)
                dofile(mac->fname,0,1) ;
            else
                dofile(mac->name,0,1) ;
            if(!(mac->hlp->l_flag & MACFILE))
                goto try_again ;
        }
        if(!silent)
            mlwrite(MWABORT,"[Can't find help on %s%s]",item,sect);
        return ABORT ;
    }
    if((wp = wpopup("*help*",(BFND_CREAT|BFND_CLEAR|WPOP_USESTR))) == NULL)
        return FALSE ;
    if((sectLen == 0) && (lp->l_text[2] != ' '))
    {
        ss = sect ;
        *ss++ = '(' ;
        *ss++ = lp->l_text[2] ;
        if(lp->l_text[3] != ' ')
            *ss++ = lp->l_text[3] ;
        *ss++ = ')' ;
        *ss = '\0' ;
    }
    
    bp = wp->w_bufp ;
    /* Add the header */
    {
        char buff[MAXBUF] ;
        sprintf(buff,"\033cD%s%s\033cA",lp->l_text+4,sect) ;
        addLineToEob(bp,buff) ;
        addLineToEob(bp,"") ;
        addLineToEob(bp,"\033lsMicroEmacs\033lm[Home]\033le \033lsCommand G\033lm[Commands]\033le \033lsVariable \033lm[Variables]\033le \033lsMacro Lan\033lm[Macro-Dev]\033le \033lsGlobal G\033lm[Glossary]\033le") ;
        memset(buff,boxChars[BCEW],78) ;
        buff[78] = '\0' ;
        addLineToEob(bp,buff) ;
        addLineToEob(bp,"") ;
    }
    while(((lp=lforw(lp)) != elp) && (lp->l_text[0] == '!'))
        ;
    while((lp != elp) && ((cc=lp->l_text[0]) != '!'))
    {
        if(cc == '|')
        {
            if(strcmp(item,(char *) lp->l_text+1))
                lp = lforw(lp) ;
        }
        else if(cc == '$')
        {
            if(lp->l_text[1] == 'a')
            {
                if(sect[1] == '5')
                {
                    char line[MAXBUF], *ss ;
                    if((ss = getval(item)) != NULL)
                    {
                        addLineToEob(bp,"") ;
                        addLineToEob(bp,"") ;
                        addLineToEob(bp,"\033cEVALUE\033cA") ;
                        addLineToEob(bp,"") ;
                        strcpy(line,"    \"") ;
                        strncpy(line+5,ss,MAXBUF-13) ;
                        line[MAXBUF-2] = '\0' ;
                        strcat(line,"\"") ;
                        addLineToEob(bp,line) ;
                    }
                }
                if(sect[1] == '2')
                {
                    if((ii = decode_fncname(item,1)) >= 0)
                    {
                        KEYTAB *ktp ;
                        char line[MAXBUF], *ss ;
                        addLineToEob(bp,"") ;
                        addLineToEob(bp,"") ;
                        addLineToEob(bp,"\033cEBINDINGS\033cA") ;
                        addLineToEob(bp,"") ;
                        strcpy(line,"    ") ;
                        ss = line+4 ;
                        for(ktp = &keytab[0] ; ktp->code != ME_INVALID_KEY ; ktp++)
                        {
                            if(ktp->index == ii)
                            {
                                *ss++ = '"' ;
                                cmdstr(ktp->code,ss);
                                ss += strlen(ss) ;
                                *ss++ = '"' ;
                                *ss++ = ' ' ;
                            }
                        }
                        if(ss == line+4)
                            strcpy(ss,"none") ;
                        else
                            *ss = '\0' ;
                        addLineToEob(bp,line) ;
                    }
                }
            }
        }
        else
            addLineToEob(bp,(char *)lp->l_text) ;
        lp = lforw(lp) ;
    }
    /* Add the footer */
    {
        extern char meCopyright[] ;
        char buff[MAXBUF] ;
        addLineToEob(bp,"") ;
        memset(buff,boxChars[BCEW],78) ;
        buff[78] = '\0' ;
        addLineToEob(bp,buff) ;
        addLineToEob(bp,"") ;
        sprintf(buff,"\033lsCopyright\033lm%s\033le",meCopyright) ;
        addLineToEob(bp,buff) ;
    }
    bp->b_dotp = lforw(bp->b_linep);
    bp->b_doto = 0 ;
    bp->line_no = 0 ;
    meModeClear(bp->b_mode,MDEDIT) ;    /* don't flag this as a change */
    meModeSet(bp->b_mode,MDVIEW) ;      /* put this buffer view mode */
    resetBufferWindows(bp) ;            /* Update the window */
    mlerase(MWCLEXEC);	                /* clear the mode line */
    return TRUE ;
}

int
help(int f, int n)
{
    return findHelpItem("MicroEmacs",0) ;
}

int	
helpItem(int f, int n)
{
    char buf[MAXBUF] ;
    
    if(mlreply("Help on item", 0, 0, buf, MAXBUF-10) != TRUE)
        return FALSE ;
    return findHelpItem(buf,0) ;
}

int	
helpCommand(int f, int n)
{
    char  buf[MAXBUF] ;
    char *ss ;
              
    if(mlreply("Help on command", MLCOMMAND, 0, buf, MAXBUF-10) != TRUE)
        return FALSE ;
    ss = buf + strlen(buf) ;
    strcpy(ss,"(2)") ;
    if(findHelpItem(buf,1) == TRUE)
        return TRUE ;
    strcpy(ss,"(3)") ;
    return findHelpItem(buf,0) ;
}

int	
helpVariable(int f, int n)
{
    char     buf[MAXBUF] ;

    if(mlreply("Help on variable", MLVARBL, 0, buf, MAXBUF-10) != TRUE)
        return FALSE ;
    strcat(buf,"(5)") ;
    return findHelpItem(buf,0) ;
}


unsigned long
cmdHashFunc (register char *cmdName)
{
    unsigned long ii = 0;
    unsigned long jj;
    char cc;
    
    while ((cc = *cmdName++) != '\0')
    {
        ii = (ii << 4) + cc;
        if ((jj = ii & 0xf0000000) != 0)
        {
            ii ^= (jj >> 24);
            ii ^= jj;
        }
    }
    return ii % cmdHashSize ;
}
     

/* match fname to a function in the names table
   and return index into table or -1 if none	
   fname;	name to attempt to match */
int
decode_fncname(register char *fname, int silent)
{
    unsigned long key ;
    meCMD *cmd ;
    
    key = cmdHashFunc(fname) ;
    for (cmd=cmdHash[key] ; cmd != NULL; cmd = cmd->hnext)
        if (strcmp (fname, cmd->name) == 0)
            return cmd->id ;
    
    if(!silent)
        mlwrite(MWABORT,"[No such command \"%s\"]", fname);
    
    return -1 ;				/* Not found - error. */
}

void
charMaskTblInit(void)
{
    int ii ;
    for(ii=0 ; ii<128 ; ii++)
    {
        charLatinUserTbl[ii] = ii ;
        charUserLatinTbl[ii] = ii ;
        charMaskTbl2[ii] &= ~(CHRMSK_LOWER|CHRMSK_UPPER|CHRMSK_HEXDIGIT|CHRMSK_SPLLEXT) ;
        charCaseTbl[ii] = ii ;
    }
    for( ; ii<256 ; ii++)
    {
        charLatinUserTbl[ii] = ii ;
        charUserLatinTbl[ii] = ii ;
        charMaskTbl2[ii] = 0x00 ;
        charCaseTbl[ii] = ii ;
    }
    for(ii=128 ; ii<160 ; ii++)
#ifdef _UNIX
        charMaskTbl1[ii] = 0x0A ;
#else
        charMaskTbl1[ii] = 0x3A ;
#endif
    for( ; ii<256 ; ii++)
        charMaskTbl1[ii] = 0x7A ;
    for(ii='0' ; ii<='9' ; ii++)
        charMaskTbl2[ii] |= CHRMSK_HEXDIGIT ;
    for(ii='A' ; ii<='F' ; ii++)
    {
        charMaskTbl2[ii] |= (CHRMSK_UPPER|CHRMSK_HEXDIGIT) ;
        charCaseTbl[ii] = (ii^0x20) ;
    }
    for( ; ii<='Z' ; ii++)
    {
        charMaskTbl2[ii] |= CHRMSK_UPPER ;
        charCaseTbl[ii] = (ii^0x20) ;
    }
    for(ii='a' ; ii<='f' ; ii++)
    {
        charMaskTbl2[ii] |= (CHRMSK_LOWER|CHRMSK_HEXDIGIT) ; ;
        charCaseTbl[ii] = (ii^0x20) ;
    }
    for( ; ii<='z' ; ii++)
    {
        charMaskTbl2[ii] |= CHRMSK_LOWER ;
        charCaseTbl[ii] = (ii^0x20) ;
    }
    charMaskTbl2[0x27] |= CHRMSK_SPLLEXT ;
    charMaskTbl2[0x2d] |= CHRMSK_SPLLEXT ;
    charMaskTbl2[0x2e] |= CHRMSK_SPLLEXT ;
    charMaskTbl2[0x5f] |= CHRMSK_USER1 ;
}


int
setCharMask(int f, int n)
{
    static   char tnkyPrompt[]="set-char-mask chars" ;
    unsigned char buf1[MAXBUF], *ss, *dd ;
    int	          c1, c2, flags, ii, jj ;
    unsigned char mask1, mask2 ;
    
    strcpy(tnkyPrompt+14,"flags") ;
    if(mlreply(tnkyPrompt,0,0,(char *) buf1,20) != TRUE)
        return FALSE ;
    /* setup the masks */
    flags = 0 ;
    ss = buf1 ;
    while((c1 = *ss++) != '\0')
        if((dd = (unsigned char *) strchr((char *) charMaskFlags,c1)) != NULL)
            flags |= 1 << (((int) dd)- ((int) charMaskFlags)) ;
    mask2 = (flags & 0x00ff) ;
    mask1 = (flags & 0x0f00) >> 4 ;
    
    if(n == 0)
    {
        ss = (unsigned char *) resultStr ;
        if(flags & 0x1c000)
        {
            if(flags & 0x4000)
                flags |= 0x18000 ;
            for(c1=0 ; c1<256 ; c1++)
            {
                c2 = charLatinUserTbl[c1] ;
                if(c1 != c2)
                {
                    if(flags & 0x8000)
                    {
                        if(c1 == 0)
                        {
                            *ss++ = 255 ;
                            c1 = 1 ;
                        }
                        else if(c1 == 255)
                            *ss++ = 255 ;
                        *ss++ = c1 ;
                    }
                    if(flags & 0x10000)
                    {
                        if(c2 == 0)
                        {
                            *ss++ = 255 ;
                            c2 = 1 ;
                        }
                        else if(c2 == 255)
                            *ss++ = 255 ;
                        *ss++ = c2 ;
                    }
                }
            }
        }
        else
        {
            if(flags & 0x1000)
                mask2 |= CHRMSK_LOWER ;
            if(flags & 0x2000)
                mask2 |= (CHRMSK_LOWER|CHRMSK_HEXDIGIT) ;
            for(c1=0 ; c1<256 ; c1++)
                if((charMaskTbl1[c1] & mask1) ||
                   (charMaskTbl2[c1] & mask2))
                {
                    if((flags & 0x3000) &&
                       ((c2 = toggleCase(c1)) != c1))
                    {
                        if(c2 == 0)
                        {
                            *ss++ = 255 ;
                            c2 = 1 ;
                        }
                        else if(c2 == 255)
                            *ss++ = 255 ;
                        *ss++ = c2 ;
                    }
                    if(c1 == 0)
                    {
                        *ss++ = 255 ;
                        c1 = 1 ;
                    }
                    else if(c1 == 255)
                        *ss++ = 255 ;
                    *ss++ = c1 ;
                }
        }
        *ss = '\0' ;
    }
    else
    {
#if	MAGIC
        /* reset the last magic search string to force a recompile as char groups may change */
        extern char mereLastPat[] ;
        mereLastPat[0] = '\0' ;
#endif
        if(flags & 0x1A003)
            return mlwrite(MWABORT,"[Cannot set flags u, l, A, L or U]");
        strcpy(tnkyPrompt+14,"chars") ;
        if(mlreply(tnkyPrompt,0,0,(char *) buf1,MAXBUF) != TRUE)
            return FALSE ;
        if(flags & 0x4000)
        {
            /* reset the tables */
            charMaskTblInit() ;
            ss = buf1 ;
            while((c1=*ss++) != '\0')
            {
                if((c1 == 255) && ((c1=*ss++) == 1))
                    c1 = 0 ;
                if((c2=*ss++) == '\0')
                    break ;
                if((c2 == 255) && ((c2=*ss++) == 1))
                    c2 = 0 ;
                charLatinUserTbl[c1] = c2 ;
                charUserLatinTbl[c2] = c1 ;
            }
            return TRUE ;
        }
        /* convert the string to user font and take out any 255 padding, 0's are now 0's so need int count */
        ss = buf1 ;
        dd = buf1 ;
        ii = 0 ;
        while((c1=*ss++) != '\0')
        {
            if((c1 == 255) && ((c1=*ss++) == 1))
                c1 = 0 ;
            *dd++ = toUserFont(c1) ;
            ii++ ;
        }
        if(n < 0)
        {
            if(flags & 0x1000)
                mask2 |= CHRMSK_ALPHA ;
            if(mask1 != 0)
            {
                mask1 ^= 0xff ;
                for(jj=0 ; jj<ii ; jj++)
                    charMaskTbl1[buf1[jj]] &= mask1 ;
            }
            if(mask2 != 0)
            {
                mask2 ^= 0xff ;
                for(jj=0 ; jj<ii ; jj++)
                    charMaskTbl2[buf1[jj]] &= mask2 ;
            }
        }
        else
        {
            if(flags & 0x1000)
            {
                for(jj=0 ; jj<ii ; )
                {
                    c1 = buf1[jj++] ;
                    c2 = buf1[jj++] ;
                    charMaskTbl2[c1] |= CHRMSK_UPPER ;
                    charMaskTbl2[c2] |= CHRMSK_LOWER ;
                    charCaseTbl[c1] = c2 ;
                    charCaseTbl[c2] = c1 ;
                }
            }
            if(mask1 != 0)
            {
                for(jj=0 ; jj<ii ; jj++)
                    charMaskTbl1[buf1[jj]] |= mask1 ;
            }
            if(mask2 != 0)
            {
                for(jj=0 ; jj<ii ; jj++)
                    charMaskTbl2[buf1[jj]] |= mask2 ;
            }
        }
    }
    return (TRUE);
}

/* bindtokey:	add a new key to the key binding table
*/

int
bindkey(char *prom, int f, int n, unsigned short *lclNoBinds, KEYTAB **lclBinds)
{
    register int	ss;
    register unsigned short cc;	/* command key to bind */
    register Fintii	kfunc;	/* ptr to the requexted function to bind to */
    register KEYTAB    *ktp;	/* pointer into the command table */
    char     outseq[40];        /* output buffer for keystroke sequence */
    int	     namidx;		/* Name index */
    char     buf[MAXBUF] ;
    unsigned int   arg ;
    unsigned short ssc ;
    
    /*---	Get the function name to bind it to */
    
    if(mlreply(prom, MLCOMMAND, 0, buf, MAXBUF) != TRUE)
        return FALSE ;
    if((namidx = decode_fncname(buf,0)) < 0)
        return FALSE ;
    kfunc = getCommandFunc(namidx) ;

    /*---	Prompt the user to type in a key to bind */
    /*---	Get the command sequence to bind */
    
    if((ss=(kfunc==(Fintii) prefixFunc)))
    {
        if((n < 1) || (n > ME_PREFIX_NUM))
            return mlwrite(MWABORT,"[Prefix number is out of range]") ;
        ssc = prefixc[n] ;
    }
    else if((ss=(kfunc==(Fintii) uniArgument)))
        ssc = reptc ;

#if LCLBIND
    if((lclNoBinds != NULL) && ss)
        return mlwrite(MWABORT,"Can not locally bind [%s]",buf) ;
#endif              
    mlwrite(MWCURSOR|MWCLEXEC,"%s [%s] to: ",prom,buf) ;
    cc = getckey(ss,1) ;
    
    /*---   Change it to something we can print as well */
    
    if (!clexec)
    {
        cmdstr(cc, &outseq[0]);
        mlwrite(0,"%s [%s] to: %s",prom,buf,outseq);
    }
    
    /*---   Calculate the argument */
    if(f)
        arg = (unsigned int) (n+0x80000000) ;
    else
        arg = 0 ;
    
    /*---	If the function is a prefix key */
    
    if(ss)
    {
	/*---	Search for an existing binding for the prefix key */
	if(ssc != ME_INVALID_KEY)
            delete_key(ssc);
        
	/*---	Reset the appropriate global prefix variable */
        if(kfunc == (Fintii) prefixFunc)
            prefixc[n] = cc ;
        else
            reptc = cc;
    }
    
    /*---	Search the table to see if it exists */
#if LCLBIND
    if(lclNoBinds != NULL)
    {
        register KEYTAB *kk=NULL ;
        
        ktp = *lclBinds ;
        for (ss=0 ; ss < *lclNoBinds ; ss++)
        {
            if(ktp[ss].code == cc)
            {
                ktp[ss].index = (unsigned short) namidx;
                ktp[ss].arg   = arg ;
                return TRUE ;
            }
            if(ktp[ss].code == 0)
                kk = ktp+ss ;
        }
        if(kk != NULL)
        {
            kk->code  = cc ;
            kk->index = (unsigned short) namidx;
            kk->arg   = arg ;
        }
        else
        {
            if((ktp = *lclBinds = 
                meRealloc(ktp,(++(*lclNoBinds))*sizeof(KEYTAB))) == NULL)
                return FALSE ;
            ktp[ss].code  = cc ;
            ktp[ss].index = (unsigned short) namidx ;
            ktp[ss].arg   = arg ;
        }
    }
    else
#endif
    {
        for (ktp = &keytab[0]; ktp->code != 0; ktp++)
            if (ktp->code == cc)
            {
                ktp->index = (unsigned short) namidx;
                ktp->arg   = arg ;
                return TRUE ;
            }
        
        /*---	Otherwise we  need to  add it  to the  end, if  we run  out of
           binding room, bitch */
        
        if(insert_key(cc, (unsigned short) namidx, arg) != TRUE)
            return mlwrite(MWABORT,"Binding table FULL!");
    }
    return(TRUE);
}


int
globalBind(int f, int n)
{
    return bindkey("global bind", f, n, NULL, NULL) ;
}

#if LCLBIND
int
bufferBind(int f, int n)
{
    return bindkey("local bind", f, n, &(curbp->nobbinds), &(curbp->bbinds)) ;
}
int
mlBind(int f, int n)
{
    return bindkey("ml bind", f, n, &mlNoBinds, &mlBinds) ;
}
#endif

/* unbindkey:	delete a key from the key binding table */

int
unbindkey(char *prom, int n, unsigned short *lclNoBinds, KEYTAB **lclBinds)
{
    register unsigned short cc;	/* command key to unbind */
    char outseq[40];	/* output buffer for keystroke sequence */
    int  rr ;
    
    /*---	Prompt the user to type in a key to unbind */
    
    /* if interactive then print prompt */
    mlwrite(MWCURSOR|MWCLEXEC,"%s unbind: ",prom);
    
    /*---	Get the command sequence to unbind */
    
    cc = getckey(n,1);		/* get a command sequence */
    
    /* If interactive then there is no really need to print the string because
     * you will probably not see it !! The TTputc() is dangerous and squirts
     * characters over the screen if the previous mlwrite() determined that
     * we were in a macro and has not moved the cursor to the message line.
     * Hence, safer bet is to re-do the mlwrite(); my preference would be
     * to remove the keyboard echo !! - Jon Green 13/03/97 */
    cmdstr(cc, &outseq[0]);		/* change to something printable */
    mlwrite(MWCURSOR|MWCLEXEC, "%s unbind: %s", prom, outseq);
#if LCLBIND
    if(lclNoBinds != NULL)
    {
        register KEYTAB *ktp ;
        register int	 ss;
        
        rr = FALSE ;
        ktp = *lclBinds ;
        ss = *lclNoBinds ;
        while(--ss >= 0)
        {
            if(ktp[ss].code == cc)
            {
                ktp[ss].code = ME_INVALID_KEY ;
                rr = TRUE ;
            }
        }
    }
    else
#endif
    {
        if(((rr = delete_key(cc)) == TRUE) && n) 
        {
            register unsigned short mask=ME_PREFIX_NUM+1;
        
            while(--mask > 0)
                if(cc == prefixc[mask])
                    break ;
            if(mask)
            {
                int ii=keyTableSize ;
                
                prefixc[mask] = ME_INVALID_KEY ;
                mask <<= ME_PREFIX_BIT ;
                while(--ii>=0)
                    if((keytab[ii].code & ME_PREFIX_MASK) == mask)
                        delete_key(keytab[ii].code) ;
#if LCLBIND
                {
                    register int ss ;
                    BUFFER *bp ;
                    
                    /* loop through all buffer bindings using prefix */
                    bp = bheadp ;
                    while(bp != NULL)
                    {
                        ss = bp->nobbinds ;
                        while(--ss >= 0)
                            if((bp->bbinds[ss].code & ME_PREFIX_MASK) == mask)
                                bp->bbinds[ss].code = ME_INVALID_KEY ;
                        bp = bp->b_bufp ;
                    }
                    /* loop through all ml bindings using prefix */
                    ss = mlNoBinds ;
                    while(--ss >= 0)
                        if((mlBinds[ss].code & ME_PREFIX_MASK) == mask)
                            mlBinds[ss].code = ME_INVALID_KEY ;
                }
#endif
            }
            else if (reptc == cc)
                reptc = ME_INVALID_KEY ;
        }
    }
    if(rr == FALSE)	/* if it isn't bound, bitch */
        return mlwrite(MWABORT|MWCLEXEC,"[%s Key \"%s\" not bound]", prom,outseq);
    return TRUE ;
}

int
globalUnbind(int f, int n)
{
    return unbindkey("Global",n==0, NULL, NULL) ;
}

#if LCLBIND
int
bufferUnbind(int f, int n)
{
    return unbindkey("Local",0, &(curbp->nobbinds), &(curbp->bbinds)) ;
}
int
mlUnbind(int f, int n)
{
    return unbindkey("ML",0, &mlNoBinds, &mlBinds) ;
}
#endif


/* list-functions
 * bring up a fake buffer and list the key bindings into it with view mode
 */
int
listCommands(int f, int n)
#if	APROP
{
    return (buildlist(TRUE, n, ""));
}

int
commandApropos(int f, int n)	/* Apropos (List functions that match a substring) */
{
    char mstring[NSTRING];	/* string to match cmd names to */
    int	 status;		/* status return */

    if ((status = mlreply("Apropos string", MLCOMMAND, 0, 
                          mstring+1, NSTRING-2)) != TRUE)
        return status ;
    mstring[0] = '*' ;
    strcat(mstring,"*") ;
    return buildlist(FALSE, n, mstring);
}


/* build a binding list (limited or full) */
/* type  	true = full list,   false = partial list */
/* mstring	match string if a partial list */
int
buildlist(int type, int n, char *mstring)
#endif
{
    meCMD    *cmd ;
    KEYTAB   *ktp;	        /* pointer into the command table */
    WINDOW   *wp;
    BUFFER   *bp;	        /* buffer to put binding list into */
    int       cpos;   		/* current position to use in outseq */
    char      _outseq[MAXBUF]; 	/* output buffer for keystroke sequence */
    char     *outseq=_outseq+4;	/* output buffer for keystroke sequence */
    int       curidx;           /* Name integer index. */
    char      op_char;		/* Output character */
    
    /* and get a buffer for it */
    if((wp = wpopup("*commands*",(BFND_CREAT|BFND_CLEAR|WPOP_USESTR))) == NULL)
        return mlwrite(MWABORT,"[Failed to create list]") ;
    bp = wp->w_bufp ;
    
    _outseq[0] = ' ' ;
    _outseq[1] = ' ' ;
    _outseq[2] = ' ' ;
    _outseq[3] = ' ' ;
    wp->w_markp = NULL;
    wp->w_marko = 0;
    
    /* build the contents of this window, inserting it line by line */
    for(cmd=cmdHead ; cmd != NULL ; cmd=cmd->anext)
    {
#if	APROP
        /*---	If we are  executing an  apropos command  and current string
        doesn't include the search string */
        if ((type == FALSE) && !regexpStrCmp(cmd->name,mstring,1))
            continue;		/* Do next loop */
#endif
        /*---	Add in the command name */
        strcpy(outseq, cmd->name);
        curidx = cmd->id ;
        cpos = strlen(outseq);
        
        /*---	Search down any keys bound to this */
        
        for (ktp = &keytab[0] ; ktp->code != ME_INVALID_KEY ; ktp++)
        {
            if (ktp->index == curidx)
            {
                /*---	Padd out some  spaces.  If we  are on a line
                with a  string then  pad  with '.'  else use
                space's */
                op_char = ((cpos) ? '.' : ' ');
                outseq[cpos++] = ' ';
                while (cpos < 31)
                    outseq[cpos++] = op_char;
                outseq[cpos++] = ' ';
                outseq[cpos++] = '"';
                /*---	Add in the command sequence */
                cmdstr(ktp->code, &outseq[cpos]);
                strcat(outseq+cpos,"\"");
                
                /* and add it as a line into the buffer */
                addLineToEob(bp,_outseq) ;
                cpos = 0;	/* and clear the line */
            }
        }
        
        /* if no key was bound, we need to dump it anyway */
        if (cpos > 0)
            addLineToEob(bp,_outseq) ;
    }
    meModeClear(bp->b_mode,MDEDIT) ;    /* don't flag this as a change */
    meModeSet(bp->b_mode,MDVIEW) ;      /* put this buffer view mode */
    wp->w_dotp = lforw(bp->b_linep);    /* back to the beginning */
    wp->w_doto = 0;
    mlerase(MWCLEXEC);          	/* clear the mode line */
    return(TRUE);
}


/*
 * listbindings
 * List ALL of the bindings held in the system.
 */
static void
showBinding(BUFFER *bp, KEYTAB *ktp)
{
    char *ss, buf[MAXBUF], outseq[40];
    int len;
    
    /* If the keyboard code is zero then the command has been unbound.
     * Situation occurs when the buffer or ml binding list is traversed.
     * We test at point of print rather than at the caller so that we do
     * not repeat the test. Speed is not a crital issue here. */
    if (ktp->code == ME_INVALID_KEY)
        return;
    
    /* The key command is non-zero therefore it is legal - print it */
    cmdstr(ktp->code, outseq);
    len = sprintf(buf, "    \"%s\" ", outseq);
    while (len < 35)
        buf[len++] = '.';
    ss = buf+len ;
    *ss++ = ' ';
    if(ktp->arg)
        sprintf(ss,"%d ",(int) (ktp->arg + 0x80000000)) ;
    else
        *ss = 0 ;
    strcat(ss,getCommandName(ktp->index)) ;
    addLineToEob(bp,buf) ;
}

int
descBindings (int f, int n)
{
    WINDOW *wp ;
    BUFFER *bp ;
    char    buf[MAXBUF] ;
    int     ii ;
    KEYTAB *ktp;
    
    if((wp = wpopup("*bindings*",(BFND_CREAT|BFND_CLEAR|WPOP_USESTR))) == NULL)
        return FALSE ;
    bp = wp->w_bufp ;
    
#if LCLBIND
    sprintf(buf,"Buffer [%s] bindings:", curbp->b_bname);
    addLineToEob(bp,buf);
    addLineToEob(bp,"") ;
    
    ii  = curbp->nobbinds ;
    ktp = curbp->bbinds ;
    while(ii--)
        showBinding(bp,ktp++) ;
    addLineToEob(bp,"") ;
    
    sprintf(buf,"Ml bindings:") ;
    addLineToEob(bp,buf);
    addLineToEob(bp,"") ;
    
    ii  = mlNoBinds ;
    ktp = mlBinds ;
    while(ii--)
        showBinding(bp,ktp++) ;
    addLineToEob(bp,"") ;
#endif    
    
    addLineToEob(bp,"Global bindings:");
    addLineToEob(bp,"") ;
    
    ii  = keyTableSize ;
    ktp = keytab ;
    while(ii--)
        showBinding(bp,ktp++) ;
    
    bp->b_dotp = lforw(bp->b_linep);
    bp->b_doto = 0 ;
    bp->line_no = 0 ;
    
    meModeClear(bp->b_mode,MDEDIT) ;    /* don't flag this as a change */
    meModeSet(bp->b_mode,MDVIEW) ;      /* put this buffer view mode */
    resetBufferWindows(bp) ;            /* Update the window */
    mlerase(MWCLEXEC);	                /* clear the mode line */
    return TRUE ;
}
     

