/*
 *	SCCS:		%W%		%G%		%U%
 *
 *	Last Modified :	<160699.0741>
 *
 * ESEARCH.H
 *
 * Defines, typdefs, and global variables that are of use for the
 * routines in search.c and isearch.c.
 *
 * July 1990. Jon Green.
 * Added the Auto Repeat structure for Magic search/replace importing.
 * 
 ****************************************************************************
 * 
 * Modifications to the original file by Jasspa. 
 * 
 * Copyright (C) 1988 - 1999, JASSPA 
 * The MicroEmacs Jasspa distribution can be copied and distributed freely for
 * any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
 * incoportated into commercial software with the expressed permission of
 * JASSPA.
 * 
 ****************************************************************************/

/*
 * PTBEG, PTEND, FORWARD, and REVERSE are all toggle-able values for
 * the scan routines.
 */
#define	PTBEG	0	/* Leave the point at the beginning on search.*/
#define	PTEND	1	/* Leave the point at the end on search.*/

#if	MAGIC

#define _GNU_REGEX 1

#ifdef _GNU_REGEX

#ifndef __STDC__
#define __STDC__ 1
#endif

#include <regex.h>

#define	MC_RRCL		'\\'	/* Replicate replace closure. (destination). */

#else
/*
 * Defines for the metacharacters in the regular expressions.
 */

#define	MCNIL		0	/* Like the '\0' for strings.*/
#define	LITCHAR		1
#define	ANY		2
#define	CCL		3
#define	NCCL		4
#define	BOL		5
#define	EOL		6
#define	SRCL		7	/* Start of replace closure. (source). */
#define	ERCL		8	/* End of replace closure. (destination). */
#define	RRCL		9	/* Replicate replace closure. (destination). */
#define	O_CLOSURE	0x0100	/* One closure. An orable value. */
#define	CLOSURE		0x0200	/* Zero closure. An or-able value.*/
#define	S_CLOSURE	0x0400	/* Single closure. An orable value. */
#define CLOSURE_MASK    0x0700  /* Closure mask */
#define	MASKCL		0x00ff

#define	MC_ANY		'.'	/* 'Any' character (except newline).*/
#define	MC_CCL		'['	/* Character class.*/
#define	MC_NCCL		'^'	/* Negate character class.*/
#define	MC_RCCL		'-'	/* Range in character class.*/
#define	MC_ECCL		']'	/* End of character class.*/
#define	MC_BOL		'^'	/* Beginning of line.*/
#define	MC_EOL		'$'	/* End of line.*/
#define	MC_CLOSURE	'*'	/* Closure - does not extend past newline.*/
#define	MC_O_CLOSURE	'+'	/* Closure of 1 - does not extend past newline */
#define	MC_S_CLOSURE	'?'	/* Closure of 0 or 1 - does not extend past newline */

#define	MC_SRCL		'('	/* Start of replace closure. (source). */
#define	MC_ERCL		')'	/* End of replace closure. (destination). */
#define	MC_RRCL		'\\'	/* Replicate replace closure. (destination). */

#define	MC_ESC		'\\'	/* Escape - suppress meta-meaning.*/

#define	meBIT(n)	(1 << (n))	/* An integer with one bit set.*/

/* HICHAR - 1 is the largest character we will deal with.
 * HIBYTE represents the number of bytes in the bitmap.
 */

#define	meHICHAR	256
#define	meHIBYTE	meHICHAR >> 3

typedef char	*meBITMAP;

typedef	struct {
    short int	 mc_type;
    union {
        int	 lchar;
        meBITMAP cclmap;
    } u;
} MC;

/*---	The auto repeat function structure - allows search strings to be
	replicated in the replacement string. There are only 10 locations
	corresponding to '0' - '9' */
typedef struct	{
    int	start_pos [10];		        /* Search string start position */
    int	end_pos [10];		        /* Search string end postion */
    int	stack [10];		        /* Search closure stack. */
    int	count;			        /* Count of number of closures */
    int	hi_water;		        /* Stack hi-water mark */
    int	do_repeat;		        /* TRUE = do repeat */
} A_RPT;

#endif
#endif

typedef struct {                       
    LINE *startline;                    /* Pattern start position */
    int   startoff;                     /* Pattern start line offset */
    int   startline_no;                 /* Pattern start line number */
    LINE *endline;                      /* Pattern end position */
    int   endoffset;                    /* Pattern end offset */
    int   endline_no;                   /* Pattern end line number */
} SCANNERPOS;

extern int iscanner (char *apat, int n, int flags, SCANNERPOS *sp);

/* search arrays and variables, defined in search.c, but exported
 * here so others can access them.
 */ 
extern char srchPat[] ;                     /* current search string          */
extern char srchRPat[] ;                    /* reverse current search string  */
#if MAGIC
#ifdef _GNU_REGEX
extern struct re_pattern_buffer mereBuffer ;
extern struct re_registers mereRegs ;
#else
extern MC   srchMcPat[] ;                   /* current search MC data         */
extern MC   srchMcRPat[] ;                  /* reverse current search MC data */
extern A_RPT srchAutoRpt;	            /* Auto repeat store              */
#endif
#endif

/* Incremental search defines.
 */
#define ISCANNER_MAGIC  0x01            /* Require magical scanner */
#define ISCANNER_PTBEG  0x02            /* Point at start of string */
#define ISCANNER_QUIET  0x04            /* Turn off pattern building errors */
#define ISCANNER_EXACT  0x08            /* Exact pattern matching */
#define ISCANNER_BACKW  0x10            /* Search backwards */
#define ISEARCH_SCANMR  0x20            /* Isearch flag to scan more */
#ifdef _IPIPES
#define ISEARCH_PIPE    0x40            /* Isearch performed in ipipe buffer */
#endif

#if	ISRCH
#define	HISTBUFSIZ	256	        /* Length of our command buffer */
#endif
