/**-*- c -*-*************************************************************
 * 
 *	SCCS:		%W%		%G%		%U%
 *
 *	Last Modified :	<990714.0751>
 *
 *	ESTRUCT:	Structure and preprocesser defined for
 *			MicroEMACS 3.7
 *
 *			written by Dave G. Conroy
 *			modified by Steve Wilhite, George Jones
 *			greatly modified by Daniel Lawrence
 *
 * Jon Green	13-07-90
 * Added Document Mode. Search for .doc or .txt extension for WRAP and 
 * indent to automatically be enabled.
 * Jon Green	17-05-91
 * Added suffix modes so that the suffix defintions may be defined in the 
 * start up file.
 * 
 ************************************************************************
 * 
 * Modifications to the original file by Jasspa. 
 * 
 * Copyright (C) 1988 - 1999, JASSPA 
 * The MicroEmacs Jasspa distribution can be copied and distributed freely for
 * any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
 * incorporated into commercial software with the expressed permission of
 * JASSPA.
 * 
 ****************************************************************************/
/*	internal constants	*/

#define NBINDS	 256			/* max # of bound keys		*/
#define NBUFN    16                     /* # of bytes, buffer name      */
#define	NSTRING	 128			/* # of bytes, string buffers	*/
#define NKBDM    256                    /* # of strokes, keyboard macro */
#define NPAT     128                    /* # of bytes, pattern          */
#define	NBLOCK	 16			/* line block chunk size	*/
#define NWINDOWS 64                     /* # of windows MAXIMUM         */
#define MAXBUF	 1024			/* size of various inputs       */
#define FILEBUF	 1024			/* size of a file name input    */
#define	TSTMPLEN 40			/* Max len of a time stamp str. */

#define	meNLCHAR   0x0a                 /* the \n char ^J, not ^M       */

#define	ME_SHIFT   0x0100		/* special key shift    	*/
#define	ME_CONTROL 0x0200		/* special key conrtol		*/
#define	ME_ALT     0x0400		/* special key alt		*/
#define	ME_SPECIAL 0x0800		/* special key (function keys)	*/
#define ME_PREFIX_NUM  16               /* the number of prefixes       */
#define ME_PREFIX_BIT  12               /* the number of prefixes       */
#define ME_PREFIX_MASK 0xf000           /* the prefix bit mask          */
#define ME_PREFIX1 0x1000               /* prefix 1 flag, or'ed in      */
#define ME_PREFIX2 0x2000               /* prefix 2 flag, or'ed in      */
#define ME_PREFIX3 0x3000               /* prefix 3 flag, or'ed in      */
#define ME_PREFIX4 0x4000               /* prefix 4 flag, or'ed in      */
#define ME_INVALID_KEY 0x0000           /* an invalid key number        */

#define FALSE   0                       /* False, no, bad, etc.         */
#define TRUE    1                       /* True, yes, good, etc.        */
#define ABORT   2                       /* Death, ^G, abort, etc.       */

#define STOP	 0			/* keyboard macro not in use	*/
#define PLAY	 1			/*		  playing	*/
#define RECORD	 2			/*		  recording	*/
#define KBD_IDLE 3                      /* keyboard not in use/idle     */

#define FORWARD 0			/* forward direction		*/
#define REVERSE 1			/* backwards direction		*/

#define FIOSUC  0                       /* File I/O, success.           */
#define FIOFNF  1                       /* File I/O, file not found.    */
#define FIOEOF  2                       /* File I/O, end of file.       */
#define FIOERR  3                       /* File I/O, error.             */
#define	FIOLNG	4			/*line longer than allowed len	*/
#define FIOFUN	5			/* File I/O, eod of file/bad line*/
#define FIOBUFSIZ 2048

#define MLHISTSIZE 20

#define CFCPCN  0x0001                  /* Last command was C-P, C-N    */
#define CFKILL  0x0002                  /* Last command was a kill      */
#define	CFYANK	0x0004			/* Last command was yank	*/
#define	CFRYANK	0x0008			/* Last command was reyank	*/
#define CFUNDO  0x0010                  /* Last command was undo        */

#ifdef _UNIX
#define PATHCHR ':'
#else
#define PATHCHR ';'
#endif

extern void  free(void *block) ;
extern void* malloc(size_t s) ;
extern void* realloc(void *block, size_t s) ;

typedef	int	(*Fintv)(void);
typedef	int	(*Fintc)(char);
typedef	int	(*Finti)(int);
typedef	int	(*Fintii)(int,int);
typedef	int	(*Fints)(char *);
typedef	int	(*Fintss)(const char *, const char *);
typedef	int	(*Fintssi)(const char *, const char *,size_t);
typedef int	(*Fintccci)(char *,char *,char *,int) ;


#define INTWIDTH (sizeof(int)*4+2)

/*
 * All text is kept in circularly linked lists of "LINE" structures. These
 * begin at the header line (which is the blank line beyond the end of the
 * buffer). This line is pointed to by the "BUFFER". Each line contains a the
 * number of bytes in the line (the "used" size), the size of the text array,
 * and the text. The end of line is not stored as a byte; its implied. Future
 * additions will include update hints, and a list of marks into the line.
 */
typedef struct LINE
{
    struct LINE	   *l_fp;		/* Link to the next line        */
    struct LINE	   *l_bp;		/* Link to the previous line    */
    unsigned short  l_size;             /* Allocated size               */
    unsigned short  l_used;             /* Used size                    */
    unsigned char   l_flag;		/* Line is marked if true	*/
    unsigned char   l_text[1];          /* A bunch of characters.       */
} LINE ;

#define LNSMASK  0x0f
#define LNNOSCHM 0x10
#define LNCHNG   0x10
#define LNMARK   0x20
#define	LNNEOL   0x40			/* Save line with no \n or \0 	*/
#define LNZERO   0x80			/* NOT USED ANY MORE FREE BIT! 	*/

#define lforw(lp)       ((lp)->l_fp)
#define lback(lp)       ((lp)->l_bp)
#define lgetc(lp, n)    ((lp)->l_text[(n)])
#define lputc(lp, n, c) ((lp)->l_text[(n)]=(c))
#define ltext(lp)       ((lp)->l_text)
#define lchng(lp)       ((lp)->l_flag &= LNCHNG)
#define llength(lp)     ((lp)->l_used)

/*
 * There is a window structure allocated for every active display window. The
 * windows are kept in a big list, in top to bottom screen order, with the
 * listhead at "wheadp". Each window contains its own values of dot and mark.
 * The flag field contains some bits that are set by commands to guide
 * redisplay; although this is a bit of a compromise in terms of decoupling,
 * the full blown redisplay is just too expensive to run for every input
 * character.
 * 
 * The physical screen extents of the window are contained in firstCol, firstRow,
 * numCols numRows. These define the physical extremities of the window, within
 * those boundaries lie the message line, scroll bar and text area. 
 * The separate window extents appear to duplicate some of the existing values,
 * (which they do), however they do serve a purpose in that they decouple the 
 * window tesselation and layout (i.e. where a window is positioned on the
 * screen) from having to have any knowledge about the exact format of the 
 * contents of the window, thereby reducing maintenance of the tesselation 
 * code which is a bit of a mind bender at times.
 * 
 * Note that the "flags" are at the top. We access these alot on the display
 * side. Positioned here ensures that they are available immediatly without
 * any indirect offset - just give that compiler a hand. 
 */

/*
 * WINDOW structure w_flag values.
 * Define states to the display drivers to update the screen. 
 */
#define WFFORCE       0x0001            /* Window needs forced reframe  */
#define WFRESIZE      0x0002            /* Reset the width of lines.    */
#define WFMOVEL       0x0004            /* Movement from line to line   */
#define WFMOVEC       0x0008            /* Movement from col to col     */
#define WFMAIN        0x0010            /* Update window main area      */
#define WFMODE        0x0020            /* Update window mode line      */
#define WFDOT         0x0040            /* Redraw windows dot line      */
#define WFREDRAW      0x0080            /* Redraw main window area      */
#define WFSBSPLIT     0x0100            /* Scroll split has changed     */
#define WFSBUP        0x0200            /* Scroll up arrow change       */
#define WFSBUSHAFT    0x0400            /* Scroll bar up shaft change   */
#define WFSBBOX       0x0800            /* Scroll bar box change        */
#define WFSBDSHAFT    0x1000            /* Scroll bar down shaft change */
#define WFSBDOWN      0x2000            /* Scroll bar down arrow change */
#define WFSBML        0x4000            /* Scroll bar mode line changed */
#define WFSELHIL      0x8000            /* Selection highlight changed  */
#define WFSELDRAWN    0x10000           /* Editing within a line        */
#define WFLOOKBK      0x20000           /* Check the hilight look-back  */
/* Composite mode to force an update of the scroll box */
#define WFSBOX        (WFSBUSHAFT|WFSBBOX|WFSBDSHAFT)    
/* Composite mode to force update of the scroll bar */
#define WFSBAR        (WFSBSPLIT|WFSBUP|WFSBDOWN|WFSBML|WFSBOX)

/* A garbage update of the screen adds following modes */
#define WFUPGAR       (WFMODE|WFRESIZE|WFSBAR)

/*
 * WINDOW structure w_mode values.
 * Modes define the sroll bar state.
 */
#define WMVWIDE 0x01                    /* PUBLIC:  Wide vert divider 2 chars    */
#define WMUP    0x02                    /* PUBLIC:  Has upper end cap            */
#define WMDOWN  0x04                    /* PUBLIC:  Has lower end cap            */
#define WMBOTTM 0x08                    /* PUBLIC:  Has a mode line character    */
#define WMSCROL 0x10                    /* PUBLIC:  Has a box on scrolling shaft */
#define WMRVBOX 0x20                    /* PUBLIC:  Reverse video on box         */
#define WMHWIDE 0x40                    /* PUBLIC:  Wide horz divider 2 chars    */
#define WMSPLIT 0x80                    /* PUBLIC:  Has a splitter               */
#define WMVBAR  0x100                   /* PUBLIC:  Window has a vertical bar    */
#define WMHORIZ 0x200                   /* PRIVATE: This is a horizontal bar     */
#define WMWIDE  0x400                   /* PRIVATE: Wide scroll-bar (internal)   */
  
/* The user interface mask for the scroll mode */
#define WMUSER  (WMVWIDE|WMUP|WMDOWN|WMBOTTM|WMSCROL|WMRVBOX|WMHWIDE|WMSPLIT|WMVBAR)
/*
 * Define the window character array. This is a fixed array of 
 * characters which define the window components
 */
/* Mode line */
#define WCMLCWSEP     0                  /* Mode line current window sep  (=)*/
#define WCMLIWSEP     1                  /* Mode line inactive window sep (-)*/
#define WCMLROOT      2                  /* Mode line root indicatior     (#)*/
#define WCMLBCHNG     3                  /* Mode line buffer changed char (*)*/
#define WCMLBVIEW     4                  /* Mode line buffer view char    (%)*/
/* Scroll Bar */
#define WCVSBSPLIT    5                  /* Buffer split character #1        */
#define WCVSBUP       6                  /* Scroll bar up character          */
#define WCVSBUSHAFT   7                  /* Scroll bar up shaft character    */
#define WCVSBBOX      8                  /* Scroll bar box character         */
#define WCVSBDSHAFT   9                  /* Scroll bar down shaft character  */
#define WCVSBDOWN    10                  /* Scroll bar down character        */
#define WCVSBML      11                  /* Scroll bar/mode line point       */
#define WCVSBSPLIT1  12                  /* Buffer split character #1        */
#define WCVSBSPLIT2  13                  /* Buffer split character #2        */
#define WCVSBUP1     14                  /* Wide scroll bar up char #1       */
#define WCVSBUP2     15                  /* Wide scroll bar up char #2       */
#define WCVSBUSHAFT1 16                  /* Wide scroll bar up shaft char #1 */
#define WCVSBUSHAFT2 17                  /* Wide scroll bar up shaft char #2 */
#define WCVSBBOX1    18                  /* Wide scroll bar box char #1      */
#define WCVSBBOX2    19                  /* Wide scroll bar box char #2      */
#define WCVSBDSHAFT1 20                  /* Wide scroll bar dn shaft char #1 */
#define WCVSBDSHAFT2 21                  /* Wide scroll bar dn shaft char #2 */
#define WCVSBDOWN1   22                  /* Wide scroll bar down char #1     */
#define WCVSBDOWN2   23                  /* Wide scroll bar down char #2     */
#define WCVSBML1     24                  /* Wide scroll bar/mode point #1    */
#define WCVSBML2     25                  /* Wide scroll bar/mode point #2    */
#define WCHSBSPLIT   26                  /* Buffer split character #1        */
#define WCHSBUP      27                  /* Scroll bar up character          */
#define WCHSBUSHAFT  28                  /* Scroll bar up shaft character    */
#define WCHSBBOX     29                  /* Scroll bar box character         */
#define WCHSBDSHAFT  30                  /* Scroll bar down shaft character  */
#define WCHSBDOWN    31                  /* Scroll bar down character        */
#define WCHSBML      32                  /* Scroll bar/mode line point       */
#define WCHSBSPLIT1  33                  /* Buffer split character #1        */
#define WCHSBSPLIT2  34                  /* Buffer split character #2        */
#define WCHSBUP1     35                  /* Wide scroll bar up char #1       */
#define WCHSBUP2     36                  /* Wide scroll bar up char #2       */
#define WCHSBUSHAFT1 37                  /* Wide scroll bar up shaft char #1 */
#define WCHSBUSHAFT2 38                  /* Wide scroll bar up shaft char #2 */
#define WCHSBBOX1    39                  /* Wide scroll bar box char #1      */
#define WCHSBBOX2    40                  /* Wide scroll bar box char #2      */
#define WCHSBDSHAFT1 41                  /* Wide scroll bar dn shaft char #1 */
#define WCHSBDSHAFT2 42                  /* Wide scroll bar dn shaft char #2 */
#define WCHSBDOWN1   43                  /* Wide scroll bar down char #1     */
#define WCHSBDOWN2   44                  /* Wide scroll bar down char #2     */
#define WCHSBML1     45                  /* Wide scroll bar/mode point #1    */
#define WCHSBML2     46                  /* Wide scroll bar/mode point #2    */
#define WCOSDTTLBR   47                  /* Osd title bar char               */
#define WCOSDTTLKLL  48                  /* Osd title bar kill char          */
#define WCOSDRESIZE  49                  /* Osd resize char                  */
#define WCOSDBTTOPN  50                  /* Osd button start char e.g. '['   */
#define WCOSDBTTCLS  51                  /* Osd button close char e.g. ']'   */
#define WCLEN        52                  /* Number of characters             */

/* The box chars BC=Box Chars, N=North, E=East, S=South, W=West,
 * So BCNS = Box Char which has lines from North & South to centre, i.e. 
 *           the vertical line |
 *    BCNESW = the + etc.
 */
#define BCNS         0
#define BCES         1
#define BCSW         2
#define BCNE         3
#define BCNW         4
#define BCESW        5
#define BCNES        6
#define BCNESW       7
#define BCNSW        8
#define BCNEW        9
#define BCEW        10
#define BCLEN       11                  /* Number of characters         */

typedef struct  WINDOW {
    struct  WINDOW *w_wndp;             /* Next window                  */
    struct  WINDOW *w_wnup;             /* Previous window              */
    struct  WINDOW *w_lvideo;           /* VVIDEO thread of windows     */
    struct  VVIDEO *w_vvideo;           /* Virtual video block          */
    struct  BUFFER *w_bufp;             /* Buffer displayed in window   */
    struct  BUFFER *l_bufp;             /* Last Buffer displayed        */
    LINE   *w_dotp;		        /* Line containing "."          */
    LINE   *w_markp;		        /* Line containing "mark"       */
    LINE   *model ;                     /* window's mode-line buffer    */
    LINE   *curLineOff ;                /* Current lines char offsets   */
    long    topLineNo ;                 /* windows top line number      */
    long    line_no;                    /* current line number          */
    long    mlineno;                    /* current mark line number     */
    unsigned long    w_flag;            /* Flags.                       */
    unsigned short   w_doto;            /* Byte offset for "."          */
    unsigned short   w_marko;           /* Byte offset for "mark"       */
    unsigned short   w_force;           /* If NZ, forcing row.          */
    unsigned short   firstRow;          /* Window starting row          */
    unsigned short   firstCol;          /* Window starting column       */
    unsigned short   numCols;           /* Window number text columns   */
    unsigned short   numRows;           /* Window number text rows      */
    unsigned short   numTxtRows;        /* # of rows of text in window  */
    unsigned short   numTxtCols;        /* Video number of columns      */
    unsigned short   w_sccol;           /* the cur line's screen column */
    unsigned short   w_scscroll;        /* cur horizontal scroll column */
    unsigned short   w_sscroll;         /* the horizontal scroll column */
    unsigned short   w_margin;          /* The margin for the window    */
    unsigned short   w_scrsiz;          /* The screen size              */
    unsigned short   w_sbpos[WCVSBML+1]; /* Scroll bar positions         */
    unsigned short   w_mode;            /* Operating mode of window     */
} WINDOW ;

typedef struct {
    long    stmtime;			/* modification time of file	*/
#ifdef _UNIX
    off_t   stsize ;                    /* File's Size                  */
    uid_t   stuid ;                     /* File's User id               */
    gid_t   stgid ;                     /* File's Group id              */
    dev_t   stdev ;                     /* Files device ID no.          */
    ino_t   stino ;                     /* Files Inode number           */
#else
    long    stsize ;                    /* File's Size                  */
#endif
    unsigned short stmode;              /* file mode flags              */
} meSTAT ;

/*
 * meABREV
 * 
 * structure to store info on an abbreviation file
 */
typedef struct meABREV {
    struct meABREV *next ;              /* Pointer to the next abrev    */
    char   *fname ;                     /* Users abrev file name        */
    LINE    hlp ;                       /* head line                    */
    unsigned char loaded ;		/* modification time of file	*/
} meABREV ;

/*	structure to hold user variables and their definitions	*/
typedef struct meVARIABLE
{
    struct meVARIABLE *next ;   /* Next pointer, MUST BE FIRST as with meVARLIST */
    char *value ;		/* value (string) */
    char  name[1] ;             /* name of user variable */
} meVARIABLE;

typedef struct meVARLIST
{
    struct meVARIABLE *head ;
    int count ;
} meVARLIST ;

/*	structure for the name binding table		*/

typedef struct meCMD {
    struct meCMD *anext ;		/* alphabetically next command */
    struct meCMD *hnext ;		/* next command in hash table */
    meVARLIST     varList ;             /* command variables list */
    char         *name;		        /* name of function key */
    int           id ;                  /* command id number */
    Fintii        func;	 	        /* function name is bound to */
} meCMD ;

typedef struct meMACRO {
    meCMD        *anext ;	       	/* alphabetically next command */
    meCMD        *hnext ;      		/* next command in hash table */
    meVARLIST     varList ;             /* command variables list */
    char         *name ;		/* name of function key */
    int           id ;                  /* command id number */
    char         *fname ;		/* file name for file-macros */
    LINE         *hlp ;		 	/* Head line of macro */
    unsigned long callback ;		/* callback time for macro */
} meMACRO ;

#define	MACHIDE  0x01			/* Hide the function		*/
#define	MACEXEC	 0x02			/* Buffer is being executed	*/
#define	MACFILE  0x04			/* macro file define            */


/*	An alphabetic mark is as follows. Alphabetic marks are implemented
	as a linked list of amark structures, with the head of the list
	being pointed to by b_amark in the buffer structure.
 */

typedef	struct	meAMARK {
    struct meAMARK   *next ;		/* pointer to next mark in list		     */
    LINE           *line ;		/* pointer to line associated with this mark */
    unsigned short  offs ;		/* line offset  */
    unsigned short  name ;		/* mark name, (letter associated with it)    */
} meAMARK;

#if NARROW
typedef	struct	meNARROW {
    struct meNARROW *next ;		/* pointer to next narrow in list	     */
    struct meNARROW *prev ;		/* pointer to previous narrow in list	     */
    LINE            *slp ;		/* pointer to narrow start line              */
    LINE            *elp ;		/* pointer to narrow end line                */
    long            noLines ;           /* Number of lines narrowed out              */
    long            sln ;               /* Narrows start line number                 */
    unsigned short  name ;		/* amark name                                */
} meNARROW ;

#endif

/*
 * Text is kept in buffers. A buffer header, described below, exists for every
 * buffer in the system. The buffers are kept in a big list, so that commands
 * that search for a buffer by name can find the buffer header. There is a
 * safe store for the dot and mark in the header, but this is only valid if
 * the buffer is not being displayed (that is, if "b_nwnd" is 0). The text for
 * the buffer is kept in a circularly linked list of lines, with a pointer to
 * the header line in "b_linep".
 * 	Buffers may be "Inactive" which means the files accosiated with them
 * have not been read in yet. These get read in at "use buffer" time.
 */
typedef struct  BUFFER {
    struct  BUFFER *b_bufp ;            /* Link to next struct BUFFER	*/
#if ABREV
    meABREV *abrevFile ;                /* Abreviation file		*/
#endif
    meAMARK *b_amark ;  		/* pointer to the mark list	*/
#if LCLBIND
    struct KEYTAB *bbinds ;		/* pointer to local bindings	*/
#endif
#if NARROW
    meNARROW *narrow ;		        /* pointer to narrow structures */
#endif
    meVARLIST varList ;               /* User local buffer variables  */
    meSTAT  stats ;                     /* file stats - inc. mod time   */
    LINE   *b_dotp;			/* Link to "." LINE structure   */
    LINE   *b_markp;			/* The same as the above two,   */
    LINE   *b_linep;			/* Link to the header LINE      */
    char   *b_fname ;                   /* File name                    */
    char   *b_bname ;                   /* Buffer name                  */
    char   *modeLineStr ;               /* buffer mode-line format      */
#if FLNEXT
    char   *nextFile ;                  /* Next's current File name     */
#endif
#if	CRYPT
    unsigned char *b_key;		/* current encrypted key	*/
#endif
#if MEUNDO
    struct  UNDOND *fUndo ;             /* First undo node              */
    struct  UNDOND *lUndo ;             /* Last undo node               */
    unsigned long undoContFlag ;        /* Was the last undo this com'd?*/ 
#endif
    unsigned long autotime;		/* auto-save time for file	*/
    long    topLineNo ;			/* Windows top line number      */
    long    line_no;                    /* current line number          */
    long    mlineno;                    /* current mark line number     */
    long    elineno;                    /* end line number              */
    int     fhook;                      /* file hook function           */
    int     bhook;                      /* b buffer hook function       */
    int     dhook;                      /* d buffer hook function       */
    int     ehook;                      /* e buffer hook function       */
    int     inputFunc;                  /* input handle function        */
#ifdef _IPIPES
    int     ipipeFunc;                  /* ipipe input handle function  */
#endif
    int     histNo;                     /* Buff switch hist no.         */
    unsigned short   b_doto;            /* Offset of "." in above LINE  */
    unsigned short   b_marko;           /* but for the "mark"           */
#if LCLBIND
    unsigned short   nobbinds;          /* but for the "mark"           */
#endif
#if COLOR
    unsigned short   b_scheme;          /* Current colour index         */
    unsigned short   lScheme[LNNOSCHM]; /* line colour scheme index     */
#endif
    meMODE           b_mode;		/* editor mode of this buffer	*/
    unsigned char    intFlag;		/* internal buffer flags	*/
    unsigned char    b_nwnd;     	/* Count of windows on buffer   */
    unsigned char    isWordMask ;       /* isWord lookup table bit mask */
    unsigned char    modeLineFlags ;    /* buffer mode-line flags       */
#if HILIGHT
    unsigned char    indent ;           /* indent number                */
    unsigned char    nextLScheme ;      /* Next line colour index       */
    unsigned char    hiLight ;          /* hilight number               */
#endif
} BUFFER ;


#define	BIFBLOW    0x01		/* Buffer is to be blown away           */
#define	BIFLOAD    0x02		/* Used on a reload to check tim        */
#define	BIFLOCK    0x04		/* Used in ipipe to flag a lock         */
#define	BIFNAME    0x08		/* The buffer name has a <?> extension  */
#define	BIFFILE    0x10		/* The buffer is a file - used at creation only */
#define	BIFNODEL   0x20		/* The buffer cannot be deleted         */
#define	BIFSERVER  0x40		/* The buffer is a windows client server*/


/*
 * The starting position of a region, and the size of the region in
 * characters, is kept in a region structure.  Used by the region commands.
 */
typedef struct  {
    long	    size;               /* Length in characters.        */
    long	    line_no;            /* Length in characters.        */
    LINE	   *linep;		/* Origin LINE address.         */
    unsigned short  offset;             /* Origin LINE offset.          */
} REGION ;


/*	structure for the table of initial key bindings		*/

typedef struct KEYTAB {
    unsigned short code ;               /* Key code                     */
    unsigned short index ;
    unsigned int   arg ;
} KEYTAB;


/*	The editor holds deleted text chunks in the KILL buffer. The
	kill buffer is logically a stream of ascii characters, however
	due to its unpredicatable size, it gets implemented as a linked
	list of chunks. (The d_ prefix is for "deleted" text, as k_
	was taken up by the keycode structure			*/

typedef	struct	KILL {
    struct KILL   *next;		/* link to next chunk, NULL if last */
    unsigned char  data[1] ;            /* First byte of the data (nil trm) */
} KILL;

/*
 * KILL chunks are pointed at by a linked list, which could be very long indeed,
 * saving all the text deleted per session. In practice the number of deleted
 * chunks is kept to a predefined number. A new (ie most recent) delete is
 * inserted into the head of the chain.
 *
 * The structure after a couple of deletes is therefore:
 *
 * 	--------------			------------------
 *	| klhead     |  ----------->	| KILL structure |  ---------> ....
 *	|            |	kl_kill		|  		 |
 *	--------------			------------------
 *		|			most recent delete
 *		|kl_next
 *		V
 * 	----------------		------------------
 *	| klhead->next |  ----------->	| KILL structure |  ---------> ....
 *	|              |		|  		 |
 *	----------------		------------------
 *		|			second most recent delete
 *		|
 *		V
 * 	------------------		------------------
 *	| klhead->next   | ----------->	| KILL structure |  ---------> ....
 *	|         ->next |		|  		 |
 *	------------------		------------------
 *		|			first delete in session
 *		|
 *		V
 *	      N U L L
 *
 * The final thing to know about klist structures is that the kl_type field
 * can take on one of two possible values, if set to CFKILL, then it contains
 * killed text, if not (ie it is zero) it contains deleted text. The important
 * difference between killed and deleted text is that killed text goes in
 * large chunks, whereas deleted text happens in single character throws. It
 * is therefore possible in single character deletions, to look to see what
 * type of text is stored in a given kill buffer and so know whether to
 * create a new buffer for the soon-to-be-deleted text.
 */

#define	NKILL	15		/* number of kills held in kill list	   */

typedef	struct	KLIST {
	KILL         *kill ;	/* pointer to kill chunk		   */
	struct KLIST *next ;	/* link to next list element, NULL if last */
} KLIST;


/**	list of recognized user functions	*/

#define MONAMIC 	1
#define DYNAMIC 	2
#define TRINAMIC	3

#if     HILIGHT

/* hilight init flags */
#define HFCASE     0x01
#define HFLOOKB    0x02

/* hilight init internal flags */
#define HFRPLCDIFF 0x8000

/* hilight token flags */
#define HLTOKEN    0x001
#define HLENDLINE  0x002
#define HLBRACKET  0x004
#define HLCONTIN   0x008
#define HLBRINEOL  0x010
#define HLSTTLINE  0x020
#define HLREPLACE  0x040
#define HLBRANCH   0x080
#define HLSOL      0x100
#define HLVALID    0x200
#define HLCOLUMN   0x400
#define HLONELNBT  0x800

/* hilight token internal flags */
#define HLRPLCDIFF 0x8000

/* indent init flags */
#define HICASE   0x01

typedef struct HILNODE {
    struct HILNODE **list ;
    unsigned char   *table ; 
    unsigned char   *close ;
    unsigned char   *rtoken ;
    unsigned char   *rclose ;
    unsigned short   type ; 
    unsigned short   color;
    unsigned char    tknSttOff ; 
    unsigned char    tknEndOff ; 
    unsigned char    clsSttOff ; 
    unsigned char    clsEndOff ; 
    unsigned char    tknSttTst ; 
    unsigned char    tknEndTst ; 
    unsigned char    ordSize ; 
    unsigned char    listSize ; 
    unsigned char    ignore ;
    unsigned char    token[1] ;
} HILNODE, *HILNODEPTR ;

/*
 * HILCOLOR
 * Hilighting screen structure. The structure contains blocks of colour
 * information used to render the screen. Each block describes the front
 * and back colour and the column on which the colour ends.
 */
typedef struct HILCOLOR {
    unsigned short ccol;                /* Color change column */
    unsigned short icol;                /* Index colour */
} HILCOLOR;    

#define TableLower (unsigned char)(' ')
#define TableUpper (unsigned char)('z')
#define TableSize  ((TableUpper - TableLower) + 3)
#define InTable(cc) (((unsigned char)(cc) >= TableLower) && \
                     ((unsigned char)(cc) <= TableUpper))
#define GetTableSlot(cc)                             \
(((unsigned char)(cc) < TableLower) ? 0:             \
 (((unsigned char)(cc) > TableUpper) ? TableSize-2:   \
  ((unsigned char)(cc)-TableLower+1)))
#define hilListSize(r) (r->ordSize + r->listSize)

/*
 * Selection Highlighting 
 * The selection highlighting buffer retains information about the 
 * selection highlighting for the current buffer. The current restriction
 * is that only one buffer may have selection hilighting enabled
 */

typedef struct SELHILIGHT {
    short  uFlags;                      /* Hilighting user flags     */
    short  flags;                       /* Hilighting flags          */
    struct BUFFER *bp;                  /* Selected hilight buffer   */
    long   dlineno;                     /* Dot line number           */
    long   mlineno;                     /* Mark line number          */
    long   dlineoff;                    /* Current line offset       */
    long   mlineoff;                    /* Current mark offset       */
    int    sline;                       /* Start line                */
    int    soff;                        /* Start offset              */
    int    eline;                       /* End line number           */
    int    eoff;                        /* End offset                */
} SELHILIGHT;

#define SELHIL_ACTIVE    0x0001         /* Buffer has been edited    */
#define SELHIL_FIXED     0x0002         /* Buffer has been edited    */
#define SELHIL_KEEP      0x0004         /* Buffer has been edited    */
#define SELHIL_CHANGED   0x0008         /* Hilighting been changed   */
#define SELHIL_SAME      0x0010         /* Highlighting is same point*/

#define SELHILU_ACTIVE   0x0001         /* Buffer has been edited    */
#define SELHILU_KEEP     0x0002         /* Buffer has been edited    */

#endif

#define HCTFORE          0              /* Foreground colour */
#define HCTBACK          1              /* Background colour */
#define HCTCURFORE       2              /* Current foreground colour */        
#define HCTCURBACK       3              /* Current background colour */        
#define HCTSELFORE       4              /* Selected foreground colour */        
#define HCTSELBACK       5              /* Selected background colour */        
#define HCTCURSELFORE    6              /* Current selected foreground color */
#define HCTCURSELBACK    7              /* Current selected background color */
#define HCTNUM           8              /* Number of colour pairs in the system */

#define HCTMSKFORE       0x01           /* Foreground colour */
#define HCTMSKCUR        0x02           /* Current colour */
#define HCTMSKSEL        0x04           /* Selection colour */

#define IS_HCTFORE(x) (((x)&1)==0)
#define IS_HCTBACK(x) ((x)&1)
#define IS_HCTCUR(x)  ((x)&2)
#define IS_HCTSEL(x)  ((x)&4)

#define VFMESSL 0x0001                  /* Message line */
#define VFMENUL 0x0002                  /* Menu line changed flag */
#define VFMODEL 0x0004                  /* Mode line */
#define VFMAINL 0x0008                  /* Main line */
#define VFCURRL 0x0010                  /* Current line */
#define VFCHNGD 0x0020                  /* Changed flag */
#define VFSHBEG 0x0100                  /* Start of the selection hilight */
#define VFSHEND 0x0200                  /* End of the selection hilight */
#define VFSHALL 0x0400                  /* Whole line is selection hilighted */

#define VFSHMSK 0x0700                  /* Mask of the flags  */
#define VFTPMSK 0x003f                  /* Mask of the line type */

typedef struct  VIDEO
{
    WINDOW        *wind ;
    LINE          *line ;
    unsigned short endp ;
    unsigned short flag ;               /* Flags */
    unsigned char  bcol ;               /* the EOL bg color */
#if     HILIGHT
    unsigned char hilno ;               /* hilight-no */
    HILNODEPTR    bracket ;
#endif
} VIDEO;


/* Virtual Video Structure.
 * Provides the VIDEO structure management for horizontal split
 * Windows. A single VVIDEO exists for each VIDEO structure.
 */

#define VVROOT   0x0001                 /* This is the root video page */
#define VVFSEPUP 0x0002                 /* Vertical sepeator is out of date */

typedef struct VVIDEO
{
    VIDEO *video;                       /* Pointer to the video block */
    WINDOW *window;                     /* Windows attached to video block */
    struct VVIDEO *next;                /* Next video item */
} VVIDEO;

/* Frame Store
 * The frame store holds the physical screen structure. This comprises the
 * text information and the colour information of each character on the
 * screen.
 *
 * The frame store is required primarily on windowing platforms such as
 * MS-Windows and X-Windows and allows the canvas to be regenerated 
 * without recomputing the contents of the window when the windowing
 * sub-system requires the window (or parts of the window) to be refreshed.
 * 
 * Hence rather than EMACS generating new lines every time we render from
 * the Frame buffer which contains a representation of the screen. (Color +
 * ASCII text information).
 * 
 * Under MS-Windows we receive a WM_PAINT message. This message may be received
 * when we are in focus or not (e.g. a window placed over ours and moved
 * will cause a WM_PAINT messaage for the area of the screen which has
 * become invalid).
 */

typedef struct
{
    unsigned char  *text ;              /* Text held on the line. */
    unsigned short *colour ;            /* Color (fore + back) of each cell */
} FRAMELINE;                            /* Line of screen text */

/*
 * Define the colour composition macros for the frame store. Store the
 * colours in a single 16 bit word, upper byte is the foreground colour
 * lower byte is the background colour.
 */
#define FScolourPair(f,b) ((((unsigned short)(f))<<8)|(unsigned short)(b))
#define FScolourFore(x)   ((x)>>8)      /* Get the foreground colour */
#define FScolourBack(x)   ((x)&0xff)    /* Get the background colour */

#ifdef _IPIPES
/* The following is structure required for unix ipipes */

typedef struct IPIPEBUF {
    BUFFER    *bp ;
    struct IPIPEBUF *next ;
    int        pid ;
#ifdef _WIN32
    HANDLE     rfd ;
    HANDLE     outWfd ;
    HANDLE     hProcess ;
    DWORD      dwProcessId ;
#else
    int        rfd ;
    int        outWfd ;
#endif
    short      noRows ;
    short      noCols ;
    short      strRow ;
    short      strCol ;
    short      curRow ;
    short      flag ;
} IPIPEBUF ;

#endif

#if MEUNDO

typedef int UNDOCOORD[2] ;

typedef struct UNDOND {
    struct UNDOND *next ;
    union {
        long       dotp ;
        UNDOCOORD *pos ;
    } udata ;
    long           count ;
    unsigned short doto ;
    unsigned char  type ;
    char           str[1] ;
} UNDOND ;

#define MEUNDO_MDEL 0x01
#define MEUNDO_MINS 0x02
#define MEUNDO_REVS 0x04
#define MEUNDO_SING 0x08
#define MEUNDO_CONT 0x10
#define MEUNDO_FRST 0x20
#define MEUNDO_REPL 0x40
#define MEUNDO_NRRW 0x80

#endif

/* The variable register (#0 - #9) uses a linked structure
 * This is so they can easily be pushed and poped
 */
#define meNUMREG 10
typedef struct meREGISTERS {
    struct meREGISTERS *prev ;
    meVARLIST *varList ;
    char  *execstr ;
    int    f, n ;
    char   reg[meNUMREG][MAXBUF] ;
} meREGISTERS ;


/* Note that the first part of meDIRLIST structure must remain
 * the same as meNAMESVAR as it is used for $file-names variable
 */
typedef struct {
    int    exact ;
    int    size ;
    char **list ;
    char  *mask ;
    int    curr ;
    char  *path ;
    int    timeStamp ;
}  meDIRLIST ;

typedef struct {
    int    exact ;
    int    size ;
    char **list ;
    char  *mask ;
    int    curr ;
} meNAMESVAR ; 

#if REGSTRY
/*
 * REGISTRY
 * ========
 * Definitions for the registry API. The structures remain private
 * within the registry source file, providing a API interface for
 * maintainability.
 */

/*
 * REG_HANDLE
 * Generates a registry ASCII handle which is overlayed into a
 * 32-bit integer. The handle is the lookup from the macro 
 * interface.
 */
#define REG_HANDLE(a,b,c,d) ((((unsigned long)(a))<<24)| \
                             (((unsigned long)(b))<<16)| \
                             (((unsigned long)(c))<<8)| \
                             ((unsigned long)(d)))

/* Registry Open types */
#define REGMODE_HIDDEN  0x001           /* Node is hidden */
#define REGMODE_FROOT   0x002           /* File root */
#define REGMODE_CHANGE  0x004           /* Tree has changed */
#define REGMODE_BACKUP  0x008           /* Perform a backup of the file */
#define REGMODE_AUTO    0x010           /* Automatic save */
#define REGMODE_DISCARD 0x020           /* Discardable entry (memory only) */
#define REGMODE_MERGE   0x040           /* Merge a loaded registry */
#define REGMODE_RELOAD  0x080           /* Reload existing registry */
#define REGMODE_CREATE  0x100           /* Create if does not exist */
#define REGMODE_QUERY   0x200           /* Query the current node */
#define REGMODE_STORE_MASK 0x03f        /* Bits actually worth storing */
/*
 * RNODE
 * Data structure to hold a hierarchy node
 */
typedef struct RNODE
{
    char *value;                        /* The value of the node */
    struct RNODE *prnt;                 /* Pointer to the parent */
    struct RNODE *chld;                 /* Pointer to the child node */
    struct RNODE *sblg;                 /* Pointer to the sibling node */
    unsigned char mode;                 /* Mode flag */
    char name[1];                       /* The name of the node */
} RNODE;
/*
 * REGHANDLE: Public anonymous typeless handle to the registry node.
 */
typedef struct RNODE *REGHANDLE;
#endif


