/*****************************************************************************
*
*	Title:		%M%
*
*	Synopsis:	Expresion evaluation functions for MicroEMACS
*
******************************************************************************
*
*	Filename:		%P%
*
*	Author:			Danial Lawrence
*
*	Creation Date:		14/05/86 12:37		<990707.0753>
*
*	Modification date:	%G% : %U%
*
*	Current rev:		%I%
*
*	Special Comments:	
*
*	Contents Description:	
*
*	Ammendments :	Jon Green of Akebia Ltd.
*
*			Token evaluation not performed correctly. Trying
*			to read from the keyboard, line not terminated by
*			a carridge return. Invoked from 'getval()' on
*			TKARG condition.
*
* Jon Green	17-05-91
* Added suffix modes so that the suffix defintions may be defined in the 
* start up file.
* 
*****************************************************************************
* 
* Modifications to the original file by Jasspa. 
* 
* Copyright (C) 1988 - 1999, JASSPA 
* The MicroEmacs Jasspa distribution can be copied and distributed freely for
* any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
* incorporated into commercial software with the expressed permission of
* JASSPA.
* 
****************************************************************************/

/*---	Include defintions */

#define	__EVALC	1		/* Define program name */

/*---	Include files */

#include "emain.h"
#include "evar.h"
#include "efunc.h"
#include "eskeys.h"

#if (defined _UNIX) || (defined _DOS) || (defined _WIN32)
#include <sys/types.h>
#include <time.h>
#ifdef _UNIX
#include <sys/stat.h>
#endif
#ifndef _WIN32
#include <sys/time.h>
#endif
#endif

char evalResult[MAXBUF];	/* resulting string */
static char machineName[]=meSYSTEM_NAME;	/* resulting string */

/*---	Local macro definitions */

/*---	External references */

#if	TIMSTMP
extern	char	time_stamp[];	/* Time stamp string */
#endif

/*---	Local type definitions */

/*---	Local variable definitions */
static char *storeResult = NULL ;
static time_t timeOffset=0 ;            /* Time offset in seconds */


#ifdef _INSENSE_CASE
meNAMESVAR buffNames={0,0} ;
meDIRLIST  fileNames={0,0} ;
#else
meNAMESVAR buffNames={1,0} ;
meDIRLIST  fileNames={1,0} ;
#endif
meNAMESVAR commNames={1,0} ;
meNAMESVAR modeNames={0,MDNUMMODES,modeName,0} ;
meNAMESVAR varbNames={1,0} ;

/* Cheesy regular expression comparer
 * 
 * reg is the regular expersion which is matched against str
 * reg's special chars are * ? [
 * in a [, a ^ can follow meaning not, and ranges are allowed
 * the exact flag indicates a case sensitive comparision
 */
int
regexpStrCmp(char *str, char *reg, int exact)
{
    char *starStr, *starReg ;
    int   starState=0 ;
    char  rc, sc, nrc ;
    
    for(;;)
    {
        rc = *reg++ ;
        if(rc == '*')
        {
            /* A wild card, if its the last char in the match then we have
             * a match, else flag we got one and where (incase we 
             * have to go back) and move on
             */
            if(*reg == '\0')
                return 1 ;
            starState = 1 ;
            starReg = reg ;
        }
        else
        {
            if(!exact)
                rc = toLower(rc) ;
            for(;;)
            {
                sc = *str++ ;
                /* is it the end of str */
                if(sc == '\0')
                {
                    /* if its the end of reg break as if a char match, otherwise
                     * we have failed
                     */
                    if(rc != '\0')
                        return 0 ;
                    break ;
                }
                if(!exact)
                    sc = toLower(sc) ;
                /* if rc is a wild char then as long as sc isn't '\0'
                 * then its a match
                 */
                if(rc == '?')
                    break ;
                /* Test for a bracketed range */
                if(rc == '[')
                {
                    int not ;
                    /* do we inverse the result ? */
                    if(*reg == '^')
                    {
                        not = 1 ;
                        reg++ ;
                    }
                    else
                        not = 0 ;
                    while((rc=*reg++) != ']')
                    {
                        if(rc == '\0')
                            /* failed to find closing ']' - abort */
                            return -1 ;
                        nrc = *reg ;
                        if(nrc == '-')
                        {
                            /* range, note that if we're done a case insensitive compare
                             * we must test the two cases of sc and not change any of the
                             * range charaters
                             */
                            reg++ ;
                            nrc = *reg++ ;
                            if((sc >= rc) && (sc <= nrc))
                                break ;
                            if(!exact)
                            {
                                char tc=toggleCase(sc) ;
                                if((tc >= rc) && (tc <= nrc))
                                    break ;
                            }
                        }
                        else
                        {
                            /* single char compare */
                            if(!exact)
                                rc = toLower(rc) ;
                            if(sc == rc)
                                break ;
                        }
                    }
                    if(rc != ']')
                    {
                        /* found match, move to the ']' and continue */
                        while((rc=*reg++) != ']')
                            if(rc == '\0')
                                return -1 ;
                        /* X-or to allow for the not */
                        not ^= 1 ;
                    }
                    /* Did we find a match ? */
                    if(not)
                        break ;
                }
                /* If the letters are the same, match. */
                else if(sc == rc)
                    break ;
                /* Character match has failed - if there's been no wild
                 * card we've failed
                 */
                if(starState == 0)
                    return 0 ;
                if(starState < 0)
                {
                    /* There has been, but we've matched chars after, so
                     * move back to the char after the first match and try again
                     * This is to cope with matching *.cpp with test.c.cpp
                     */
                    starState = 1 ;
                    str = starStr ;
                    reg = starReg ;
                    rc = *reg++ ;
                    if(!exact)
                        rc = toLower(rc) ;
                }
            }
            /* we matched the char, if it was the null terminator then we've
             * got a match
             */
            if(sc == '\0')
                return 1 ;
            /* if the previous char was a'*' then store this location incase
             * we fail later on and we can come back to here and try again
             */
            if(starState > 0)
            {
                starState = -1 ;
                starStr = str ;
            }
        }
    }
}

/*	itoa:	integer to ascii string.......... This is too
    inconsistant to use the system's	*/
char *
meItoa(int i)
/* i : integer to translate to a string */
{
    register char *sp;		/* pointer into result */
    register int sign;		/* sign of resulting number */
    
    if(i == 0)			/* eliminate the trivial 0  */
        return falsem ;
    else if(i == 1)		/* and trivial 1            */
        return truem ;
    
    if((sign = (i < 0)))	/* record the sign...*/
        i = -i;
    
    /* and build the string (backwards!) */
    sp = &(evalResult[INTWIDTH]);
    *sp = 0;
    do
        *(--sp) = '0' + (i % 10);	/* Install the new digit */
    while (i /= 10);			/* Remove digit and test for 0 */
    
    if (sign != 0)			/* and fix the sign */
        *(--sp) = '-';			/* and install the minus sign */
    
    return sp ;
}


static char *
gtenv(char *vname)   /* vname	name of environment variable to retrieve */
{
    int ii ;
    register char  *ret ;
    register meNAMESVAR *mv ;
    
    /* scan the list, looking for the referenced name */
#ifdef _MACRO_COMP
    if((ii = (unsigned char) *vname) >= 128)
        ii -= 128 ;
    else
#endif
        ii = biChopFindString(vname,14,envars,NEVARS) ;
    switch(ii)
    {
        /* Fetch the appropriate value */
    case EVAUTOTIME:	return (meItoa(autotime));
    case EVKEPTVERS:	return (meItoa(keptVersions));
    case EVBOXCHRS:     return (char *) boxChars;
    case EVDELAYTIME:   return (meItoa(delaytime));
    case EVIDLETIME:    return (meItoa(idletime));
    case EVREPEATTIME:  return (meItoa(repeattime));
    case EVMODELINE:	return modeLineStr ;
    case EVMACHINE:     return machineName ;
    case EVPROGNM:      return progName ;
    case EVCURSORX:	return (meItoa(mwCol));
    case EVCURSORY:	return (meItoa(mwRow));
#if MOUSE
    case EVMOUSE:       return (meItoa(meMouseCfg));
    case EVMOUSEPOS:    return (meItoa(mouse_pos));
    case EVMOUSEX:	return (meItoa(mouse_X));
    case EVMOUSEY:	return (meItoa(mouse_Y));
#endif
    case EVSYSTEM:      return (meItoa(meSystemCfg));
    case EVFILLBULLET:  return fillbullet;
    case EVFILLBULLETLEN:return (meItoa(fillbulletlen));
    case EVFILLCOL:	return (meItoa(fillcol));
    case EVFILLEOS:     return filleos;
    case EVFILLEOSLEN:  return (meItoa(filleoslen));
    case EVFILLIGNORE:  return fillignore;
    case EVFILLMODE:
        evalResult[0] = fillmode ;
        evalResult[1] = '\0' ;
        return evalResult ;
    case EVBNAMES:
        mv = &buffNames ;
        goto handle_namesvar ;
    case EVCNAMES:
        mv = &commNames ;
        goto handle_namesvar ;
    case EVFNAMES:
        mv = (meNAMESVAR *) &fileNames ;
        goto handle_namesvar ;
    case EVMNAMES:
        mv = &modeNames ;
        goto handle_namesvar ;
    case EVVNAMES:
        mv = &varbNames ;
handle_namesvar:
        if(mv->mask == NULL)
            return abortm ;
        while(mv->curr < mv->size)
        {
            char *ss = mv->list[(mv->curr)++] ;
            if(regexpStrCmp(ss,mv->mask,mv->exact))
                return ss ;
        }
        return emptym ;
    case EVTEMPNAME:
        mkTempName (evalResult, NULL,NULL);
        return evalResult ;
    case EVFINDWORDS:   return findWordsNext() ;
    case EVRECENTKEYS:
        {
            int ii, jj, kk ;
            unsigned short cc ;
            for(ii=100,jj=TTnextKeyIdx,kk=0 ; --ii>0 ; )
            {
                if(((cc=TTkeyBuf[jj++]) == 0) ||
                   ((kk+=cmdchar(cc,evalResult+kk)) > MAXBUF-20))
                    break ;
                evalResult[kk++] = ' ' ;
                if(jj == KEYBUFSIZ)
                    jj = 0 ;
            }
            evalResult[kk] = '\0' ;
            return evalResult;
        }
    case EVRESULT:      return resultStr;
    case EVTIME:
        {
            struct meTimeval tp ;           /* Time interval for current time */
            struct tm *time_ptr;            /* Pointer to time frame. */
            time_t clock;                   /* Time in machine format. */
            
            /* Format the result. Format is:-
             * 
             *    YYYYCCCMMDDWhhmmssSSS
             * 
             * Where:
             *     YYYY - Year i.e. 1997
             *     CCC  - Day of the week since January 1st (0-365).
             *     MM   - Month of year 1-12
             *     DD   - Day of the month 1-31
             *     W    - Weekday since sunday 0-6 (sunday == 0)
             *     hh   - Hours 0-23
             *     mm   - Minutes 0-59
             *     ss   - Seconds 0-59
             *     SSS  - Milliseconds 0-999
             */
            gettimeofday(&tp,NULL) ;
            clock = tp.tv_sec + timeOffset ;              /* Get system time */
            time_ptr = (struct tm *) localtime (&clock);  /* Get time frame */
            /* SWP - Took out the 0 padding because of macro use.
             *       ME interprets numbers of the form 0## as octagon based numbers
             *       so in september the number is 09 which is illegal etc.
             */
            sprintf (evalResult, "%4d%3d%2d%2d%1d%2d%2d%2d%3d",
                     time_ptr->tm_year+1900,/* Year - 2000 should be ok !! */
                     time_ptr->tm_yday,     /* Year day */
                     time_ptr->tm_mon+1,    /* Month */
                     time_ptr->tm_mday,     /* Day of the month */
                     time_ptr->tm_wday,     /* Day of the week */
                     time_ptr->tm_hour,     /* Hours */
                     time_ptr->tm_min,      /* Minutes */
                     time_ptr->tm_sec,      /* Seconds */
                     (int) (tp.tv_usec/1000));/* Milliseconds */
            return evalResult;
        }
    case EVMATCHLEN:    return (meItoa(matchlen));
    case EVPAGELEN:	return (meItoa(TTnrow + 1));
    case EVRANDOM:	return (meItoa(rand()));
    case EVCURCOL:	return (meItoa(curwp->w_doto));
    case EVABSLINE:     return (meItoa(gotoAbsLine(-1)+1));
    case EVCURLINE:     return (meItoa(curwp->line_no+1));
    case EVWINCOL:	return (meItoa(getcwcol()));
    case EVWINCHRS:     return (char *) windowChars;
    case EVWINLINE:     return (meItoa(curwp->line_no-curwp->topLineNo));
    case EVBMDLINE:
        if(curbp->modeLineStr == NULL)
            return modeLineStr ;
        return curbp->modeLineStr ;
    case EVWMDLINE:     return (meItoa(curwp->firstRow+curwp->numTxtRows));
    case EVWSBAR:       return (meItoa(curwp->firstCol+curwp->numTxtCols));
    case EVWDEPTH:      return (meItoa(curwp->numTxtRows));
    case EVWWIDTH:      return (meItoa(curwp->numTxtCols));
    case EVCURWIDTH:    return (meItoa(TTncol));
    case EVCBUFNAME:    return (curbp->b_bname);
#ifdef _UNIX
    case EVBUFFMOD:
        sprintf(evalResult,"0%o",(int) curbp->stats.stmode);
        return evalResult ;
    case EVGLOBFMOD:
        sprintf(evalResult,"0%o",(int) meUmask);
        return evalResult ;
#else
    case EVBUFFMOD:	return (meItoa(curbp->stats.stmode));
    case EVGLOBFMOD:	return (meItoa(meUmask));
#endif
    case EVCFNAME:	return ((curbp->b_fname == NULL) ? "":curbp->b_fname);
    case EVDEBUG:	return (meItoa(macbug));
    case EVSTATUS:	return (meLtoa(cmdstatus));
#if	TIMSTMP
    case EVTIMSTMP:	return time_stamp ;
#endif
    case EVTABSIZE: 	return (meItoa(tabsize));
    case EVTABWIDTH:	return (meItoa(tabwidth));
    case EVSRCHPATH:    return searchPath ;
    case EVHOMEDIR:     return homedir ;
#if	CFENCE
    case EVSTATEMENTINDENT:
        return meItoa(statementIndent) ;
    case EVCONTINUEINDENT:
        return meItoa(continueIndent) ;
    case EVBRACEINDENT: return meItoa(braceIndent) ;
    case EVCASEINDENT:  return meItoa(caseIndent) ;
    case EVCONTCOMM:    return commentCont ;
    case EVECMARGIN:    return meItoa(commentMargin) ;
#endif
    case EVSHWMDS:
        {
            ret = evalResult;
            for (ii=0; ii < MDNUMMODES ; ii++) 
                *ret++ = (meModeTest(modeLineDraw,ii)) ? '1':'0' ;
            *ret = '\0' ;
            return evalResult;
        }
    case EVSHWRGN:      return meItoa(selhilight.uFlags);
    case EVCURSORBLK:   return meItoa(cursorBlink);
#if COLOR
    case EVCURSORCOL:   return meItoa(cursorColor);
    case EVOSDSCHM:     return meItoa(osdScheme/HCTNUM);
    case EVMLSCHM:      return meItoa(mlScheme/HCTNUM);
    case EVMDLNSCHM:    return meItoa(mdLnScheme/HCTNUM);
    case EVSBARSCHM:    return meItoa(sbarScheme/HCTNUM);
    case EVGLOBSCHM:    return meItoa(globScheme/HCTNUM);
    case EVBUFSCHM:     return meItoa(curbp->b_scheme/HCTNUM);
    case EVLINESCHM:    return meItoa((curwp->w_dotp->l_flag & LNSMASK) ? 
                                    (curbp->lScheme[curwp->w_dotp->l_flag & LNSMASK]/HCTNUM):-1) ;
    case EVSBAR:        return meItoa(gsbarmode);
#endif  /* COLOR  */
    case EVBUFIPIPE:
#if _IPIPES
        ii = curbp->ipipeFunc ;
#else
        ii = -1 ;
#endif
        goto hook_jump ;
    case EVBUFINP:
        ii = curbp->inputFunc ;
        goto hook_jump ;
    case EVBUFBHK:
        ii = curbp->bhook ;
        goto hook_jump ;
    case EVBUFDHK:
        ii = curbp->dhook ;
        goto hook_jump ;
    case EVBUFEHK:
        ii = curbp->ehook ;
        goto hook_jump ;
    case EVBUFFHK:    
        ii = curbp->fhook ;
hook_jump:
        if(ii < 0)
            evalResult[0] = '\0';
        else
            strcpy(evalResult,getCommandName(ii)) ;
        return evalResult ;

#if HILIGHT
    case EVBUFHIL:      return meItoa(curbp->hiLight);
    case EVBUFIND:      return meItoa(curbp->indent);
#endif
    case EVBUFMASK:
        {
            char *ss=evalResult ;
            int   ii ;
            for(ii=0 ; ii<8 ; ii++)
                if(isWordMask & (1<<ii))
                    *ss++ = charMaskFlags[ii] ;
            *ss = '\0' ;
            return evalResult ;
        }
    case EVFILEIGNORE:  return fileIgnore ;
#if FLNEXT
    case EVFILETEMP:    return flNextFileTemp ;
    case EVLINETEMP:    return flNextLineTemp ;
#endif
#if DORCS
    case EVRCSFILE:     return rcsFile ;
    case EVRCSCOCOM:    return rcsCoStr ;
    case EVRCSCOUCOM:   return rcsCoUStr ;
    case EVRCSCICOM:    return rcsCiStr ;
    case EVRCSCIFCOM:   return rcsCiStr ;
#endif
    case EVSCROLL:      return meItoa(scrollFlag) ;
    }
    if((ret = getenv(vname)) == NULL)
        ret = errorm ; /* errorm on bad reference */
    
    return ret ;
}

/* look up a user var's value */
/* vname - name of user variable to fetch */
char *
getUsrLclCmdVar(char *vname, register meVARLIST *varList)
{
    register meVARIABLE *vptr, *vp ;
    register int ii, jj ;
    
    vptr = varList->head ;
    ii = varList->count ;
    /* scan the list looking for the user var name */
    while(ii)
    {
        signed char *s1, *s2, ss ;
        
        jj = ii>>1 ;
        vp = vptr ;
        while(--jj >= 0)
            vp = vp->next ;
        
        s1 = (signed char *) vp->name ;
        s2 = (signed char *) vname ;
        for( ; ((ss=*s1++) == *s2) ; s2++)
            if(ss == 0)
                return(vp->value);
        if(ss > *s2)
            ii = ii>>1 ;
        else
        {
            ii -= (ii>>1) + 1 ;
            vptr = vp->next ;
        }
    }
    return errorm ;		/* return errorm on a bad reference */
}

/* The following horrid global variable solves a horrid problem, consider:
   define-macro l3
      set-variable @1 @2
   !emacro
   define-macro l2
      l3 @1 @2
   !emacro
   define-macro l1
      l2 @1 @2
   !emacro
   define-macro Test
      l1 #l0 @#
      ml-write #l0
   !emacro
   500 Test
 * The above example should write out "500".
 * its not good enough for @1 in l3 to return #l0, it must return Test's #l0
 * and thats horrid, hence the variable. Similarly it should be set to Test's
 * @#, not l3's which will be 1.
 */
static meREGISTERS *gmaLocalRegPtr=NULL ;

static char *
getMacroArg(int index)
{
    meREGISTERS *crp ;
    char *oldestr, *ss ;
    int oldF, oldN ;
    
    /* move the register pointer to the parent as any # reference
     * will be w.r.t the parent
     */
    crp = meRegCurr ;
    if((ss=crp->execstr) == NULL)
        return NULL ;
    oldestr = execstr ;
    oldF    = thisF ;
    oldN    = thisN ;
    execstr = ss ;
    thisF = crp->f ;
    thisN = crp->n ;
    meRegCurr = crp->prev ;
    if(alarmState & 0x02)
        gmaLocalRegPtr = meRegCurr ;
    for( ; index>0 ; index--)
    {
        execstr = token(execstr, evalResult) ;
        ss = getval(evalResult) ;
    }
    execstr   = oldestr ;
    thisF     = oldF ;
    thisN     = oldN ;
    meRegCurr = crp ;
    return ss ;
}


char *
getval(char *tkn)	/* find the value of a token */
{
    switch (getMacroTypeS(tkn)) 
    {
    case TKARG:
        if(tkn[1] == 'w')
        {
            if(alarmState & 0x02)
                return tkn ;
            if(tkn[2] == 'l')
            {
                register int blen ;
                char *ss, *dd ;
                /* Current buffer line fetch */
                blen = curwp->w_dotp->l_used;
                if (blen >= MAXBUF)
                    blen = MAXBUF-1 ;
                ss = (char *) curwp->w_dotp->l_text ;
                dd = evalResult ;
                while(--blen >= 0)
                    *dd++ = *ss++ ;
                *dd = '\0' ;
            }
            else
            {
                /* Current Buffer character fetch */
                evalResult[0] = getCurChar(curwp) ;
                evalResult[1] = '\0';
            }
        }
        else if(tkn[1] == '#')
        {
            if(alarmState & 0x02)
                return tkn ;
            return meItoa(thisN) ;
        }
        else if(tkn[1] == '?')
        {
            if(alarmState & 0x02)
                return tkn ;
            return meItoa(thisF) ;
        }
        else if(isDigit(tkn[1]))
        {
            extern char *curFuncName ;
            int index ;
            char *ss ;
            
            if((index = meAtoi(tkn+1)) == 0)
                ss = curFuncName ;
            else
                ss = getMacroArg(index) ;
            if(ss != NULL)
                return ss ;
            return abortm ;
        }
        else if(tkn[1] == 'y')
        {
            /* Don't use the X or windows clipboard in this case */
            register char *ss, *dd, cc ;
            register int   ii=MAXBUF-1 ;
            KILL          *killp;
            
#ifdef _CLIPBRD
            if(clipState & CLIP_TRY_GET)
                TTgetClipboard() ;
#endif
            if(klhead == (KLIST*) NULL)
                return abortm ;
            dd = evalResult ;
            killp = klhead->kill;
            while(ii && (killp != NULL))
            {
                ss = (char *) killp->data ;
                while(((cc=*ss++) != '\0') && (--ii))
                    *dd++ = cc ;
                killp = killp->next;
            }
            *dd = '\0' ;
        }
        else if(tkn[1] == 'c')
        {
            unsigned short key ;
            int kk, index ;
                        
            kk = (tkn[3] == 'k') ;
            if(tkn[2] == 'c')
            {
                if(alarmState & 0x02)
                    return tkn ;
                key = (unsigned short) thisCommand ;
                index = thisIndex ;
            }
            else if(tkn[2] == 'l')
            {
                if(alarmState & 0x02)
                    return tkn ;
                key = (unsigned short) lastCommand ;
                index = lastIndex ;
            }
            else
            {
                if(tkn[2] == 'q')
                {
                    int cc ;
                    key = tgetc() ;
                    /* intercative single char read which will be quoted */
                    if((cc=quoteKeyToChar(key)) > 0)
                    {
                        if(tkn[3] == 'k')
                        {
                            evalResult[0] = cc ;
                            evalResult[1] = '\0';
                            return evalResult ;
                        }
                        key = cc ;
                    }
                }
                else
                    key = getkeycmd(FALSE, 1, 0) ;
                if(!kk)
                {
                    unsigned int arg ;
                    index = decode_key(key,&arg) ;
                }
            }
            if(kk)
                cmdstr(key,evalResult) ;
            else if(index < 0)
            {
                if(key > 0xff)
                    return errorm ;
                evalResult[0] = (char) key ;
                evalResult[1] = '\0';
            }
            else
                strcpy(evalResult,getCommandName(index)) ;
        }
        else if(tkn[1] == 'm')
        {
            if(tkn[2] == 'l')
            {
                /* interactive argument */
                /* note: the result buffer cant be used directly as it is used
                 *       in modeline which is is by update which can be used
                 *       by mlgetstring
                 */
                static char **strList=NULL ;
                static int    strListSize=0 ;
                char cc, *ss, buff[MAXBUF], prompt[MAXBUF] ;
                int  option=0 ;
                int  flag, defH ;
                char divChr ;
                
                if((flag=tkn[3]) != '\0')
                {
                    flag -= '0' ;
                    if(((cc=tkn[4]) > '0') && (cc<='9'))
                        cc -= '0' ;
                    else if(cc == 'a')
                        cc = 10 ;
                    else
                        cc = 11 ;
                    if(cc == 1)
                        option = MLFILE ;
                    else
                    {
                        if(cc == 7)
                        {
                            mlgsStrList = modeName ;
                            mlgsStrListSize = MDNUMMODES ;
                        }
                        else if(cc == 9)
                        {
                            flag |= 0x100 ;
                            cc = 7 ;
                        }
                        else if(cc == 10)
                        {
                            flag |= 0x300 ;
                            cc = 7 ;
                        }
                        if(cc < 11)
                            option = 1 << (cc-2) ;
                    }
#if (MLFILECASE != MLFILE)
                    if(option & MLFILE)
                        option |= MLFILECASE ;
#if (MLBUFFERCASE != MLBUFFER)
                    else if(option & MLBUFFER)
                        option |= MLBUFFERCASE ;
#endif
#endif
                }
                else
                    cc = 0 ;
                
                /* Get and evaluate the next argument - this is the prompt */
                execstr = token(execstr, buff) ;
                if((ss = getval(buff)) == abortm)
                    return abortm ;
                strcpy(prompt,ss) ;
                if(flag & 0x01)
                {
                    /* Get and evaluate the next argument - this is the default
                     * value
                     */
                    execstr = token(execstr, buff) ;
                    if((ss = getval(buff)) == abortm)
                        return abortm ;
                    addHistory(option,ss) ;
                    defH = 1 ;
                }
                else
                    defH = 0 ;
                if(flag & 0x02)
                {
                    /* Get and evaluate the next argument - this is the ml
                     * line init value
                     */
                    execstr = token(execstr, buff) ;
                    if((ss = getval(buff)) == abortm)
                        return abortm ;
                    if(ss != buff)
                        strcpy(buff,ss) ;
                    option |= MLNORESET ;
                }
                if(flag & 0x04)
                    /* set the first key to 'tab' to expand the input according 
                     * to the completion list */
                    mlfirst = ME_SPECIAL|SKEY_tab ;
                
                if(flag & 0x100)
                {
                    /* Get and evaluate the next argument - this is the
                     * completion list
                     */
                    char comp[MAXBUF], *tt ;
                    execstr = token(execstr,comp) ;
                    if((ss = getval(comp)) == abortm)
                        return abortm ;
                    if(flag & 0x200)
                    {
                        BUFFER *bp ;
                        LINE *lp ;
                        flag &= ~0x100 ;
                        if((bp = bfind(ss,0)) == NULL)
                            return abortm ;
                        if(bp->elineno > strListSize)
                        {
                            strListSize = bp->elineno ;
                            if((strList = meRealloc(strList,strListSize*sizeof(char *))) == NULL)
                            {
                                strListSize = 0 ;
                                return abortm ;
                            }
                        }
                        lp = bp->b_linep ;
                        mlgsStrListSize = 0 ;
                        while((lp=lforw(lp)) != bp->b_linep)
                            strList[mlgsStrListSize++] = (char *) lp->l_text ;
                    }
                    else
                    {
                        /* if the resultant string is going to the same
                         * location as we are getting the competion list
                         * then we MUST not restore the list after, so remove
                         * the flag
                         */
                        if((ss >= evalResult) && (ss < evalResult+MAXBUF))
                            flag &= ~0x100 ;
                        divChr = *ss++ ;
                        mlgsStrListSize = 0 ;
                        while((tt = strchr(ss,divChr)) != NULL)
                        {
                            *tt++ = '\0' ;
                            if(mlgsStrListSize == strListSize)
                            {
                                if((strList = meRealloc(strList,(strListSize+8)*sizeof(char *))) == NULL)
                                {
                                    strListSize = 0 ;
                                    return abortm ;
                                }
                                strListSize += 8 ;
                            }
                            strList[mlgsStrListSize++] = ss ;
                            ss = tt ;
                        }
                    }
                    mlgsStrList = strList ;
                }
                if(cc == 1)
                {
                    if(!(option & MLNORESET))
                        getFilePath(curbp->b_fname,buff) ;
                    clexec = FALSE ;
                    if((cc = mlreply(prompt,(MLFILECASE|MLNORESET|MLMACNORT),0,buff,MAXBUF)) == TRUE)
                        fileNameCorrect(buff,evalResult,NULL) ;
                    clexec = TRUE ;
                }
                else
                {
                    /* Note that buffer result cannot be used directly as in the process
                     * of updating the screen the result buffer can be used by meItoa
                     * so use a temp buffer and copy across. note that inputFileName
                     * above doesn't suffer from this as the function uses a temp buffer
                     */
                    cc = mlgetstring(prompt,option,defH,buff,MAXBUF) ;
                    strncpy(evalResult,buff,MAXBUF) ;
                    evalResult[MAXBUF-1] = '\0' ;
                }
                if(flag & 0x100)
                {
                    /* glue the completion string back together */
                    while(--mlgsStrListSize >= 0)
                    {
                        char *s1 = mlgsStrList[mlgsStrListSize] ;
                        s1[strlen(s1)] = divChr ;
                    }
                }
                if(cc == ABORT)
                    return abortm ;
            }
            else if(tkn[2] == 'c')
            {
                char *ss, buff[MAXBUF], prompt[MAXBUF] ;
                int  ret ;
                
                if((ret=tkn[3]) != '\0')
                    ret -= '0' ;
                
                /* Get and evaluate the next argument - this is the prompt */
                execstr = token(execstr,buff) ;
                if((ss = getval(buff)) == abortm)
                    return abortm ;
                strcpy(prompt,ss) ;
                if(ret & 0x01)
                {
                    /* Get and evaluate the next argument - this is valid
                     * values list
                     */
                    execstr = token(execstr, buff) ;
                    if((ss = getval(buff)) == abortm)
                        return abortm ;
                    if((ss != buff) || (ss != buff+1))
                    {
                        strcpy(buff,ss) ;
                        ss = buff ;
                    }
                }
                else
                    ss = NULL ;
                
                clexec = FALSE ;
                ret = mlCharReply(prompt,0,ss,NULL) ;
                clexec = TRUE ;
                if(ret < 0)
                    return abortm ;
                evalResult[0] = ret ;
                evalResult[1] = 0 ;
            }
            else
                return abortm ;
                    
        }
        else
        {
            mlwrite(MWABORT|MWWAIT,"[Unknown argument %s]",tkn);
            return abortm ;
        }
        return evalResult ;
        
    case TKREG:
        if(alarmState & 0x02)
            return tkn ;
        {
            meREGISTERS   *rp ;
            unsigned char  cc ;
            
            if((cc=tkn[1]) == 'l')
                rp = meRegCurr ;
            else if(cc == 'p')
                rp = meRegCurr->prev ;
            else
                rp = meRegHead ;
            cc = tkn[2] - '0' ;
            if(cc < ('0'+meNUMREG))
                return rp->reg[cc] ;
            break ;
        }
    
    case TKVAR:
        if(alarmState & 0x02)
            return tkn ;
        return getUsrVar(tkn+1) ;
    
    case TKLVR:
        if(alarmState & 0x02)
            return tkn ;
        return getUsrLclCmdVar(tkn+1,&(curbp->varList)) ;
        
    case TKCVR:
        {
            char *ss, *tt, cc ;
            if(alarmState & 0x02)
                return tkn ;
            tt = tkn+1 ;
            for(ss=tt ; (cc=*++ss) != '\0' ; )
            {
                if(cc == '.')
                {
                    int ii ;
                    *ss++ = '\0' ;
                    if((ii=decode_fncname(tt,1)) < 0)
                        /* not a command - error */
                        return errorm ;
                    return getUsrLclCmdVar(ss,&(cmdTable[ii]->varList)) ;
                }
            }
            if(meRegCurr->varList == NULL)
                return abortm ;
            return getUsrLclCmdVar(tkn+1,meRegCurr->varList) ;
        }
    case TKENV:
        if(alarmState & 0x02)
            return tkn ;
        return gtenv(tkn+1) ;
    
    case TKFUN:
        return gtfun(tkn+1) ;
    
    case TKSTR:
        tkn++ ;
    case TKCMD:
    case TKLIT:
        return tkn ;
    }
    /* If here then it is either TKDIR, TKLBL or TKNUL. Doesn't matter which
     * cause something has gone wrong and we return abortm
     */
    return abortm ;
}

char *
queryMode(char *name, unsigned char *mode)
{
    int ii ;
    
    for (ii=0; ii < MDNUMMODES ; ii++) 
        if(strcmp(name,modeName[ii]) == 0) 
            return meLtoa(meModeTest(mode,ii)) ;
    return abortm ;
}

char *
gtfun(char *fname)	/* evaluate a function given name of function */
{
    register int fnum;	        /* index to function to eval */
    char arg1[MAXBUF];		/* value of first argument */
    char arg2[MAXBUF];		/* value of second argument */
    char arg3[MAXBUF];		/* value of third argument */
    char *store ;
    
    /* look the function up in the function table */
#ifdef _MACRO_COMP
    if((fnum = (unsigned char) *fname) >= 128)
        fnum -= 128 ;
    else
#endif
        if((fnum = biChopFindString(fname,3,funcNames,NFUNCS)) < 0)
    {
        mlwrite(MWABORT|MWWAIT,"[Unknown function &%s]",fname);
        return abortm ;
    }
    if(fnum == UFKBIND)
    {
        unsigned int   arg ;
        unsigned short ii ;
        int jj ;
        
        ii = getckey(0,1) ;
        if((jj = decode_key(ii,&arg)) < 0)
            return errorm ;
        strcpy(evalResult,getCommandName(jj)) ;
        return evalResult ;
    }
    store = execstr ;
    /* retrieve the arguments */
    {
        short alarmStateStr=alarmState ;
        
        alarmState &= ~0x02 ;
        
        if((macarg(arg1) != TRUE) ||
           ((funcTypes[fnum] >= DYNAMIC) &&
            ((macarg(arg2) != TRUE) ||
             ((funcTypes[fnum] >= TRINAMIC) &&
              (macarg(arg3) != TRUE)) ) ) )
        {
            mlwrite(MWABORT|MWWAIT,"[Failed to get argument for function &%s]",fname);
            return abortm ;
        }
        alarmState = alarmStateStr ;
    }
    
    /* and now evaluate it! */
    switch(fnum)
    {
    case UFABS:        return meItoa(abs(meAtoi(arg1)));
    case UFADD:        return meItoa(meAtoi(arg1) + meAtoi(arg2));
    case UFSUB:        return meItoa(meAtoi(arg1) - meAtoi(arg2));
    case UFMUL:        return meItoa(meAtoi(arg1) * meAtoi(arg2));
    case UFDIV:
    {
        int ii ;
        ii  = meAtoi(arg1) ;
        ii /= meAtoi(arg2) ;
        return meItoa(ii) ;
    }
    case UFMOD:        return meItoa(meAtoi(arg1) % meAtoi(arg2));
    case UFNEG:        return meItoa(0-meAtoi(arg1));
    case UFATOI:       return meItoa((unsigned char) arg1[0]) ;
    case UFITOA:
        evalResult[0] = (unsigned char) meAtoi(arg1) ;
        evalResult[1] = '\0' ;
        return evalResult ;
    case UFCAT:
        {
            char *dd, *ss ;
            int   ii = MAXBUF-1 ;
            
            /* first string can be copied, second we must check the left */
            dd = evalResult ;
            ss = arg1 ;
            while((*dd++ = *ss++))
                ii-- ;
            dd-- ;
            ss = arg2 ;
            do {
                if(--ii <= 0)
                {
                    *dd = '\0' ;
                    break ;
                }
            } while((*dd++ = *ss++)) ;
            return evalResult ;
        }
    case UFLEFT:
        {
            int ii=meAtoi(arg2) ;
            char *dd, *ss ;
            dd = evalResult ;
            ss = arg1 ;
            for( ; ii ; ii--)
                if(!(*dd++ = *ss++))
                    break ;
            *dd = '\0' ;
            return evalResult ;
        }
    case UFRIGHT:
        {
            char *dd, *ss ;
            dd = evalResult ;
            ss = arg1+meAtoi(arg2) ;
            while((*dd++ = *ss++))
                ;
            return evalResult ;
        }
    case UFMID:
        {
            int ii=meAtoi(arg3) ;
            char *dd, *ss ;
            dd = evalResult ;
            ss = arg1+meAtoi(arg2) ;
            for( ; ii ; ii--)
                if(!(*dd++ = *ss++))
                    break ;
            *dd = '\0' ;
            return evalResult ;
        }
    case UFLEN:
        {
            int   ii=0 ;
            char *ss ;
            ss = arg1 ;
            while((*ss++))
                ii++ ;
            return meItoa(ii) ;
        }
    case UFSLOWER:
        {
            char cc, *dd, *ss ;
            dd = evalResult ;
            ss = arg1 ;
            do {
                cc = *ss++ ;
                *dd++ = toLower(cc) ;
            } while(cc) ;
            return evalResult ;
        }
    case UFSUPPER:
        {
            char cc, *dd, *ss ;
            dd = evalResult ;
            ss = arg1 ;
            do {
                cc = *ss++ ;
                *dd++ = toUpper(cc) ;
            } while(cc) ;
            return evalResult ;
        }
    case UFSIN:
    case UFRSIN:
    case UFISIN:
    case UFRISIN:
        {
            Fintssi cmpIFunc ;
            char *ss=arg2, *lss=NULL ;
            int   len ;
            len = strlen(arg1) ;
            if((fnum == UFSIN) || (fnum == UFRSIN))
                cmpIFunc = strncmp ;
            else
                cmpIFunc = icstrncmp ;
            
            do
                if(!cmpIFunc(arg1,ss,len))
            {
                lss = ss ;
                if((fnum == UFSIN) || (fnum == UFISIN))
                    break ;
            }
            while(*ss++ != '\0') ;
            if(lss != NULL)
                return meItoa(lss-arg2+1);
            return(meLtoa(0)) ;
        }
    case UFNBMODE:
        {
            BUFFER *bp ;
            if((bp = bfind(arg1,0)) == NULL)
                return abortm ;
            return queryMode(arg2,bp->b_mode) ;
        }
    case UFBMODE:      return queryMode(arg1,curbp->b_mode) ;
    case UFGMODE:      return queryMode(arg1,globMode) ;
    case UFBAND:       return(meItoa(meAtoi(arg1) & meAtoi(arg2)));
    case UFBNOT:       return(meItoa(~meAtoi(arg1))) ;
    case UFBOR:        return(meItoa(meAtoi(arg1) |  meAtoi(arg2)));
    case UFBXOR:       return(meItoa(meAtoi(arg1) ^  meAtoi(arg2)));
    case UFNOT:        return(meLtoa(meAtol(arg1) == FALSE));
    case UFAND:        return(meLtoa(meAtol(arg1) && meAtol(arg2)));
    case UFOR:         return(meLtoa(meAtol(arg1) || meAtol(arg2)));
    case UFEQUAL:      return(meLtoa(meAtoi(arg1) == meAtoi(arg2)));
    case UFLESS:       return(meLtoa(meAtoi(arg1) <  meAtoi(arg2)));
    case UFGREATER:    return(meLtoa(meAtoi(arg1) >  meAtoi(arg2)));
    case UFSEQUAL:     return(meLtoa(strcmp(arg1,arg2) == 0));
    case UFSLESS:      return(meLtoa(strcmp(arg1,arg2) < 0));
    case UFSGREAT:     return(meLtoa(strcmp(arg1,arg2) > 0));
    case UFINWORD:     return(meLtoa(isWord(arg1[0])));
    case UFISEQUAL:    return(meLtoa(icstrcmp(arg1,arg2) == 0));
    case UFXSEQ:       return(meLtoa(regexpStrCmp(arg1,arg2,1) == 1));
    case UFXISEQ:      return(meLtoa(regexpStrCmp(arg1,arg2,0) == 1));
    case UFLDEL:
        {
            int  index=meAtoi(arg2) ;
            char cc, *s1, *s2 ;
            if(index > 0)
            {
                s2 = arg1 ;
                cc = *s2 ;
                do {
                    s1 = s2+1 ;
                    if((s2 = strchr(s1,cc)) == NULL)
                        break ;
                } while(--index > 0) ;
            }
            else
                s2 = NULL ;
            if(s2 == NULL)
                strcpy(evalResult,arg1) ;
            else
            {
                index = (int) (s1 - arg1) ;
                strncpy(evalResult,arg1,index) ;
                strcpy(evalResult+index,s2+1) ;
            }
            return evalResult ;
        }
    case UFLFIND:
        {
            int  index ;
            char cc, *s1, *s2 ;
            s2 = arg1 ;
            cc = *s2 ;
            for(index=1 ; ; index++)
            {
                s1 = s2+1 ;
                if((s2 = strchr(s1,cc)) == NULL)
                    return falsem ;
                *s2 = '\0' ;
                if(!strcmp(s1,arg2))
                    return meItoa(index) ;
            }
        }
    case UFLGET:
        {
            int  index ;
            char cc, *s1, *s2 ;
            if((index=meAtoi(arg2)) <= 0)
                return emptym ;
            s2 = arg1 ;
            cc = *s2 ;
            do {
                s1 = s2+1 ;
                if((s2 = strchr(s1,cc)) == NULL)
                    return emptym ;
            } while(--index > 0) ;
            index = (int) (s2 - s1) ;
            strncpy(evalResult,s1,index) ;
            evalResult[index] = '\0' ;
            return evalResult ;
        }
    case UFLINS:
    case UFLSET:
        {
            int  index=meAtoi(arg2), ii ;
            char cc, *s1, *s2 ;
            s2 = arg1 ;
            cc = *s2 ;
            if(index <= 0)
            {
                if(index < 0)
                    s2 += strlen(s2)-1 ;
                s1 = s2+1 ;
            }
            else
            {
                do {
                    s1 = s2+1 ;
                    if((s2 = strchr(s1,cc)) == NULL)
                    {
                        s2 = s1-1 ;
                        break ;
                    }
                } while(--index > 0) ;
            }
            index = (int) (s1 - arg1) ;
            strncpy(evalResult,arg1,index) ;
            ii = strlen(arg3) ;
            if(ii+index < MAXBUF)
            {
                strncpy(evalResult+index,arg3,ii) ;
                index += ii ;
                if(fnum == UFLINS)
                    s2 = s1-1 ;
                ii = strlen(s2) ;
                if(ii+index < MAXBUF)
                {
                    strcpy(evalResult+index,s2) ;
                    index += ii ;
                }
            }
            evalResult[index] = '\0' ;
            return evalResult ;
        }
    case UFFIND:
        {
            if(!fileLookup(arg1,arg2,meFL_CHECKDOT|meFL_USESRCHPATH,evalResult))
                return errorm ;
            return evalResult ;
        }
    case UFWHICH:
        {
            if(!executableLookup(arg1,evalResult))
                return errorm ;
            return evalResult ;
        }
    case UFCOND:
        {
            if(meAtol(arg1))
                strcpy(evalResult,arg2) ;
            else
                strcpy(evalResult,arg3) ;
            return evalResult ;
        }
    case UFTRIMB:
        {
            /* Trim both */
            char *ss;
            char cc;
            int ii;
            
            /* Trim left */
            for (ss = arg1; (((cc = *ss) != '\0') && isSpace(cc)) ; ss++)
                ;
            /* Trim right */
            ii = strlen (ss);
            while (ii > 0)
            {
                cc = ss[ii-1];
                if (isSpace(cc))
                    ii--;
                else
                    break;
            }
            strncpy (evalResult, ss, ii);
            evalResult[ii] = '\0';
            return evalResult;
        }
    case UFTRIML:
        {
            /* Trim left */
            char *ss;
            char cc;
            for (ss = arg1; (((cc = *ss) != '\0') && isSpace(cc)) ; ss++)
                ;
            strcpy (evalResult, ss);
            return evalResult;
        }
    case UFTRIMR:
        {
            /* Trim right */
            int ii;
            ii = strlen (arg1);
            while (ii > 0)
            {
                char cc;
                cc = arg1[ii-1];
                if (isSpace(cc))
                    ii--;
                else
                    break;
            }
            strncpy (evalResult, arg1, ii);
            evalResult[ii] = '\0';
            return evalResult;
        }
    case UFSET:
        {
            char *ss;
            ss = execstr;
            storeResult = arg2;
            execstr = store;
            setVariable (FALSE, 1);
            execstr = ss;
            storeResult = NULL;
            strcpy (evalResult, arg2);
            return evalResult;
        }
    case UFDEC:
        {
            char *ss, *rr ;
            rr = meItoa(meAtoi(arg1) - meAtoi(arg2));
            ss = execstr ;
            storeResult = rr ;
            execstr = store ;
            setVariable(FALSE,1) ;
            execstr = ss ;
            storeResult = NULL ;
            return rr ;
        }
    case UFINC:
        {
            char *ss, *rr ;
            rr = meItoa(meAtoi(arg1) + meAtoi(arg2));
            ss = execstr ;
            storeResult = rr ;
            execstr = store ;
            setVariable(FALSE,1) ;
            execstr = ss ;
            storeResult = NULL ;
            return rr ;
        }
    case UFPDEC:
        {
            char *ss, *rr ;
            rr = meItoa(meAtoi(arg1) - meAtoi(arg2));
            ss = execstr ;
            storeResult = rr ;
            execstr = store ;
            setVariable(FALSE,1) ;
            execstr = ss ;
            storeResult = NULL ;
            strcpy(evalResult,arg1) ;
            return evalResult ;
        }
    case UFPINC:
        {
            char *ss, *rr ;
            rr = meItoa(meAtoi(arg1) + meAtoi(arg2));
            ss = execstr ;
            storeResult = rr ;
            execstr = store ;
            setVariable(FALSE,1) ;
            execstr = ss ;
            storeResult = NULL ;
            strcpy(evalResult,arg1) ;
            return evalResult ;
        }
    case UFREGISTRY:
        {
            REGHANDLE reg;
            char *p = arg3;
            /*
             * Read a value from the registry.
             * 
             * Arg1 is the path,
             * Arg2 is the subkey
             * Arg3 is the default value
             * 
             * If the node cannot be found then return the result.
             */
            if (((reg = regFind (NULL, arg1)) != NULL) &&
                ((reg = regFind (reg,  arg2)) != NULL))
                p = regGetString (reg, p) ;
            /* Set up the result string */
            if (p == NULL)
                evalResult [0] = '\0';
            else
                strcpy (evalResult, p);
            return evalResult;
        }
    case UFSPRINT:                         /* Monamic function !! */
        {
            /*
             * Do a sprintf type of function. 
             * &sprintf "This is a %s %s." "arg1" "arg2"
             *
             * Note that we build the result in a buffer (arg3) on the stack
             * which allows us to effectivelly do recursive invocations 
             * of &sprintf.
             * 
             * Note that this is a monamic function, the first argument is the
             * format string. The rest of the arguments are dynamic and are
             * feteched as required.
             * 
             * Note - need to sort the '"' character. Not sure how macro process
             * it when specified in string.
             * 
             * Syntax defined as follows:-
             * 
             * sprintf "%s" arg
             * sprintf "%n" count arg
             * sprintf "%d" 10
             * 
             * e.g.
             *     sprintf "Foo [%s%s]" "a" "b"
             * generates:-
             *     "Foo [ab]"
             * 
             *     sprintf "Foo [%n%s]" 10 "a" "b"
             * generates:-
             *     "Foo [aaaaaaaaaab]"
             */
            
            char *p;                            /* Temp char pointer */
            char  c;                            /* Current char */
            char *s = arg1;                     /* String pointer */
            int   count, nn, index=0 ;          /* Count of repeat */
            
            while ((c = *s++) != '\0') 
            {
                if (c == '%')                   /* Check for control */
                {
                    p = s-1 ;
                    count = 1 ;
get_flag:
                    switch((c = *s++))
                    {
                    case 'n':                   /* Replicate string */
                        if (macarg (arg2) != TRUE) 
                            return abortm;
                        count *= meAtoi(arg2);
                    case 's':                   /* String */
                        if (macarg (arg2) != TRUE)
                            return abortm;
                        nn = strlen(arg2) ;
                        while(--count >= 0)
                        {
                            if(index+nn >= MAXBUF)
                                /* break if we dont have enough space */
                                break ;
                            strcpy(arg3+index,arg2) ;
                            index += nn ;
                        }
                        break;
                    case 'X':                   /* Hexadecimal */
                    case 'x':                   /* Hexadecimal */
                    case 'd':                   /* Decimal */
                    case 'o':                   /* Octal */
                        if (macarg (arg2) != TRUE) 
                            return abortm;
                        nn = meAtoi(arg2) ;
                        c = *s ;
                        *s = '\0' ;
                        if(index < MAXBUF-16)
                            /* Only do it if we have enough space */
                            index += sprintf(arg3+index,p,nn) ;
                        *s = c ;
                        break;
                    default:
                        if(isDigit(c))
                        {
                            count = c - '0' ;
                            while((c = *s++),isDigit(c)) 
                                count = (count*10) + (c-'0') ;
                            s-- ;
                            goto get_flag ;
                        }
                        else if(index >= MAXBUF)
                            break ;
                        else
                            arg3[index++] = c;  /* Just a literal char - pass in */
                    }
                }
                else if(index >= MAXBUF)
                    break ;
                else
                    arg3[index++] = c;          /* Just a literal char - pass in */
            }
            arg3[index] = '\0';                 /* make sure sting terminated */
            strcpy(evalResult,arg3) ;
            return evalResult ;
        }
    case UFSTAT:
        {
            static int typeRet[] = {4,1,2,0,4,3} ;
            meSTAT stats ;
            int    ftype ;
            /*
             * 0 -> 4 - If file is nasty
             * 1 -> 1 - If file is regular
             * 2 -> 2 - If file is a directory
             * 3 -> 0 - If file doesn't exist
             * 5 -> 3 - If file is a url
             */
            ftype = getFileStats(arg2,1,&stats,evalResult) ;
            switch(arg1[0])
            {
            case 'a':
                /* absolute name - check for symbolic link */
                if(*evalResult == '\0')
                    fileNameCorrect(arg2,evalResult,NULL) ;
                return evalResult ;
                
            case 'd':
                /* file date stamp */
                ftype = stats.stmtime ;
                break ;
                
            case 'r':
                /* File read permission */
                /* If its a nasty or it doesn't exist - no */
                if((ftype == 0) || (ftype == 3))
                    return meLtoa(0) ;
                /* If its a url then do we support url ? */
#if URLAWAR
                if(ftype == 5)
#ifdef _URLSUPP
                    return meLtoa(1) ;
#else
                    return meLtoa(0) ;
#endif
#endif
                /* If its a DOS/WIN directory - yes */
#if (defined _DOS) || (defined _WIN32)
                if(ftype == 2)
                    return meLtoa(1) ;
#endif
                /* normal test */
                return meLtoa(meStatTestRead(stats)) ;
                
            case 's':
                /* File size - return -1 if not a regular file */
                if(ftype == 1)
                    ftype = stats.stsize ;
                else
                    ftype = -1 ;
                break ;
            case 't':
                /* File type - use look up table, see first comment */
                ftype = typeRet[ftype] ;
                break ;
            
            case 'w':
                /* File write permission */
                /* If nasty or url - no */
                if((ftype == 0)
#if URLAWAR
                   || (ftype == 5)
#endif
                   )
                    return meLtoa(0) ;
                /* if it doesnt exist or its an DOS/WIN directory - yes */
                if((ftype == 3)
#if (defined _DOS) || (defined _WIN32)
                   || (ftype == 2)
#endif
                   )
                    return meLtoa(1) ;
                return meLtoa(meStatTestWrite(stats)) ;
            
            case 'x':
                /* File execute permission */
                /* If nasty or doesnt exist, or url then no */
                if((ftype == 0) || (ftype == 3) 
#if URLAWAR
                   || (ftype == 5)
#endif
                   )
                    return meLtoa(0) ;
#if (defined _DOS) || (defined _WIN32)
                /* If directory, simulate unix style execute flag and return yes */
                if(ftype == 2)
                    return meLtoa(1) ;
                /* Must test for .exe, .com, .bat, .btm extensions etc */
                return meLtoa(meTestExecutable(arg2)) ;
#else
                /* Test the unix execute flag */
                return meLtoa((((stats.stuid == meUid) && (stats.stmode & S_IXUSR)) ||
                               ((stats.stgid == meGid) && (stats.stmode & S_IXGRP)) ||
                               (stats.stmode & S_IXOTH))) ;
#endif
            default:
                return abortm ;
            }                
            return meItoa(ftype) ;
        }
    }
    /* if we've got here then it is a &ind */
    return(getval(arg1));
}


/* numeric arg can overide prompted value */
int
unsetVariable(int f, int n)		/* Delete a variable */
{
    register meVARLIST  *varList ;	/* User variable pointer */
    register meVARIABLE *vptr, *prev;  	/* User variable pointer */
    register int   vnum;	        /* ordinal number of var refrenced */
    char	   var[NSTRING] ;  	/* name of variable to fetch */
    char          *vv ;
    
    /* First get the variable to set.. */
    /* set this flag to indicate that the variable name is required, NOT
     * its value */
    alarmState |= 0x02 ;
    /* horrid global variable, see notes at definition */
    gmaLocalRegPtr = meRegCurr ;
    vnum = mlreply("Unset variable", MLVARBL, 0, var,
                     NSTRING) ;
    alarmState &= ~0x02 ;
    if(vnum != TRUE)
        return vnum ;
    
    /* Check the legality and find the var */
    vv = var+1 ;
    vnum = getMacroTypeS(var) ; 
    if(vnum == TKVAR) 
        varList = &usrVarList ;
    else if(vnum == TKLVR) 
        varList = &curbp->varList ;
    else if(vnum == TKCVR)
    {
        if((vv=strchr(vv,'.')) != NULL)
        {
            int idx ;
            
            *vv++ = '\0' ;
            if((idx = decode_fncname(var+1,0)) < 0)
                return mlwrite(MWABORT,"[No such variable]");
            varList = &(cmdTable[idx]->varList) ;
        }
        else
        {
            if((varList = gmaLocalRegPtr->varList) == NULL)
                return mlwrite(MWABORT,"[No such variable]");
            vv = var+1 ;
        }
    }
    else
        return mlwrite(MWABORT,"[User variable required]");
    
    /*---	Check for existing legal user variable */
    prev = NULL ;
    vptr = varList->head ;
    while(vptr != NULL)
    {
        if(!(vnum=strcmp(vptr->name,vv)))
        {
            /* link out the variable to unset and free it */
            if(prev == NULL)
                varList->head = vptr->next ;
            else
                prev->next = vptr->next ;
            varList->count-- ;
            meNullFree(vptr->value) ;
            meFree(vptr) ;
            return TRUE ;			/* True exit */
        }
        if(vnum > 0)
            break ;
        prev = vptr ;
        vptr = vptr->next ;
    }
    
    /* If its not legal....bitch */
    return mlwrite(MWABORT|MWCLEXEC,"[No such variable]") ;
}

int
descVariable(int f, int n)		/* describe a variable */
{
    char  var[NSTRING];	/* name of variable to fetch */
    char *ss ;
    int   status ;
    
    /* first get the variable to describe */
    if((status = mlreply("Show variable",MLVARBL,0,var,NSTRING)) != TRUE)
        return(status);
    if((ss = getval(var)) == NULL)
        return mlwrite(MWABORT,"Unknown variable type") ;
    mlwrite(0,"Current setting is \"%s\"", ss) ;
    return TRUE ;
} 


int
setVariable(int f, int n)		/* set a variable */
{
    meVARLIST     *varList ;            /* Pointer to the root */
    register int   status ;       	/* status return */
    register int   vtype;	        /* type of variable to set */
    char          *nn, var[NSTRING];	/* name of variable to fetch */
    char          *value, vv[MAXBUF];	/* value to set variable to */
    
    /* set this flag to indicate that the variable name is required, NOT
     * its value */
    alarmState |= 0x02 ;
    /* horrid global variable, see notes at definition */
    gmaLocalRegPtr = meRegCurr ;
    status = mlreply("Set variable", MLVARBL, 0, var,
                     NSTRING) ;
    alarmState &= ~0x02 ;
    if(status != TRUE)
        return status ;
    
    /* check the legality and find the var */
    vtype = getMacroTypeS(var) ; 
    nn = var+1 ;
    if(vtype == TKREG)
    {
        meREGISTERS   *rp ;
        unsigned char  cc ;
        if((cc=*nn) == 'l')
            rp = gmaLocalRegPtr ;
        else if(cc == 'p')
            rp = gmaLocalRegPtr->prev ;
        else
            rp = meRegHead ;
        cc = nn[1] - '0' ;
        if(cc >= ('0'+meNUMREG))
            return mlwrite(MWABORT,"[No such register %s]",var);
        value = rp->reg[cc] ;
    }
    else
    {
        if(vtype == TKVAR) 
            varList = &usrVarList ;
        else if(vtype == TKLVR) 
            varList = &(curbp->varList) ;
        else if(vtype == TKCVR)
        {
            char *ss, cc ;
            for(ss=nn ; (((cc=*++ss) != '\0') && (cc != '.')) ; )
                ;
            if(cc == '.')
            {
                int ii ;
                *ss++ = '\0' ;
                
                if((ii=decode_fncname(nn,1)) < 0)
                    /* not a command - abort */
                    return mlwrite(MWABORT,"[No such variable]");
                varList = &(cmdTable[ii]->varList) ;
                nn = ss ;
            }
            else if((varList = gmaLocalRegPtr->varList) == NULL)
                return mlwrite(MWABORT,"[No such variable]");
        }
        else if((vtype != TKENV) && (vtype != TKARG))
            /* else its not legal....bitch */
            return mlwrite(MWABORT,"[No such variable]");
        value = &(vv[0]) ;
    }
    /* get the value for that variable */
    if (f == TRUE)
        strcpy(value, meItoa(n));
    else if(storeResult != NULL)
        strcpy(value, storeResult);
    else if((status = mlreply("Value", 0, 0, value,MAXBUF)) != TRUE)
        return status ;

    /* and set the appropriate value  */
    if(vtype == TKREG) 
        return TRUE ;
    else if(vtype == TKARG)
    {
        if(*nn == 'w')
        {
            if((status=bchange()) != TRUE)
                return status ;
            if(nn[1] == 'l')
            {
                int ll=0 ;
                char cc ;
                curwp->w_doto = 0 ;
#if MEUNDO
                meUndoAddDelChars(llength(curwp->w_dotp)) ;
#endif
                mldelete(llength(curwp->w_dotp),NULL) ;
                while((cc=*value++))
                {
                    if (cc == 0x0a)
                        status = lnewline();
                    else
                        status = linsert(1,cc);
                    if (status != TRUE)
                        return status ;
                    ll++ ;
                }
#if MEUNDO
                meUndoAddReplaceEnd(ll) ;
#endif
            }
            else
            {
                lchange(WFMAIN);
#if MEUNDO
                meUndoAddRepChar() ;
#endif
                lputc(curwp->w_dotp,curwp->w_doto,value[0]) ;
            }
        }
        else if(*nn == 'c')
        {
            int ii ;
            if(nn[2] == 'k')
            {
                int savcle;
                char *saves ;
                          
                /* use the macro string to key evaluator to get the value,
                 * must set this up carefully and restore state */
                savcle = clexec;		/* save execution mode */
                clexec = TRUE;		/* get the argument */
                saves = execstr ;
                execstr = value ;
                ii = getckey(0,1) ;
                execstr = saves ;
                clexec = savcle;		/* restore execution mode */
                if(nn[1] == 'c')
                    thisCommand = ii ;
                else
                    lastCommand = ii ;
            }
            else
            {
                ii = decode_fncname(value,1) ;
                if(nn[1] == 'l')
                {
                    switch(ii)
                    {
                    case CK_BAKLIN:
                    case CK_FORLIN:
                        thisflag = CFCPCN;
                        break ;
                    case CK_DELBAK:
                    case CK_DELFOR:
                        if(!meModeTest(curbp->b_mode,MDLETTR))
                            break ;
                    case CK_DELWBAK:
                    case CK_CPYREG:
                    case CK_DELFWRD:
                    case CK_KILEOL:
                    case CK_KILPAR:
                    case CK_KILREG:
                        thisflag = CFKILL;
                        break ;
                    case CK_REYANK:
                        thisflag = CFRYANK ;
                        break ;
                    case CK_YANK:
                        thisflag = CFYANK ;
                        break ;
                    case CK_UNDO:
                        /* ZZZZ - test for mode */
                        thisflag = CFUNDO ;
                        break ;
                    }
                    lastIndex = ii ;
                }
                else
                    thisIndex = ii ;
            }
        }
        else if(*nn == '?')
            thisF = meAtoi(value) ;
        else
            thisN = meAtoi(value) ;
        return TRUE ;
    }
    if(vtype != TKENV) 
    {
        /* set a user variable */
#if 0
        register meVARIABLE *vptr, *prev ;	/* User variable pointer */
                    
        vptr = varList->head ;
        prev = NULL ;
        while(vptr != NULL)
        {
            signed char *s1, *s2, ss ;
        
            s1 = (signed char *) vptr->name ;
            s2 = (signed char *) nn ;
            for( ; ((ss=*s1++) == *s2) ; s2++)
                if(ss == 0)
                {
                    /* found it, replace value */
                    meNullFree(vptr->value);
                    goto found_ulcvar ;
                }
            if(ss > *s2)
            {
                vptr = NULL ;
                break ;
            }
            prev = vptr ;
            vptr = vptr->next ;
        }
#endif
        register meVARIABLE *vptr, *vp ;
        register int ii, jj ;
        
        vptr = (meVARIABLE *) varList ;
        ii = varList->count ;
        /* scan the list looking for the user var name */
        while(ii)
        {
            signed char *s1, *s2, ss ;
            
            jj = (ii>>1)+1 ;
            vp = vptr ;
            while(--jj >= 0)
                vp = vp->next ;
            
            s1 = (signed char *) vp->name ;
            s2 = (signed char *) nn ;
            for( ; ((ss=*s1++) == *s2) ; s2++)
                if(ss == 0)
                {
                    /* found it, replace value */
                    meNullFree(vp->value);
                    goto found_ulcvar ;
                }
            if(ss > *s2)
                ii = ii>>1 ;
            else
            {
                ii -= (ii>>1) + 1 ;
                vptr = vp ;
            }
        }
        
        /* Not found so create a new one */
        if((vp = (meVARIABLE *) meMalloc(sizeof(meVARIABLE)+strlen(nn))) == NULL)
            return FALSE ;
        strcpy(vp->name,nn) ;
        vp->next = vptr->next ;
        vptr->next = vp ;
        varList->count++ ;
found_ulcvar:
        vp->value = strdup(value) ;
        return TRUE ;
    }
    /* set an environment variable */
#ifdef _MACRO_COMP
    if((status = (unsigned char) *nn) >= 128)
        status -= 128 ;
    else
#endif
        status = biChopFindString(nn,14,envars,NEVARS) ;
    switch(status)
    {
    case EVAUTOTIME:
        autotime = meAtoi(value);
        break;
    case EVKEPTVERS:
        if(!(meSystemCfg & meSYSTEM_DOSFNAMES) && 
           ((keptVersions = meAtoi(value)) < 0))
            keptVersions = 0 ;
        break;
    case EVBOXCHRS:
        {
            unsigned char cc ;
            int len;
            len = strlen (value);
            if (len > BCLEN)
                len = BCLEN;
            while(--len >= 0)
            {
                cc = value[len] ;
                if(!isPokable(cc))
                    cc = '.' ;
                boxChars[len] = cc ;
            }
        }
        break;
    case EVIDLETIME:
        idletime = (unsigned long) meAtoi(value);
        if (idletime < 10)
            idletime = 10;
        break;
    case EVDELAYTIME:
        delaytime = (unsigned long) meAtoi(value);
        if (delaytime < 10)
            delaytime = 10;
        break;
    case EVREPEATTIME:
        repeattime = (unsigned long) meAtoi(value);
        if (repeattime < 10)
            repeattime = 10;
        break;
    case EVMODELINE:
        if(modeLineStr != orgModeLineStr)
            meFree(modeLineStr) ;
        modeLineStr = strdup(value) ;
        modeLineFlags = assessModeLine(value) ;
        addModeToWindows(WFMODE) ;
        break ;
    case EVBMDLINE:
        meNullFree(curbp->modeLineStr) ;
        curbp->modeLineStr = strdup(value) ;
        curbp->modeLineFlags = assessModeLine(value) ;
        addModeToWindows(WFMODE) ;
        break ;
#if MOUSE
    case EVMOUSE:
        meMouseCfg = meAtoi(value) ;
        TTinitMouse() ;
        break;
#endif
    case EVSYSTEM:
        {
            meSystemCfg = (meSystemCfg & ~meSYSTEM_MASK) | 
                      (meAtoi(value)&meSYSTEM_MASK) ;
            if(meSystemCfg & meSYSTEM_DOSFNAMES)
                keptVersions = 0 ;
#ifdef _UNIX
            /* on unix, if using x-window then can't set ANSI || XANSI bits
             * else must redraw incase these bits are changed */
#ifdef _XTERM
            if(!(meSystemCfg & meSYSTEM_CONSOLE))
                meSystemCfg &= ~(meSYSTEM_ANSICOLOR|meSYSTEM_XANSICOLOR|meSYSTEM_FONTS) ;
            else
#endif
            {
                if(meSystemCfg & meSYSTEM_FONTS)
                    meSystemCfg &= ~(meSYSTEM_ANSICOLOR|meSYSTEM_XANSICOLOR) ;
                sgarbf = TRUE ;
            }
#endif
#ifdef _CLIENTSERVER
            /* open or close the client server files */
            if(meSystemCfg & meSYSTEM_CLNTSRVR)
                TTopenClientServer ();
            else
                TTkillClientServer ();
#endif
            break;
        }
    case EVFILLBULLET:
        strncpy(fillbullet,value,15);
        fillbullet[15] = '\0' ;
        break;
    case EVFILLBULLETLEN:
        fillbulletlen = meAtoi (value);
        break;
    case EVFILLCOL:
        fillcol = meAtoi(value);
        break;
    case EVFILLEOS:
        strncpy(filleos,value,15);
        filleos[15] = '\0' ;
        break;
    case EVFILLEOSLEN:
        filleoslen = meAtoi (value);
        break;
    case EVFILLIGNORE:
        strncpy(fillignore,value,15);
        fillignore[15] = '\0' ;
        break;
    case EVFILLMODE:
        fillmode = *value ;
        break;
    case EVTIME:
        timeOffset = meAtoi(value) ;
        break;
    case EVMATCHLEN:
        matchlen = meAtoi (value);
        break;
    case EVPAGELEN:
        return changeScreenDepth(TRUE, meAtoi(value));
    case EVRANDOM:
        srand(time(NULL)) ;
        break ;
    case EVCURCOL:
        return (((status=meAtoi(value)) > llength(curwp->w_dotp)) ?
                ((curwp->w_doto = llength(curwp->w_dotp)),FALSE):
                ((curwp->w_doto = status),TRUE)) ;
    case EVABSLINE:
        return gotoAbsLine(meAtoi(value));
    case EVCURLINE:
        return gotoLine(TRUE, meAtoi(value));
    case EVWINCOL:
        return setcwcol(meAtoi(value));
    case EVWINCHRS:
        {
            unsigned char cc ;
            int len;
            len = strlen (value);
            if (len > WCLEN)
                len = WCLEN;
            while(--len >= 0)
            {
                cc = value[len] ;
                if(!isPokable(cc))
                    cc = '.' ;
                windowChars[len] = cc ;
            }
        }
        break;
    case EVWINLINE:
        return gotoLine(TRUE,curwp->topLineNo+meAtoi(value));
    case EVCURWIDTH:
        return changeScreenWidth(TRUE, meAtoi(value));
    case EVCBUFNAME:
        unlinkBuffer(curbp) ;
        if((value[0] == '\0') || (bfind(value,0) != NULL))
        {
            /* if the name is used */
            linkBuffer(curbp) ;
            return ABORT ;
        }
        meFree(curbp->b_bname) ;
        curbp->b_bname = strdup(value);
        addModeToWindows(WFMODE) ;
        linkBuffer(curbp) ;
        break;
    case EVBUFFMOD:
        curbp->stats.stmode = meAtoi(value) ;
        break ;
    case EVGLOBFMOD:
        meUmask = meAtoi(value) ;
        break ;
    case EVCFNAME:
        meNullFree(curbp->b_fname) ;
        curbp->b_fname = strdup(value);
        addModeToWindows(WFMODE) ;
        break;
    case EVDEBUG:
        macbug = meAtoi(value);
        break;
#if	TIMSTMP
    case EVTIMSTMP:
        strncpy(&time_stamp[0], value, TSTMPLEN-1);
        time_stamp[TSTMPLEN-1] = '\0';
        break;
#endif
    case EVTABSIZE:
        tabsize = meAtoi(value);
        break;
    case EVTABWIDTH:
        tabwidth = meAtoi(value);
        break;
    case EVSRCHPATH:
        meNullFree(searchPath) ;
        searchPath = strdup(value) ;
        break;
    case EVHOMEDIR:
        meNullFree(homedir) ;
        homedir = strdup(value) ;
        break;
#if	CFENCE
    case EVSTATEMENTINDENT:
        statementIndent = meAtoi (value);
        break;
    case EVCONTINUEINDENT:
        continueIndent = meAtoi (value);
        break;
    case EVBRACEINDENT:
        braceIndent = meAtoi (value);
        break;
    case EVCASEINDENT:
        caseIndent = meAtoi (value);
        break;
    case EVCONTCOMM:
        meNullFree(commentCont) ;
        commentCont = strdup(value) ;
        break ;
    case EVECMARGIN:
        commentMargin = meAtoi(value);
        break;
#endif
    case EVSHWMDS:
        {
            char cc ;
            int nn ;
            for(nn=0; nn < MDNUMMODES ; nn++) 
            {
                if((cc=*value++) == '\0')
                    break ;
                else if(cc == '0')
                    meModeClear(modeLineDraw,nn) ;
                else
                    meModeSet(modeLineDraw,nn) ;
            }
            addModeToWindows(WFMODE) ;
            break ;
        }
    case EVSHWRGN:
        selhilight.uFlags = meAtoi(value);
        break;
    case EVCURSORBLK:
        timerKill(CURSOR_TIMER_ID) ;
        if((cursorBlink = meAtoi(value)) > 0)
            /* kick off the timer */
            TThandleBlink(1) ;
        break ;
#if COLOR
    case EVCURSORCOL:
        if((cursorColor = meAtoi(value)) >= noColors)
            cursorColor = TTdefFCol ;
        break ;
    case EVOSDSCHM:
        osdScheme = hilValidateScheme(meAtoi(value),
                                       osdScheme/HCTNUM)*HCTNUM;
        break ;
    case EVMLSCHM:
        mlScheme = hilValidateScheme(meAtoi(value),
                                    mlScheme/HCTNUM)*HCTNUM;
        break ;
    case EVMDLNSCHM:
        mdLnScheme = hilValidateScheme(meAtoi(value),mdLnScheme/HCTNUM)*HCTNUM;
        addModeToWindows(WFMODE) ;
        break ;
    case EVGLOBSCHM:
        globScheme = hilValidateScheme(meAtoi(value),globScheme/HCTNUM)*HCTNUM;
        vvideo.video [TTnrow].endp = TTncol ;
        sgarbf = TRUE ;
        TTsetBgcol() ;
        break ;
    case EVBUFSCHM:
        curbp->b_scheme = hilValidateScheme(meAtoi(value),curbp->b_scheme/HCTNUM)*HCTNUM;
        addModeToWindows(WFRESIZE) ;
        break ;
    case EVLINESCHM:
        {
            register int ii, schm ;
            
            schm = meAtoi(value) ;
            if(schm >= 0)
            {
                schm = hilValidateScheme(schm,curbp->b_scheme/HCTNUM)*HCTNUM;
                for(ii=1 ; ii<LNNOSCHM ; ii++)
                    if(curbp->lScheme[ii] == schm)
                        break ;
                if(ii == LNNOSCHM)
                {
                    if((ii=curbp->nextLScheme+1) == LNNOSCHM)
                        ii = 1 ;
                    curbp->lScheme[ii] = schm ;
                    curbp->nextLScheme = (ii+1) & LNSMASK ;
                }
                curwp->w_dotp->l_flag = (curwp->w_dotp->l_flag & ~LNSMASK) | ii ;
            }
            else if(curwp->w_dotp->l_flag & LNSMASK)
                curwp->w_dotp->l_flag = (curwp->w_dotp->l_flag & ~LNSMASK) ;
            lchange(WFMAIN) ;
            break ;
        }
    case EVSBAR:    
        gsbarmode = meAtoi(value) & WMUSER;
        resizeAllWnd (TRUE, 0);         /* Force window update */
        break;
    case EVSBARSCHM:
        sbarScheme = hilValidateScheme(meAtoi(value),sbarScheme/HCTNUM)*HCTNUM;
        addModeToWindows(WFRESIZE|WFSBAR) ;
        break;
#endif
#if _IPIPES
    case EVBUFIPIPE:
        curbp->ipipeFunc = decode_fncname(value,1) ;
        break ;
#endif
    case EVBUFINP:
        curbp->inputFunc = decode_fncname(value,1) ;
        break ;
    case EVBUFBHK:
        curbp->bhook = decode_fncname(value,1) ;
        break ;
    case EVBUFDHK:
        curbp->dhook = decode_fncname(value,1) ;
        break ;
    case EVBUFEHK:
        curbp->ehook = decode_fncname(value,1) ;
        break ;
    case EVBUFFHK:    
        curbp->fhook = decode_fncname(value,1) ;
        break ;
#if HILIGHT
    case EVBUFHIL:
        if(((curbp->hiLight = meAtoi(value)) >= noHilights) ||
           (hilights[curbp->hiLight] == NULL))
            curbp->hiLight = 0 ;
        addModeToWindows(WFRESIZE) ;
        break ;
    case EVBUFIND:
        if(((curbp->indent = meAtoi(value)) >= noIndents) ||
           (indents[curbp->indent] == NULL))
            curbp->indent = 0 ;
        break ;
#endif
    case EVBUFMASK:
        {
            int ii ;
            isWordMask = 0 ;
            for(ii=0 ; ii<8 ; ii++)
                if(strchr(value,charMaskFlags[ii]) != NULL)
                    isWordMask |= 1<<ii ;
            curbp->isWordMask = isWordMask ;
            break ;
        }
    case EVFILEIGNORE:
        meNullFree(fileIgnore) ;
        fileIgnore = strdup(value) ;
        break ;
#if FLNEXT
    case EVFILETEMP:
        meNullFree(flNextFileTemp) ;
        flNextFileTemp = strdup(value) ;
        break ;
    case EVLINETEMP:
        meNullFree(flNextLineTemp) ;
        flNextLineTemp = strdup(value) ;
        break ;
#endif
#if DORCS
    case EVRCSFILE:
        meNullFree(rcsFile) ;
        rcsFile = strdup(value) ;
        break ;
    case EVRCSCICOM:
        meNullFree(rcsCiStr) ;
        rcsCiStr = strdup(value) ;
        break ;
    case EVRCSCIFCOM:
        meNullFree(rcsCiFStr) ;
        rcsCiFStr = strdup(value) ;
        break ;
    case EVRCSCOCOM:
        meNullFree(rcsCoStr) ;
        rcsCoStr = strdup(value) ;
        break ;
    case EVRCSCOUCOM:
        meNullFree(rcsCoUStr) ;
        rcsCoUStr = strdup(value) ;
        break ;
#endif
    case EVSCROLL:
        scrollFlag = (unsigned char) meAtoi(value) ;
        break ;
    case EVBNAMES:
        meNullFree(buffNames.list) ;
        meNullFree(buffNames.mask) ;
        if((buffNames.mask = strdup(value)) != NULL)
            buffNames.size = createBuffList(&buffNames.list,0) ;
        buffNames.curr = 0 ;
        break ;
    case EVCNAMES:
        meNullFree(commNames.list) ;
        meNullFree(commNames.mask) ;
        if((commNames.mask = strdup(value)) != NULL)
        {
            commNames.size = createCommList(&commNames.list,1) ;
            sortStrings(commNames.size,commNames.list,0,strcmp) ;
        }
        commNames.curr = 0 ;
        break ;
    case EVFNAMES:
        {
            char buff[FILEBUF], *mask ;
            
            meNullFree(fileNames.mask) ;
            getFilePath(curbp->b_fname,buff) ;
            strcat(buff,value) ;
            fileNameCorrect(buff,resultStr,&mask) ;
            if((fileNames.mask = strdup(mask)) != NULL)
            {
                *mask = '\0' ;
                getDirectoryList(resultStr,&fileNames) ;
            }
            fileNames.curr = 0 ;
            break ;
        }
    case EVMNAMES:
        meNullFree(modeNames.mask) ;
        modeNames.mask = strdup(value) ;
        modeNames.curr = 0 ;
        break ;
    case EVVNAMES:
        if(varbNames.list != NULL)
            freeFileList(varbNames.size,varbNames.list) ;
        meNullFree(varbNames.mask) ;
        if((varbNames.mask = strdup(value)) != NULL)
            varbNames.size = createVarList(&varbNames.list) ;
        varbNames.curr = 0 ;
        break ;
    case EVFINDWORDS:
        findWordsInit(value) ;
        break ;
    case EVRESULT:                      /* Result string - assignable */
        strcpy (resultStr, value);
        break;
    case -1:
        {
            /* unknown so set an environment variable */
            char buff[MAXBUF+NSTRING] ;
            sprintf(buff,"%s=%s",nn,value) ;
            putenv(strdup(buff)) ;
            break ;
        }
    default:
        /* default includes EVMOUSEX EVMOUSEY EVSTATUS EVMACHINE 
         * EVSYSRET EVUSEX, EVWMDLINE or system dependant vars
         * where this isn't the system (e.g. use-x) which cant be set
         */
        return FALSE ;
    }
    return TRUE ;
}

/*
 * showVariable
 * Format a variable for display. This is a helper function for 
 * "listVariables()" used to format and display each of the values.
 * 
 * Arguments:-
 *    bp     - The buffer to be rendered into.
 *    prefix - The prefix character(s) of the variable
 *    name   - The name of the variable.
 *    value  - The value assigned to variable.
 * 
 * Produce a line in the form:-
 * 
 *    name......... "value"
 */
static void
showVariable (BUFFER *bp, char prefix, char *name, char *value)
{
    char buf [MAXBUF];
    int len;
    int slen;
    
    if (value == NULL)
        value = "ERROR";
    
    len = sprintf (buf, "    %c%s ", prefix, name);
    while (len < 35)
        buf[len++] = '.';
    buf[len++] = ' ';
    buf[len++] = '"';
    slen = strlen (value);
    if (slen > (MAXBUF - 3 - len))
        slen = (MAXBUF - 3 - len);
    strncpy (&buf[len], value, slen);
    len += slen;
    buf[len++] = '"';
    buf[len] = '\0';
    addLineToEob(bp,buf);
}

/*
 * listVariables
 * List ALL of the variables held in the system.
 */

int
listVariables (int f, int n)
{
    meVARIABLE *tv ;
    WINDOW *wp ;
    BUFFER *bp ;
    char    buf[MAXBUF] ;
    int     ii ;
    
    if((wp = wpopup("*variables*",(BFND_CREAT|BFND_CLEAR|WPOP_USESTR))) == NULL)
        return FALSE ;
    bp = wp->w_bufp ;
    
    addLineToEob(bp,"Register variables:");
    addLineToEob(bp,"") ;
    
    buf[0] = 'g';
    buf[2] = '\0';
    for(ii = 0; ii<meNUMREG ; ii++)
    {
        buf[1] = (int)('0') + ii;
        showVariable (bp, '#', buf, meRegHead->reg[ii]);
    }
    addLineToEob(bp,"") ;
    sprintf (buf,"Buffer [%s] variables:", curbp->b_bname);
    addLineToEob(bp,buf);
    addLineToEob(bp,"") ;
    tv = curbp->varList.head ;
    while(tv != NULL)
    {
        showVariable (bp, ':', tv->name, tv->value);
        tv = tv->next ;
    }
    
    addLineToEob(bp,"") ;
    addLineToEob(bp,"System variables:");
    addLineToEob(bp,"") ;
    for (ii = 0; ii < NEVARS; ii++)
        showVariable (bp, '$', envars [ii], gtenv (envars [ii]));
    
    addLineToEob(bp,"") ;
    addLineToEob(bp,"Global variables:");
    addLineToEob(bp,"") ;
    for (tv=usrVarList.head ; tv != NULL ; tv = tv->next)
        showVariable (bp, '%', tv->name, tv->value);
    
    bp->b_dotp = lforw(bp->b_linep);
    bp->b_doto = 0 ;
    bp->line_no = 0 ;
    meModeClear(bp->b_mode,MDEDIT) ;    /* don't flag this as a change */
    meModeSet(bp->b_mode,MDVIEW) ;      /* put this buffer view mode */
    resetBufferWindows(bp) ;            /* Update the window */
    mlerase(MWCLEXEC);	                /* clear the mode line */
    return TRUE ;
}
