/*****************************************************************************
*
*	Title:		%M%
*
*	Synopsis:	Command execution.
*
******************************************************************************
*
*	Filename:		%P%
*
*	Author:			Danial Lawrence
*
*	Creation Date:		14/05/86 12:37		<990714.0911>
*
*	Modification date:	%G% : %U%
*
*	Current rev:		%I%
*
*	Special Comments:	
*
*	Contents Description:	
*
*	This file is for functions dealing with execution of
*	commands, command lines, buffers, files and startup files
*
*****************************************************************************
* 
* Modifications to the original file by Jasspa. 
* 
* Copyright (C) 1988 - 1999, JASSPA 
* The MicroEmacs Jasspa distribution can be copied and distributed freely for
* any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
* incorporated into commercial software with the expressed permission of
* JASSPA.
* 
****************************************************************************/

/*---	Include defintions */

#define	__EXECC			/* Define program name */

#define KEY_TEST 0

/*---	Include files */

#include "emain.h"
#include "efunc.h"
#include "eskeys.h"
#include "evar.h"

/*---	Local macro definitions */

#define DRRABORT     0x00
#define DRRCONTIN    0x60
#define DRRDONE      0x61
#define DRRELIF      0xc8
#define DRRELSE      0xc9
#define DRRENDIF     0x08
#define DRRGOTO      0x62
#define DRRIF        0x38
#define DRRREPEAT    0x20
#define DRRRETURN    0x05
#define DRRUNTIL     0x31
#define DRRUNTILF    0xb3
#define DRRWHILE     0x37
#define DRRWHILEF    0xb9

#define DRRJUMP      0x80

/*---	External references */

/*---	Local type definitions */

/*---	Local variable definitions */

int relJumpTo ;

/*	docmd:	take a passed string as a command line and translate
		it to be executed as a command. This function will be
		used by execute-line and by all source and
		startup files.

	format of the command line is:

		{# arg} <command-name> {<argument string(s)>}

	Directives start with a "!" and include:

	!endm		End a macro
	!if (cond)	conditional execution
	!else
	!endif
	!return 	Return (terminating current macro)
	!goto <label>	Jump to a label in the current macro

	Line Labels begin with a "*" in column 1, like:

	*LBL01
*/

int
biChopFindString(register char *ss, register int len, register char **tbl,
                 register int size)
{
    register int hi, low, mid, ll ;
    
    hi  = size-1;		/* Set hi water to end of table */
    low = 0;			/* Set low water to start of table */
    
    do
    {
        signed char *s1, *s2, cc ;
        mid = (low + hi) >> 1;		/* Get mid value. */
        
        ll = len ;
        s1 = (signed char *) tbl[mid] ;
        s2 = (signed char *) ss ;
        for( ; ((cc=*s1++) == *s2) ; s2++)
            if(!--ll || (cc == 0))
                return mid ;
        if(cc > *s2)
            hi = mid - 1;		/* Discard bottom half */
        else
            low = mid + 1;		/* Discard top half */
    } while(low <= hi);			/* Until converges */
    
    return -1 ;
}

/* token:	chop a token off a string
		return a pointer past the token
*/
/* src - source string, tok - destination token string */

char *
token(char *src, char *tok)
{
    register int  quotef;
    register unsigned char cc, *ss, *dd ;
    unsigned short key ;
    
    ss = (unsigned char *) src ;
    dd = (unsigned char *) tok ;
    
    /*---	First scan past any whitespace in the source string */

    while(((cc = *ss) == ' ') || (cc == '\t'))
        ss++ ;
    
    /*---	Scan through the source string */

    for(quotef=0 ; ; cc=*++ss)
    {
        switch(cc)
        {
        case '"':
            if(quotef)
            {
                /* Null Terminate the token and exit */
                ss++;
                *dd = 0;
                return (char *) ss ;
            }
            quotef = 1 ;
            *dd++ = '"' ;
            break ;
        case '\t':
        case ' ':
        case ';':
            if(quotef)
            {
                *dd++ = cc;
                break ;
            }
            /* no break - this is the end of the token */
        case '\0':
            /* Null Terminate the token and exit */
            *dd = 0;
            return (char *) ss ;
            
        case '\\':
            /* Process special characters */
            switch ((cc = *++ss))
            {
            case '\0':
                /* this probably should be an error */
                *dd = 0;
                return (char *) ss ;
                
            case '0':
                cc = *(ss+1) ;
                if(isHexDigit(cc) && isHexDigit(*(ss+2)))
                {
                    if((cc = (hexToNum(cc) << 4) | hexToNum(*(ss+2))) == 0)
                    {
                        *dd++ = 0xff ;
                        *dd++ = 0x01 ;
                    }
                    else if(cc == 0xff)
                    {
                        *dd++ = 0xff ;
                        *dd++ = 0xff ;
                    }
                    else
                        *dd++ = cc ;
                    ss += 2 ;
                }
                break;
            case 'C':
                /* Control key - \C? */
                *dd++ = *++ss - '@'; 
                break;
            case 'E':
                /* Escape key - replace with esc */
                key = ME_SPECIAL|SKEY_esc ;
                goto quote_spec_key ;
            case 'I':
                /* backward-delete-tab - replace with S-tab */
                key = ME_SPECIAL|ME_SHIFT|SKEY_tab ;
                goto quote_spec_key ;
            case 'N':
                /* Return key - replace with return */
                key = ME_SPECIAL|SKEY_return ;
                goto quote_spec_key ;
            case 'P':
                /* Go to set position, defined by \p - replace with \CXAP */
                *dd++ = 'X' - '@';
                *dd++ = 'A' ;
                *dd++ = 'P';
                break;
            case 'T':
                /* Tab key - replace with tab */
                key = ME_SPECIAL|SKEY_tab ;
quote_spec_key:
                *dd++ = 0xff ;
                *dd++ = 2 ;
                *dd++ = key >> 8 ;
                *dd++ = key & 0xff ;
                break;
            case 'b':	*dd++ = 0x08; break;
            case 'd':	*dd++ = 0x7f; break;
            case 'e':	*dd++ = 0x1b; break;
            case 'f':	*dd++ = 0x0c; break;
            case 'i':
                *dd++ = 0xff;
                *dd++ = 2;  
                *dd++ = '@';
                *dd++ = 'I' - '@';
                break;
            case 'n':	*dd++ = 0x0a; break;
            case 'p':
                *dd++ = 'X' - '@';
                *dd++ = 'A' - '@';
                *dd++ = 'P';
                break;
            case 'r':	*dd++ = 0x0d; break;
            case 's':	
                *dd++ = 0xff;
                *dd++ = 0x02;  
                break;
            case 't':	*dd++ = 0x09; break;
            default:
                *dd++ = cc;
            }
            break ;
            
        case 255:
            *dd++ = 0xff;
            /* no break - must store 2 0xff's */
        default:
            /* record the character */
            *dd++ = cc;
        }
    }
    
}

/*	mlreply:	get the next argument	*/
/* prompt  - prompt to use if we must be interactive */
/* option  - options for getstring */
/* defnum  - Default history number */
/* buffer  - buffer to put token into */
/* size	   - size of the buffer */
int
mlreply(char *prompt, int option, int defnum, char *buffer, int size)
{
    /* if we are not interactive, go get it! */
    if (clexec == TRUE)
    {
        char buff[MAXBUF], *res, *ss ;
        
        /* grab token and advance past */
        ss = execstr ;
        execstr = token(ss, buff);
        
        if((buff[0] != '@') || (buff[1] != 'm') || (buff[2] != 'n'))
        {
            /* evaluate it */
            res = getval(buff) ;
            if(res == abortm)
            {
                *buffer = '\0' ;
                return FALSE ;
            }
            ss = buffer ;
            if(option & MLMACNORT)
            {
                while((*ss))
                {
                    ss++ ;
                    size-- ;
                }
            }
            for(; ((*ss++ = *res++) != '\0') ; )
                if((--size) == 0)
                {
                    ss[-1] = '\0' ;
                    break ;
                }
            return TRUE ;
        }
        /* if @mna (get all input from user) then rewind the execstr */
        if(buff[3] == 'a')
            execstr = ss ;
    }
    return mlgetstring(prompt, option, defnum, buffer, size) ;
}

int
macarg(char *tok)		/* get a macro line argument */
{
    int savcle;			/* buffer to store original clexec */
    int status;
    
    savcle = clexec;		/* save execution mode */
    clexec = TRUE;		/* get the argument */
    status = mlreply("", MLNOHIST, 0, tok, MAXBUF) ;
    clexec = savcle;		/* restore execution mode */
    
    return(status);
}

#if KEY_TEST
/*---	Little function to test the  alphabetic state of the table.  Invoked
     	via "!test" primative.  */

int
fnctest(void)
{
    extern unsigned long cmdHashFunc (register char *cmdName) ;
    register KEYTAB	*ktp;			/* Keyboard character array */
    register meCMD	*cmd ;			/* Names pointer */
    unsigned long        key ;
    char                 outseq[12];
    int	                 count=0, ii;		/* Counter of errors */

    /* test the command hash table */
    for (ii=0; ii < CK_MAX; ii++)
    {
        key = cmdHashFunc(getCommandName(ii)) ;
        cmd = cmdHash[key] ;
        while((cmd != NULL) && (cmd != cmdTable[ii]))
            cmd = cmd->hnext ;
        if(cmd == NULL)
        {
            count++;
            printf("cmdHash Error: [%s] should be in position %d\n", 
                   getCommandName(ii),key) ;
        }
    }
    
    /* test the command alphabetically ordered list */
    cmd = cmdHead ;
    while(cmd->anext != NULL)
    {
        if(strcmp(cmd->anext->name,cmd->name) < 0)
        {
            count++;
            printf("cmdHead Error: [%s] should be before [%s]\n", 
                   cmd->anext->name,cmd->name) ;
        }
        cmd = cmd->anext ;
    }
    /* Scan through the key table looking for the character code.  If found
     * then return the index into the names table.  */

    for (ktp = &keytab[0]; ktp[1].code != 0; ktp++)
        if (ktp->code > ktp[1].code)
        {
            count++;
            cmdstr(ktp->code, outseq);
            printf("[%s] key out of place\n",outseq);
        }
        
    if(count)
        mlwrite(MWWAIT,"!test: %d Error(s) Detected",count);
    return (count);
}
#endif

int
domstore(char *cline)
{
    register char cc ;		/* Character */
    
    if(mcStore == 2)
    {
        /* check to see if this line ends definition */
        if(strncmp(cline, "!ehe",4) == 0)
        {
            mcStore = 0 ;
            lpStore = NULL;
            return TRUE ;
        }
        return addLine(lpStore,cline) ;
    }
    cc = *cline ;
    /* eat leading spaces */
    while((cc == ' ') || (cc == '\t'))
        cc = *++cline ;
    
    /* dump comments and empties here */
    if((cc == ';') || (cc == '\0'))
        return TRUE ;
#ifdef _MACRO_COMP
    /* check to see if this line turns macro storage off */
    if((cc == '!') &&
       ((((unsigned char) cline[1]) == DREMACRO+128) ||
        (strncmp(cline+1, "ema", 3) == 0)) && 
       !execlevel--)
    {
        mcStore = FALSE;
        lpStore = NULL;
        execlevel = 0 ;
        return TRUE ;
    }
    if(macbug == -3)
    {
        LINE           *lp, *ilp ;
        unsigned char  *ss ;
        register int    status ;
        int             len=0 ;
        
        ss = (unsigned char *) cline ;
        while((*ss++))
            len++ ;
        if ((lp=lalloc(len)) == NULL)
            return FALSE ;
        ilp = lpStore ;
        ilp->l_bp->l_fp = lp;
        lp->l_bp = ilp->l_bp;
        ilp->l_bp = lp;
        lp->l_fp = ilp ;
        ss = lp->l_text ;
        for(;;)
        {
            unsigned char dd, *s1 ;
            int           idx ;
            
            while(((dd = (unsigned char) *cline++) == ' ') || (dd == '\t'))
                ;
            if((dd == '\0') || (dd == ';'))
            {
                *--ss = '\0' ;
                lp->l_used = ss-lp->l_text ;
                return TRUE ;
            }
            s1 = (unsigned char *) cline ;
            cc = dd ;
            idx = 0 ;
            for(;;)
            {
                if(cc == '"')
                    idx ^= 1 ;
                else if((cc == '\\') && (*cline != '\0'))
                    cline++ ;
                if(!idx && ((cc == ' ') || (cc == '\t') || (cc == '\0')))
                    break ;
                cc = *cline++ ;
            }
            *--cline = '\0' ;
            switch(dd)
            {
            case '!':
                if((*s1 >= 128) ||
                   ((status = biChopFindString((char *) s1,3,derNames,NDERIV)) < 0))
                    goto cpy_str ;
                *ss++ = '!' ;
                *ss++ = status + 128 ;
                break ;
            case '$':
                if((*s1 >= 128) ||
                   ((status = biChopFindString((char *) s1,14,envars,NEVARS)) < 0))
                    goto cpy_str ;
                *ss++ = '$' ;
                *ss++ = status + 128 ;
                break ;
            case '&':
                if((*s1 >= 128) ||
                   ((status = biChopFindString((char *) s1,3,funcNames,NFUNCS)) < 0))
                    goto cpy_str ;
                *ss++ = '&' ;
                *ss++ = status + 128 ;
                break ;
            default:
                if(!isAlpha(dd) ||
                   ((idx = __decode_fncname((char *) s1-1,1)) < 0) ||
                   (idx >= CK_MAX))
                    goto cpy_str ;
                if(idx == CK_DEFMAC)
                    execlevel++ ;
                *ss++ = '^' ;
                if(((++idx) == '\n') || (idx == '\r') || (idx == 26))
                    ss += expandchar(idx,(char *) ss,1) ;
                else
                {
                    if((idx == ' ') || (idx == '\\') || (idx == '\t'))
                        *ss++ = '\\' ;
                    *ss++ = idx ;
                }
                break ;
                
cpy_str:
                if((dd == '^') &&
                   (*s1 == CK_DEFMAC+1))
                    execlevel++ ;
                if((dd == '"') && (cline[-1] == '"') && 
                   ((long) cline != (long) s1+1))
                {
                    unsigned char *s2, ee ;
                    
                    if((ee=*s1) == '\\')
                        ee = s1[1] ;
                    if(getMacroType(ee) == TKCMD)
                    {
                        s2 = s1 ;
                        while(((ee = *s2++) != '\0') && (ee != ' ') && (ee != '\t'))
                            ;
                        if(ee == '\0')
                        {
                            while((dd = *s1++))
                                *ss++ = dd ;
                            ss-- ;
                            break ;
                        }
                    }   
                }
                do {
                    *ss++ = dd ;
                } while((dd = *s1++)) ;
                break ;
            }
            *cline = cc ;
            *ss++ = ' ' ;
        }
    }
    else
    {
        if(((cc == '^') && (cline[1] == CK_DEFMAC+1)) ||
           ((cc == 'd') && (strncmp(cline, names[CK_DEFMAC].name,12) == 0)))
            execlevel++ ;
        return addLine(lpStore,cline) ;
    }
#else
    /* check to see if this line turns macro storage off */
    if((cc == '!') && !strncmp(cline+1, "ema", 3) && !execlevel--)
    {
        mcStore = FALSE;
        lpStore = NULL;
        execlevel = 0 ;
        return TRUE ;
    }
    if((cc == 'd') && !strncmp(cline, getCommandName(CK_DEFMAC),12))
        execlevel++ ;
    return addLine(lpStore,cline) ;
#endif
}


int
docmd(char *cline)
/* cline	command line to execute */
{
    register int  f ;		/* default argument flag */
    register int  n ;		/* numeric repeat value */
    register char cc ;		/* Character */
    register int  status ;	/* return status of function */
    register int  force ;	/* force TRUE result? */
    register int  nmacro ;	/* run as it not a macro? */
    char tkn[MAXBUF] ;		/* next token off of command line */
    
    cc = *cline ;
    /* eat leading spaces */
    while((cc == ' ') || (cc == '\t'))
        cc = *++cline ;
    
    /* dump comments and empties here */
    if((cc == ';') || (cc == '\0'))
        return TRUE ;
    
    /* dump labels here */
    if(cc == '*')
        return TRUE ;
    
    /* first set up the default command values */
    force = 0 ;
    nmacro = FALSE;
    execstr = cline;	/* and set this one as current */
    f = FALSE;
    n = 1;
    
try_again:
    execstr = token(execstr, tkn);

    /* process leadin argument */
    if((status=getMacroTypeS(tkn)) == TKDIR)
    {
        register int derType ;
        /* look the derivative up in the name table */
#ifdef _MACRO_COMP
        if((status = (unsigned char) tkn[1]) >= 128)
            status -= 128 ;
        else
#endif
            if((status = biChopFindString(tkn+1,3,derNames,NDERIV)) < 0)
                return mlwrite(MWABORT|MWWAIT,"[Unknown directive %s]",tkn);
        if(execlevel)
        {
            switch(status)
            {
            case DRIF:
            case DRREPEAT:
            case DRWHILE:
                execlevel += 2 ;
                break ;
            
            case DRDONE:
            case DRUNTIL:
                execlevel -= 2 ;
                break ;
            
            case DRELIF:
            case DRELSE:
                if(execlevel == 1)
                {
                    --execlevel;
                    if(status == DRELIF)
                    {
                        derType = DRRIF ;
                        goto elif_jump ;
                    }
                }
                break ;
                
            case DRENDIF:
                if(--execlevel)
                    --execlevel ;
            }
        }
        else if((derType=derTypes[status]) == 0)
        {
            switch(status)
            {
#if KEY_TEST
            case DRTEST:
                /* Test directive. */
                return ((fnctest() == 0));
#endif
            case DRFORCE:
                force++ ;
                goto try_again;
            case DRNMACRO:
                nmacro = TRUE;
                goto try_again;
            case DRABORT:
                return FALSE ;
            case DRBELL:
                TTdoBell(n) ;
                break ;
            case DRRETURN:
                return (n) ? DRRRETURN:FALSE ;
            }
        }
        else
        {
elif_jump:
            if(derType & 0x10)
            {
                if(macarg(tkn) != TRUE)
                    return FALSE ;
                if(!meAtol(tkn))
                    derType = (derType | 0x80)+2 ;
            }
            if(f && (derType & 0x80))
            {
                /* FAILED IFs (jumped ELIF), WHILEs, all ELSEs ELIFs */
                if(derType & 0x01)
                    /* FAILED WHILEs & ELSE (skips DONE & ENDIF) */
                    n++ ;
                else if(derType & 0x02)
                    /* FAILED IFs (jumped ELIF) */
                    execlevel++ ;
                relJumpTo = n ;
                return DRRJUMP ;
            }
            if(derType & 0x08)
            {
                if((derType & 0x90) == 0x90) 
                    execlevel += 1 + (derType & 0x01) ;
                else if((derType & 0xc0) == 0xc0) 
                    execlevel += 2 ;
                return TRUE ;
            }
            if(f && (derType & 0x20))
            {
                if(derType & 0x40)
                {
                    relJumpTo = n ;
                    return DRRJUMP ;
                }
                return TRUE ;
            }
            return derType ;
        }
        return TRUE ;
    }
    else if(status == TKCMD)
    {
        register int idx ; /* index to function to execute */
        
        if(execlevel)
            return TRUE ;

        /* and match the token to see if it exists */
        if ((idx = decode_fncname(tkn,0)) < 0)
            return(FALSE);
        if(nmacro)
            clexec = FALSE ;
        cmdstatus = (execFunc(idx,f,n) == TRUE) ;	/* call the function */
        clexec = TRUE ;
        if(cmdstatus == TRUE)
            status = TRUE ;
        else if(TTbreakFlag)
        {
            if(force > 1)
            {
                /* A double !force - macro says \CG is okay, 
                 * remove all input (must remove the \CG) and rest flag */
                TTinflush() ;
                TTbreakFlag = 0 ;
                status = TRUE ;
            }
            else
                status = FALSE ;
        }
        else if(force)
            status = TRUE ;
        else
            status = FALSE ;
        return status ;
    }
    
    {
        char *tmp ;
        
        if((tmp = getval(tkn)) == abortm)
            return FALSE ;
        n = meAtoi(tmp) ;
        f = TRUE;
        goto try_again;
    }
}

/*	dobuf:	execute the contents of the buffer pointed to
		by the passed BP				*/
/* bp - buffer to execute */
char *curFuncName=errorm ;
static char __dobufStr1[]="(^G)Abort, (^L)redraw, (V)ariable, (!)Continue, (S)tep, <any>=next ?" ;
static LINE *errorLine=NULL ;

static int
dobuf(LINE *hlp)
{
    int   status;	        /* status return */
    LINE *lp;		        /* pointer to line to execute */
    LINE *wlp;                  /* line to while */
    LINE *rlp;		        /* line to repeat */
    char *tline;		/* Temp line */
    char  c;			/* Temp charcter */
#if	DEBUGM
    signed char debug=0;
#endif
    
    clexec = TRUE;			/* in cline execution */
    execstr = NULL ;
    execlevel = 0;			/* Reset execution level */
    
    /* starting at the beginning of the buffer */
    wlp = NULL;
    rlp = NULL;
    lp = hlp->l_fp;
    while (lp != hlp)
    {
        tline = (char *) lp->l_text;
#if	DEBUGM
        /* if $debug == TRUE, every line to execute
           gets echoed and a key needs to be pressed to continue
           ^G will abort the command */
        
        if(macbug > 0)
        {
            if((macbug > 1) || (macbug && !execlevel))
            {
                LINE *tlp=hlp ;
                int  lno=0 ;
                char outline[MAXBUF];   /* string to hold debug line text */
                
                /*---	Generate the debugging line for the 'programmer' !! */
                do 
                    lno++ ;
                while ((tlp=lforw(tlp)) != lp) ;
                sprintf(outline,"%s:%d:%d [%s] ?",curFuncName,lno,execlevel,tline) ;
loop_round2:
                mlwrite(MWSPEC,outline);	/* Write out the debug line */
                /* Cannot do update as if this calls a macro then
                 * we will end up in an infinite loop, the best we can do is
                 * call the screenUpdate function
                 */
                screenUpdate(TRUE,sgarbf) ;
                /* reset garbled status */
                sgarbf = FALSE ;
loop_round:
                /* and get the keystroke */
            
                switch (tgetc())
                {
                case 7:			/* Abort */
                    /* Must exit the usual way so that macro storing and execlevel
                     * are corrected */
                    ctrlg(FALSE,1) ;
                    status = ABORT ;
                    errorLine = lp ;
                    goto dobuf_exit ;
                case '?':
                    mlwrite(MWSPEC,__dobufStr1) ;	/* Write out the debug line */
                    goto loop_round ;
                case 'l'-'@':
                    goto loop_round2 ;
                case 'v':
                    {
                        unsigned char mlStatusStore ;
                        mlStatusStore = mlStatus ;
                        mlStatus = 0 ;
                        clexec = FALSE ;
                        descVariable(FALSE,1) ;
                        clexec = TRUE ;
                        mlStatus = mlStatusStore ;
                        goto loop_round ;
                    }
                default:
                    debug = macbug ;
                case '!':
                    macbug = 0 ;
                case 's':
                    break;
                }
            }
        }
#endif
        if(TTbreakTest(0))
        {
            mcStore = FALSE;
            lpStore = NULL;
            execlevel = 0 ;
            status = ctrlg(FALSE,1) ;
        }
        else
        {
            if(mcStore)
                status = domstore(tline) ;
            else
                status = docmd(tline) ;
#if	DEBUGM
            if(debug)
            {
                macbug = debug ;
                debug = 0 ;
            }
#endif
            switch(status)
            {
            case DRRGOTO:
                {
                    register int linlen;	/* length of line to execute */
                    register LINE *glp;		/* line to goto */
                    
                    /* if it is a !GOTO directive, deal with it */
                    while (*execstr == ' ' || *execstr == '\t')
                        ++execstr;
                    for(linlen = 0; ((c = execstr[linlen]) != '\0') ; linlen++)
                        if (c == '\t' || c == ' ')
                            break;
                    
                    glp = hlp->l_fp;
                    while (glp != hlp)
                    {
                        if (*glp->l_text == '*' &&
                            (strncmp((char *) glp->l_text+1,execstr,linlen) == 0))
                        {
                            lp = glp;
                            status = TRUE;
                            break ;
                        }
                        glp = glp->l_fp;
                    }
                    
                    if (status == DRRGOTO)
                    {
                        mlwrite(MWABORT|MWWAIT,"No such label");
                        status = FALSE;
                        break;
                    }
                    break;
                }
            case DRRREPEAT:			/* REPEAT */
                if (rlp != NULL)
                {
                    mlwrite(MWABORT|MWWAIT,"Nested Repeat");
                    status = FALSE;
                    break;
                }
                rlp = lp;		/* Save line */
                status = TRUE;
                break;
            case DRRUNTIL:			/* TRUE UNTIL */
                if (rlp == NULL)
                {
                    mlwrite(MWABORT|MWWAIT,"No repeat set");
                    status = FALSE;
                    break;
                }
                rlp = NULL;
                status = TRUE;
                break;	
            case DRRUNTILF:			/* FALSE UNTIL */
                if (rlp == NULL)
                {
                    mlwrite(MWABORT|MWWAIT,"No repeat set");
                    status = FALSE;
                    break;
                }
                lp = rlp;		/* Back to 'repeat' */
                status = TRUE;
                break; 
            case DRRCONTIN:
                if (wlp == NULL)
                {
                    mlwrite(MWABORT|MWWAIT,"No while");
                    status = FALSE;
                    break;
                }
                lp = wlp;
                continue;
                
            case DRRDONE:	
                if (wlp == NULL)
                {
                    mlwrite(MWABORT|MWWAIT,"No while");
                    status = FALSE;
                    break;
                }
                lp = wlp;
                wlp = NULL;
                continue;
                
            case DRRWHILE:			
                wlp = lp;
                status = TRUE;
                break;
                
            case DRRJUMP:
                if(relJumpTo < 0)
                    for( ; relJumpTo < 0 ; relJumpTo++)
                        lp = lp->l_bp ;
                else
                    for( ; relJumpTo > 0 ; relJumpTo--)
                        lp = lp->l_fp ;
                continue;
                
            case DRRRETURN:	/* if it is a !RETURN directive...do so */
                status = TRUE ;
                goto dobuf_exit ;
            }
        }
        /* check for a command error */
        if (status != TRUE)
        {
            /* in any case set the buffer . */
            errorLine = lp;
            goto dobuf_exit ;
        }
        
        /* on to the next line */
        lp = lp->l_fp;
    }
    if(mcStore)
        mlwrite(MWABORT|MWWAIT,"[Missing !emacro termination]");
    else
        status = TRUE ;

    /* exit the current function */
dobuf_exit:
    
    if(mcStore)
    {
        mcStore = 0 ;
        lpStore = NULL ;
        execlevel = 0 ;
        errorLine = hlp;
        status = FALSE ;
    }

    return status ;
}


static int
donbuf(LINE *hlp, meVARLIST *varList, int f, int n)
{
    meREGISTERS  rp ;
    unsigned char oldcle ;
    int oldexec, status ;
    
    /* save the arguments */
    oldcle = clexec;
    oldexec = execlevel ;
    /* Setup these argument and move the register stack */
    rp.prev = meRegCurr ;
    rp.execstr = execstr ;
    rp.varList = varList ;
    rp.f = thisF ;
    rp.n = thisN ;
    meRegCurr = &rp ;
    thisF = f ;
    thisN = n ;
    
    status = dobuf(hlp) ;
    
    /* restore old arguments & stack */
    meRegCurr = rp.prev ;
    execstr = rp.execstr ;
    thisF = rp.f ;
    thisN = rp.n ;
    execlevel = oldexec ;
    clexec = oldcle;
    
    cmdstatus = status ;
    return status ;
}

int
execFunc(int index, int f, int n)
{
    register int status;
    
    if(index < CK_MAX)
    {
        thisflag = 0 ;
        status = cmdTable[index]->func(f,n) ;
        lastflag = thisflag;
        if(selhilight.uFlags)
        {
            if((status == TRUE) || isComSelIgnFail(index))
            {
                if(isComSelKill(index))
                {
                    selhilight.bp = NULL ;
                    selhilight.flags = 0 ;
                }
                else
                {
                    if(isComSelStart(index))
                    {
                        selhilight.bp = curbp;                  /* Select the current buffer */
                        selhilight.flags = SELHIL_ACTIVE|SELHIL_CHANGED ;
                    }
                    if(selhilight.flags & SELHIL_FIXED)
                    {
                        if(!(selhilight.flags & SELHIL_KEEP) && isComSelDelFix(index))
                            selhilight.flags &= ~SELHIL_ACTIVE ;
                    }
                    else if(selhilight.flags & SELHIL_ACTIVE)
                    {
                        if(isComSelSetMark(index) &&
                           ((selhilight.mlineoff != curwp->w_marko) ||
                            (selhilight.mlineno != curwp->mlineno)))
                        {
                            selhilight.flags |= SELHIL_CHANGED ;
                            selhilight.mlineoff = curwp->w_marko;   /* Current mark offset */
                            selhilight.mlineno = curwp->mlineno;    /* Current mark line number */
                        }
                        if(isComSelSetDot(index) &&
                           ((selhilight.dlineoff != curwp->w_doto) ||
                            (selhilight.dlineno != curwp->line_no)))
                        {
                            selhilight.flags |= SELHIL_CHANGED ;
                            selhilight.dlineoff = curwp->w_doto;   /* Current mark offset */
                            selhilight.dlineno = curwp->line_no;    /* Current mark line number */
                        }
                        if(isComSelSetFix(index))
                        {
                            selhilight.flags |= SELHIL_FIXED ;
                            if(selhilight.uFlags & SELHILU_KEEP)
                                selhilight.flags |= SELHIL_KEEP ;
                        }
                        if(isComSelStop(index))
                            selhilight.flags &= ~SELHIL_ACTIVE ;
                        if(selhilight.flags & SELHIL_CHANGED)
                            addModeToBufferWindows(selhilight.bp, WFSELHIL);
                    }
                }
            }
            else if((selhilight.flags & (SELHIL_ACTIVE|SELHIL_KEEP)) == SELHIL_ACTIVE)
                selhilight.flags &= ~SELHIL_ACTIVE ;
        }
    }
    else
    {
        meMACRO *mac ;
        LINE    *hlp ;
        char *oldFuncName ;
        char firstExec ;                    /* set if this is first */
        
        mac = getMacro(index) ;
        hlp = mac->hlp ;
        if(!(hlp->l_flag & MACEXEC))
        {
            if(hlp->l_flag & MACFILE)
            {
                if(mac->fname != NULL)
                    dofile(mac->fname,0,1) ;
                else
                    dofile(mac->name,0,1) ;
                hlp = mac->hlp ;
                if(hlp->l_flag & MACFILE)
                    return mlwrite(MWABORT,"[file-macro %s not defined]",mac->name);
            }
            hlp->l_flag |= MACEXEC ;
            firstExec = 1 ;
        }
        else
            firstExec = 0 ;
        
        oldFuncName = curFuncName ;
        curFuncName = mac->name ;
        
        status = donbuf(hlp,&mac->varList,f,n) ;
        
        curFuncName = oldFuncName ;
        
        if(firstExec)
        {
            if(mac->hlp != hlp)
                freeLineLoop(hlp,1) ;
            else
                hlp->l_flag &= ~MACEXEC ;
        }
    }
    
    return status ;
}

/* Execute the given command, but is a hidden way, i.e.
 * the 'key' is not recorded
 */
void
execFuncHidden(int keyCode, int index, unsigned int arg)
{
    int tc, ti ;
    unsigned char tf, lf, cs;
    int f, n ;
    
    cs = cmdstatus ;
    tc = thisCommand ;
    ti = thisIndex ;
    tf = thisflag ;
    lf = lastflag ;
    thisCommand = keyCode ;
    thisIndex = index ;
    if(curbp->inputFunc >= 0)
        index = curbp->inputFunc ;
    if(arg != 0)
    {
        f = 1 ;
        n = (int) (arg + 0x80000000) ;
    }
    else
    {
        f = 0 ;
        n = 1 ;
    }
    execFunc(index,f,n);
    thisflag = tf;
    lastflag = lf;
    thisIndex = ti ;
    thisCommand = tc ;
    cmdstatus = cs ;
}

void
execBufferFunc(BUFFER *bp, int index, int flags, int n)
{
    int movePos = ((flags & meEBF_HIDDEN) && (bp == curbp)) ; 
    int tc, ti ;
    unsigned char tf, lf, cs;
    
    cs = cmdstatus ;
    tc = thisCommand ;
    ti = thisIndex ;
    tf = thisflag ;
    lf = lastflag ;
    thisIndex = index ;
    if(movePos)
    {
        /* drop an alpha mark and return back to it after */
        alphaMarkSet(curbp,meAM_EXECBUFF,curwp->w_dotp,curwp->w_doto,1) ;
        curwp->w_dotp = curbp->b_dotp ;
        curwp->w_doto = curbp->b_doto ;
        curwp->line_no = curbp->line_no ;
    }
    if(bp != curbp)
    {
        /* swap this buffer into the current window, execute the function
         * and then swap back out */
	register BUFFER *tbp = curbp ;
        unsigned long w_flag;
        long topLineNo ;
	
	storeWindBSet(tbp,curwp) ;
        w_flag = curwp->w_flag ;
        topLineNo = curwp->topLineNo ;
        
        tbp->b_nwnd-- ;
	curbp = curwp->w_bufp = bp ;
	restoreWindBSet(curwp,bp) ;
        isWordMask = bp->isWordMask ;
        bp->b_nwnd++ ;
        
	execFunc(index,(flags & meEBF_ARG_GIVEN),n) ;
	
        bp->b_nwnd-- ;
	storeWindBSet(bp,curwp) ;
	curbp = curwp->w_bufp = tbp ;
        tbp->b_nwnd++ ;
        curwp->topLineNo = topLineNo ;
        /* force a check on update, just incase the Func has done something behind our back */
        curwp->w_flag |= (w_flag|WFMOVEL|WFMAIN) ;
	restoreWindBSet(curwp,tbp) ;
        isWordMask = tbp->isWordMask ;
    }
    else
	execFunc(index,(flags & meEBF_ARG_GIVEN),n) ;
    if(movePos)
    {
        /* return back to the alpha mark */
        alphaMarkGet(curbp,meAM_EXECBUFF) ;
        curwp->w_dotp = curbp->b_dotp ;
        curwp->w_doto = curbp->b_doto ;
        curwp->line_no = curbp->line_no ;
        curwp->w_flag |= WFMOVEL ;
    }
    thisflag = tf;
    lastflag = lf;
    thisIndex = ti ;
    thisCommand = tc ;
    cmdstatus = cs ;
}

long
macroPrintError(LINE *hlp, char *macName)
{
    LINE *lp ;
    long  ln ;
    
    if(errorLine == NULL)
        /* Quit if there is no error line */
        return -1 ;
    
    lp=lforw(hlp) ;
    ln=0 ;
    while(lp != errorLine)
    {
        lp = lforw(lp) ;
        ln++ ;
    }
    mlwrite(MWABORT|MWWAIT,"[Error executing %s, line %d]",macName,ln+1) ;
    return ln ;
}

/*	execBuffer:	Execute the contents of a buffer of commands	*/
int
execBuffer(int f, int n)
{
    register BUFFER *bp;		/* ptr to buffer to execute */
    register int s;			/* status return */
    meVARLIST varList={NULL,0} ;
    char *oldFuncName ;
    char  bufn[MAXBUF];   		/* name of buffer to execute */
    long  ln ;

    /* find out what buffer the user wants to execute */
    if((s = getBufferName("Execute buffer", 0, 1, bufn)) != TRUE)
        return s ;
    
    /* find the pointer to that buffer */
    if ((bp=bfind(bufn, 0)) == NULL)
        return mlwrite(MWABORT,"No such buffer");

    oldFuncName = curFuncName ;
    curFuncName = bp->b_bname ;
    
    /* and now execute it as asked */
    if(((s = donbuf(bp->b_linep,&varList,f,n)) != TRUE) &&
       ((ln = macroPrintError(bp->b_linep,bufn)) >= 0))
    {
        /* the execution failed lets go to the line that caused the grief */
        bp->b_dotp  = errorLine ;
        bp->line_no = ln ;
        bp->b_doto  = 0 ;
        /* look if buffer is showing */
        if(bp->b_nwnd)
        {
            register WINDOW *wp;	/* ptr to windows to scan */
            
            wp = wheadp;
            while (wp != NULL)
            {
                if (wp->w_bufp == bp)
                {
                    /* and point it */
                    wp->w_dotp = errorLine ;
                    wp->w_doto = 0 ;
                    wp->line_no = ln ;
                    wp->w_flag |= WFMOVEL ;
                }
                wp = wp->w_wndp;
            }
        }
    }
    curFuncName = oldFuncName ;
    /* free off any command variables */
    if(varList.head != NULL)
    {
        meVARIABLE *cv=varList.head, *ncv ;
        while(cv != NULL)
        {
            ncv = cv->next ;
            meNullFree(cv->value) ;
            free(cv) ;
            cv = ncv ;
        }
    }
    return s ;
}

/*	dofile:	yank a file into a buffer and execute it
		if there are no errors, delete the buffer on exit */
/* Note for Dynamic file names
 * This ensures that nothing is done to fname
 */
int
dofile(char *fname, int f, int n)
{
    char          fn[FILEBUF] ;
    meVARLIST     varList={NULL,0} ;
    char         *oldFuncName ;
    LINE          hlp ;
    register int  status ;	/* results of various calls */
    
    hlp.l_fp = &hlp ;
    hlp.l_bp = &hlp ;
    /* use a new buffer to ensure it doesn't mess with any loaded files */
    if(!fileLookup(fname,".emf",meFL_CHECKDOT|meFL_USESRCHPATH,fn) ||
       (ffReadFile(fn,READ_SILENT,NULL,&hlp) == ABORT))
        return mlwrite(MWABORT|MWCLEXEC,"[Failed to load file %s]", fname);
    
    /* go execute it! */
    oldFuncName = curFuncName ;
    curFuncName = fname ;
    if((status = donbuf(&hlp,&varList,f,n)) != TRUE)
        /* the execution failed lets go to the line that caused the grief */
        macroPrintError(&hlp,fname) ;
    freeLineLoop(&hlp,0) ;
    curFuncName = oldFuncName ;
    /* free off any command variables */
    if(varList.head != NULL)
    {
        meVARIABLE *cv=varList.head, *ncv ;
        while(cv != NULL)
        {
            ncv = cv->next ;
            meNullFree(cv->value) ;
            free(cv) ;
            cv = ncv ;
        }
    }
    
    return status ;
}

int
execFile(int f, int n)	/* execute a series of commands in a file */
{
    int  status;
    char filename[MAXBUF];	/* Filename */

    if ((status=mlreply("Execute File",MLFILECASE, 0, filename, 
                        MAXBUF)) != TRUE)
        return status ;
    
    /* otherwise, execute it */
    return dofile(filename,f,n) ;
}

/* execute the startup file */
/* sfname	name of startup file (null if default) */

int
startup(char *sfname)
{
    /* look up the startup file */
    if(sfname == NULL)
        sfname = "me" ;

    return dofile(sfname,0,1) ;
}


/* execCommand:	execute a named command even if it is not bound */

int
execCommand(int f, int n)
{
    register int idx ;	     /* index to the requested function */
    char buf[MAXBUF];        /* buffer to hold tentative command name */
    char prm[30] ;
    
    /* setup prompt */
    if(f==TRUE)
        sprintf(prm,"Arg %d: esc x ",n) ;
    else
        strcpy(prm,"esc x ") ;
    
    /* if we are executing a command line get the next arg and match it */
    if((idx = mlreply(prm, MLCOMMAND, 1, buf, MAXBUF)) != TRUE)
        return idx ;
    
    /* decode the function name, if -ve then duff */
    if((idx = decode_fncname(buf,0)) < 0)
        return(FALSE);
    
    /* and then execute the command */
    return execFunc(idx,f,n) ;
}


/*	execLine:	Execute a command line command to be typed in
			by the user					*/

/*ARGSUSED*/
int
lineExec (int f, int n, char *cmdstr)
{
    register int status;		/* status return */
    char *oldestr;		/* original exec string */
    unsigned char oldcle ;      /* old contents of clexec flag */
    int oldexec, oldF, oldN ;
    
    /* save the arguments */
    oldcle  = clexec;
    oldexec = execlevel ;
    oldestr = execstr;
    oldF    = thisF ;
    oldN    = thisN ;
    
    clexec = TRUE;			/* in cline execution */
    execstr = NULL ;
    execlevel = 0;			/* Reset execution level */
    thisF = f ;
    thisN = n ;
    
    status = docmd(cmdstr) ;
    
    /* restore old arguments */
    execlevel = oldexec ;
    clexec    = oldcle;
    execstr   = oldestr ;
    thisF     = oldF ;
    thisN     = oldN ;
    
    return status ;
}    


int
execLine(int f, int n)
{
    register int status;		/* status return */
    char cmdstr[MAXBUF];		/* string holding command to execute */
    
    /* get the line wanted */
    if ((status = mlreply(":", 0, 0, cmdstr, MAXBUF)) != TRUE)
        return status ;

    return lineExec(f,n,cmdstr) ;
}

