/*****************************************************************************
*
*	Title:		%M%
*
*	Synopsis:	File handling.
*
******************************************************************************
*
*	Filename:		%P%
*
*	Author:			Danial Lawrence
*
*	Creation Date:		14/05/86 12:37		<990702.2105>
*
*	Modification date:	%G% : %U%
*
*	Current rev:		%I%
*
*	Special Comments:
*
*	Contents Description:
*
*	The routines in this file handle the reading, writing
*	and lookup of disk files.  All of details about the
*	reading and writing of the disk are in "fileio.c".
*
* Jon Green	13-07-90
* Added Document Mode. Search for .doc or .txt extension for WRAP and
* indent to automatically be enabled.
*
* Jon Green	17-05-91
* Added suffix modes so that the suffix defintions may be defined in the
* start up file.
*****************************************************************************
* 
* Modifications to the original file by Jasspa. 
* 
* Copyright (C) 1988 - 1999, JASSPA 
* The MicroEmacs Jasspa distribution can be copied and distributed freely for
* any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
* incorporated into commercial software with the expressed permission of
* JASSPA.
* 
****************************************************************************/

/*---	Include defintions */

#define __FILEC 1                  /* Define the filename */

/*---	Include files */

#include "emain.h"
#include "efunc.h"
#if (defined _UNIX) || (defined _DOS) || (defined _WIN32)
#include <errno.h>
#include <limits.h>                     /* Constant limit definitions */
#include <sys/types.h>
#include <sys/stat.h>
#ifdef _WIN32 
#include <direct.h>                     /* Directory entries */
#elif (defined _SYSV42)
#include <dirent.h>
#else
#include <sys/dir.h>
#endif
#include <time.h>
#endif

#ifdef _DOS
#include <sys/file.h>
#include <dir.h>
#include <errno.h>
#include <pc.h>
#include <dos.h>

/* attribute stuff */
#define	A_RONLY		0x01
#define	A_HIDDEN	0x02
#define	A_SYSTEM	0x04
#define	A_LABEL		0x08
#define	A_DIR		0x10
#define	A_ARCHIVE	0x20

/* dos call values */
#define	DOSI_GETDRV	0x19
#define	DOSI_SDTA	0x1A
#define	DOSI_GDTA	0x2F
#define DOSI_GETDIR     0x47
#define	DOSI_FINDF	0x4E
#define	DOSI_FINDN	0x4F

typedef struct
{
    char d_buf[21];
    char d_attrib;
    unsigned short d_time;
    unsigned short d_date;
    long d_size;
    char d_name[13];
} Dta_buf;

#endif  /* _DOS */


/*
 * Return:
 * 0 - If file is nasty
 * 1 - If file is regular
 * 2 - If file is a directory
 * 3 - If file doesn't exist
 * 5 - If file is a url
 */
int
getFileStats(char *file, int flag, meSTAT *stats, char *lname)
{
    if(lname != NULL)
        *lname = '\0' ;
    /* setup the default stat values */
    if(stats != NULL)
    {
        stats->stmtime= -1 ;
        stats->stsize = 0 ;
#ifdef _UNIX
        stats->stuid  = meUid ;
        stats->stgid  = meGid ;
        stats->stdev  = -1 ;
        stats->stino  = -1 ;
#endif
    }
    if(!strncmp(file,"http://",7) || !strncmp(file,"ftp://",6))
    {
        return 5 ;
    }
#ifdef _DOS
    {
        union REGS reg ;		/* cpu register for use of DOS calls */
        int        len ;
        
        if(((len = strlen(file)) == 0) ||
           (strchr(file,'*') != NULL) ||
           (strchr(file,'?') != NULL))
        {
            if(flag & 1)
                mlwrite(MWABORT|MWPAUSE,"[%s illegal name]", file);
            return 0 ;
        }
        if((file[len-1] == DIR_CHAR) || ((len == 2) && (file[1] == _DRV_CHAR)))
            goto gft_directory ;
        reg.x.ax = 0x4300 ;
        reg.x.dx = ((unsigned long) file) ;
        intdos(&reg, &reg);
        
        if(reg.x.cflag)
            return 3 ;
        if(stats != NULL)
        {
            struct ffblk fblk ;
            stats->stmode = (reg.x.cx|0x020) ;
            if(!findfirst(file, &fblk, FA_RDONLY|FA_HIDDEN|FA_SYSTEM|FA_DIREC))
            {
                stats->stmtime = (((unsigned long) fblk.ff_ftime) & 0x0ffff) | (((unsigned long) fblk.ff_fdate) << 16) ;
                stats->stsize = fblk.ff_fsize ;
            }
        }
        if(reg.x.ax & 0x010)
        {
gft_directory:
            if(flag & 2)
                mlwrite(MWABORT|MWPAUSE,"[%s directory]", file) ;
            return 2 ;
        }
        return 1 ;
    }
#else
#ifdef _WIN32
    {
        DWORD status;
        int   len ;
        
        if(((len = strlen(file)) == 0) || 
           (strchr(file,'*') != NULL) || (strchr(file,'?') != NULL))
        {
            if(flag & 1)
                mlwrite(MWABORT|MWPAUSE,"[%s illegal name]", file);
            return 0 ;
        }
        if((file[len-1] == DIR_CHAR) || ((len == 2) && (file[1] == _DRV_CHAR)))
            goto gft_directory ;
        
        if(stats != NULL)
        {
            HANDLE          *fh ;
            WIN32_FIND_DATA  fd ;
            
            fh = FindFirstFile(file,&fd) ;
            if(fh == INVALID_HANDLE_VALUE)
                return 3 ;
            status = fd.dwFileAttributes ;
            /* Dangerously ingoring the nFileSizeHigh for now */
            stats->stsize = fd.nFileSizeLow ;
            stats->stmtime = ((fd.ftLastWriteTime.dwLowDateTime  >> 24) & 0x000000ff) |
                      ((fd.ftLastWriteTime.dwHighDateTime <<  8) & 0x7fffff00) ;
            FindClose(fh) ;
            stats->stmode = (unsigned short) status | FILE_ATTRIBUTE_ARCHIVE ;
        }
        else if((status = GetFileAttributes(file)) == 0xFFFFFFFF)
            return 3 ;
        if (status & FILE_ATTRIBUTE_DIRECTORY)
        {
gft_directory:
            if(flag & 2)
                mlwrite(MWABORT|MWPAUSE,"[%s directory]", file);
            return 2 ;
        }
#if 0
        /* FILE_ATTRIBUTE_SYSTEM
         * The file or directory is part of, or is used exclusively by,
         * the operating system. */
        else if (status & FILE_ATTRIBUTE_SYSTEM)
        {
            if(flag & 1)
                mlwrite(MWABORT|MWPAUSE,"[%s operating system file]", file);
            return 0 ;
        }
#endif
        /* FILE_ATTRIBUTE_ARCHIVE
         * The file or directory is an archive file or directory.
         * Applications use this flag to mark files for backup or removal.
         * 
         * FILE_ATTRIBUTE_HIDDEN
         * The file or directory is hidden.
         * It is not included in an ordinary directory listing.
         * 
         * FILE_ATTRIBUTE_READONLY
         * The file or directory is read-only.
         * Applications can read the file but cannot write to it or delete it.
         * In the case of a directory, applications cannot delete it.
         * 
         * FILE_ATTRIBUTE_SYSTEM
         * The file or directory is part of, or is used exclusively by,
         * the operating system.
         * 
         * FILE_ATTRIBUTE_NORMAL
         * The file or directory has no other attributes set.
         * This attribute is valid only if used alone. */
        return 1 ;
    }
#else
#ifdef _UNIX
    {
        struct stat statbuf;
        long stmtime = -1 ;
        
        if((lname == NULL) && (stats == NULL))
        {
            if(stat(file, &statbuf) != 0)
                return 3 ;
        }
        else if(lstat(file, &statbuf) != 0)
            return 3 ;
        else if(S_ISLNK(statbuf.st_mode))
        {
            char lbuf[FILEBUF], buf[FILEBUF], *ss ;
            int  ii, jj ;
            
            ii = strlen(file) ;
            strncpy(lbuf,file,ii) ;
            do {
                if(file[ii-1] == DIR_CHAR)
                    ii-- ;
                lbuf[ii] = '\0' ;
                if(statbuf.st_mtime > stmtime)
                    stmtime = statbuf.st_mtime ;
                if((jj=readlink(lbuf, buf, FILEBUF)) <= 0)
                {
                    if(flag & 1)
                        mlwrite(MWABORT|MWPAUSE,"[%s symbolic link problems]", file);
                    return 0 ;
                }
                buf[jj] = '\0' ;
                if((buf[0] == DIR_CHAR) || ((ss=strrchr(lbuf,DIR_CHAR)) == NULL))
                    strcpy(lbuf,buf) ;
                else
                {
                    ss++ ;
                    ii = ((int) ss) - ((int) lbuf) ;
                    strcpy(ss,buf) ;
                    ii += jj ;
                }
            } while(((jj=lstat(lbuf, &statbuf)) == 0) && S_ISLNK(statbuf.st_mode)) ;
            if(lname != NULL)
                fileNameCorrect(lbuf,lname,NULL) ;
            if(jj)
            {
                if(flag & 1)
                    mlwrite(MWABORT|MWPAUSE,"[%s dangling symbolic link]",lbuf);
                if(stats != NULL)
                    stats->stmtime = stmtime ;
                return 3 ;
            }
        }
        
        if(stats != NULL)
        {
            stats->stmode = statbuf.st_mode ;
            stats->stuid  = statbuf.st_uid ;
            stats->stgid  = statbuf.st_gid ;
            stats->stdev  = statbuf.st_dev ;
            stats->stino  = statbuf.st_ino ;
            stats->stsize = statbuf.st_size ;
            if(statbuf.st_mtime > stmtime)
                stats->stmtime= statbuf.st_mtime ;
            else
                stats->stmtime= stmtime ;
        }
        if(S_ISREG(statbuf.st_mode))
        {
            return 1 ;
        }
        if(S_ISDIR(statbuf.st_mode))
        {
            if(flag & 2)
                mlwrite(MWABORT|MWPAUSE,"[%s directory]", file);
            return 2 ;
        }
        if(flag & 1)
        {
#ifdef	S_IFIFO
            if(S_ISFIFO(statbuf.st_mode))
                mlwrite(MWABORT|MWPAUSE,"[%s is a FIFO]", file);
#endif
#ifdef	S_IFCHR
            else if(S_ISCHR(statbuf.st_mode))
                mlwrite(MWABORT|MWPAUSE,"[%s is character special]", file);
#endif
#ifdef	S_IFBLK
            else if(S_ISBLK(statbuf.st_mode))
                mlwrite(MWABORT|MWPAUSE,"[%s is block special]", file);
#endif
            else
                mlwrite(MWABORT|MWPAUSE,"[%s not a regular file]", file);
        }
        return 0 ;
    }
#else
    return 1 ;
#endif /* _UNIX */
#endif /* _WIN32 */
#endif /* _DOS */
}

int
fnamecmp(char *f1, char *f2)
{
    if((f1==NULL) || (f2==NULL))
        return 1 ;
#ifdef _INSENSE_CASE
    return icstrcmp(f1,f2) ;
#else
    return strcmp(f1,f2) ;
#endif
}

/*---	Set up directories. */

void	
set_dirs(void)
{
#if (defined _UNIX) || (defined _DOS) || (defined _WIN32)
#ifdef _UNIX
    static char spath[2]= { PATHCHR, '\0' } ;
    static char lpath[23] = { PATHCHR,'/','u','s','r','/','l','o','c','a','l','/','m','i','c','r','o','e','m','a','c','s','\0' } ;
#endif    
    char	*ss;				/* Home directory */
    
    ss = getenv("HOME");			/* Get home directory */
    if (ss != NULL)			        /* Null ?? */
        homedir = strdup(ss) ;
    else
        homedir = NULL ;			/* Set to NULL */
#ifdef _UNIX
    ss = getenv("MEPATH");			/* Get home directory */
    if(ss == NULL)
    {
        ss = getenv("PATH");			/* Get home directory */
        if(ss == NULL)
            searchPath = strdup(lpath+1) ;
        else
            searchPath = nstrcat(3,lpath+1,spath,ss);
    }
    else
        searchPath = nstrcat(2,ss,lpath);
#else
    ss = getenv("MEPATH");			/* Get home directory */
    if(ss == NULL)
        ss = getenv("PATH");			/* Get home directory */
    if(ss == NULL)
        searchPath = NULL ;
    else
        searchPath = strdup(ss);
#endif
#else
    homedir = NULL ;			        /* Set to NULL */
    searchPath = NULL ;		        /* Set to NULL */
#endif
#ifdef _UNIX
    {
        char *pwd;
        struct stat dotstat, pwdstat;

        /* If PWD is accurate, use it instead of calling gwd.
         * This solves problems with bad ~/ as the shell will
         * store as /user/.... etc.
         */
        if(((pwd = getenv ("PWD")) != NULL) &&
           (pwd[0] == DIR_CHAR) && (stat(pwd,&pwdstat) == 0) &&
           (stat(".", &dotstat) == 0) && (dotstat.st_ino == pwdstat.st_ino) &&
           (dotstat.st_dev == pwdstat.st_dev))
        {
            int  ll=strlen(pwd) ;
            char cc, *dd ;
            
            curdir = dd = meMalloc(ll+2) ;
            while((cc=*pwd++) != '\0')
                *dd++ = cc ;
            if(dd[-1] != DIR_CHAR)
                *dd++ = DIR_CHAR ;
            *dd = 0 ;
        }
        else
            curdir = gwd(0) ;
    }
#else    
    curdir = gwd(0) ;
#endif
    if(curdir == NULL)
    {
        printf("Failed to get cwd\n") ;
        meExit(1);
    }
}	/* End of "set_dirs" () */

/*	Look up the existance of a file along the normal or PATH
	environment variable. Look first in the HOME directory if
	asked and possible
*/

int
fileLookup(char *fname, char *ext, unsigned char flags, char *outName)
{
    register char *path;	/* environmental PATH variable */
    register char *sp;	/* pointer into path spec */
    register int   ii; 	/* index */
    char nname[FILEBUF] ;
    char buf[FILEBUF] ;
 
    if(ext != NULL)
    {
        if((sp=strrchr(fname,DIR_CHAR)) == NULL)
            sp = fname ;
        /* Jon 990608 - If the extension is the same then we do not
         * concatinate it, otherwise we do !!. Compare the complete extension
         * with the end of the string, allowing double extensions to be
         * handled properly. */
        if (((ii = strlen (sp) - strlen (ext)) < 0) ||
 #ifdef _INSENSE_CASE
              icstrcmp(&sp[ii], ext)
 #else
              strcmp(&sp[ii], ext)
 #endif
            )
        {
            strcpy(nname,fname) ;
            strcat(nname,ext);
            fname = nname ;
        }
    }
    if(flags & meFL_CHECKDOT)
    {
        fileNameCorrect(fname,outName,NULL) ;
#ifdef _UNIX
        if(flags & meFL_EXEC)
        {
            if(!meTestExec(outName) &&
               (getFileStats(outName,0,NULL,NULL) == 1))
            return 1 ;
        }
        else
#endif
            if(!meTestExist(outName))
                return 1 ;
    }
    if(flags & meFL_USEPATH)
        path=getenv("PATH") ;
    else if(flags & meFL_USESRCHPATH)
        path=searchPath ;
    else
        return 0 ;
    while(path != NULL)
    {
        /* build next possible file spec */
        sp = path ;
        if((path = strchr(++path,PATHCHR)) != NULL)
        {
            ii = path++ - sp ;
            strncpy(buf,sp,ii) ;
        }
        else
        {
            strcpy(buf,sp) ;
            ii = strlen(buf) ;
        }
        /* Check for zero length path */
        if (ii <= 0)
            continue;
        /* Add a directory separator if missing */
        if (buf[ii-1] != DIR_CHAR)
            buf[ii++] = DIR_CHAR ;
        strcpy(buf+ii,fname) ;               /* Concatinate the name */
        
        /* and try it out */
        fileNameCorrect(buf,outName,NULL) ;
#ifdef _UNIX
        if(flags & meFL_EXEC)
        {
            if(!meTestExec(outName) &&
               (getFileStats(outName,0,NULL,NULL) == 1))
            return 1 ;
        }
        else
#endif
            if(!meTestExist(outName))
                return 1 ;
    }
    return 0 ;	/* no such luck */
}


int
executableLookup(char *fname, char *outName)
{
#if (defined _WIN32) || (defined _DOS)
    char *ss ;
    int ll=strlen(fname) ;
    int ii ;
    
    
    if(ll > 4)
    {
        ss = fname+ll-4 ;
        ii = noExecExtensions ;
        while(--ii >= 0)
#ifdef _INSENSE_CASE
            if(!icstrncmp(ss,execExtensions[ii],4))
#else
            if(!strncmp(ss,execExtensions[ii],4))
#endif
            {
                if(fileLookup(fname,NULL,meFL_CHECKDOT|meFL_USEPATH,outName))
                    return ii+1 ;
                return 0 ;
            }
    }
    ii=noExecExtensions ;
    while(--ii >= 0)
        if(fileLookup(fname,execExtensions[ii],meFL_CHECKDOT|meFL_USEPATH,outName))
            return ii+1 ;
#endif
#ifdef _UNIX
    unsigned char flags ;
    
    flags = (strchr(fname,DIR_CHAR) != NULL) ? meFL_CHECKDOT|meFL_EXEC:meFL_USEPATH|meFL_EXEC ;
    if(fileLookup(fname,NULL,flags,outName))
        return 1 ;
#endif
    return 0 ;
}

int
bufferOutOfDate(BUFFER *bp)
{
    meSTAT stats ;
    if((bp->b_fname != NULL) && getFileStats(bp->b_fname,0,&stats,NULL) &&
       (stats.stmtime > bp->stats.stmtime))
        return TRUE ;
    return FALSE ;
}

/* get current working directory
*/
char *
gwd(char drive)
{
    /*
     * Get the current working directory into a static area and return
     * a pointer to it.
     *
     * This routine accounts for the differences between BSD getwd() and
     * System V getcwd().
     *
     * Return NULL on error or if we dont have a routine for the current
     * system (true for VMS, MS-DOS, etc etc until someone writes them).
     */
#if (defined _BSD) || (defined _UNIX) || (defined _DOS) || (defined _WIN32)

    char dir[MAXBUF] ;
    register int ll ;

#ifdef _BSD
    /* MAXBUF must be > 512 */

    char*  getwd(char *) ;

    getwd(dir) ;

#endif  /* _BSD */
#ifdef _UNIX

/*    char* getcwd APRAM((char *,int)) ;*/

    getcwd(dir, MAXBUF-2) ;

#endif  /* _UNIX */

#ifdef _WIN32
    /* Change drive to the destination drive.
     * remeber so that we can get back. Note we
     * do no test for fails on the drive. We will
     * pick up the current working directory if
     * it fails. */
    if (drive != 0)
    {
        int curDrive;

        curDrive = _getdrive ();
        drive = tolower (drive) - 'a' + 1;
        if ((drive == curDrive) || (_chdrive (drive) != 0))
            drive = 0;             /* Failed drive change or same drive */
        else
            drive = curDrive;      /* Save to restore */
    }

    /* Pick up the directory information */
    GetCurrentDirectory (MAXBUF, dir);
    if (strlen(dir) > 2)
    {
        char *p;                   /* Local character pointer */
    
        /* convert all '\\' to '/' */
        p = dir+2 ;
        while((p=strchr(p,'\\')) != NULL)    /* got a '\\', -> '/' */
            *p++ = DIR_CHAR ;
    }
    /* change the drive back - dont care if it fails cos theres nothing we can do! */
    if (drive != 0)
        _chdrive (drive) ;
#endif /* _WIN32 */

#ifdef _DOS
    union REGS reg ;		/* cpu register for use of DOS calls */

    if(drive == 0)
    {
        reg.h.ah = DOSI_GETDRV ;
        intdos (&reg, &reg);
        drive = reg.h.al + 'a' ;
    }
    else if(drive < 'a')
        drive += 'a' - 'A' ;
    dir[0] = drive ;
    dir[1] = ':' ;
    dir[2] = DIR_CHAR ;

    reg.h.ah = DOSI_GETDIR ;
    reg.h.dl = drive - 'a' + 1 ;
    reg.x.si = (unsigned long) (dir+3) ;
    (void) intdos(&reg, &reg);

#if DIR_CHAR != '/'
    /* convert all '/' to '\' for dos etc */
    p1 = dir+3 ;
    while((p1=strchr(p1,'/')) != NULL)  /* got a '/', -> '\' */
        *p1++ = DIR_CHAR ;
#endif
#endif  /* _DOS */

    if((ll = strlen(dir)) > 0)
    {
        if(dir[ll-1] != DIR_CHAR)
        {
            dir[ll] = DIR_CHAR ;
            dir[ll+1] = '\0' ;
        }
        return strdup(dir) ;
    }

#endif  /* UNKNOWN */
    /*
    ** An unknown system to me - dont know what to do here.
    */
    return NULL ;

}

char *
getFileBaseName(char *fname)
{
    char cc, *p, *p1 ;
    p = p1 = fname ;
    while((cc=*p1++) != '\0')
    {
        if((cc == DIR_CHAR) && (*p1 != '\0'))
            p = p1 ;
    }
    return p ;
}

void
getFilePath(char *fname, char *path)
{
    if(fname != NULL)
    {
        char *ss ;
        strcpy(path,fname) ;
        if((ss=strrchr(path,DIR_CHAR)) == NULL)
            path[0] = '\0' ;
        else
            ss[1] = '\0' ;
    }
    else
        strcpy(path,curdir) ;
}

int
inputFileName(char *str, char *fn, int corFlag)
{
    char tmp[FILEBUF], *buf ;
    int  s ;
    
    buf = (corFlag) ? tmp:fn ;

    getFilePath(curbp->b_fname,buf) ;
    s = mlreply(str,(MLFILECASE|MLNORESET|MLMACNORT), 0, buf, FILEBUF) ;
    if(corFlag && (s == TRUE))
        fileNameCorrect(tmp,fn,NULL) ;
    return s ;
}


#if	CRYPT
int
resetkey(BUFFER *bp)	/* reset the encryption key if needed */
{
    BUFFER *obp ;
    register int s; /* return status */

    /* if we are in crypt mode */
    if(meModeTest(bp->b_mode,MDCRYPT))
    {
        if(bp->b_key == NULL)
        {
            obp = curbp ;
            curbp = bp ;
            s = setCryptKey(FALSE, 0);
            curbp = obp ;
            if (s != TRUE)
                return(s);
        }

        /* and set up the key to be used! */
        meCrypt(NULL, 0);
        meCrypt(bp->b_key, strlen((char *) bp->b_key));

    }

    return(TRUE);
}
#endif

typedef struct FILENODE {
    struct FILENODE *next;              /* Next node in the list */
    char   *fname;                        /* Name of the file */
    char   *lname;                        /* Linkname */
    char    attrib[4] ;
    long    size ;
#ifdef _UNIX
    time_t  mtime ;
#endif
#ifdef _DOS
    time_t  mtime ;
#endif
#ifdef _WIN32
    FILETIME mtime ;
#endif
} FILENODE;

FILENODE *curHead ;

FILENODE *
addFileNode(FILENODE *cfh, FILENODE *cf)
{
    FILENODE *pf, *nf ;
    if(cfh == NULL)
    {
        cf->next = NULL ;
        return cf ;
    }
    pf = NULL ;
    nf = cfh ;
    if(cf->attrib[0] != 'd')
    {
        /* skip all the directories */
        while((nf != NULL) && (nf->attrib[0] == 'd'))
        {
            pf = nf ;
            nf = nf->next ;
        }
        /* now find the correct alpha place */
        while((nf != NULL) && (strcmp(nf->fname,cf->fname) < 0))
        {
            pf = nf ;
            nf = nf->next ;
        }
    }
    else
    {
        /* now find the correct alpha place */
        while((nf != NULL) && (nf->attrib[0] == 'd') &&
              (strcmp(nf->fname,cf->fname) < 0))
        {
            pf = nf ;
            nf = nf->next ;
        }
    }
    cf->next = nf ;
    if(pf == NULL)
        return cf ;
    pf->next = cf ;
    return cfh ;
}

#if (defined _WIN32) || (defined _DOS)
int
meTestExecutable(char *fileName)
{
    int ii ;
    if((ii = strlen(fileName)) > 4)
    {
        char  *ee ;
        ee = fileName+ii-4 ;
        ii = noExecExtensions ;
        while(--ii >= 0)
#ifdef _INSENSE_CASE
            if(!icstrncmp(ee,execExtensions[ii],4))
#else
            if(!strncmp(ee,execExtensions[ii],4))
#endif
                return 1 ;
    }
    return 0 ;
}
#endif

FILENODE *
getDirectoryInfo(char *fname)
{
    char   bfname[MAXBUF];               /* Working file name buffer */
    char  *fn;
    FILENODE *curFile ;
    FILENODE *cfh=NULL ;
    int    noFiles=0;                     /* No files count */
    int    ii;                            /* Local loop counter */
    
#ifdef _UNIX
    DIR    *dirp ;
#if (defined _SYSV42)
    struct  dirent *dp ;
#else
    struct  direct *dp ;
#endif
    struct stat sbuf ;
    char  *ff;
    int    llen;
    
    /* Render the list of files. */
    curHead = NULL ;
    strcpy(bfname,fname) ;
    fn = bfname+strlen(bfname) ;
    
    if((dirp = opendir(fname)) != NULL)
    {
        while((dp = readdir(dirp)) != NULL)
        {
            if(((noFiles & 0x0f) == 0) &&
               ((curHead = (FILENODE *) meRealloc(curHead,sizeof(FILENODE)*(noFiles+16))) == NULL))
                return NULL ;
            curFile = &(curHead[noFiles++]) ;
            if((ff = (char *) strdup(dp->d_name)) == NULL)
                return NULL ;
            curFile->lname = NULL ;
            curFile->fname = ff ;
            strcpy(ff,dp->d_name) ;
            strcpy(fn,dp->d_name) ;
            stat(bfname,&sbuf) ;
            if(sbuf.st_uid != meUid)
            {
                /* remove the user modes */
                sbuf.st_mode &= ~00700 ;
                if(sbuf.st_gid != meGid)
                    /* replace with group */
                    sbuf.st_mode |= ((sbuf.st_mode & 00070) << 3)  ;
                else
                    /* replace with other */
                    sbuf.st_mode |= ((sbuf.st_mode & 00007) << 6)  ;
            }
            {
                /* Check if its a symbolic link */
                char   link[1024];
                if((llen=readlink(bfname,link,1024)) > 0)
                {
                    link[llen] = '\0' ;
                    curFile->lname = strdup(link) ;
                }
            }
            /* construct attribute string */
            if(S_ISREG(sbuf.st_mode))
                curFile->attrib[0] = '-' ;
            else if(S_ISDIR(sbuf.st_mode))
                curFile->attrib[0] = 'd' ;
            else
                curFile->attrib[0] = 's' ;
            curFile->attrib[1] = (sbuf.st_mode & 00400) ? 'r' : '-' ;
            curFile->attrib[2] = (sbuf.st_mode & 00200) ? 'w' : '-' ;
            curFile->attrib[3] = (sbuf.st_mode & 00100) ? 'x' : '-' ;
            curFile->size = sbuf.st_size ;
            curFile->mtime = sbuf.st_mtime ;
        }
        closedir(dirp) ;
    }
#endif
#ifdef _DOS
    static char *searchString = "*.*";
    struct ffblk fblk ;
    char  *ff ;
    int    done ;

    /* Render the list of files. */
    curHead = NULL ;
    strcpy(bfname,fname) ;
    fn = bfname+strlen(bfname) ;
    strcpy(fn,searchString) ;
    done = findfirst(bfname,&fblk,FA_RDONLY|FA_HIDDEN|FA_SYSTEM|FA_DIREC) ;
    while(!done)
    {
        if(((noFiles & 0x0f) == 0) &&
           ((curHead = (FILENODE *) meRealloc(curHead,sizeof(FILENODE)*(noFiles+16))) == NULL))
            return NULL ;
        curFile = &(curHead[noFiles++]) ;
        if((ff = (char *) strdup(fblk.ff_name)) == NULL)
            return NULL ;
        makestrlow(ff) ;
        curFile->fname = ff ;
        curFile->lname = NULL ;
        curFile->size  = fblk.ff_fsize ;
        curFile->mtime = (((unsigned long) fblk.ff_ftime) & 0x0ffff) | (((unsigned long) fblk.ff_fdate) << 16) ;
        /* construct attribute string */
        if(fblk.ff_attrib & FA_DIREC)
            strncpy(curFile->attrib,"drwx",4) ;
        else
        {
            curFile->attrib[0] = '-' ;
            curFile->attrib[1] = 'r' ;
            curFile->attrib[2] = (fblk.ff_attrib & FA_RDONLY) ? '-' : 'w' ;
            curFile->attrib[3] = '-' ;
            if(meTestExecutable(ff))
                curFile->attrib[3] = 'x' ;
        }
        done = findnext(&fblk) ;
    }
#endif
#ifdef _WIN32
    static char *searchString = "*.*";
    HANDLE          *fh ;
    WIN32_FIND_DATA  fd ;
    char            *ff ;
    
    /* Render the list of files. */
    curHead = NULL ;
    strcpy(bfname,fname) ;
    fn = bfname+strlen(bfname) ;
    strcpy(fn,searchString) ;
    if((fh = FindFirstFile(bfname,&fd)) != INVALID_HANDLE_VALUE)
    {
        do
        {
            if(((noFiles & 0x0f) == 0) &&
               ((curHead = (FILENODE *) meRealloc(curHead,sizeof(FILENODE)*(noFiles+16))) == NULL))
                return NULL ;
            curFile = &(curHead[noFiles++]) ;
            if((ff = (char *) strdup(fd.cFileName)) == NULL)
                return NULL ;
            curFile->fname = ff ;
            curFile->lname = NULL ;
            curFile->size  = (((unsigned long) fd.nFileSizeHigh) << 16) + ((unsigned long) fd.nFileSizeLow) ;
            curFile->mtime.dwLowDateTime  = fd.ftLastWriteTime.dwLowDateTime ;
            curFile->mtime.dwHighDateTime = fd.ftLastWriteTime.dwHighDateTime ;
            /* construct attribute string */
            if(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
                strncpy(curFile->attrib,"drwx",4) ;
            else
            {
                curFile->attrib[0] = '-' ;
                curFile->attrib[1] = 'r' ;
                curFile->attrib[2] = (fd.dwFileAttributes & FILE_ATTRIBUTE_READONLY) ? '-' : 'w' ;
                curFile->attrib[3] = '-' ;
                if(meTestExecutable(ff))
                    curFile->attrib[3] = 'x' ;
            }
        } while (FindNextFile(fh, &fd));
        FindClose(fh);
    }
#endif
    curFile = curHead ;
    /* Build a profile of the current directory. */
    for(ii=0 ; ii<noFiles ; ii++)
    {
        cfh = addFileNode(cfh,curFile) ;
        curFile++ ;
    }
    return cfh ;
}

static int
readDirectory(BUFFER *bp, char *fname)
{
    FILENODE *fnode, *fn ;
#ifdef _UNIX
    struct tm *tmp;
#endif
#ifdef _WIN32
    SYSTEMTIME tmp;
#endif
    char buf[MAXBUF];                  /* Working line buffer */
    int  len, dirs=0, files=0 ;        /* Length of the rendered line */
    long totSize=0 ;
    char cc;

    if((fnode=getDirectoryInfo(fname)) == NULL)
        return ABORT ;
    sprintf (buf, "Directory listing of: %s", fname);
    addLineToEob(bp,buf);
    fn = fnode ;
    while(fn != NULL)
    {
        totSize += fn->size ;
        if(fn->attrib[0] == 'd')
            dirs++ ;
        else
            files++ ;
        fn = fn->next ;
    }
    buf[0] = ' ' ;
    buf[1] = ' ' ;
    len = 2 ;
    if (totSize > 1024*1024*1024)
        len += sprintf (buf+len, "%7ldM ", totSize/(1024*1024));
    else if (totSize > 1024*1024)
        len += sprintf (buf+len, "%7ldK ", totSize/1024);
    else
        len += sprintf (buf+len, "%7ld  ", totSize);
    sprintf(buf+len, "used in %d files and %d dirs",files,dirs) ;
    addLineToEob(bp,buf);
    addLineToEob(bp,"");                   /* Blank line */
    while(fnode != NULL)
    {
        len = 0;                /* Reset to the start of the line */
        buf[len++] = ' ';
        strncpy(buf+len,fnode->attrib,4) ;
        len += 4 ;
        buf[len++] = ' ' ;
        /* Add the file statistics */
        if (fnode->size > 1024*1024*1024)
            len += sprintf(buf+len, "%7ldM ", fnode->size/(1024*1024));
        else if (fnode->size > 1024*1024)
            len += sprintf(buf+len, "%7ldK ", fnode->size/1024);
        else
            len += sprintf(buf+len, "%7ld  ", fnode->size);
        
#ifdef _UNIX
        if ((tmp = localtime(&fnode->mtime)) != NULL)
            len += sprintf(buf+len, "%02d/%02d/%02d %02d:%02d. ",
                           tmp->tm_year, tmp->tm_mon+1, tmp->tm_mday,
                           tmp->tm_hour, tmp->tm_min);

        else
#endif
#ifdef _DOS
        if((fnode->mtime & 0x0ffff) != 0x7fff)
            len += sprintf(buf+len, "%02d/%02d/%02d %02d:%02d. ",
                           (int) ((fnode->mtime >> 25) & 0x007f)+80,
                           (int) ((fnode->mtime >> 21) & 0x000f),
                           (int) ((fnode->mtime >> 16) & 0x001f),
                           (int) ((fnode->mtime >> 11) & 0x001f),
                           (int) ((fnode->mtime >>  5) & 0x003f)) ;
        else
#endif
#ifdef _WIN32
        if(FileTimeToSystemTime(&fnode->mtime,&tmp))
            len += sprintf(buf+len, "%02d/%02d/%02d %02d:%02d. ",
                           (tmp.wYear%100),tmp.wMonth,tmp.wDay,tmp.wHour,tmp.wMinute) ;
        else
#endif
            len += sprintf(buf+len, "--/--/-- --:--. ") ;
            
        if(strchr(fnode->fname,' ') != NULL)
            cc = '"' ;
        else
            cc = ' ' ;
        buf[len++] = cc ;
        len += sprintf (&buf[len], "%s", fnode->fname);
        if(fnode->attrib[0] == 'd')
            buf[len++] = '/' ;
        buf[len++] = cc ;
        
        if(fnode->lname != NULL)
        {
            sprintf(buf+len, " -> %s", fnode->lname);
            free(fnode->lname) ;
        }
        else
            buf[len] = '\0' ;
            
        addLineToEob(bp,buf);
        free(fnode->fname) ;
        fnode = fnode->next ;
    }
    free(curHead) ;
    meModeSet(bp->b_mode,MDVIEW) ;
    meModeClear(bp->b_mode,MDATSV) ;
    meModeClear(bp->b_mode,MDUNDO) ;
    return TRUE ;
}



/*
 * Read file "fname" into the current
 * buffer, blowing away any text found there. Called
 * by both the read and find commands. Return the final
 * status of the read. Also called by the mainline,
 * to read in a file specified on the command line as
 * an argument. If the filename ends in a ".c", CMODE is
 * set for the current buffer.
 */
/* Note for Dynamic file names
 * readin sets the buffer file name to point to fname (freeing old if not
 * NULL and bp->b_fname != fname) so fname should be dynamically allocated
 * or the file name reset after readin.
 */
/* bp        buffer to load into */
/* fname     name of file to read */
/* lockfl    check for file locks? */
/* frompipe  input coming from pipe. Dont open file */
int
readin(register BUFFER *bp, char *fname)
{
    int     ss=ABORT ;
    char    lfn[FILEBUF], afn[FILEBUF], *fn=fname ;
    
#if	CRYPT
    if(resetkey(bp) != TRUE)
        return ABORT ;
#endif
    if(bp->b_fname != fname)
    {
        meNullFree(bp->b_fname) ;
        bp->b_fname = fname ;
    }
    if(fn != NULL)
    {
        meSTAT stats ;
        int ft ;
        if((ft=getFileStats(fn,1,&(bp->stats),lfn)) != 5)
        {
            if(ft == 0)
                goto error_end ;
            if(lfn[0] != '\0')
                /* something was found, don't want to do RCS on this,
                 * the link may be dangling which returns 3 */
                fn = lfn ;
#if DORCS
            else if((ft == 3) &&       /* file not found */
               (rcsFilePresent(fn) == TRUE))
            {
                if(doRcsCommand(fn,rcsCoStr) != TRUE)
                    goto error_end ;
                if((ft=getFileStats(fn,1,&(bp->stats),lfn)) == 0)
                    goto error_end ;
            }
#endif
            if(autotime && !createBackupName(afn,fn,'#') &&
               (getFileStats(afn,0,&stats,NULL) == 1) && (stats.stmtime > bp->stats.stmtime))
            {
                char *tt, prompt[200] ;
                tt = getFileBaseName(afn) ;
                sprintf(prompt,"Recover newer %s",tt) ;
                if(mlyesno(prompt) == TRUE)
                {
                    fn = afn ;
                    ft = 1 ;
                    memcpy(&bp->stats,&stats,sizeof(meSTAT)) ;
                }
            }
            if(ft == 2)
            {
                if(readDirectory(bp,fn) != TRUE)
                    goto error_end ;
                ss = TRUE ;
                goto newfile_end;
            }
            if(ft == 3)
            {   /* File not found.      */
                
                /* WHY IS THIS JUST UNIX DOES THIS NOT APPLY TO ALL PLATFORMS ?? */
                char dirbuf [FILEBUF];

                /* See if we can write to the directory. */
                getFilePath (fn, dirbuf);
                if((getFileStats(dirbuf,0,NULL,NULL) != 2)
#ifdef _UNIX
                   || meTestWrite(dirbuf)
#endif
                   )
                {
                    /* READ ONLY DIR */
                    mlwrite(MWPAUSE,"%s: %s", dirbuf, sys_errlist[errno]);
                    /* Zap the filename - it is invalid.
                       We only want a buffer */
                    mlwrite (0, "[New buffer %s]", getFileBaseName(fname));
                    meNullFree(bp->b_fname) ;
                    bp->b_fname = NULL;
                    ss = ABORT;
                }
                else
                {
                    mlwrite(0,"[New file %s]", fname);
                    ss = TRUE ;
                }
                /* its not linked to a file so change the flag */
                bp->intFlag &= ~BIFFILE ;
                goto newfile_end;
            }
            
            /* Make sure that we can read the file. If we are root then
             * we do not test the 'stat' bits. Root is allowed to read
             * anything */
            if ((meTestRead (fn)) ||
                (
#ifdef _UNIX
                 (meUid != 0) &&
#endif
                 (!meStatTestRead(bp->stats))))
            {
                /* We are not allowed to read the file */
#if ((defined _UNIX) || (defined _DOS))
                mlwrite(MWABORT,"[%s: %s]", fn, sys_errlist[errno]) ;
#else
                mlwrite(MWABORT,"[cannot read: %s]", fn) ;
#endif
                /* Zap the filename - it is invalid.
                   We only want a buffer */
                meNullFree(bp->b_fname) ;
                bp->b_fname = NULL;
                /* its not linked to a file so change the flag */
                bp->intFlag &= ~BIFFILE ;
                goto newfile_end;
            }
        }
        
#ifdef _WIN32
        if(!meStatTestSystem(bp->stats))
        {
            /* if windows system file read in a readonly */
            meModeSet(bp->b_mode,MDVIEW) ;
            mlwrite(MWCURSOR|MWCLEXEC,"[Reading %s (system file)]", fn);
        }
        else
#endif
        /* Depending on whether we have write access set the buffer into
         * view mode. Again root gets the privelidge of being able to
         * write if possible. */
        if ((meTestWrite (fn)) || 
            (
#ifdef _UNIX
             (meUid != 0) &&
#endif
             (!meStatTestWrite(bp->stats))))
        {
            meModeSet(bp->b_mode,MDVIEW) ;
            mlwrite(MWCURSOR|MWCLEXEC,"[Reading %s (readonly)]", fn);
        }
        else
            mlwrite(MWCURSOR|MWCLEXEC,"[Reading %s]", fn);
    }
    ss = ffReadFile(fn,0,bp,bp->b_linep) ;
    
    /*
    ** Set up the modification time field of the buffer structure.
    */
    if(ss != ABORT)
    {
        mlwrite(MWCLEXEC,"[Read %d line%s]",bp->elineno,(bp->elineno==1) ? "":"s") ;
        if(fn == afn)
            /* this is a recovered file so flag the buffer as changed */
            meModeSet(bp->b_mode,MDEDIT) ;
        ss = TRUE ;
    }
    else
        meModeSet(bp->b_mode,MDVIEW) ;

newfile_end:

    bp->b_dotp = lforw(bp->b_linep);
    bp->line_no = 0 ;
    bp->b_doto = 0 ;
    
error_end:
    return ss ;
}

/*
 * Insert file "fname" into the current
 * buffer, Called by insert file command. Return the final
 * status of the read.
 */
/* must have the buffer line no. correct */
int
ifile(BUFFER *bp, char *fname, int flags)
{
    register WINDOW *wp ;
    register LINE   *lp ;
    register int     ii, ss ;
    register long    nline ;
    meMODE md ;
    
    meModeSet(bp->b_mode,MDEDIT) ;			/* we have changed	*/
    meModeCopy(md,bp->b_mode) ;
              
#if	CRYPT
    if(resetkey(bp) != TRUE)
        return FALSE ;
#endif
    if(!(flags & READ_SILENT))
        mlwrite(MWCURSOR|MWCLEXEC,"[Inserting file]");
    
    nline = bp->elineno ;
    lp = bp->b_dotp ;
    ss = ffReadFile(fname,flags,bp,lp) ;
    nline = bp->elineno-nline ;
    /* restore the mode */
    meModeCopy(bp->b_mode,md) ;

    if(ss != ABORT)
    {
        ss = TRUE ;
        if(!(flags & READ_SILENT))
           mlwrite(MWCLEXEC,"[inserted %d line%s]",nline,(nline==1) ? "":"s") ;
    }
    for (wp=wheadp; wp!=NULL; wp=wp->w_wndp)
        if (wp->w_bufp == bp)
        {
            if(wp->line_no >= bp->line_no)
                wp->line_no += nline ;
            if(wp->mlineno >= bp->line_no)
                wp->mlineno += nline ;
            wp->w_flag |= WFMAIN|WFMOVEL ;
        }
#if MEUNDO
    if((bp == curbp) && meModeTest(bp->b_mode,MDUNDO))
    {
        curwp->w_doto = 0 ;
        ii = 0 ;
        while(nline--)
        {
            lp = lback(lp) ;
            ii += llength(lp) + 1 ;
        }
        meUndoAddInsChars(ii) ;
    }
#endif
    return ss ;
}

/*
 * Insert a file into the current
 * buffer. This is really easy; all you do it
 * find the name of the file, and call the standard
 * "insert a file into the current buffer" code.
 * Bound to "C-X C-I".
 */
int
insFile(int f, int n)
{
    register int s;
    char fname[FILEBUF] ;

    if((s=inputFileName("Insert file",fname,1)) != TRUE)
        return s ;
    /* Allow existing or url files */
    if((getFileStats(fname,3,NULL,NULL) & 0x3) != 1)
        return ABORT ;
    if((s=bchange()) != TRUE)               /* Check we can change the buffer */
        return s ;
    /* set mark to previous line so it doesnt get moved down */
    curwp->w_markp = lback(curwp->w_dotp) ;
    curwp->w_marko = 0 ;
    curwp->mlineno = curwp->line_no-1 ;

    /* store current line in buffer */
    curbp->b_dotp = curwp->w_dotp ;
    curbp->line_no = curwp->line_no ;   /* must have the line no. correct */
    
    for(; n>0 ; n--)
        if((s = ifile(curbp,fname,0)) != TRUE)
            break ;

    /* move the mark down 1 line into the correct place */
    curwp->w_markp = lforw(curwp->w_markp);
    curwp->mlineno++ ;
    return s ;
}


/*
 * Take a file name, and from it
 * fabricate a buffer name. This routine knows
 * about the syntax of file names on the target system.
 * I suppose that this information could be put in
 * a better place than a line of code.
 */
unsigned char
makename(char *bname, char *fname)
{
    register char   *cp1 ;
    register char   *cp2 ;
    int              i ;

    cp1 = getFileBaseName(fname) ;
    cp2 = bname ;
    while((*cp2++=*cp1++) != '\0')
        ;
    cp2-- ;
    /*
     * Now we've made our buffer name, check that it is unique.
     */
    for(i = 1 ; (bfind(bname, 0) != NULL) ; i++)
    {
        /*
         * There is another buffer with this name. Try sticking a
         * numeric "i" on the end of the buffer.
         */
        sprintf(cp2, "<%d>", i);
    }
    return (i == 1) ? 0:BIFNAME ;
}

#define fileNameWild(fileName)                                               \
((strchr(fileName,'*') != NULL) || (strchr(fileName,'?') != NULL) ||         \
 (strchr(fileName,'[') != NULL))

int
findFileSingle(char *fname, int bflag, long lineno)
{
#ifdef _UNIX
    meSTAT  stats ;
#endif
    BUFFER *bp ;
    int   gft ;
                
#ifdef _UNIX
    gft = getFileStats(fname,1,&stats,NULL) ;
#else
    gft = getFileStats(fname,1,NULL,NULL) ;
#endif
    if(!gft || ((gft == 3) && !(bflag & BFND_CREAT)))
        return 0 ;
    bflag |= BFND_CREAT ;
    
    /* if this is a directory then add the '/' */
    if(gft == 2)
    {
        char *ss=fname+strlen(fname) ;
        if(ss[-1] != DIR_CHAR)
        {
            ss[0] = DIR_CHAR ;
            ss[1] = '\0' ;
        }
    }
    for (bp=bheadp; bp!=NULL; bp=bp->b_bufp)
    {
        if((bp->b_bname != NULL) && (bp->b_bname[0] != '*') && 
#ifdef _UNIX
           (!fnamecmp(bp->b_fname,fname) || 
            ((stats.stdev != (dev_t)(-1)) &&
             (bp->stats.stdev == stats.stdev) && 
             (bp->stats.stino == stats.stino) &&
             (bp->stats.stsize == stats.stsize))))
#else
           !fnamecmp(bp->b_fname,fname))
#endif
            break ;
    }
    if(bp == NULL)
    {
        bp = bfind(fname,bflag);
        bp->b_fname = strdup(fname) ;
        bp->intFlag |= BIFFILE ;
        if(gft == 2)
        {
            /* if its a directory then add dir mode and remove binary */
            meModeSet(bp->b_mode,MDDIR) ;
            meModeClear(bp->b_mode,MDBINRY) ;
        }
        bp->line_no = lineno ;
        bp->histNo = bufHistNo ;
    }
    else
    {
        if(!meModeTest(bp->b_mode,MDNACT))
            bp->intFlag |= BIFLOAD ;
        bp->histNo = bufHistNo ;
    }
    return 1 ;
}

int
findFileList(char *fname, int bflag, long lineno)
{
    register int nofiles=0, ii ;
    char fileName[FILEBUF], *baseName ;
    
    bufHistNo++ ;
    fileNameCorrect(fname,fileName,&baseName) ;
              
#if URLAWAR
    if(isUrlLink(fileName))
        nofiles += findFileSingle(fileName,bflag,lineno) ;
    else
#endif
        if(fileNameWild(baseName))
    {
        char mask[FILEBUF] ;
        strcpy(mask,baseName) ;
        *baseName = '\0' ;
        getDirectoryList(fileName,&curDirList) ;
        for(ii=0 ; ii<curDirList.size ; ii++)
        {
            char *ss = curDirList.list[ii] ;
#ifdef _INSENSE_CASE
            if(regexpStrCmp(ss,mask,0))
#else
            if(regexpStrCmp(ss,mask,1))
#endif
            {
                strcpy(baseName,ss) ;
                nofiles += findFileSingle(fileName,bflag,lineno) ;
            }
        }
    }
    else
    {
#ifdef _WIN32_NOT_NEEDED_ANY_MORE
        /* with windows naff file name case insensitivity we must get
         * the correct case by using FindFirstFile*/
        /* ZZZZZZ - must check whether ..../xxxx/ gets XXXX or xxxx/\*.* */
        HANDLE *handle;
        WIN32_FIND_DATA fd;
        if((handle = FindFirstFile(fileName,&fd)) != INVALID_HANDLE_VALUE)
        {
            strcpy(baseName,fd.cFileName) ;
            if(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
            {
                char *ss ;
                ss = baseName + strlen(baseName) ;
                *ss++ = DIR_CHAR ;
                *ss   = '\0' ;
            }
            FindClose (handle);
        }
#endif
        nofiles += findFileSingle(fileName,bflag,lineno) ;
    }
    return nofiles ;
}

int
findSwapFileList(char *fname, int bflag, long lineno)
{
    BUFFER *bp ;
    int     ret ;

    bufHistNo++ ;
    if(!findFileList(fname,bflag,lineno))
        return mlwrite(MWABORT,"[Failed to find file %s]",fname);
    for(bp=bheadp ; bp->histNo!=bufHistNo ; bp=bp->b_bufp)
        ;
    bufHistNo-=2 ;
    ret = swbuffer(curwp,bp) ;	/* make buffer BP current */
    bufHistNo++ ;
    return ret ;
}

/*
 * Select a file for editing.
 * Look around to see if you can find the
 * fine in another buffer; if you can find it
 * just switch to the buffer. If you cannot find
 * the file, create a new buffer, read in the
 * text, and switch to the new buffer.
 * Bound to C-X C-F.
 */
int
findFile(int f, int n)
{
    register int s ;
    char	 fname[FILEBUF] ;
    
    if((s=inputFileName("Find file",fname,0)) != TRUE)
        return s ;
    if(n)
        n = BFND_CREAT ;
    n |= BFND_MKNAM ;
    return findSwapFileList(fname,n,0) ;
}

/*
 * Select a binary file for editing.
 * Look around to see if you can find the
 * fine in another buffer; if you can find it
 * just switch to the buffer. If you cannot find
 * the file, create a new binary buffer, read in the
 * text, and switch to the new buffer.
 */
int
findBFile(int f, int n)
{
    register int s;		  /* status return	    */
    char	 fname[FILEBUF];  /* file user wishes to find */

    if((s=inputFileName("Find binary file",fname,0)) != TRUE)
        return s ;
    if(n)
        n = BFND_CREAT ;
    n |= BFND_MKNAM|BFND_BINARY ;
    return findSwapFileList(fname,n,0) ;
}

/* Select a crypt file for editing. */
int
findCFile(int f, int n)
{
    register int s;		   /* status return	    */
    char	 fname[FILEBUF];   /* file user wishes to find */

    if((s=inputFileName("Find crypt file",fname,0)) != TRUE)
        return s ;
    if(n)
        n = BFND_CREAT ;
    n |= BFND_MKNAM|BFND_CRYPT ;
    return findSwapFileList(fname,n,0) ;
}

/*
 * Find file into other window. This is trivial since all the hard stuff is
 * done by the routines splitWindVert() and filefind().
 *
 * The count is used to put the "found file" into the upper or lower window.
 *
 * Normally mapped to ^X4. (or C-X 4 if you prefer).
 */
int
nextWndFindFile(int f, int n)
{
    char fname[FILEBUF];	/* file user wishes to find */
    register int s;		/* status return */

    if ((s=inputFileName("Find file",fname,0)) != TRUE)
        return(s);
    if(wpopup(NULL,WPOP_MKCURR) == NULL)
        return FALSE ;
    if(n)
        n = BFND_CREAT ;
    n |= BFND_MKNAM ;
    return findSwapFileList(fname,n,0) ;
}

int
readFile(int f, int n)
{
    char fname[FILEBUF];	/* file user wishes to find */
    register int s;		/* status return */

    if ((s=inputFileName("Read file", fname,0)) != TRUE)
        return s ;
    if(n)
        n = BFND_CREAT ;
    n |= BFND_MKNAM ;
    if((s=zotbuf(curbp,clexec)) == TRUE)
        s = findSwapFileList(fname,n,0) ;
    return s ;
}

int
viewFile(int f, int n)	/* visit a file in VIEW mode */
{
    char fname[FILEBUF];	/* file user wishes to find */
    register int ss, vv;	/* status return */

    if ((ss=inputFileName("View file", fname,0)) != TRUE)
        return ss ;
    if(n)
        n = BFND_CREAT ;
    n |= BFND_MKNAM ;
    /* Add view mode globally so any new buffers will be created
     * with view mode */
    vv = meModeTest(globMode,MDVIEW) ;
    meModeSet(globMode,MDVIEW) ;
    ss = findSwapFileList(fname,n,0) ;
    /* if view mode was not set globally restore it */
    if(!vv)
        meModeClear(globMode,MDVIEW) ;
    return ss ;
}

/*
 * This function performs an auto writeout to disk for the given buffer
 */
void
autowriteout(register BUFFER *bp)
{
    char fn[FILEBUF], lname[FILEBUF], *ff ;
    int  ss ;
    
    bp->autotime = 0xffffffff ;
    
    if(bp->b_fname == NULL)
        ss = ABORT ;
    else
    {
        getFileStats(bp->b_fname,0,NULL,lname) ;
        ff = (lname[0] == '\0') ? bp->b_fname:lname ;
        if(createBackupName(fn,ff,'#'))
            ss = ABORT ;
        else if((ss=ffWriteFileOpen(fn,WRITE_AUTOSAVE,bp)) == TRUE)
        {    
            LINE *lp ;
            
            lp = lforw(bp->b_linep);            /* First line.          */
            while((lp != bp->b_linep) &&
                  ((ss=ffWriteFileWrite(llength(lp),ltext(lp),
                                    !(lp->l_flag & LNNEOL))) == TRUE))
            {
                lp = lforw(lp) ;
                if(TTbreakTest(1))
                {
                    ss = ABORT ;
                    break ;
                }
            }
            ffWriteFileClose(WRITE_AUTOSAVE,bp) ;
        }
    }
    if(ss == TRUE)
        mlerase(MWCLEXEC) ;
    else
        mlwrite(0,"[Auto-writeout failure for buffer %s]",bp->b_bname) ;
}

/*
 * This function removes an auto writeout file for the given buffer
 */
void
autowriteremove(register BUFFER *bp)
{
    char fn[FILEBUF] ;

    if(autotime && bufferNeedSaving(bp) &&
       !createBackupName(fn,bp->b_fname,'#') &&
       !meTestExist(fn))
        /* get rid of any tempory file */
        meUnlink(fn) ;
    bp->autotime = 0xffffffff ;
}

/*
 * writeCheck
 * This performs some sime access checks to determine if we
 * can write the file.
 */
static int 
writeCheck (char *pathname, int flags, meSTAT *statp)
{
    /* Quick test for read only. */
    if (pathname != NULL)
    {
        char dirbuf [FILEBUF];
#ifdef _UNIX
        /* READ ONLY directory test
           See if we can write to the directory. */
        getFilePath (pathname, dirbuf);
        if (meTestWrite (dirbuf))
            return mlwrite (MWABORT, "Read Only Directory: %s", dirbuf);
#endif        
        /* See if there is an existing file */
        if ((flags & 1) == 0)           /* Validity check enabled ?? */
            return TRUE;                /* No - quit. */
        if (meTestExist (pathname))     /* Does it exist ?? */
            return TRUE;                /* No - quit */
        if ((statp != NULL) && !meStatTestWrite(*statp))
        {
            /* No - advised read only - see if the users wants to write */
            sprintf (dirbuf, "%s is read only. Try anyway", pathname);
            if (mlyesno(dirbuf) != TRUE)
                return mlwrite (MWABORT, "File not written.");
        }
    }   
    return TRUE;
}

/*
 * This function performs the details of file
 * writing. Uses the file management routines in the
 * "fileio.c" package. The number of lines written is
 * displayed. Sadly, it looks inside a LINE; provide
 * a macro for this. Most of the grief is error
 * checking of some sort.
 */
int
writeOut(register BUFFER *bp, int flags, char *fn)
{
    register int    s;
    
#if (defined _UNIX) || (defined _DOS) || (defined _WIN32)
    /* Add write permission to backup file. */
    if(meModeTest(bp->b_mode,MDBACK))
    {
#ifdef _DOS
        s = bp->stats.stmode & 0x0fe ;
#endif
#ifdef _WIN32
        s = bp->stats.stmode & ~(FILE_ATTRIBUTE_HIDDEN|FILE_ATTRIBUTE_READONLY) ;
#endif
#ifdef _UNIX
        s = bp->stats.stmode | S_IWUSR ;
#endif
        if(s == bp->stats.stmode)
            s = 0 ;
        flags |= WRITE_BACKUP | s ;
    }
#endif
#if	TIMSTMP
    set_timestamp(bp);			/* Perform time stamping */
#endif
    if(ffWriteFile(fn,flags,bp) != TRUE)
        return FALSE ;
    /* No write error.      */
    /* must be done before buffer is flagged as not changed */
#if MEUNDO
    if(meSystemCfg & meSYSTEM_KEEPUNDO)
    {
        /* go through undo list removing any unedited flags */
        UNDOND *nn ;
        nn = bp->fUndo ;
        while(nn != NULL)
        {
            if(nn->type & MEUNDO_FRST)
                nn->type |= MEUNDO_MDEL ;
            nn = nn->next ;
        }
    }
    else
        meUndoRemove(bp) ;
#endif
    autowriteremove(bp) ;
    meModeClear(bp->b_mode,MDEDIT) ;
    meModeClear(bp->b_mode,MDSAVE) ;

    if(fn != NULL)
    {
#ifndef _WIN32
        /* set the right file attributes */
        if(bp->stats.stmode != meUmask)
            meChmod(fn,bp->stats.stmode) ;
#ifdef _UNIX
        /* If we are the root then restore the existing settings on the 
         * file. This should really be done using a $system flag - but
         * one assumes that this is the default action for the super user. */
        if ((meUid == 0) && 
            ((bp->stats.stuid != meUid) || (bp->stats.stgid != meGid)))
            chown (fn, bp->stats.stuid, bp->stats.stgid);
#endif
#endif
        /* Read in the new stats */
        getFileStats(fn,0,&bp->stats,NULL) ;
        /* Change the view status back to the file's permissions. 
         * For root we let the super user do as they wish so no 
         * read protection is added. */
        if(
#if (defined _UNIX)
           (meUid != 0) &&
#endif
           (!meStatTestWrite(bp->stats)))
            meModeSet(curbp->b_mode,MDVIEW) ;
        else
            meModeClear(curbp->b_mode,MDVIEW) ;
    }
    return TRUE ;
}

/*
 * This function performs the details of file writing. Uses the file
 * management routines in the "fileio.c" package. The number of lines written
 * is displayed. Sadly, it looks inside a LINE; provide a macro for this. Most
 * of the grief is error checking of some sort.
 */
int
writeout(register BUFFER *bp, int flags, char *fname)
{
    char lname[FILEBUF], *fn ;
    if(!strcmp(bp->b_bname,"*stdin*"))
        fn = NULL ;
    else if((bp->b_bname[0] == '*') || (fname == NULL))
        return mlwrite(MWABORT,"[No file name set for buffer %s]",bp->b_bname);
    else
    {
        meSTAT stats ;
        /*
         * Check that the file has not been modified since it was read it in. This
         * is a bit of a problem since the info we require is hidden in a buffer
         * structure which may not necessarily be the current buffer. If it is not
         * in the current buffer then dont bother to look for it.
         */
        getFileStats(fname,0,&stats,lname) ;
        if((flags & 1) && (stats.stmtime > bp->stats.stmtime))
        {
            char prompt[MAXBUF];
            sprintf(prompt, "%s modified, write", fname);
            if(mlyesno(prompt) != TRUE)
                return ctrlg(FALSE,1);
        }
        fn = (lname[0] == '\0') ? fname:lname ;
        /* Quick check on the file write condition */
        if (writeCheck (fn, flags, &stats) != TRUE)
            return ABORT;
    }
    
    flags = (flags & 0x02) ? WRITE_IGNRNRRW : 0 ;
    return writeOut(bp,flags,fn) ;
}

/*
 * Ask for a file name, and write the contents of the current buffer to that
 * file. Update the remembered file name and clear the buffer changed flag.
 * This handling of file names is different from the earlier versions, and is
 * more compatable with Gosling EMACS than with ITS EMACS. Bound to "C-X C-W".
 */
int
writeBuffer(int f, int n)
{
    meSTAT stats ;
    register int s;
    char         fname[FILEBUF], lname[FILEBUF], *fn ;

    if ((s=inputFileName("Write file",fname,1)) != TRUE)
        return s ;

    if (!((s=getFileStats(fname,3,&stats,lname)) & 0x01))
        return ABORT ;
    fn = (lname[0] == '\0') ? fname:lname ;
    if(n & 0x01)
    {
        BUFFER *bp ;
        
        /* Check for write-out filename problems */
        if((s == 1) &&
           (mlyesno("File already exists, overwrite") != TRUE))
            return ctrlg(FALSE,1);
        
        /* Quick check on the file write condition */
        if (writeCheck (fn, n, &stats) != TRUE)
            return ABORT;
        /*
         * Check to see if the new filename conflicts with the filenames for any
         * other buffer and produce a warning if so.
         */
        bp=bheadp ;
        while(bp != NULL)
        {
            if((bp != curbp) &&
#ifdef _UNIX
               (!fnamecmp(bp->b_fname,fn) || 
                ((stats.stdev != (dev_t)(-1)) &&
                 (bp->stats.stdev == stats.stdev) &&
                 (bp->stats.stino == stats.stino) &&
                 (bp->stats.stsize == stats.stsize))))
#else
               !fnamecmp(bp->b_fname,fn))
#endif
            {
                char prompt[MAXBUF];
                sprintf(prompt, "Buffer %s is the same file, write",bp->b_bname) ;
                if(mlyesno(prompt) != TRUE)
                    return ctrlg(FALSE,1);
            }
            bp = bp->b_bufp ;
        }
    }
    s = (n & 0x02) ? WRITE_IGNRNRRW : 0 ;

    if(!writeOut(curbp,s,fn))
        return FALSE ;

    if(fnamecmp(curbp->b_fname,fname))
    {
        char buf[MAXBUF], *bb ;
        meNullFree(curbp->b_fname) ;
        curbp->b_fname = strdup(fname) ;
        unlinkBuffer(curbp) ;
        curbp->intFlag = (curbp->intFlag & ~BIFNAME) | makename(buf, fname) ;
        if((bb = meMalloc(strlen(buf)+1)) != NULL)
        {
            meFree(curbp->b_bname) ;
            strcpy(bb,buf) ;
            curbp->b_bname = bb ;
        }
        linkBuffer(curbp) ;
    }
    addModeToWindows(WFMODE) ; /* and update ALL mode lines */

    return TRUE ;
}

/*
 * Save the contents of the current buffer in its associatd file. No nothing
 * if nothing has changed (this may be a bug, not a feature). Error if there
 * is no remembered file name for the buffer. Bound to "C-X C-S". May get
 * called by "C-Z".
 */
int
saveBuffer(int f, int n)
{
    register int    s;
    
    /* Note that we check for existance here just incase sombody has
     * deleted it under our feet. There is nothing more annoying than the 
     * editor bitching there are no changes when it's been zapped !! */
    /* Further note: the file name can be NULL, e.g. theres no file name
     * so this must be tested before meTestExist, BUT the file name can be
     * NULL and still be saved, e.g. the buffer name is *stdin* - so be careful */
    if((n & 0x01) && (curbp->b_fname != NULL) &&  /* Are we Checking ?? */
       (meTestExist (curbp->b_fname) == 0) &&     /* Does file actually exist ? */
       !meModeTest(curbp->b_mode,MDEDIT))         /* Have we edited buffer ? */
        /* Return, no changes.  */
        return mlwrite(0,"[No changes made]") ;
    if((s=writeout(curbp,n,curbp->b_fname)) == TRUE)
        addModeToWindows(WFMODE) ;  /* and update ALL mode lines */
    return (s);
}



int
appendBuffer(int f, int n)
{
    register int s;
    char         fname[FILEBUF], lname[FILEBUF], *fn ;

    if ((s=inputFileName("Append to file",fname,1)) != TRUE)
        return s ;

    if(!((s=getFileStats(fname,3,NULL,lname)) & 0x01))
        return ABORT ;
    fn = (lname[0] == '\0') ? fname:lname ;
    if(s != 1)
    {
        if(n & 0x01)
        {
            char prompt[MAXBUF];
            sprintf(prompt, "%s does not exist, create",fn) ;
            if(mlyesno(prompt) != TRUE)
                return ctrlg(FALSE,1);
        }
        s = 0 ;
    }
    else if(n & 0x04)
        s = WRITE_OPENTRUNC ;
    else
        s = WRITE_OPENEND ;
    if(n & 0x02)
        s |= WRITE_IGNRNRRW ;
    return ffWriteFile(fname,s,curbp) ;
}
    
/*
 * save-some-buffers, query saves all modified buffers
 */
int
saveSomeBuffers(int f, int n)
{
    register BUFFER *bp;    /* scanning pointer to buffers */
    register int status=TRUE ;
    char     tmp[MAXBUF] ;

    bp = bheadp;
    while (bp != NULL)
    {
        if(bufferNeedSaving(bp))
        {
            if(n & 1)
            {
                if(bp->b_fname != NULL)
                    sprintf(tmp, "Save file %s", bp->b_fname) ;
                else
                    sprintf(tmp, "Save buffer %s", bp->b_bname) ;
                if((status = mlyesno(tmp)) == ABORT)
                    return ABORT ;
            }
            if((status == TRUE) &&
               (writeout(bp, 0, bp->b_fname) != TRUE))
                return ABORT ;
        }
        bp = bp->b_bufp;            /* on to the next buffer */
    }
    addModeToWindows(WFMODE) ;  /* and update ALL mode lines */
    return TRUE ;
}

/*
 * The command allows the user
 * to modify the file name associated with
 * the current buffer. It is like the "f" command
 * in UNIX "ed". The operation is simple; just zap
 * the name in the BUFFER structure, and mark the windows
 * as needing an update. You can type a blank line at the
 * prompt if you wish.
 */
int
changeFileName(int f, int n)
{
    register int s ;
    char         fname[FILEBUF], lname[FILEBUF], *fn ;

    if ((s=inputFileName("New file name",fname,1)) == ABORT)
        return s ;

    if((getFileStats(fname,3,NULL,lname) & 0x05) != 1)
        return ABORT ;
    fn = (lname[0] == '\0') ? fname:lname ;
    meNullFree(curbp->b_fname) ;

    if (s == FALSE)
        curbp->b_fname = NULL ;
    else
        curbp->b_fname = strdup(fname) ;

#if (defined _UNIX) || (defined _DOS) || (defined _WIN32)
    if(meTestWrite(fn))
        meModeSet(curbp->b_mode,MDVIEW) ;
    else
        meModeClear(curbp->b_mode,MDVIEW) ;	/* no longer read only mode */
#endif
    addModeToWindows(WFMODE) ;  /* and update ALL mode lines */

    return (TRUE);
}

#ifdef _DOS
int
_meChdir(char *path)
{
    register int s ;
    int len ;

    len = strlen (path)-1 ;

    if((len > 1) && (path[1] == _DRV_CHAR))
    {
        union REGS reg ;		/* cpu register for use of DOS calls */

        reg.h.ah = 0x0e ;
        reg.h.dl = path[0] - 'a' ;
        intdos(&reg, &reg);
    }
    if((len > 3) && (path[len] == DIR_CHAR))
        path[len] = '\0' ;
    else
        len = -1 ;
    s = chdir(path) ;
    if(len > 0)
        path[len] = DIR_CHAR ;
    return s ;
}
#endif

int
changeDir(int f, int n)
{
    /*
     * Change directory.
     *
     * Alain D D Williams, Sept 1986.
     *
     * Hacked by Tony Bamford to change relative filenames to absolute
     * paths, Nov 1986.
     * Completely hack by Steve Phillips to make it not to bloody
     * dangerous.
     * Done it by making any files loaded imediately absolute (only
     * sane way of doing it) and disallowing operation if cant be done.
     */
#if (defined _UNIX) || (defined _DOS) || (defined _WIN32)
    register int    s;
    char *dd, dname[FILEBUF];		/* directory to change to   */

    if((s = inputFileName("Directory Name ",dname,1)) != TRUE)
        return(s);

    if(meChdir(dname) == -1)
        return mlwrite(MWABORT,"directory %s: %s", dname, sys_errlist[errno]);
    if((dd = strdup(dname)) != NULL)
    {
        meFree(curdir) ;
        curdir = dd ;
    }
    return(TRUE);
#else
    return mlwrite(MWABORT,"Cant change directory as cant get absolute path") ;
#endif
}


/************************ New file routines *****************************/

void
pathNameCorrect(char *oldName, char *newName, char **baseName)
{
    register char *p, *p1 ;
    int            flag ;
#if (defined _DOS) || (defined _WIN32)
    char          *gwdbuf ;
#endif

#ifdef _CONVDIR_CHAR
    /* convert all '\\' to '/' for dos etc */
    p1 = oldName ;
    while((p1=strchr(p1,_CONVDIR_CHAR)) != NULL)  /* got a '\\', -> '/' */
        *p1++ = DIR_CHAR ;
#endif
    
    p = p1 = oldName ;
#if URLAWAR
    if(isUrlLink(p1))
        flag = 2 ;
    else
#endif
#ifndef _UNIX
        /* the following picks up on windows network drive paths, like
         * //foo/root/temp/... etc unfortunately this causes grief on unix
         * when loading absolute files from root (i.e. /) e.g. normally
         *      /tmp//etc/... -> /etc/...
         * but from root:
         *      //etc/...     -> //etc/...
         * Cause it thinks its a network drive. So I've disabled it for unix!
         */
        if((p1[0] == DIR_CHAR) && (p1[1] == DIR_CHAR))
        {
            flag = 1 ;
            p1++ ;
        }
    else
#endif
    {
        flag = 0 ;
#if URLAWAR
        if(isUrlFile(p1))
            p1 += 5 ;
#endif
    }
        /* search for
         * 1) set to root,  xxxx/http:// -> http://  (for urls)
         * 2) set to root,  xxxx/ftp://  -> ftp://   (for urls)
         * 3) set to root,  xxxx///yyyyy -> //yyyyy  (for network drives)
         * 4) set to root,  xxxx//yyyyy  -> /yyyyy
         * 5) set to Drive, xxxx/z:yyyyy -> z:yyyyy
         * 6) set to home,  xxxx/~yyyyy  -> ~yyyyy
         */
    while((p1=strchr(p1,DIR_CHAR)) != NULL)
    {
#if URLAWAR
        if(isUrlLink(p1+1))
        {
            flag = 2 ;
            p = p1+1 ;
            break ;
        }
        else if(isUrlFile(p1+1))
        {
            flag = 0 ;
            p1 += 6 ;
            p = p1 ;
        }
        else
#endif
            if(flag != 2)
        {
            if(p1[1] == DIR_CHAR)
            {
                if(p1[2] == DIR_CHAR)
                {
                    /* Got a xxxx///yyyyy -> //yyyyy must not optimise further */
                    flag = 1 ;
                    p1++ ;
                    while(p1[2] == DIR_CHAR)
                        p1++ ;
                    p = p1 ;
                    p1++ ;
                }
                else
                {
                    /* Got a xxxx//yyyyy -> /yyyyy */
                    p = ++p1 ;
                    flag = 0 ;
                }
            }
#ifdef _DRV_CHAR
            else if((isAlpha(p1[1]) || (p1[1] == '?')) && (p1[2] == _DRV_CHAR))
            {
                /* got a Drive, xxxx/z:yyyyy -> z:yyyyy */
                p = ++p1 ;
                flag = 0 ;
            }
#endif
            else if((homedir != NULL) && (p1[1] == '~'))
            {
                /* Got a home,  xxxx/~yyyyy  -> ~yyyyy */
                p = ++p1 ;
                flag = 0 ;
            }
            else
                p1++ ;
        }
        else
            p1++ ;
    }
    
    if(flag)
        strcpy(newName,p) ;
    else if((homedir != NULL) && (p[0] == '~'))
    {
        strcpy(newName,homedir) ;
        p1 = newName + strlen(newName) ;
        if(p[1] != DIR_CHAR)
        {
            *p1++ = DIR_CHAR ;
            *p1++ = '.' ;
            *p1++ = '.' ;
            *p1++ = DIR_CHAR ;
        }
        strcpy(p1,p+1) ;
    }
#ifdef _DRV_CHAR
    /* case for /xxxx... */
    else if(p[0] == DIR_CHAR)
    {
        if((p != oldName) && (oldName[1] == _DRV_CHAR))
            /* file name was D:xxxxxx//yyyyyy, convert to D:/yyyyyy */ 
            newName[0] = oldName[0] ;
        else
            /* file name is /yyyyyy, convert to D:/yyyyyy */ 
            newName[0] = curdir[0] ;
        newName[1] = _DRV_CHAR ;
        strcpy(newName+2,p) ;
    }
    /* case for xxxx... */
    else if(p[1] != _DRV_CHAR)
#else
    /* case for xxxx... */
    else if(p[0] != DIR_CHAR)
#endif
    {
        strcpy(newName,curdir) ;
        strcat(newName,p) ;
    }
#ifdef _DRV_CHAR
    /* case for D:xxxxxx */
    else if((p[1] == _DRV_CHAR) && (p[2] != DIR_CHAR) &&
            (p[0] != '?') && ((gwdbuf = gwd(p[0])) != NULL))
    {
        strcpy(newName,gwdbuf) ;
        strcat(newName,p+2) ;
        meFree(gwdbuf) ;
    }
#endif
    else
        strcpy(newName,p) ;

    if(flag != 2)
    {
        p = NULL ;
        p1 = newName ;
        while((p1=strchr(p1,DIR_CHAR)) != NULL)
        {
            if((p1[1] == '.') && (p1[2] == '.') &&         /* got /../YYYY */
               (p1[3] == DIR_CHAR))
            {
                if(p == NULL)        /* got /../YYYY */
                    p = newName ;
                /* else got /XXX/../YYYY */
                
                p1 += 3 ;
                while((*p++ = *p1++) != '\0')
                    ;
                p = NULL ;
                p1 = newName ;
            }
            else if((p1[1] == '.') && (p1[2] == DIR_CHAR))      /* got /./YYYY */
            {
                char *tt ;
                tt = p1+2 ;
                while((*p1++ = *tt++) != '\0')
                    ;
                if(p == NULL)
                    p1 = newName ;
                else
                    p1 = p ;
            }
            else
            {
                p = p1 ;
                p1++ ;
            }
        }
#ifdef _SINGLE_CASE
        makestrlow(newName) ;
#endif
    }
    {
        char cc ;
        p = p1 = newName ;
        while((cc=*p1++) != '\0')
        {
            if((cc == DIR_CHAR) && (*p1 != '\0'))
                p = p1 ;
            else if((cc == '[') || (cc == '?') || (cc == '*'))
                /* this is base of a wild file base name break */
                break ;
        }
    }
    if(baseName != NULL)
        *baseName = p ;
}

#ifdef _WIN32
void
fileNameCorrect(char *oldName, char *newName, char **baseName)
{
    char *bn ;
    
    pathNameCorrect(oldName,newName,&bn) ;
    if(baseName != NULL)
        *baseName = bn ;
    
    if(
#if URLAWAR
       !isUrlLink(newName) && 
#endif
       !fileNameWild(bn))
    {
        /* with windows naff file name case insensitivity we must get
         * the correct a single name letter case by using FindFirstFile*/
        HANDLE *handle;
        WIN32_FIND_DATA fd;
        if((handle = FindFirstFile(newName,&fd)) != INVALID_HANDLE_VALUE)
        {
            strcpy(bn,fd.cFileName) ;
            /* If a directory that append the '/' */
            if(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
            {
                bn += strlen(bn) ;
                *bn++ = DIR_CHAR ;
                *bn = '\0' ;
            }
            FindClose (handle);
        }
    }
}
#endif

void
getDirectoryList(char *pathName, meDIRLIST *dirList)
{
    char **fls ;
    int    noFiles ;
#ifdef _UNIX
    struct stat statbuf;
    long   timeStamp ;
    
    if(!stat(pathName,&statbuf))
        timeStamp = statbuf.st_mtime ;
    else
        timeStamp = 0 ;
#endif
    
    if((dirList->path != NULL) &&
       !strcmp(dirList->path,pathName) &&
#ifdef _UNIX
       (timeStamp <= dirList->timeStamp)
#else
       !dirList->timeStamp
#endif
       )
        return ;
    
    /* free off the old */
    meNullFree(dirList->path) ;
    freeFileList(dirList->size,dirList->list) ;
    fls = NULL ;
    noFiles = 0 ;
                  
#ifdef _DOS
    if(pathName[0] == '\0')
    {
        union REGS reg ;		/* cpu register for use of DOS calls */
        char  *ff ;
        int    ii ;
        
        for (ii = 1; ii <= 26; ii++)    /* Drives are a-z (1-26) */
        {
#if 0
            reg.h.ah = 0x36 ;
            reg.h.dl = ii ;
            if((ii <= 2) || 
               (intdos(&reg,&reg),(reg.x.ax != 0xffff)))
#else
            reg.x.ax = 0x440e ;
            reg.h.bl = ii ;
            intdos(&reg,&reg) ;
            if((reg.x.cflag == 0) || (reg.x.ax != 0x0f))
#endif
            {
                if(((fls = meRealloc(fls,sizeof(char *) * (noFiles+1))) == NULL) ||
                   ((ff = meMalloc(4*sizeof(char))) == NULL))
                {
                    fls = NULL ;
                    noFiles = 0 ;
                    break ;
                }
                fls[noFiles++] = ff ;
                *ff++ = ii + 'a' - 1 ;
                *ff++ = _DRV_CHAR ;
                *ff++ = DIR_CHAR ;
                *ff   = '\0' ;
            }
        }
    }
    else
    {
        struct ffblk fblk ;
        char  *ff, *ee, es[4] ;
        int    done ;
        
        /* append the *.* - Note that this function assumes the pathName has a '/' and
         * its an array with 3 extra char larger than the string size */
        ee = pathName + strlen(pathName) ;
        ee[0] = '*' ;
        es[1] = ee[1] ;
        ee[1] = '.' ;
        es[2] = ee[2] ;
        ee[2] = '*' ;
        es[3] = ee[3] ;
        ee[3] = '\0' ;
        done = findfirst(pathName,&fblk,FA_RDONLY|FA_HIDDEN|FA_SYSTEM|FA_DIREC) ;
        ee[0] = '\0' ;
        ee[1] = es[1] ;
        ee[2] = es[2] ;
        ee[3] = es[3] ;
        while(!done)
        {
            if(((noFiles & 0x0f) == 0) &&
               ((fls = meRealloc(fls,sizeof(char *) * (noFiles+16))) == NULL))
            {
                noFiles = 0 ;
                break ;
            }
            if((ff = meMalloc(strlen(fblk.ff_name)+2)) == NULL)
            {
                fls = NULL ;
                noFiles = 0 ;
                break ;
            }
            fls[noFiles++] = ff ;
            strcpy(ff,fblk.ff_name) ;
            makestrlow(ff) ;
            if(fblk.ff_attrib & FA_DIREC)
            {
                ff += strlen(ff) ;
                *ff++ = DIR_CHAR ;
                *ff   = '\0' ;
            }
            done = findnext(&fblk) ;
        }
        dirList->timeStamp = 0 ;
    }
#endif
#ifdef _WIN32
    if(pathName[0] == '\0')
    {
        char  *ff ;
        int    ii ;
        DWORD  dwAvailableDrives, dwMask;
            
        /* Get the drives */
        dwAvailableDrives = GetLogicalDrives ();

        /* Drives are a-z (bit positions 0-25) */
        for (ii = 1, dwMask = 1; ii <= 26; ii++, dwMask <<= 1)
        {
            if((ii == 1) || (dwAvailableDrives & dwMask)) /* Drive exists ?? */
            {
                /* Yes - add to the drive list */
                if(((fls = meRealloc(fls,sizeof(char *) * (noFiles+1))) == NULL) ||
                   ((ff = meMalloc(4*sizeof(char))) == NULL))
                {
                    fls = NULL ;
                    noFiles = 0 ;
                    break ;
                }
                fls[noFiles++] = ff ;
                *ff++ = ii + 'a' - 1;
                *ff++ = _DRV_CHAR ;
                *ff++ = DIR_CHAR ;
                *ff   = '\0' ;
            }
        }
    }
    else
    {
        HANDLE *handle;
        WIN32_FIND_DATA fd;
        char  *ff, *ee, es[4] ;
        
        /* append the *.* - Note that this function assumes the pathName has a '/' and
         * its an array with 3 extra char larger than the string size */
        ee = pathName + strlen(pathName) ;
        ee[0] = '*' ;
        es[1] = ee[1] ;
        ee[1] = '.' ;
        es[2] = ee[2] ;
        ee[2] = '*' ;
        es[3] = ee[3] ;
        ee[3] = '\0' ;
        handle = FindFirstFile(pathName,&fd) ;
        ee[0] = '\0' ;
        ee[1] = es[1] ;
        ee[2] = es[2] ;
        ee[3] = es[3] ;
        if(handle != INVALID_HANDLE_VALUE)
        {
            do
            {
                if(((noFiles & 0x0f) == 0) &&
                   ((fls = meRealloc(fls,sizeof(char *) * (noFiles+16))) == NULL))
                {
                    noFiles = 0 ;
                    break ;
                }
                if((ff = meMalloc(strlen(fd.cFileName)+2)) == NULL)
                {
                    fls = NULL ;
                    noFiles = 0 ;
                    break ;
                }
                fls[noFiles++] = ff ;
                strcpy(ff,fd.cFileName) ;
                if(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
                {
                    ff += strlen(ff) ;
                    *ff++ = DIR_CHAR ;
                    *ff   = '\0' ;
                }
            } while (FindNextFile (handle, &fd));
            FindClose (handle);
        }
        dirList->timeStamp = 0 ;
    }
#endif
#ifdef _UNIX
    {
        DIR    *dirp ;
        
        if((dirp = opendir(pathName)) != NULL)
        {
#if (defined _SYSV42)
            struct  dirent *dp ;
#else
            struct  direct *dp ;
#endif
            struct stat statbuf;
            char *ff, *bb, fname[FILEBUF] ;
            
            strcpy(fname,pathName) ;
            bb = fname + strlen(fname) ;
            
            while((dp = readdir(dirp)) != NULL)
            {
                if(((noFiles & 0x0f) == 0) &&
                   ((fls = meRealloc(fls,sizeof(char *) * (noFiles+16))) == NULL))
                {
                    noFiles = 0 ;
                    break ;
                }
                if((ff = meMalloc(strlen(dp->d_name)+2)) == NULL)
                {
                    fls = NULL ;
                    noFiles = 0 ;
                    break ;
                }
                fls[noFiles++] = ff ;
                strcpy(ff,dp->d_name) ;
                if((ff[0] == '.') && ((ff[1] == '\0') || ((ff[1] == '.') && (ff[2] == '\0'))))
                {
                    ff += (ff[1] == '\0') ? 1:2 ;
                    *ff++ = DIR_CHAR ;
                    *ff   = '\0' ;
                }
#ifdef DT_DIR
                else if(dp->d_type == DT_DIR)
                {
                    ff += strlen(ff) ;
                    *ff++ = DIR_CHAR ;
                    *ff   = '\0' ;
                }
                else if(dp->d_type == DT_UNKNOWN)
#else
                else
#endif
                {
                    strcpy(bb,dp->d_name) ;
                    if(!stat(fname,&statbuf) && S_ISDIR(statbuf.st_mode))
                    {
                        ff += strlen(ff) ;
                        *ff++ = DIR_CHAR ;
                        *ff   = '\0' ;
                    }
                }
            }
            closedir(dirp) ;
        }
        dirList->timeStamp = timeStamp ;
    }
#endif  /* _UNIX */
    dirList->path = strdup(pathName) ;
    dirList->size = noFiles ;
    dirList->list = fls ;
    if(noFiles > 1)
#ifdef _INSENSE_CASE
        sortStrings(noFiles,fls,0,icstrdif) ;
#else
        sortStrings(noFiles,fls,0,strcmp) ;
#endif
}

/*
** frees the file list created by getFileList
*/
void
freeFileList(int noStr, char **files)
{
    if(files == NULL)
        return ;

    while (--noStr >= 0)
        meFree(files[noStr]) ;

    meFree((void *) files) ;
}


