/*****************************************************************************
*
*	Title:		%M%
*
*	Synopsis:	Display routines.
*
******************************************************************************
*
*	Filename:		%P%
*
*	Author:			Steven Phillips
*
*	Creation Date:		19/2/95
*
*	Modification date:	%G% : %U%	<160699.0734>
*
*	Current rev:		%I%
*
*	Special Comments:	
*
*	Contents Description:	
*
*
* The functions in this file handle hilighting, used by the display and the 
* printer.
*
*
******************************************************************************
* 
* Copyright (C) 1995 - 1999, JASSPA 
* The MicroEmacs Jasspa distribution can be copied and distributed freely for
* any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
* incorporated into commercial software with the expressed permission of
* JASSPA.
* 
****************************************************************************/

#define	__HILIGHTC				/* Name file */

#include "emain.h"

/*
 * HilFunc
 * This function is invoked whenever the source text reaches a defined postion,
 * presently defined at the start and end of the highlight selection blocks.
 * The call back function deals with the change in colour that results. 
 * */
struct HILDATA;                         /* Forward reference structure */
typedef void (*HilFunc)(int, struct HILDATA *);

/*
 * HILDATA
 * This structure contains data that is exported from hilightLine to the
 * worker functions. It contains enough information to allow the selection
 * colour modifications to be performed.
 */
typedef struct HILDATA {
    HILNODEPTR root;                    /* Root of the hilighting */
    HILNODEPTR node;                    /* Current node used in hilighting */
    HILNODEPTR cnode;                   /* Current working node */
    HILNODEPTR hnode;                   /* Current hilighting node */
    HILCOLOR *colp;                     /* Color pointer */
    int noColChng;                      /* Number of colour changes */
    int srcPos;                         /* Source position */
    unsigned short srcOff;              /* Offset in source to find */
    HilFunc hfunc;                      /* Callback function */
    unsigned short len;                 /* Transient length information */
    unsigned short flag;                /* Video structure flag */
    unsigned char colno;                /* Color number */
} HILDATA;    

#define meHIL_NOFIXEDTEST 10
#define meHIL_NOTEST      0x00
#define meHIL_TESTTOKEN   0x9a

static char *meHilTestNames[meHIL_NOFIXEDTEST]={
    "space", "digit", "hex", "lower", "upper", "alpha", "alphanum", "sword", "word", "any", 
} ;

#define meHIL_MODECASE  0x01
#define meHIL_MODESTART 0x02
#define meHIL_MODESTTLN 0x04
#define meHIL_MODETOEOL 0x08
#define meHIL_MODEEXIT  0x10

#define HLASOL      0x2000
#define HLASTTLINE  0x4000

/* mlerase relegated to here due to its use of mlwrite */
/*
 * Erase the message line. This is a special routine because the message line
 * is not considered to be part of the virtual screen. It always works
 * immediately; the terminal buffer is flushed via a call to the flusher.
 */
void
mlerase(int flag)
{
    if(clexec && (flag & MWCLEXEC))
        return ;
    if((mlStatus & MLSTATUS_RESTORE) && !(flag & MWERASE))
    {
        mlCol = mlStoreCol ;
        mlwrite(((mlStatus & MLSTATUS_POSML) ? MWSPEC|MWUSEMLCOL|MWCURSOR:MWSPEC),mlStore) ;
        mlStatus = (mlStatus & ~MLSTATUS_RESTORE) | MLSTATUS_KEEP ;
    }
    else if(!(mlStatus & MLSTATUS_KEEP))
    {
        mlwrite(MWSPEC|(flag&MWCURSOR),"") ;
        mlStatus &= ~MLSTATUS_CLEAR ;
    }
}

/*
 * hilValidateScheme
 * Validate a hilighting colour.
 */
int
hilValidateScheme (int n, int defaultRtn)
{
    if ((n < 0) || (n >= colSchmTblSize))
    {
        mlwrite (MWABORT|MWPAUSE,"[Invalid color-scheme index %d]", n);
        return defaultRtn ;
    }
    return n;
}

#if HILIGHT

static unsigned char varTable[11] ;

void
freeToken(HILNODEPTR root)
{
    unsigned char ii, top ;
    
    top = hilListSize(root) ;
    for(ii=0 ; ii<top ; ii++)
        freeToken(root->list[ii]) ;
    meNullFree(root->table) ;
    meNullFree(root->list) ;
    meNullFree(root->close) ;
    meNullFree(root->rtoken) ;
    meNullFree(root->rclose) ;
    free(root) ;
}

HILNODEPTR
createHilight(unsigned char hilno, unsigned char *noHils , HILNODEPTR **hTbl)
{
    HILNODEPTR    root ;
    
    if(hilno >= *noHils)
    {
        if((*hTbl = (HILNODEPTR *) 
             meRealloc(*hTbl,(hilno+1)*sizeof(HILNODEPTR))) == NULL)
            return NULL ;
        memset((*hTbl)+(*noHils),0,(hilno+1-(*noHils))*sizeof(HILNODEPTR)) ;
        *noHils = hilno+1 ;
    }
    else
    {
        if((root=(*hTbl)[hilno]) != NULL)
        {
            HILNODEPTR hn ;
            while((hn = (HILNODEPTR) root->rclose) != NULL)
            {
                root->rclose = hn->rclose ;
                meFree(hn) ;
            }
            root->rtoken = NULL ;
            freeToken(root) ;
        }
        (*hTbl)[hilno] = NULL ;
    }
    if((root = meMalloc(sizeof(HILNODE))) == NULL)
        return NULL ;
    memset(root,0,sizeof(HILNODE)) ;
    return root ;
}

void
ListToTable(HILNODEPTR root)
{
    unsigned char pos, ii, size ;
    unsigned char cc ;
    
    if(root->listSize < 8)
        return ;
    
    root->table = (unsigned char *) meMalloc(TableSize) ;
    
    root->table[0] = root->ordSize ;
    pos  = root->ordSize ;
    size = hilListSize(root) ;
    for(; (pos<size) && ((unsigned char)(root->list[pos]->token[0]) < TableLower) ; pos++)
        ;
    cc = TableLower ;
    for(ii=1 ; ii<TableSize ; ii++,cc++)
    {
        root->table[ii] = pos ;
        for(; (pos<size) && (root->list[pos]->token[0] == cc) ; pos++)
            ;
    }
    root->table[TableSize-1] = size;
}

#define ADDTOKEN_REMOVE   0x001
#define ADDTOKEN_DUMMY    0x002
#define ADDTOKEN_MAINTOK  0x004
#define ADDTOKEN_MAINRPL  0x008
#define ADDTOKEN_CLOSETOK 0x010
#define ADDTOKEN_CLOSERPL 0x020
#define ADDTOKEN_SOL      HLASOL           /* 0x2000 */
#define ADDTOKEN_STTLINE  HLASTTLINE       /* 0x4000 */

#define tokenSetType(node,tt) \
((node)->type = (((node)->type & (HLASOL|HLASTTLINE)) | (tt)))


HILNODEPTR
addTokenNode(HILNODEPTR root, unsigned char *token, int flags)
{
    HILNODEPTR    node, nn ;
    unsigned char c1=1, ii, jj, top, spec ;
    int len ;
    
    if((*token == 0xff) && (*++token != 0xff))
    {
        /* special test */
        ii  = 0 ;
        top = root->ordSize ;
        spec = 1 ;
    }
    else
    {
        ii = root->ordSize ;
        top = hilListSize(root) ;
        spec = 0 ;
    }
    for( ; ii<top ; ii++)
    {
        unsigned char c2, *s1, *s2 ;
        node = root->list[ii] ;
        s1 = token ;
        s2 = node->token ;
        while((c1 = *s1) == (c2 = *s2))
        {
            if(c1 == '\0')
            {
                if(flags & ADDTOKEN_DUMMY)
                {
                    /* if adding a dummy, the node already exists as a legit, 
                     * the only thinkg to do is correct the ASTTLINE & ASOL flags,
                     * otherwise leave it alone */
                    node->type &= (~(HLASOL|HLASTTLINE)) | flags ; 
                    return node ;
                }
                /* if removing token or readding, this node must
                 * be reset
                 */
                if(node->rtoken != NULL)
                {
                    meFree(node->rtoken) ;
                    node->rtoken = NULL ;
                }
                if(node->close != NULL)
                {
                    meFree(node->close) ;
                    node->close = NULL ;
                }
                if(node->rclose != NULL)
                {
                    meFree(node->rclose) ;
                    node->rclose = NULL ;
                }
                tokenSetType(node,(flags & ADDTOKEN_REMOVE) ? HLVALID:0) ;
                return node ;
            }
            s1++ ;
            s2++ ;
        }
        if(c1 == '\0')
        {
            if(flags & ADDTOKEN_REMOVE)
                return NULL ;
            break ;
        }
        if(c2 == '\0')
        {
            node->type &= (~(HLASOL|HLASTTLINE)) | flags ; 
            return addTokenNode(node,s1,flags) ;
        }
        if(c2 > c1)
            break ;
    }
    if(flags & ADDTOKEN_REMOVE)
        return NULL ;
                
    /* ii points to position to insert top must now be the table size */
    top = hilListSize(root) ;
    /* going to be added to this list, check that we have enough room */
    if(top++ == 0xff)
    {
        mlwrite(MWABORT|MWWAIT,"Table full, cant add [%s]. Optimise!",token) ;
        return NULL ;
    }
    len = strlen((char *) token) ;
    if(((root->list = (HILNODEPTR *) meRealloc(root->list,top*sizeof(HILNODEPTR))) == NULL) ||
       ((nn = meMalloc(sizeof(HILNODE)+len)) == NULL))
        return NULL ;
    memset(nn,0,sizeof(HILNODE)) ;
    nn->type = ((flags & ADDTOKEN_DUMMY) ?
                (HLVALID|(flags & (HLASOL|HLASTTLINE))):HLASOL|HLASTTLINE) ;
    strcpy((char *) nn->token,(char *) token) ;
    
    if(c1 == 0)
    {
        int kk, ll ;
        for(jj=0,kk=ii ;  kk<top-1 ; kk++,jj++)
            if(strncmp((char *) token,(char *) root->list[kk]->token,len))
                break ;
        if((nn->list = (HILNODEPTR *) meMalloc(jj*sizeof(HILNODEPTR))) == NULL)
            return NULL ;
        for(jj=0,ll=ii ;  ll<kk ; ll++,jj++)
        {
            nn->list[jj] = root->list[ll] ;
            strcpy((char *) nn->list[jj]->token,(char *) nn->list[jj]->token+len) ;
            if((nn->list[jj]->type & (HLASOL|HLSOL))  != (HLASOL|HLSOL))
                nn->type &= ~HLASOL ;
            if((nn->list[jj]->type & (HLASTTLINE|HLSTTLINE))  != (HLASTTLINE|HLSTTLINE))
                nn->type &= ~HLASTTLINE ;
        }
        nn->listSize = jj ;
        root->list[ii++] = nn ;
        for( ; ll<top ; ii++,ll++)
            root->list[ii] = root->list[ll] ;
        jj = 1 - jj ;
    }
    else
    {
        for(jj=top-1 ; jj>ii ; jj--)
            root->list[jj] = root->list[jj-1] ;
        root->list[ii] = nn ;
        jj = 1 ;
    }
    if(spec)
        (root->ordSize) += jj ;
    else
        (root->listSize) += jj ;
        
    if(root->table == NULL)
        ListToTable(root) ;
    else
    {
        if(spec)
            ii = 0 ;
        else
            ii = GetTableSlot((unsigned char)(*token)) + 1 ;
        for(; ii<TableSize ; ii++)
            (root->table[ii]) += jj ;
    }
    return nn ;
}


HILNODEPTR
addTokenString(HILNODEPTR root, HILNODEPTR node, unsigned char *token, int flags)
{
    unsigned char buff[MAXBUF], *dd, *ss, cc ;
    int l1=0, l2=-1, len=0 ;
    
    ss = token ;
    dd = buff ;
    if(flags & (ADDTOKEN_MAINTOK|ADDTOKEN_CLOSETOK))
    {
        unsigned char chgCase=(root->type & HFCASE) ;
        int charSlot = 1 ;
        while((cc=*ss++) != '\0')
        {
            if(cc == '[')
            {
                unsigned char *dds, test ;
                int ii ;
                
                if((flags & ADDTOKEN_MAINTOK) && (dd != buff) &&
                   ((*dd = '\0'), (addTokenNode(root,buff,flags|ADDTOKEN_DUMMY) == NULL)))
                    return NULL ;
                dds = dd ;
                dd += 4 ;
                if(*ss == '^')
                {
                    test = 0x90 ;
                    ss++ ;
                }
                else
                    test = 0x80 ;
                while((cc=*ss++) != ']')
                {
                    if((cc == '\\') && ((cc=*ss++) == '-') &&
                       (dd != dds+4) && (dd[-1] != '-') && (*ss != '-'))
                    {
                        unsigned char *ddt=dd ;
                        do
                            ddt[0] = ddt[-1] ;
                        while (--ddt != dds+4) ;
                        ddt[0] = '-' ;
                        dd++ ;
                    }
                    else if(cc == '\0')
                        return NULL ;
                    else
                        *dd++ = cc ;
                }
                *dd = 0 ;
                for(ii=0 ; ii<meHIL_NOFIXEDTEST ; ii++)
                    if(!strcmp((char *) dds+4,meHilTestNames[ii]))
                        break ;
                if(ii < meHIL_NOFIXEDTEST)
                    ii += 2 ;
                else if((dds+9 == dd) && isDigit(dds[8]) && !strncmp((char *) dds+4,"char",4))
                    ii = 0 ;
                else
                    ii = 1 ;
                *dds++ = 0xff ;
                if((cc=*ss) == '?')
                {
                    *dds++ = 0x20|test|ii ;
                    *dds++ = 10 ;
                    ss++ ;
                }
                else if(cc == '+')
                {
                    *dds++ = 0x40|test|ii ;
                    *dds++ = 10 ;
                    ss++ ;
                }
                else if(cc == '*')
                {
                    *dds++ = 0x60|test|ii ;
                    *dds++ = 10 ;
                    ss++ ;
                }
                else
                {
                    *dds++ = test|ii ;
                    *dds++ = charSlot  ;
                    if(charSlot < 10)
                        charSlot++ ;
                }                    
                if(ii == 0)
                {
                    *dds++ = dd[-1] - '0' ;
                    dd = dds ;
                }
                else if(ii == 1)
                    *dds = ((int) dd) - ((int) dds) - 1 ;
                else
                    dd = dds ;
                len++ ;
            }
            else if(cc == '\\')
            {
                cc = *ss++ ;
                if(cc == '<')
                    l1 = len ;
                else if(cc == '>')
                    l2 = len ;
                else
                {
                    if(chgCase)
                        cc = toLower(cc) ;
                    *dd++ = cc ;
                    len++ ;
                }
            }
            else
            {
                if(cc == 0xff)
                {
                    *dd++ = 0xff ;
                    cc = *ss++ ;
                }
                if(chgCase)
                    cc = toLower(cc) ;
                *dd++ = cc ;
                len++ ;
            }
        }
    }
    else
    {
        while((cc=*ss++) != '\0')
        {
            if(cc == '\\')
            {
                cc = *ss++ ;
                if((cc > '0') && (cc <= '9'))
                {
                    *dd++ = 0xff ;
                    cc -= '0' ;
                }
            }
            else if(cc == 0xff)
            {
                *dd++ = 0xff ;
                cc = *ss++ ;
            }
            *dd++ = cc ;
            len++ ;
        }
    }
    *dd = 0 ;
    if(l2 < 0)
        l2 = len ;
    l2 = len - l2 ;
    len -= l1 + l2 ;
    
    if(flags & ADDTOKEN_MAINTOK)
    {
        if((node = addTokenNode(root,buff,flags)) == NULL)
            return NULL ;
        node->tknSttOff = l1 ;
        node->tknEndOff = l2 ;
        /* store the len temporarily into color for len tests on the replace strings */
        node->color = len ;
    }
    else
    {
        if((dd = (unsigned char *) strdup((char *) buff)) == NULL)
            return NULL ;

        if(flags & ADDTOKEN_MAINRPL)
        {
            node->rtoken = dd ;
            if(len != node->color)
            {
                node->type |= HLRPLCDIFF ;
                root->type |= HFRPLCDIFF ;
            }
        }
        else if(flags & ADDTOKEN_CLOSETOK)
        {
            node->close = dd ;
            node->clsSttOff = l1 ;
            node->clsEndOff = l2 ;
            /* store the len temporarily into color for len tests on the replace strings */
            node->color = len ;
        }
        else if(flags & ADDTOKEN_CLOSERPL)
        {
            node->rclose = dd ;
            if(len != node->color)
            {
                node->type |= HLRPLCDIFF ;
                root->type |= HFRPLCDIFF ;
            }
        }
    }
    return node ;
}

HILNODEPTR
addToken(HILNODEPTR root, char *token, unsigned short type, int flags)
{
    HILNODEPTR     node ;
    unsigned char  cc, *ss, *tss ;
    unsigned char  sttTst, endTst ;
    int            ii, jj ;
    
    ss = (unsigned char *) token ;
    
    /* Count the number of \\< & \\> and set tss to the end of the token */
    ii = 0 ;
    jj = 0 ;
    tss=ss ;
    while((cc=*tss++) != '\0')
        if(cc == '\\')
        {
            if((cc=*tss++) == '<')
                ii++ ;
            else if(cc == '>')
                jj++ ;
        }
    tss-- ;
    
    /* process the start of the token */
    if(*ss == '^')
    {
        /* replace a starting "^..." with the HLSOL flag and
         * "^[space]*...\\<..." with the HLSTTLINE flag */
        ss++ ;
        if(ii && !strncmp((char *) ss,"[space]*",8))
        {
            type |= HLSTTLINE ;
            ss += 8 ;
        }
        else
            type |= HLSOL ;
    }
    /* setup the start Test flag */
    sttTst = meHIL_NOTEST ;
    if(type & HLTOKEN)
        sttTst = meHIL_TESTTOKEN ;
    else if(ii && (*ss == '['))
    {
        unsigned char *s1=ss+1, *s2, test ;
        if(*s1 == '^')
        {
            test = 0x90 ;
            s1++ ;
        }
        else
            test = 0x80 ;
        s2 = s1 ;
        while((cc=*s2++) != ']')
            if(cc == '\0')
                return NULL ;
        if(((cc=*s2) != '?') && (cc != '+') && (cc != '*'))
        {
            int kk ;
            *--s2 = 0 ;
            for(kk=0 ; kk<meHIL_NOFIXEDTEST ; kk++)
                if(!strcmp((char *) s1,meHilTestNames[kk]))
                {
                    sttTst = test|(kk+2) ;
                    ss = s2+1 ;
                    break ;
                }
            *s2 = ']' ;
        }
    }
    if(ii && (*ss == '\\') && (ss[1] == '<'))
    {
        /* ditch a \\< at the start */
        ss += 2 ;
        ii-- ;
    }

    /* process the end of the token */
    if(jj && (tss[-1] == '>') && (tss[-2] == '\\'))
    {
        /* ditch a \\> at the end */
        tss -= 2 ;
        jj-- ;
    }
    if(!jj && !strncmp((char *) tss-6,"[any]*",6))
    {
        /* replace an ending "...[any]*" with the HLENDLINE - 
         * only if theres no "\\<" */
        type |= HLENDLINE ;
        tss -= 6 ;
    }
    /* setup the start Test flag */
    endTst = meHIL_NOTEST ;
    if(type & HLTOKEN)
        endTst = meHIL_TESTTOKEN ;
    else if(jj && (tss[-1] == ']'))
    {
        unsigned char *s1=tss-1, *s2, *s3, test ;
        int kk ;
        s2 = s1 ;
        while(--s1 != ss)
        {
            if(*s1 == '[')
            {
                s3 = s1++ ;
                if(*s1 == '^')
                {
                    test = 0x90 ;
                    s1++ ;
                }
                else
                    test = 0x80 ;
                *s2 = 0 ;
                for(kk=0 ; kk<meHIL_NOFIXEDTEST ; kk++)
                    if(!strcmp((char *) s1,meHilTestNames[kk]))
                    {
                        endTst = test|(kk+2) ;
                        tss = s3 ;
                        break ;
                    }
                *s2 = ']' ;
                break ;
            }
        }
    }
    if(jj && (tss[-1] == '>') && (tss[-2] == '\\'))
    {
        /* ditch a \\> at the end */
        tss -= 2 ;
        jj-- ;
    }
    *tss = '\0' ;
    
    /* have we got anything left? if not best fail as addTokenNode will fail */
    if(*ss == '\0')
        return NULL ;
    
    if(type & HLSOL)
        flags |= ADDTOKEN_SOL|ADDTOKEN_STTLINE ;
    else if(type & HLSTTLINE)
        flags |= ADDTOKEN_STTLINE ;
    
    if(((node = addTokenString(root,root,ss,flags)) == NULL) || (node == NULL))
        return NULL ;
    
    node->tknSttTst = sttTst ;
    node->tknEndTst = endTst ;
    
    if(type & (HLBRANCH|HLBRACKET))
        type &= ~HLCONTIN ;
    tokenSetType(node,type) ;
    
    return node ;
}

int
hilight(int f, int n)
{
    HILNODEPTR     root, node ;
    char           buf[MAXBUF] ;
    unsigned short type ;
    unsigned char  hilno ;
    int            ii ;
    
    if(n == 0)
    {
        if((mlreply("Enter hi-light number",0,0,buf,MAXBUF-1) != TRUE) ||
           ((hilno = (unsigned char) meAtoi(buf)) == 0) || 
           (mlreply("Flags",0,0,buf,MAXBUF-1) != TRUE))
            return ABORT ;
        type = (unsigned short) meAtoi(buf) ;
        
        if((root = createHilight(hilno,&noHilights,&hilights)) == NULL)
            return FALSE ;
        
        root->type = type ;
        if(type & HFLOOKB)
        {
            if(mlreply("lines",0,0,buf,MAXBUF) != TRUE)
                return ABORT ;
            root->ignore = (unsigned char) meAtoi(buf) ;
        }
        if(mlreply("default color index",0,0,buf,MAXBUF) != TRUE)
            return ABORT ;
        root->color = (unsigned short) meAtoi(buf) ;
        if (root->color != hilValidateScheme (root->color, -1))
            return (FALSE);
        root->color *= HCTNUM;;
        hilights[hilno] = root ;
        return TRUE ;
    }
    n = (n < 0) ? ADDTOKEN_REMOVE:ADDTOKEN_MAINTOK ;
        
    if((mlreply("Hi number",0,0,buf,MAXBUF-1) != TRUE) ||
       ((hilno = (unsigned char) meAtoi(buf)) == 0) ||
       (hilno > noHilights) ||
       ((root  = hilights[hilno]) == NULL) ||
       (mlreply("Type",0,0,buf,MAXBUF-1) != TRUE))
        return FALSE ;
    type = (unsigned short) meAtoi(buf) ;
    if(type & HLCOLUMN)
    {
        int fmCol, toCol ;
        
        if((mlreply("From",0,0,buf,MAXBUF-1) != TRUE) ||
           ((fmCol = (unsigned char) meAtoi(buf)) < 0) ||
           (!(n & ADDTOKEN_REMOVE) &&
            ((mlreply("To",0,0,buf,MAXBUF-1) != TRUE) ||
             ((toCol = (unsigned char) meAtoi(buf)) < fmCol))))
            return FALSE ;
        while(((node = (HILNODEPTR) root->rclose) != NULL) && (((int) node->close) < fmCol))
            root = node ;
        if((node != NULL) && ((int) node->close) == fmCol)
        {
            if(n & ADDTOKEN_REMOVE)
            {
                root->rclose = node->rclose ;
                meFree(node) ;
                return TRUE ;
            }
        }
        else if(n & ADDTOKEN_REMOVE)
            return TRUE ;
        else if((node = meMalloc(sizeof(HILNODE))) == NULL)
            return ABORT ;
        else
        {
            memset(node,0,sizeof(HILNODE)) ;
            node->type = HLCOLUMN ;
            node->close = (unsigned char *) fmCol ;
            node->rclose = root->rclose ;
            root->rclose = (unsigned char *) node ;
        }
        node->rtoken = (unsigned char *) toCol ;
        goto get_color ;
    }
    
    if(mlreply("Token",0,0,buf,MAXBUF-1) != TRUE)
        return FALSE ;
    
    /* add the token */
    node = addToken(root,buf,type,n) ;
    if(n & ADDTOKEN_REMOVE)
        return TRUE ;
    if(node == NULL)
        return FALSE ;
    type = node->type ;
    
    if(!(type & HLVALID))
    {
        if(((type & HLREPLACE) && 
            ((mlreply("Replace",0,0,buf,MAXBUF-1) != TRUE) ||
             (addTokenString(root,node,(unsigned char *) buf,ADDTOKEN_MAINRPL) == NULL))) ||
           ((type & HLBRACKET) &&
            ((mlreply("Close",0,0,buf,MAXBUF-1) != TRUE) ||
             (addTokenString(root,node,(unsigned char *) buf,ADDTOKEN_CLOSETOK) == NULL) ||
             ((type & HLREPLACE) && 
              ((mlreply("Replace",0,0,buf,MAXBUF-1) != TRUE) ||
               (addTokenString(root,node,(unsigned char *) buf,ADDTOKEN_CLOSERPL) == NULL))) ||
             (mlreply("Ignore",0,0,buf,MAXBUF-1) != TRUE) ||
             ((node->ignore = buf[0]) > 255))) ||
           ((type & HLCONTIN) &&
            ((mlreply("Continue",0,0,buf,MAXBUF-1) != TRUE) ||
             ((node->close = (unsigned char *) strdup(buf)) == NULL) ||
             ((type & HLREPLACE) && 
              ((mlreply("Replace",0,0,buf,MAXBUF-1) != TRUE) ||
               ((node->rclose = (unsigned char *) strdup(buf)) == NULL))))) ||
           ((type & HLBRANCH) &&
            ((mlreply("hilno",0,0,buf,MAXBUF-1) != TRUE) ||
             ((node->ignore = (unsigned char) meAtoi(buf)) > 255))))
            return FALSE;
get_color:        
        /* Get the colour index. This uses the root hilight 
         * colour if not specified. */
        ii = mlreply("Color Index",0,0,buf,MAXBUF-1);
        if (ii == ABORT)
            return (FALSE);
        
        if (ii == TRUE) 
        {
            node->color = (unsigned short) meAtoi(buf);
            if (node->color != hilValidateScheme (node->color, -1))
                return (FALSE);
            node->color *= HCTNUM;
        }
        else
            node->color = root->color;
    }
    return TRUE ;
}


#define findTokenCharTest(ret,lastChar,srcText,tokTest,testStr)              \
do {                                                                         \
    unsigned char *__ts, ftctcc=*srcText++ ;                                 \
                                                                             \
    switch(tokTest & 0x0f)                                                   \
    {                                                                        \
    case 0:                                                                  \
        __ts = testStr ;                                                     \
        ret = (varTable[*testStr++] == ftctcc) ;                             \
        break ;                                                              \
    case 1:                                                                  \
        {                                                                    \
            unsigned char rc, nrc ;                                          \
                                                                             \
            __ts = testStr ;                                                 \
            for(ret=*testStr++ ; ret>0 ; ret--)                              \
            {                                                                \
                rc=*testStr++ ;                                              \
                nrc = *testStr ;                                             \
                if(nrc == '-')                                               \
                {                                                            \
                    /* range */                                              \
                    testStr++ ;                                              \
                    nrc = *testStr++ ;                                       \
                    ret -= 2 ;                                               \
                    if((ftctcc >= rc) && (ftctcc <= nrc))                    \
                    {                                                        \
                        testStr += ret - 1 ;                                 \
                        break ;                                              \
                    }                                                        \
                }                                                            \
                /* single char compare */                                    \
                else if(ftctcc == rc)                                        \
                {                                                            \
                    testStr += ret - 1 ;                                     \
                    break ;                                                  \
                }                                                            \
            }                                                                \
            break ;                                                          \
        }                                                                    \
    case 2:                                                                  \
        ret = isSpace(ftctcc) ;                                              \
        break ;                                                              \
    case 3:                                                                  \
        ret = isDigit(ftctcc) ;                                              \
        break ;                                                              \
    case 4:                                                                  \
        ret = isHexDigit(ftctcc) ;                                           \
        break ;                                                              \
    case 5:                                                                  \
        ret = isLower(ftctcc) ;                                              \
        break ;                                                              \
    case 6:                                                                  \
        ret = isUpper(ftctcc) ;                                              \
        break ;                                                              \
    case 7:                                                                  \
        ret = isAlpha(ftctcc) ;                                              \
        break ;                                                              \
    case 8:                                                                  \
        ret = isAlphaNum(ftctcc) ;                                           \
        break ;                                                              \
    case 9:                                                                  \
        ret = isSpllWord(ftctcc) ;                                           \
        break ;                                                              \
    case 10:                                                                 \
        ret = isWord(ftctcc) ;                                               \
        break ;                                                              \
    case 11:                                                                 \
        ret = 1 ;                                                            \
        break ;                                                              \
    }                                                                        \
    if(tokTest & 0x10)                                                       \
        ret = !ret ;                                                         \
    if(!ret)                                                                 \
    {                                                                        \
        if(tokTest & 0x20)                                                   \
        {                                                                    \
            ret = 1 ;                                                        \
            srcText-- ;                                                      \
        }                                                                    \
        tokTest = 0 ;                                                        \
    }                                                                        \
    else if((lastChar=ftctcc) == '\0')                                       \
        tokTest = 0 ;                                                        \
    else if(tokTest & 0x40)                                                  \
    {                                                                        \
        if(!(tokTest & 0x0e))                                                \
            testStr = __ts ;                                                 \
        tokTest |= 0x20 ;                                                    \
    }                                                                        \
} while(tokTest & 0x40)

int
__findTokenSingleCharTest(unsigned char cc, unsigned char tokTest)
{
    int ret ;
    switch(tokTest & 0x0f)
    {
    case 2:
        ret = isSpace(cc) ;
        break ;
    case 3:
        ret = isDigit(cc) ;
        break ;
    case 4:
        ret = isHexDigit(cc) ;
        break ;
    case 5:
        ret = isLower(cc) ;
        break ;
    case 6:
        ret = isUpper(cc) ;
        break ;
    case 7:
        ret = isAlpha(cc) ;
        break ;
    case 8:
        ret = isAlphaNum(cc) ;
        break ;
    case 9:
        ret = isSpllWord(cc) ;
        break ;
    case 10:
        ret = isWord(cc) ;
        break ;
    case 11:
        ret = 1 ;
        break ;
    }
    if(tokTest & 0x10)
        ret = !ret ;
    return ret ;
}    

#define findTokenSingleCharTest(cc,tokTest)                                  \
((tokTest == meHIL_NOTEST) || __findTokenSingleCharTest(cc,tokTest))

HILNODEPTR
findToken(HILNODEPTR root, unsigned char *text, unsigned char mode, 
          unsigned char lastChar, unsigned short *len)
{
    HILNODEPTR     nn, node ;
    unsigned char  cc, *s1, *s2;
    int            status;
    int            hi, low, mid ;
    unsigned short type ;
    
    if(*text == '\0')
    {
        /* one test is needed here for the one char token "\n"
         * This is either the \n char or \s11 white space char
         */
        low = root->listSize ;
        hi = root->ordSize ;
        while((--low >= 0) && ((nn=root->list[hi++]),((cc=nn->token[0]) <= meNLCHAR)))
        {
            if((cc == meNLCHAR) && (nn->token[1] == 0) &&
               !((type=nn->type) & HLVALID) && (nn->tknEndTst == meHIL_NOTEST) &&
               (!(mode & meHIL_MODETOEOL) || (type & HLBRINEOL)) &&
               (!(type & HLSTTLINE) || (mode & meHIL_MODESTTLN)) &&
               (!(type & HLSOL) || (mode & meHIL_MODESTART)) &&
               findTokenSingleCharTest(lastChar,nn->tknSttTst))
            {
                *len = (nn->tknEndOff) ? 1:0 ;
                return nn ;
            }
        }
        if((hi = root->ordSize) > 0)
        {
            s1 = text ;
            while(--hi >= 0)
            {
                nn = root->list[hi] ;
                if(!((type=nn->type) & HLVALID) && (nn->tknEndTst == meHIL_NOTEST) &&
                   (((mode & meHIL_MODETOEOL) == 0) || (type & HLBRINEOL)) &&
                   (!(type & HLSTTLINE) || (mode & meHIL_MODESTTLN)) &&
                   (!(type & HLSOL) || (mode & meHIL_MODESTART)) &&
                   findTokenSingleCharTest(lastChar,nn->tknSttTst))
                {
                    unsigned char rr, tt, lstc ;
                    s2 = nn->token ;
                    tt = *s2++ ;
                    s2++ ;
                    findTokenCharTest(rr,lstc,s1,tt,s2) ;
                    if(rr && (*s2 == '\0'))
                    {
                        varTable[nn->token[1]] = '\0' ;
                        *len = (nn->tknEndOff) ? 1:0 ;
                        return nn ;
                    }
                }
            }
        }
        return NULL ;
    }
    if(root->table == NULL)
    {
        low = root->ordSize ;
        hi  = low+root->listSize ;
    }
    else
    {
        cc = *text ;
        if(mode & meHIL_MODECASE)
            cc = toLower(cc) ;
        low = GetTableSlot(cc) ;
        hi  = root->table[low+1] ;
        low = root->table[low] ;
    }
    if(low != hi)
    {
        hi-- ;
        if(mode & meHIL_MODECASE)
        {
            do
            {
                mid = (low + hi) >> 1;		/* Get mid value. */
                s1 = text ;
                nn = root->list[mid] ;
                s2 = (unsigned char *) nn->token ;
                do {
                    cc = *s1++ ;
                    if((status=((int) toLower(cc)) - ((int) *s2++)) != 0)
                        break ;
                } while(*s2 != '\0') ;
                if(status == 0)
                    break ;
                else if (status < 0)
                {
                    if((status == '\0' - meNLCHAR) && (cc == '\0') && (*s2 == '\0') &&
                       (nn->tknEndTst == meHIL_NOTEST))
                    {
                        if(!nn->tknEndOff)
                            s1-- ;
                        break ;
                    }
                    hi  = mid-1 ;		/* Discard bottom half */
                }
                else
                    low = mid+1 ;		/* Discard top half */
            } while(low <= hi) ;		/* Until converges */
        }
        else
        {
            do
            {
                mid = (low + hi) >> 1;		/* Get mid value. */
                s1 = text ;
                nn = root->list[mid] ;
                s2 = (unsigned char *) nn->token ;
                do {
                    if((status=((int) (cc=*s1++)) - ((int) *s2++)) != 0)
                        break ;
                } while(*s2 != '\0') ;
                if(status == 0)
                    break ;
                else if (status < 0)
                {
                    if((status == '\0' - meNLCHAR) && (cc == '\0') && (*s2 == '\0') &&
                       (nn->tknEndTst == meHIL_NOTEST))
                    {
                        if(!nn->tknEndOff)
                            s1-- ;
                        break ;
                    }
                    hi  = mid-1 ;		/* Discard bottom half */
                }
                else
                    low = mid+1 ;		/* Discard top half */
            } while(low <= hi) ;		/* Until converges */
        }
        if(low <= hi)
        {
            type = nn->type ;
            if((cc != '\0') && (hilListSize(nn) > 0) &&
               (!(type & HLASTTLINE) || (mode & meHIL_MODESTTLN)) &&
               (!(type & HLASOL) || (mode & meHIL_MODESTART)) &&
               ((node = findToken(nn,s1,mode,lastChar,len)) != NULL))
            {
                *len += s1-text ;
                return node ;
            }
            if(!(type & HLVALID) && 
               (!(mode & meHIL_MODETOEOL) || (type & HLBRINEOL)) &&
               (!(type & HLSTTLINE) || (mode & meHIL_MODESTTLN)) &&
               (!(type & HLSOL) || (mode & meHIL_MODESTART)) &&
               findTokenSingleCharTest(lastChar,nn->tknSttTst) &&
               findTokenSingleCharTest(*s1,nn->tknEndTst))
            {
                *len = s1-text ;
                return nn ;
            }
        }
    }
    if((hi = root->ordSize) > 0)
    {
        /* note that findTokenCharTest may not set lstc even if it does 
         * match as the ? & * may match 0 chars - so theres no last!
         * Therefore initialize to a sensible value */
        unsigned char dd, lstc=1 ;
        mid = 0 ;
        
        for( ; mid < hi ; mid++)
        {
            nn = root->list[mid] ;
            s2 = nn->token ;
            dd = *s2++ ;
            s2++ ;
            s1 = text ;
            findTokenCharTest(cc,lstc,s1,dd,s2) ;
            if(cc)
            {
                if(lstc != '\0')
                {
                    if(mode & meHIL_MODECASE)
                    {
                        while((dd=*s2++) != '\0')
                        {
                            cc = *s1++ ;
                            if((toLower(cc)) != dd)
                                break ;
                        }
                    }
                    else
                    {
                        while((dd=*s2++) != '\0')
                            if((cc=*s1++) != dd)
                                break ;
                    }
                    if((dd == meNLCHAR) && (cc == '\0') &&
                       (nn->tknEndTst == meHIL_NOTEST))
                    {
                        if(!nn->tknEndOff)
                            s1-- ;
                        dd = *s2 ;
                    }
                }
                else
                {
                    if(!nn->tknEndOff)
                        s1-- ;
                    dd = (nn->tknEndTst == meHIL_NOTEST) ? *s2:0 ;
                    cc = '\0' ;
                }
                if(dd == '\0')
                {
                    varTable[nn->token[1]] = lstc ;
                    type = nn->type ;
                    if((cc != '\0') && (hilListSize(nn) > 0) &&
                       (!(type & HLASTTLINE) || (mode & meHIL_MODESTTLN)) &&
                       (!(type & HLASOL) || (mode & meHIL_MODESTART)) &&
                       ((node = findToken(nn,s1,mode,lastChar,len)) != NULL))
                    {
                        *len += s1-text ;
                        return node ;
                    }
                    if(!(type & HLVALID) && 
                       (!(mode & meHIL_MODETOEOL) || (type & HLBRINEOL)) &&
                       (!(type & HLSTTLINE) || (mode & meHIL_MODESTTLN)) &&
                       (!(type & HLSOL) || (mode & meHIL_MODESTART)) &&
                       findTokenSingleCharTest(lastChar,nn->tknSttTst) &&
                       findTokenSingleCharTest(*s1,nn->tknEndTst))
                    {
                        *len = s1-text ;
                        return nn ;
                    }
                }
            }
        }
    }
    return NULL ;
}

#define __hilCopyChar(dstPos,cc)                                             \
{                                                                            \
    /* the largest character size is a tab which is user definable */        \
    if(dstPos >= disLineSize)                                                \
    {                                                                        \
        disLineSize += 512 ;                                                 \
        disLineBuff = realloc(disLineBuff,disLineSize+32) ;                  \
    }                                                                        \
    if(isDisplayable(cc))                                                    \
        disLineBuff[dstPos++] = cc ;                                         \
    else if(cc == TAB)                                                       \
    {                                                                        \
        int ii=get_tab_pos(dstPos) + 1 ;                                     \
        do                                                                   \
            disLineBuff[dstPos++] = ' ' ;                                    \
        while(--ii) ;                                                        \
    }                                                                        \
    else if(cc < 0x20)                                                       \
    {                                                                        \
        disLineBuff[dstPos++] = '^' ;                                        \
        disLineBuff[dstPos++] = cc ^ 0x40 ;                                  \
    }                                                                        \
    else                                                                     \
    {                                                                        \
        /* Its a nasty character */                                          \
        disLineBuff[dstPos++] = '\\' ;                                       \
        disLineBuff[dstPos++] = '0' ;                                        \
        disLineBuff[dstPos++] = hexdigits[cc/0x10] ;                         \
        disLineBuff[dstPos++] = hexdigits[cc%0x10] ;                         \
    }                                                                        \
}

/*
 * hilColorChange.
 * This function manages the colour changes. Each change in colour results
 * in a call. The changes of colour are added to the "hd->colp" structure.
 * In addition the last colour node is retained in "hd->hnode" which is the
 * current hilighting colour. The selection hilighting uses this node to 
 * switch the selection colours on an off. The correct state is mainated by
 * this function.
 */
static void
hilColorChange (HILNODEPTR node, HILDATA *hd)
{
    register HILCOLOR *colp;
    
    if (hd->noColChng > 0)
    {
        /* Re-allocate a new colour block - run out of room */
        if(hd->noColChng == hilColS)
        {
            hilColS += 20 ;
            hilCol = (HILCOLOR *) realloc (hilCol, hilColS*sizeof(HILCOLOR)) ;
        }
        colp = hilCol + hd->noColChng;
    }
    else
    {
        colp = hilCol;
        colp->ccol = 0;
    }
    /* Apply the new change of colour to the new enry created. */
    hd->noColChng++ ;
    colp->icol = node->color + (unsigned short)(hd->colno);
    hd->hnode = node;
    hd->colp = colp;
}

static int
hilCopyChar(register int dstPos, register unsigned char cc, HILDATA *hd)
{
    /* Change the selection hilight if required. */
    if ((hd->srcOff != 0xffff) && (hd->srcPos >= hd->srcOff))
        hd->hfunc (dstPos, hd);
    /* Copy the character. */
    __hilCopyChar(dstPos,cc)
    return dstPos ;
}

static int
hilCopyReplaceString(int sDstPos, register unsigned char *srcText,
                     int len, HILDATA *hd)
{
    int dstPos = sDstPos ;
    int hoff, dstLen ;
    unsigned char cc ;
    
    if((hd->srcOff != 0xffff) && (hd->srcPos+len >= hd->srcOff))
    {
        /* as the source len and destination len could be different, the best
         * we can do is set the hilight change position to be the ratio
         * between the two */
        unsigned char *ss=srcText ;
        dstLen = 0 ;
        while((cc = *ss++) != 0)
        {
            if(cc == 0xff)
                ss++ ;
            dstLen++ ;
        }
        if(len)
            hoff = sDstPos+((hd->srcOff-hd->srcPos)*dstLen/len) ;
        else
            hoff = sDstPos+(hd->srcOff-hd->srcPos)*dstLen ;
    }
    else
        hoff = 0x7fffffff ;
    while((cc = *srcText++) != '\0')
    {
        if((cc == 0xff) && ((cc=*srcText++) != 0xff))
            cc = varTable[cc] ;
        if(dstPos >= hoff)
        {
            hd->hfunc (dstPos, hd);
            /* we may have another offset to cope with */
            if((hd->srcOff != 0xffff) && (hd->srcPos+len >= hd->srcOff))
            {
                if(len)
                    hoff = sDstPos+((hd->srcOff-hd->srcPos)*dstLen/len) ;
                else
                    hoff = sDstPos+(hd->srcOff-hd->srcPos)*dstLen ;
            }
            else
                hoff = 0x7fffffff ;
        }
        __hilCopyChar(dstPos,cc)
    }
    /* This is not quite right - but will have to do for present. Change
     * the hilighting according to the ammount of text used. */
    return dstPos ;
}

static int
hilCopyString(register int dstPos, register unsigned char *srcText,HILDATA *hd)
{
    unsigned char cc ;
    /* Handle the selection hilighting if enabled. */
    if (hd->srcOff != 0xffff)
    {
        unsigned short srcPos;
        
        srcPos = hd->srcPos;
        while((cc = *srcText++) != '\0')
        {
            if (hd->srcOff <= srcPos)
                (hd->hfunc)(dstPos, hd);
            __hilCopyChar(dstPos,cc);
            srcPos++;
        }
    }
    else
    {
        while((cc = *srcText++) != '\0')
            __hilCopyChar(dstPos,cc);
    }
    return dstPos ;
}

int
hilCopyLenString(register int dstPos, register unsigned char *srcText,
                 register int len, HILDATA *hd)
{
    unsigned char cc ;
    
    /* Handle the selection hilighting if enabled. */
    if ((hd->srcOff != 0xffff) && ((hd->srcOff - hd->srcPos) < len))
    {
        unsigned short srcPos;
        
        srcPos = hd->srcPos;
        while (--len >= 0)
        {
            if (hd->srcOff <= srcPos)
                (hd->hfunc)(dstPos, hd);
            
            cc = *srcText++ ;
            __hilCopyChar(dstPos,cc);
            srcPos++;
        }
    }
    else
    {
        while(len--)
        {
            cc = *srcText++ ;
            __hilCopyChar(dstPos,cc);
        }
    }
    return dstPos ;
}


/*
 * hilHandleCallback
 * This function handles the callbacks when the text has reached a designated
 * position in the source line.
 */
static void
hilHandleCallback (int dstPos, struct HILDATA *hd)
{
    /* Handle the start of the selection hilight */
    if (hd->flag & VFSHBEG)
    {
        /* Change to selection hilighting */
        hd->colno += HCTSELFORE;        /* To new index */
        hd->colp->ccol = dstPos;        /* Terminate the current colour */
        hilColorChange (hd->hnode, hd); /* New colour node */
        hd->colp->ccol = dstPos;        /* New end of this colour */
        hd->flag &= ~VFSHBEG;           /* Knock off the background mode */
        
        /* Reprime the callback with the location of the end of hilight */
        if (hd->flag & VFSHEND)
            hd->srcOff = selhilight.eoff;
        else
            hd->srcOff = 0xffff ; /* Set to unreachable value */
    }
    else if (hd->flag & VFSHEND)
    {
        /* Restore the existing hilighting */
        hd->colno -= HCTSELFORE;        /* To new index */
        hd->colp->ccol = dstPos;        /* Terminate current colour */
        hilColorChange (hd->hnode, hd); /* New colour node */
        hd->colp->ccol = dstPos;        /* New end of this colour */
        hd->flag &= ~VFSHEND;           /* Remove the end flag */
        hd->srcOff = 0xffff ;           /* Set to unreachable value */
    }
}

unsigned short 
hilightLine(VIDEO *vp1)
{
    unsigned char *srcText=vp1->line->l_text ;
    unsigned char  hilno = vp1->hilno, cc=meNLCHAR, mode ;
    HILNODEPTR     node;
    unsigned short len ;
    register int   srcWid=vp1->line->l_used, dstPos=0 ;
    HILDATA hd;
    
    /* Initialise the hilight data structure */
    hd.root = hilights[hilno];          /* Root of the hilighting */
    hd.flag = vp1->flag;                /* Cache the video flag */
    hd.noColChng = 0;                   /* Reset number of colour changes */
    hd.srcPos = 0;                      /* Reset the source position */
    hd.srcOff = 0xffff;                 /* No callback required */
    hd.colp = hilCol;                   /* Color pointer */
    
    /* Determine if we are processing a forground or hilighted line.
     * Determined by the current line flag */
    hd.colno = HCTFORE;
    if (hd.flag & VFCURRL) 
        hd.colno += HCTCURFORE ;

    /* Determine if we are handling a selection hilighted line.
     * If so then set up the starting background colour in accordance
     * with the start/end/all selection hilighting flags in the video
     * structure. */
    if (hd.flag & VFSHMSK)              /* Selection hilight to be done ? */
    {
        hd.colno += HCTSELFORE;         /* Assume selection hilight enabled */
        
        if (hd.flag & VFSHALL)          /* Whole line hilighted ?? */
            hd.flag &= ~VFSHMSK;        /* Yes - kill off the flag */
        else
        {
            /* Check for the beginning of a slection hilight. If it is at the 
             * beginning of the line then use this as the starting colour. */
            if (hd.flag & VFSHBEG)
            {
                if (selhilight.soff!=0) /* Starts in column 0 ?? */
                {   /* Does not start in column 0 */
                    hd.colno -= HCTSELFORE ;   /* remove selection hilight */
                    hd.srcOff = selhilight.soff;
                    hd.hfunc = hilHandleCallback;
                }
                else
                    hd.flag &=~VFSHBEG; /* Knock off the begin - found */
            }
        
            /* Check for the end of a selection hilight. If it is at the 
             * beginning of the line then it effectivelly disables the 
             * hilight. */
            if ((hd.flag & VFSHEND) && ((hd.flag & VFSHBEG) == 0))
            {
                if (selhilight.eoff==0) /* End's in column 0 ?? */
                {
                    hd.colno -= HCTSELFORE; /* Yes - use normal colour */
                    hd.flag &=~VFSHEND;     /* Knock off end - found */
                }
                else
                {
                    hd.srcOff = selhilight.eoff;
                    hd.hfunc = hilHandleCallback;
                }
            }
        }
    }
    
    mode = (meHIL_MODESTTLN|meHIL_MODESTART|(hd.root->type&HFCASE)) ;
    if(vp1->bracket != NULL)
    {
        node = vp1->bracket ;
        hilColorChange ((hd.cnode = node), &hd);
        goto BracketJump ;
    }
    hilColorChange ((hd.cnode = hd.root), &hd);
    
    while(!(mode & meHIL_MODEEXIT))
    {
        node = (HILNODEPTR) hd.root->rclose ;
        while(node != NULL)
        {
            if((int) node->close == hd.srcPos)
            {
                if((len = (int) node->rtoken) > srcWid)
                    len = srcWid ;
                len -= hd.srcPos ;
                goto column_token ;
            }
            else if((int) node->close > hd.srcPos)
                break ;
            node = (HILNODEPTR) node->rclose ;
        }
        if((node = findToken(hd.root,srcText+hd.srcPos,mode,cc,&len)) != NULL)
        {
column_token:
            if(len == 0)
                /* the match string was 0 in length - must be the terminating \0 (i.e. the \n)
                 * MUST exit this time else we will spin! */
                mode |= meHIL_MODEEXIT ;
            if(node->tknSttOff)
            {
                dstPos = hilCopyLenString(dstPos,srcText+hd.srcPos,node->tknSttOff,&hd) ;
                hd.srcPos += node->tknSttOff ;
                len -= node->tknSttOff ;
            }
            if(node->tknEndOff)
                len -= node->tknEndOff ;
            if(hd.colp->ccol == dstPos)
            {
                hd.colp->icol = node->color + hd.colno ;
                hd.hnode = node;
            }
            else
            {
                hd.colp->ccol = dstPos ;
                hilColorChange (node, &hd);
                hd.colp->ccol = dstPos ;
            }
            if(node->type & HLREPLACE)
                dstPos = hilCopyReplaceString(dstPos, (unsigned char *) node->rtoken,len,&hd) ;
            else if(len)
                dstPos = hilCopyLenString(dstPos,srcText+hd.srcPos,len,&hd) ;
            
            hd.srcPos += len ;
            
BracketJump:
            if(node->type & HLBRACKET)
            {
                unsigned char ignore, tt=1, *c1, *c2, ss, *s1, *s2 ;
                
                c1 = (unsigned char *) node->close ;
                ignore = node->ignore ;
                s1 = srcText+hd.srcPos ;
                for( ; ; hd.srcPos++)
                {
                    c2 = c1 ;
                    s2 = s1 ;
                    while((tt = *c2++) != '\0')
                    {
                        if((tt == 0xff) && ((tt = *c2++) != 0xff))
                        {
                            unsigned char ret, vv ;
                            
                            vv = *c2++ ;
                            findTokenCharTest(ret,ss,s2,tt,c2) ;
                            if(!ret)
                            {
                                tt = 1 ;
                                break ;
                            }
                            varTable[vv] = ss ;
                        }
                        else
                        {
                            ss = *s2++ ;
                            if(mode & meHIL_MODECASE)
                                ss = toLower(ss) ;
                            if(tt != ss)
                            {
                                if((tt != meNLCHAR) || (ss != '\0'))
                                    break ;
                            }
                        }
                        if(ss == '\0')
                        {
                            if((tt=*c2) != '\0')
                                break ;
                            if(!node->clsEndOff)
                                s2-- ;
                        }
                    }
                    if((tt == '\0') || (hd.srcPos == srcWid))
                        break ;
                    ss = *s1++ ;
                    dstPos = hilCopyChar(dstPos,ss, &hd) ;
                    if(ss == ignore)
                    {
                        if(++hd.srcPos == srcWid)
                            break ;
                        dstPos = hilCopyChar(dstPos,*s1++,&hd) ;
                    }
                }
                if(tt != '\0')
                {
                    hd.colp->ccol = dstPos ;
                    goto hiline_exit ;
                }
                len = s2-s1 ;
                if(node->clsSttOff)
                {
                    dstPos = hilCopyLenString(dstPos,s1,node->clsSttOff,&hd) ;
                    hd.srcPos += node->clsSttOff ;
                    s1 += node->clsSttOff ;
                    len -= node->clsSttOff ;
                }
                if(node->clsEndOff)
                    len -= node->clsEndOff ;
                if(node->type & HLREPLACE)
                    dstPos = hilCopyReplaceString(dstPos,(unsigned char *) node->rclose,
                                                  len,&hd) ;
                else
                    dstPos = hilCopyLenString(dstPos,s1,len,&hd) ;
                hd.srcPos += len ;
                hd.cnode = hd.root ;
            }
            else if(node->type & HLCONTIN)
            {
                char *ss ;
                int   pos ;
                
                ss  = (char *) node->close ;
                len = strlen(ss) ;
                pos = srcWid ;
                while((srcText[--pos] == ' ') && (pos > 0))
                    ;
                pos -= len-1 ;
                if((pos >= hd.srcPos) &&
                   !strncmp((char *)srcText+pos,ss,len))
                {
                    dstPos = hilCopyString(dstPos,srcText+hd.srcPos, &hd) ;
                    hd.colp->ccol = dstPos ;
                    goto hiline_exit ;
                }
            }
            if(node->type & HLENDLINE)
            {
                if(!(node->type & HLBRINEOL) ||
                   (hd.srcPos == srcWid))
                {
                    dstPos = hilCopyString(dstPos,srcText+hd.srcPos, &hd) ;
                    hd.colp->ccol = dstPos ;
                    node = NULL ;
                    goto hiline_exit ;
                }
                hd.colp->ccol = dstPos ;
                mode |= meHIL_MODETOEOL ;
                hd.cnode = node;
            }
            else
            {
                if(node->type & HLBRANCH)
                {
                    if((hilno = node->ignore) == 0)
                        hilno = vp1->wind->w_bufp->hiLight ;
                    hd.root = hilights[hilno] ;
                    hd.cnode = hd.root;
                }
                if(hd.colp->ccol == dstPos)
                {
                    hd.colp->icol = hd.cnode->color + hd.colno ;
                    hd.hnode = node;
                }
                else
                {
                    hd.colp->ccol = dstPos ;
                    hilColorChange(hd.cnode, &hd);
                    hd.colp->ccol = dstPos ;
                }
            }
            if(dstPos)
            {
                cc = disLineBuff[dstPos-1] ;
                mode &= ~(meHIL_MODESTTLN|meHIL_MODESTART) ;
            }
        }
        else
        {
            if((cc=srcText[hd.srcPos]) == 0)
                break ;
            if(mode & (meHIL_MODESTART|meHIL_MODESTTLN))
                mode &= ~((isSpace(cc)) ? meHIL_MODESTART:(meHIL_MODESTART|meHIL_MODESTTLN)) ;
            dstPos = hilCopyChar(dstPos,cc, &hd) ;
            hd.srcPos++ ;
        }
    }

    if(hd.colp->ccol != dstPos)
        hd.colp->ccol = dstPos ;
    node = NULL ;
hiline_exit:
    
    if((hd.flag & (VFSHBEG|VFSHEND)) == VFSHEND)
    {
        /* must end the selection hilighting */
        hd.colno -= HCTSELFORE ;
        hilColorChange (hd.hnode, &hd);
        hd.colp->ccol = dstPos;
    }
    else if((hd.flag & (VFSHBEG|VFSHEND)) == VFSHBEG)
    {
        /* must end the selection hilighting */
        hd.colno += HCTSELFORE ;
        hilColorChange (hd.hnode, &hd);
        hd.colp->ccol = dstPos;
    }

    if((vp1[1].hilno != hilno) || (vp1[1].bracket != node))
    {
        vp1[1].flag |= VFCHNGD ;
        vp1[1].hilno = hilno ;
        vp1[1].bracket = node ;
    }
    return hd.noColChng ;
}

void
hilLookBack(WINDOW *wp)
{
    VIDEO  *vptr ;
    BUFFER *bp ;
    register LINE *lp, *blp ;
    unsigned char  hilno ;
    HILNODEPTR     root, bracket ;
    int            ii, jj ;
    
    bp = wp->w_bufp ;
    hilno=bp->hiLight ;
    root = hilights[hilno] ;
    bracket = NULL ;
    
    ii = wp->topLineNo - wp->line_no ;
    jj = root->ignore ;
    lp = wp->w_dotp ;
    while(ii < jj)
    {
        if((blp=lback(lp)) == bp->b_linep)
            break ;
        lp = blp ;
        ii++ ;
    }
    if(ii > 0)
    {
        VIDEO vps[2] ;
        
        /* the flag is used only to set the current line, do the select hilighting
         * and flag the next line as changed if in a bracket, therefore init to 
         * 0 and forget about */
        vps[0].wind = wp ;
        vps[0].hilno = hilno ;
        vps[0].bracket = NULL ;
        vps[0].flag = 0 ;
        vps[1].hilno = hilno ;
        vps[1].bracket = NULL ;
        vps[1].flag = 0 ;
        for(;;)
        {
            vps[0].line = lp ;
            if((hilightLine(vps) > 1) && (vps[0].bracket != NULL) &&
               (vps[0].bracket == vps[1].bracket) &&
               (vps[0].bracket->type & HLONELNBT))
                vps[1].bracket = NULL ;
            if(--ii == 0)
                break ;
            vps[0].hilno = vps[1].hilno ;
            vps[0].bracket = vps[1].bracket ;
            lp = lforw(lp) ;
        }
        hilno = vps[1].hilno ;
        bracket = vps[1].bracket ;
    }
    vptr  = wp->w_vvideo->video + wp->firstRow ;  /* Video block */
    if((vptr->hilno != hilno) || (vptr->bracket != bracket))
    {
        vptr->hilno = hilno ;
        vptr->bracket = bracket ;
        vptr->flag |= VFCHNGD ;
    }
}

#define hilOffsetChar(off,dstPos,dstJmp,cc)                                  \
{                                                                            \
    int ii ;                                                                 \
    if(isDisplayable(cc))                                                    \
        ii = 1 ;                                                             \
    else if(cc == TAB)                                                       \
        ii = get_tab_pos(dstPos) + 1 ;                                       \
    else if(cc < 0x20)                                                       \
        ii = 2 ;                                                             \
    else                                                                     \
        ii = 4 ;                                                             \
    if(dstJmp)                                                               \
    {                                                                        \
        *off++ = ii+dstJmp ;                                                 \
        dstJmp = 0 ;                                                         \
    }                                                                        \
    else                                                                     \
        *off++ = ii ;                                                        \
    dstPos += ii ;                                                           \
}

#define hilOffsetString(lastcc,off,dstPos,dstJmp,srcText)                    \
{                                                                            \
    unsigned char __cc, *__ss=(srcText) ;                                    \
    while((__cc = *__ss++) != '\0')                                          \
    {                                                                        \
        hilOffsetChar(off,dstPos,dstJmp,__cc)                                \
        lastcc = __cc ;                                                      \
    }                                                                        \
}

#define hilOffsetLenString(lastcc,off,dstPos,dstJmp,srcText,len)             \
{                                                                            \
    unsigned short __ll=len ;                                                \
    unsigned char *__ss=(srcText) ;                                          \
    while(__ll--)                                                            \
    {                                                                        \
        lastcc = *__ss++ ;                                                   \
        hilOffsetChar(off,dstPos,dstJmp,lastcc)                              \
    }                                                                        \
}

static int
hilOffsetReplaceString(register unsigned char **offPtr, register int dstPos,
                       int *dstJmp, register unsigned char *srcText, int len)
{
    unsigned char *off=*offPtr, cc, lastcc ;
    int            rlen=dstPos, ii, dd=*dstJmp ;
    
    while((cc = *srcText++) != '\0')
    {
        if((cc == 0xff) && ((cc=*srcText++) != 0xff))
            cc = varTable[cc] ;
        lastcc = cc ;
        hilOffsetChar(off,dstPos,dd,cc)
    }
    
    rlen = dstPos-rlen ;
    if(len)
    {
        dd = *dstJmp ;
        off = *offPtr ;
        while(len)
        {
            ii = rlen / len-- ;
            if(dd)
            {
                *off++ = ii+dd ;
                dd = 0 ;
                *dstJmp = 0 ;
            }
            else
                *off++ = ii ;
            rlen -= ii ;
        }
        *offPtr = off ;
    }
    else
        *dstJmp += rlen ;
    *off = lastcc ;
    return dstPos ;
}

void
hilightCurLineOffsetEval(WINDOW *wp)
{
    unsigned char *srcText=wp->w_dotp->l_text ;
    unsigned char *off=wp->curLineOff->l_text ;
    unsigned char  hilno=wp->w_bufp->hiLight, cc=meNLCHAR, mode ;
    HILNODEPTR     node, root ;
    unsigned short len ;
    register int   srcWid=wp->w_dotp->l_used, srcPos=0, dstPos=0 ;
    int dstJmp=0 ;
    
    root = hilights[hilno];          /* Root of the hilighting */
    
    mode = (meHIL_MODESTTLN|meHIL_MODESTART|(root->type&HFCASE)) ;
/*    if(vp1->bracket != NULL)*/
/*    {*/
/*        node = vp1->bracket ;*/
/*        hilColorChange ((hd.cnode = node), &hd);*/
/*        goto BracketJump ;*/
/*    }*/
    while(!(mode & meHIL_MODEEXIT))
    {
        node = (HILNODEPTR) root->rclose ;
        while(node != NULL)
        {
            if((int) node->close == srcPos)
            {
                if((len = (int) node->rtoken) > srcWid)
                    len = srcWid ;
                len -= srcPos ;
                goto column_token ;
            }
            else if((int) node->close > srcPos)
                break ;
            node = (HILNODEPTR) node->rclose ;
        }
        if((node = findToken(root,srcText+srcPos,mode,cc,&len)) != NULL)
        {
column_token:
            if(len == 0)
                /* the match string was 0 in length - must be the terminating \0 (i.e. the \n)
                 * MUST exit this time else we will spin! */
                mode |= meHIL_MODEEXIT ;
            if(node->tknEndOff)
                len -= node->tknEndOff ;
            if(node->type & HLREPLACE)
            {
                if(node->tknSttOff)
                {
                    hilOffsetLenString(cc,off,dstPos,dstJmp,srcText+srcPos,node->tknSttOff) ;
                    srcPos += node->tknSttOff ;
                    len -= node->tknSttOff ;
                }
                dstPos = hilOffsetReplaceString(&off,dstPos,&dstJmp,(unsigned char *) node->rtoken,len) ;
                cc = *off ;
            }
            else
            {
                hilOffsetLenString(cc,off,dstPos,dstJmp,srcText+srcPos,len) ;
            }
            srcPos += len ;

            /*BracketJump:*/
            if(node->type & HLBRACKET)
            {
                unsigned char ignore, tt=1, *c1, ss, *c2, *s1, *s2 ;
                
                c1 = (unsigned char *) node->close ;
                ignore = node->ignore ;
                s1 = srcText+srcPos ;
                for( ; ; srcPos++)
                {
                    c2 = c1 ;
                    s2 = s1 ;
                    while((tt = *c2++) != '\0')
                    {
                        if((tt == 0xff) && ((tt = *c2++) != 0xff))
                        {
                            unsigned char ret, vv ;
                            
                            vv = *c2++ ;
                            findTokenCharTest(ret,ss,s2,tt,c2) ;
                            if(!ret)
                            {
                                tt = 1 ;
                                break ;
                            }
                            varTable[vv] = ss ;
                        }
                        else if((ss = *s2++) != tt)
                        {
                            if((tt != meNLCHAR) || (ss != '\0'))
                                break ;
                        }
                        if(ss == '\0')
                        {
                            if((tt=*c2) != '\0')
                                break ;
                            if(!node->clsEndOff)
                                s2-- ;
                        }
                    }
                    if((tt == '\0') || (srcPos == srcWid))
                        break ;
                    cc = *s1++ ;
                    hilOffsetChar(off,dstPos,dstJmp,cc) ;
                    if(ss == ignore)
                    {
                        if(++srcPos == srcWid)
                            break ;
                        cc = *s1++ ;
                        hilOffsetChar(off,dstPos,dstJmp,cc) ;
                    }
                }
                if(tt != '\0')
                    goto hiline_exit ;
                
                len = s2-s1 ;
                    
                if(node->clsEndOff)
                    len -= node->clsEndOff ;
                
                if(node->type & HLREPLACE)
                {
                    if(node->clsSttOff)
                    {
                        hilOffsetLenString(cc,off,dstPos,dstJmp,s1,node->clsSttOff) ;
                        srcPos += node->clsSttOff ;
                        len -= node->clsSttOff ;
                    }
                    dstPos = hilOffsetReplaceString(&off,dstPos,&dstJmp,(unsigned char *) node->rclose,len) ;
                    cc = *off ;
                }
                else
                {
                    hilOffsetLenString(cc,off,dstPos,dstJmp,s1,len) ;
                }
                srcPos += len ;
            }
            else if(node->type & HLCONTIN)
            {
                char *ss ;
                int   pos ;
                
                ss  = (char *) node->close ;
                len = strlen(ss) ;
                pos = srcWid ;
                while((srcText[--pos] == ' ') && (pos > 0))
                    ;
                pos -= len-1 ;
                if((pos >= srcPos) &&
                   !strncmp((char *)srcText+pos,ss,len))
                {
                    hilOffsetString(cc,off,dstPos,dstJmp,srcText+srcPos) ;
                    goto hiline_exit ;
                }
            }
            if(node->type & HLENDLINE)
            {
                if(!(node->type & HLBRINEOL) || (srcPos == srcWid))
                {
                    hilOffsetString(cc,off,dstPos,dstJmp,srcText+srcPos) ;
                    goto hiline_exit ;
                }
                mode |= meHIL_MODETOEOL ;
            }
            else
            {
                if(node->type & HLBRANCH)
                {
                    if((hilno = node->ignore) == 0)
                        hilno = wp->w_bufp->hiLight ;
                    root = hilights[hilno] ;
                }
            }
            if(dstPos)
                mode &= ~(meHIL_MODESTTLN|meHIL_MODESTART) ;
        }
        else
        {
            if((cc=srcText[srcPos]) == 0)
                break ;
            if(mode & (meHIL_MODESTART|meHIL_MODESTTLN))
                mode &= ~((isSpace(cc)) ? meHIL_MODESTART:(meHIL_MODESTART|meHIL_MODESTTLN)) ;
            hilOffsetChar(off,dstPos,dstJmp,cc) ;
            srcPos++ ;
        }
    }
hiline_exit:
    *off = '\0' ;
}


#define HIGOTCONT     0x8000

#define INDINDCURLINE 0x0100
#define INDINDONWARD  0x0200
#define INDGOTIND     0x0400

#define INDBRACKETOPEN   (0x1000)
#define INDBRACKETCLOSE  (0x1800)  
#define INDCONTINUE      (0x0800|INDGOTIND)
#define INDEXCLUSION     (0x2000)
#define INDFIXED         (INDGOTIND)
#define INDIGNORE        (0x2800)
#define INDNEXTONWARD    (INDINDONWARD|INDGOTIND)
#define INDCURONWARD     (INDINDCURLINE|INDINDONWARD|INDGOTIND)
#define INDSINGLE        (INDINDCURLINE|INDGOTIND)

#define noINDTYPES 8
int
indent(int f, int n)
{
    static char           typesChar[(noINDTYPES*2)+1]="bcefinosBCEFINOS" ;
    static unsigned short typesFlag[noINDTYPES]=
    { INDBRACKETOPEN, INDCONTINUE, INDEXCLUSION, INDFIXED, INDIGNORE, INDNEXTONWARD, INDCURONWARD, INDSINGLE } ;
    static unsigned short typesType[noINDTYPES]=
    { 0x00, 0x00, HLBRACKET, HLENDLINE|HLSTTLINE, HLENDLINE, 0x00, 0x00, 0x00 } ;
    HILNODEPTR     root, node ;
    char           buf[MAXBUF] ;
    unsigned char  indno ;
    int            itype ;
    unsigned short htype ;
    
    if(n == 0)
    {
        if((mlreply("Enter indent no",0,0,buf,MAXBUF-1) != TRUE) ||
           ((indno = (unsigned char) meAtoi(buf)) == 0) || 
           (mlreply("Flags",0,0,buf,MAXBUF-1) != TRUE) ||
           ((itype = meAtoi(buf)),
            (mlreply("look back",0,0,buf,MAXBUF) != TRUE)))
            return mlwrite(MWABORT|MWPAUSE,"Invalid init-indent entry") ;
        if((indents[indno] = createHilight(indno,&noIndents,&indents)) == NULL)
            return FALSE ;
        indents[indno]->ignore = (unsigned char) meAtoi(buf) ;
        indents[indno]->type   = itype ;
    
        return TRUE ;
    }
    
    n = (n < 0) ? ADDTOKEN_REMOVE:ADDTOKEN_MAINTOK ;
    
    if((mlreply("Ind no",0,0,buf,MAXBUF-1) != TRUE) ||
       ((indno = (unsigned char) meAtoi(buf)) > noIndents) ||
       ((root  = indents[indno]) == NULL) ||
       ((itype=mlCharReply("Type: ",0,typesChar,NULL)) == -1) ||
       (mlreply("Token",0,0,buf,MAXBUF-1) != TRUE))
        return FALSE ;
    
    itype = (int) (strchr(typesChar,itype) - typesChar) ;
    if(itype >= noINDTYPES)
    {
        itype -= noINDTYPES ;
        htype = (HLTOKEN|typesType[itype]) ;
    }
    else
        htype = typesType[itype] ;
    
    /* add the token */
    node = addToken(root,buf,htype,n) ;
    if(n & ADDTOKEN_REMOVE)
    {
        if(itype == 0)
        {
            if(mlreply("Close",0,0,buf,MAXBUF-1) != TRUE)
                return FALSE ;
            addToken(root,buf,htype,n) ;
        }
        return TRUE ;
    }
    if(node == NULL)
        return FALSE ;
    
    node->color = typesFlag[itype] ;
    
    if(itype == 1)
        root->type |= HIGOTCONT ;
    else if(itype <= 2)
    {
        if(mlreply("Close",0,0,buf,MAXBUF-1) != TRUE)
            return FALSE ;
        if(itype == 0)
        {
            if((node = addToken(root,buf,htype,n)) == NULL)
                return FALSE ;
            node->color = INDBRACKETCLOSE ;
        }
        else
        {
            if((addTokenString(root,node,(unsigned char *) buf,ADDTOKEN_CLOSETOK) == NULL) ||
               (mlreply("Ignore",0,0,buf,MAXBUF-1) != TRUE))
                return FALSE;
            node->ignore = buf[0] ;
        }
    }
    if(typesFlag[itype] & INDGOTIND)
    {
        if(mlreply("Indent",0,0,buf,MAXBUF-1) != TRUE)
            return FALSE ;
        node->color |= ((signed char) meAtoi(buf)) & 0x00ff ;
    }
    return TRUE ;
}

int
indentLine(void)
{
    HILNODEPTR     *bhis ;
    VIDEO          vps[2] ;
    unsigned short noColChng, fnoz, ii ;
    unsigned char  *ss ;
    int            ind, cind, coff ;
    
    vps[0].wind = curwp ;
    vps[0].hilno = curbp->indent ;
    vps[0].bracket = NULL ;
    vps[0].flag = 0 ;
    vps[0].line = curwp->w_dotp ;
    
    bhis = hilights ;
    hilights = indents ;
    noColChng = hilightLine(vps) ;
    hilights = bhis ;
/*    printf("Got %d colour changes\n",noColChng) ;*/
/*    for(ii=0 ; ii<noColChng ; ii++)*/
/*        printf("Colour change %d is to 0x%x at column %d\n",ii,hilCol[ii].icol,hilCol[ii].ccol) ;*/
    ss = disLineBuff ;
    ss[hilCol[noColChng-1].ccol] = '\0' ;
    while(*ss == ' ')
        ss++ ;
    cind = ss - disLineBuff ;
    if(hilCol[0].icol != 0x00)
        fnoz = hilCol[0].icol ;
    else if(noColChng > 1)
        fnoz = hilCol[1].icol ;
    else
        fnoz = 0 ;
    if((fnoz & 0xff00) == INDFIXED)
        ind = ((signed char) (fnoz & 0x0ff)) ;
    else
    {
        int brace=0, jj, contFlag=0, aind, lind, nind ;
        
        if((fnoz & INDINDCURLINE) &&
           ((fnoz == hilCol[0].icol) || (cind >= hilCol[0].ccol)))
            aind = ((signed char) (fnoz & 0x0ff)) ;
        else
            aind = 0 ;
        hilights = indents ;
        jj = indents[curbp->indent]->ignore ;
        for(ind=0 ; (--jj >=0) ; )
        {
            if((vps[0].line = lback(vps[0].line)) == curbp->b_linep)
                break ;
            noColChng = hilightLine(vps) ;
            if((hilCol[0].icol & 0xff00) == INDFIXED)
                continue ;
            ss = disLineBuff ;
            ss[hilCol[noColChng-1].ccol] = '\0' ;
            while(*ss == ' ')
                ss++ ;
            if(*ss == '\0')
                continue ;
/*            printf("Got prev line breakdown %d colour changes ind %d\n",noColChng,ind) ;*/
            for(ii=0 ; ii<noColChng ; ii++)
            {
                fnoz = (hilCol[ii].icol & 0xff00) ;
                if(fnoz == INDNEXTONWARD)
                    ind += ((signed char) (hilCol[ii].icol & 0x0ff)) ;
                else if((fnoz == INDSINGLE) && 
                        ((ii == 0) || ((ii == 1) && ((ss-disLineBuff) >= hilCol[0].ccol))))
                    ind -= ((signed char) (hilCol[ii].icol & 0x0ff)) ;
                else if((fnoz == INDCURONWARD) && (ii != 0) &&
                         ((ii != 1) || ((ss-disLineBuff) < hilCol[0].ccol)))
                    ind += ((signed char) (hilCol[ii].icol & 0x0ff)) ;
                else if(fnoz == INDBRACKETOPEN)
                {
                    if(brace >= 0)
                        hilCol[brace].ccol = hilCol[ii].ccol ;
                    brace++ ;
                }
                else if(fnoz == INDBRACKETCLOSE)
                    brace-- ;
                else if(fnoz == INDCONTINUE)
                {
                    contFlag |= 0x03 ;
                    if(!(contFlag & 0x04))
                        ind += ((signed char) (hilCol[ii].icol & 0x0ff)) ;
                }
/*                printf("Colour change %d is to 0x%x at column %d, ind %d\n",ii,hilCol[ii].icol,hilCol[ii].ccol,ind) ;*/
            }
            if(brace < 0)
            {
                if((contFlag == 0) && (indents[curbp->indent]->type & HIGOTCONT))
                    contFlag = 0x07 ;
                nind = ind = 0 ;
                continue ;
            }
            if((contFlag & 0x03) == 0x02)
                ind = lind ;
            else if(brace)
                ind = hilCol[brace-1].ccol ;
            else
            {
                if(contFlag == 7)
                    ind += nind ;
                nind = ind ;
                ind += ss - disLineBuff + aind ;
                if((contFlag == 0) && (indents[curbp->indent]->type & HIGOTCONT))
                    contFlag = 0x07 ;
            }
            if(ind < 0)
                ind = 0 ;
            if(!brace && (contFlag & 0x01))
            {
                lind = ind ;
                ind = 0 ;
                contFlag &= ~0x01 ;
                continue ;
            }
            break ;
        }
        hilights = bhis ;
    }
    /*    printf("\nIndent line to %d\n\n",ind) ;*/
    /* Always do the doto change so the tab on the left hand edge moves
     * the cursor to the first non-white char
     */
    if((coff = getccol()-cind) < 0)
        coff = 0 ;
    coff += ind ;
    /* Now change the indent if required */
    if(cind != ind)
    {
        register int rr ;
        if((rr=bchange()) != TRUE)               /* Check we can change the buffer */
            return rr ;
        curwp->w_doto = 0 ;
        ss = curwp->w_dotp->l_text ;
        while((*ss == ' ') || (*ss == '\t'))
            ss++ ;
        cind = ss - curwp->w_dotp->l_text ;
        ldelete(cind,6) ;
        if(meModeTest(curbp->b_mode,MDTAB))
            rr = 0 ;
        else
        {
            rr = ind / tabwidth ;
            ind -= rr * tabwidth ;
            linsert(rr,'\t') ;
        }
        linsert(ind,' ') ;
#if MEUNDO
        if(meModeTest(curbp->b_mode,MDUNDO))
        {
            meUndoAddReplaceEnd(ind+rr) ;
            curbp->fUndo->doto = ind+rr ;
            curbp->fUndo->type |= MEUNDO_REVS ;
        }
#endif
    }
    setccol(coff) ;
    return TRUE ;
}

#endif /* HILIGHT */
