/****************************************************************************
 *
 *
 *  			Copyright 1999 Jasspa
 *
 *  System        : MicroEmacs Jasspa Distribution
 *  Module        : 
 *  Object Name   : $RCSfile$
 *  Revision      : $Revision$
 *  Date          : $Date$
 *  Author        : Steve Philips
 *  Created       : 1995
 *  Last Modified : <020699.1622>
 *
 *  Description	
 *      Handle the session history.
 * 
 *  Notes
 *
 *  History
 *	
 *  $Log$
 *
 ****************************************************************************
 *
 * Copyright (C) 1995 - 1999, JASSPA 
 *
 * The MicroEmacs Jasspa distribution can be copied and distributed freely for
 * any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
 * incorporated into commercial software with the expressed permission of
 * JASSPA.
 *  
 ****************************************************************************/

#define	__HISTORYC			/* Define the name of the file */

#include "emain.h"
#include "evers.h"

static char histVerId[] = meDATE ;      /* History version is the date */
char *defHistFile=NULL ;

/* Constants */
static char *regNames[5] = 
{
    "string",
    "buffer",
    "command",
    "file",
    "search"
};
static unsigned char *numList[5]=
{
    &numStrHist,
    &numBuffHist,
    &numCommHist,
    &numFileHist,
    &numSrchHist
};
    

void
initHistory(void)
{
    /* Must malloc the 20 history slots for the 5 history types */
    if((strHist = (char **) meMalloc(sizeof(char *) * 5 * MLHISTSIZE)) == NULL)
        meExit(1) ;
    /* Initialise the array to NULLS so we know they don't point to a history
     * string
     */
    memset(strHist,0,sizeof(char *) * 5 * MLHISTSIZE) ;
    buffHist = strHist + MLHISTSIZE ;
    commHist = buffHist + MLHISTSIZE ;
    fileHist = commHist + MLHISTSIZE ;
    srchHist = fileHist + MLHISTSIZE ;
}

int
setupHistory(int option, unsigned char **numPtr, char ***list)
{
    if(option & MLBUFFER)
    {
        *numPtr = &numBuffHist ;
        *list = buffHist ;
    }
    else if(option & MLCOMMAND)
    {
        *numPtr = &numCommHist ;
        *list = commHist ;
    }
    else if(option & MLFILE)
    {
        *numPtr = &numFileHist ;
        *list = fileHist ;
    }
    else if(option & MLSEARCH)
    {
        *numPtr = &numSrchHist ;
        *list = srchHist ;
    }
    else
    {
        *numPtr = &numStrHist ;
        *list = strHist ;
    }
    return **numPtr ;
}


void
addHistory(int option, char *str)
{
    unsigned char   *numPtr, numHist,i ;
    char **history, *buf ;

    numHist = setupHistory(option, &numPtr, &history) ;

    if((numHist > 0) && !strcmp(str,history[0]))
        return ;
    if((buf = (char *) meMalloc(strlen(str)+1)) == NULL)
        return ;
    if(numHist == MLHISTSIZE)
        meFree(history[numHist-1]) ;
    else
    {
        numHist++ ;
        (*numPtr)++ ;
    }
    for(i=numHist-1 ; i>0 ; i--)
        history[i] = history[i-1] ;
    strcpy(buf,str) ;
    history[0] = buf ;
}

/*
 * Read in the history.
 */

int
readHistory(int f, int n)
{
    BUFFER *bp ;                        /* Buffer pointer */
    REGHANDLE regRoot;                  /* Root of the registry */
    REGHANDLE reg;                      /* temp registry pointer */
    char filename[FILEBUF] ;	        /* Filename */
    char *fname;	                /* Resulting file name to execute */
    int ii;                             /* Loop counter */
    
    if(n == 0)
    {
        if(defHistFile == NULL)
            return mlwrite(MWABORT|MWCLEXEC,"[No default history file]") ;
        fname = defHistFile ;
    }
    else
    {
        if (mlreply("Read history", MLFILECASE,0,filename,FILEBUF) != TRUE)
            return ABORT ;
        fname = filename ;
    }
    
    /* Write the strings into the registry */
    if ((regRoot = regRead ("history", fname, REGMODE_RELOAD|REGMODE_BACKUP)) == NULL)
        return ABORT ;
    
    if((f == TRUE) && (n != 0))
    {
        /* Store the default history file name */
        meNullFree(defHistFile) ;
        defHistFile = strdup(regRoot->value) ;
    }
    if ((regRoot = regFind (regRoot, "history")) != NULL)
    {
        char *ss;                       /* temp string pointer */
        
        /* Load in the history */
        for (ii = 0; ii < 5; ii++)
        {
            int numHistory = 0;
            
            if((reg = regFind (regRoot, regNames [ii])) != NULL)
               reg = regGetChild(reg) ;
            while (reg != NULL)
            {
                meNullFree(strHist[(ii*MLHISTSIZE)+numHistory]) ;
                if((strHist[(ii*MLHISTSIZE)+numHistory] = reg->value) != NULL)
                {
                    reg->value = NULL;
                    numHistory++;
                }
                reg = regGetNext(reg) ;
            }
            /* Assign the number of history items */
            *(numList[ii]) = numHistory;
            while(numHistory < MLHISTSIZE)
            {
                if(strHist[(ii*MLHISTSIZE)+numHistory] != NULL)
                {
                    free(strHist[(ii*MLHISTSIZE)+numHistory]) ;
                    strHist[(ii*MLHISTSIZE)+numHistory] = NULL ;
                }
                numHistory++ ;
            }
        }
        /*
         * Read in the buffer history provided that the history file state is 
         * defined correctly and the versions are not different
         */
        if(HistNoFilesLoaded < 0)
        {
            /* Get the version information */
            if (((reg = regFind (regRoot, "version")) != NULL) &&
                ((ss=regGetString(reg,NULL)) != NULL) && !strcmp(ss,histVerId))
            {
                /* Set up to extract the 'active' configuration */
                /* Reset the files loaded counter */
                ii = 0;
                if((reg = regFind (regRoot, "active")) != NULL)
                    reg = regGetChild(reg) ;
                while (reg != NULL)
                {
                    meSTAT stats ;
                    char   fnbuf[FILEBUF] ;
                    int    flags ;
                    
                    /* Get the filename */
                    /* If file is nasty, or it doesnt exist, ignore it */
                    /* NOTE: must correct the file name as duel systems such as window & dos
                     * have different naming conventions */
                    if(((fname = reg->value) != NULL) &&
                       ((flags = getFileStats(fname,0,&stats,NULL)) != 0) && (flags != 3) &&
                       (fileNameCorrect(fname,fnbuf,NULL),((fname=strdup(fnbuf)) != NULL)))
                    {
                        int    histNo, lineNo, m0, m1, m2, m3, m4 ;
                        meMODE mode ;
                        
                        /* now extract the buffer info */
                        sscanf(reg->name,"%d,%d,%d,%d,%d,%d,%d,%d",
                               &flags,&histNo,&lineNo,&m0,&m1,&m2,&m3,&m4) ;
                        mode[0] = (unsigned char) m0 ;
                        mode[1] = (unsigned char) m1 ;
                        mode[2] = (unsigned char) m2 ;
                        mode[3] = (unsigned char) m3 ;
                        mode[4] = (unsigned char) m4 ;
                        
                        /* set this to inactive */
                        bp = bfind(fname,(BFND_CREAT|BFND_MKNAM|
                                          (meModeTest(mode,MDBINRY) ? BFND_BINARY:0)));
                        bp->intFlag |= BIFFILE ;
                        bp->b_fname = fname ;
                        bp->line_no = lineNo ;
                        bp->histNo = histNo ;
                        
                        meModeCopy(bp->b_mode,mode) ;
                        meModeClear(bp->b_mode,MDEDIT) ;
                        meModeClear(bp->b_mode,MDNRRW) ;
                        meModeSet(bp->b_mode,MDNACT) ;
                        /* copy across the stat info - really important for unix
                         * to avoid the posibility of load the file in twice
                         */
                        memcpy(&(bp->stats),&stats,sizeof(meSTAT)) ;
                        bufHistNo = 2 ;
                        if(histNo)
                            ii++ ;
                    }
                    reg = regGetNext(reg) ;
                }
                /* Fix up the history */
                if(!ii && (bufHistNo == 2))
                    ii++ ;
                HistNoFilesLoaded = ii ;
            }
            else
                /* MUST reset the HistNoFilesLoaded value or a call to save-history
                 * followed by a read-history will really screw things up! */
                HistNoFilesLoaded = 0 ;
        }
        /* Delete the history from memory - we have finished with it */
        regDelete (regRoot);
        mlwrite(MWCLEXEC,"[History loaded]") ;
    }
    return TRUE ;
}

/*
 * saveHistory
 * Save the history back to the registry and to file.
 */
int
saveHistory(int f, int n)
{
    int ii, jj;                         /* Local loop counters */
    BUFFER *bp ;                        /* Pointer to buffers */
    REGHANDLE regRoot;                  /* Root of the registry */
    REGHANDLE regHistory;               /* Handle into the registry */
    char name[2] = "0";                 /* Automatic name seed */
    char filename[FILEBUF] ;            /* Filename */
    char *fname;                        /* Resulting file name to execute */

    if(n == 0)
    {
        if((fname = defHistFile) == NULL)
            return mlwrite(MWABORT|MWCLEXEC,"[No default history file]") ;
    }
    else
    {
        char tmp[FILEBUF] ;	/* Filename */
        if(mlreply("Save history",MLFILECASE,0,tmp,FILEBUF) != TRUE)
            return FALSE ;
        if(!fileLookup(tmp,".erf",meFL_CHECKDOT|meFL_USESRCHPATH,filename))
            strcpy(filename,tmp) ;
        fname = filename ;
    }
    /* Write the strings into the registry */
    if ((regRoot = regRead ("history", NULL, REGMODE_BACKUP)) == NULL)
        return mlwrite (MWABORT|MWPAUSE, "Cannot write history");
    
    /* Delete any other history information */
    if((regHistory = regFind (regRoot, "history")) != NULL)
        regDelete (regHistory);
    
    /* Construct a new history entry */
    if ((regHistory = regSet (regRoot, "history", "")) != 0)
    {
        REGHANDLE regSection;
        
        /* hide from this node down as its junk */
        regHistory->mode |= REGMODE_HIDDEN ;
        
        /* Write the version information in */
        regSet (regHistory, "version", histVerId);
        
        for (ii = 0; ii < 5; ii++)
        {
            /* Construct the new node container */
            if ((regSection = regSet (regHistory, regNames [ii], "")) != NULL)
            {
                int numHistory = *(numList[ii]);
                
                for (jj = 0; jj < numHistory; jj++)
                {
                    name[0] = 'A' + jj ;
                    regSet (regSection, name,
                            strHist[(ii*MLHISTSIZE)+jj]);
                }
            }
        }
        
                        
        /* Construct the buffer history list container and make it
         * current. */
        if ((regHistory = regSet (regHistory, "active", "")) != NULL)
        {
            int bufno=0 ;
            /* Write the buffer information */
            for (bp = bheadp; bp != NULL; bp = bp->b_bufp)
            {
                if((bp->b_bname[0] != '*') && (bp->b_fname != NULL))
                {
                    char buff[40] ;
                    int histno ;
                    long lineNo ;
                    
                    if(bp->b_nwnd)
                    {
                        if(bp == curbp)
                        {
                            histno = 2 ;
                            lineNo = curwp->line_no+1;
                        }
                        else
                        {
                            WINDOW *wp ;
                            histno = 1 ;
                            wp = wheadp ;
                            while(wp->w_bufp != bp)
                                wp = wp->w_wndp ;
                            lineNo = wp->line_no+1;
                        }
                    }
                    else
                    {
                        histno = 0 ;
                        lineNo = bp->line_no ;
                    }
                    sprintf(buff,"%d,%d,%ld,%d,%d,%d,%d,%d",bufno++,histno,lineNo,
                            bp->b_mode[0],bp->b_mode[1],bp->b_mode[2],
                            bp->b_mode[3],bp->b_mode[4]) ;
                    /* Write file name = state info. */
                    regSet (regHistory, buff, bp->b_fname) ;
                }
            }
        }
    }
    
    /* Save the history to file. */
    if (regSave (regRoot, fname))
    {
        mlwrite(MWCLEXEC,"[History written in %s]", fname) ;
        return TRUE ;
    }
    return FALSE;
}
