/*****************************************************************************
*
*	Title:		%M%
*
*	Synopsis:	Emacs Main.
*
******************************************************************************
*
*	Filename:		%P%
*
*	Author:			Jon Green
*
*	Creation Date:		03/05/91 17:19		<990714.0824>
*
*	Modification date:	%G% : %U%
*
*	Current rev:		%I%
*
*	Special Comments:	
*
*	Contents Description:	
*
*	MicroEMACS 3.8
*			written by Dave G. Conroy.
*			greatly modified by Daniel M. Lawrence
*
*	(C)opyright 1987 by Daniel M. Lawrence
*	MicroEMACS 3.8 can be copied and distributed freely for any
*	non-commercial purposes. MicroEMACS 3.8 can only be incorporated
*	into commercial software with the permission of the current author.
*
*****************************************************************************
* 
* Modifications to the original file by Jasspa. 
* 
* Copyright (C) 1988 - 1999, JASSPA 
* The MicroEmacs Jasspa distribution can be copied and distributed freely for
* any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
* incorporated into commercial software with the expressed permission of
* JASSPA.
* 
****************************************************************************/

/*---	Include defintions */

#define	__MAINC			/* Define program name */

/*---	Include files */
#define	INC_MODE_DEF

#include "emain.h"

#define	maindef		/* Make main defintions - cannot define this at the top
                         * because all the main defs are needed to init edef's vars */

#include "efunc.h"	/* function declarations and name table	*/
#include "eskeys.h"     /* Special key names - Must include before edef.h & ebind.h */
#include "edef.h"	/* The main global variables		*/
#include "ebind.h"	/* default bindings			*/
#include "evar.h"	/* env. vars and user funcs		*/
#include "evers.h"	/* Variables */
#include "esearch.h"	/* defines PTBEG PTEND			*/

#if (defined _UNIX) || (defined _DOS) || (defined _WIN32)
#include <sys/types.h>
#include <time.h>
#ifndef _WIN32
#include <sys/time.h>
#endif
#ifdef _UNIX
#include <fcntl.h>
#include <sys/wait.h>
#include <sys/stat.h>
#endif
#endif

#ifdef _MEVMS
#include        <ssdef.h>
#endif

/*---	Local macro definitions */

/*---	External references */

/*---	Local type definitions */

/*---	Global variable defintions */

/*---   Local variable definitions */

/*
 * Initialize all of the buffers and windows. The buffer name is passed down
 * as an argument, because the main routine may have been told to read in a
 * file by default, and we want the buffer name to be right.
 */
static void
edinit(char *bname)
{
    register BUFFER *bp;
    register WINDOW *wp;
    register LINE   *lp, *off;
    
    if(((bp = bfind(bname,BFND_CREAT)) == NULL) ||
       ((wp = (WINDOW *) meMalloc(sizeof(WINDOW))) == NULL) ||
       ((lp =lalloc(TTmcol)) == NULL) ||
       ((off=lalloc(TTmcol)) == NULL))
        meExit(1);
    curbp  = bp;                        /* Make this current    */
    wheadp = wp;
    curwp  = wp;
    numWindows   = 1;                   /* Number of windows */
    memset(wp,0,sizeof(WINDOW)) ;
    wp->numCols  = TTncol;              /* Window # columns     */
    wp->numRows  = TTnrow;              /* Window # rows        */
    wp->model    = lp ;
    off->l_fp    = NULL ;
    wp->curLineOff= off ;
    wp->w_bufp   = bp;
    wp->w_dotp   = bp->b_linep;
    wp->w_flag   = WFMODE|WFRESIZE|WFSBAR;  /* Full + resize        */
    bp->b_nwnd   = 1;                   /* Displayed.           */
    fixWindowTextSize (wp);             /* Fix the text window  */
    vvideoAttach (&vvideo, wp);         /* Attach to video      */
}

/*
 * insertChar
 * 
 * Insert n * 'c' chars into the current buffer
 * Copes with insert and undo modes correctly
 */
int
insertChar(register int c, register int n)
{
    /*---    If we are  in overwrite mode,  not at eol,  and next char is
       not a tab or we are at a tab stop, delete a char forword */
    
    if(meModeTest(curbp->b_mode,MDOVER))
    {
        int index, ii=n ;
        for(index=0,n=0 ; (ii>0) && (curwp->w_doto < curwp->w_dotp->l_used) ; ii--)
        {
            if((lgetc(curwp->w_dotp, curwp->w_doto) != TAB) ||
               (at_tab_pos(getccol()+index+1) == 0))
            {
                lchange(WFMAIN);
#if MEUNDO
                meUndoAddRepChar() ;
#endif
                lputc(curwp->w_dotp,curwp->w_doto++,c) ;
                index = 0 ;
            }
            else
            {
                index++ ;
                n++ ;
            }
        }
        n += ii ;
    }
    /*---    Do the appropriate insertion */
    if(n > 0)
    {
#if MEUNDO
        if(n == 1)
        {
            meUndoAddInsChar() ;
            if(linsert(1,c) != TRUE)
                return FALSE ;
        }
        else
        {
            if(linsert(n, c) != TRUE)
                return FALSE ;
            meUndoAddInsChars(n) ;
        }
#else
        if(linsert(n, c) != TRUE)
            return FALSE ;
#endif
    }
    return TRUE ;
}

/*
 * This is the general command execution routine. It handles the fake binding
 * of all the keys to "self-insert". It also clears out the "thisflag" word,
 * and arranges to move it to the "lastflag", so that the next command can
 * look at it. Return the status of command.
 */
int
execute(register int c, register int f, register int n)
{
    register int index;
    unsigned int arg ;
    int      ii ;
    
    if (kbdmode != PLAY)
    {
#ifdef _CLIPBRD
        clipState |= CLIP_TRY_GET|CLIP_TRY_SET ;
#endif
#if MEUNDO
        undoContFlag++ ;
#endif
    }
    lastCommand = thisCommand ;
    lastIndex = thisIndex ;
    thisCommand = c ;
        
    if(((index = decode_key((unsigned short) c,&arg)) >= 0) && (arg != 0))
    {
        f = 1 ;
        n *= (int) (arg + 0x80000000) ;
    }
    thisIndex = index ;
    /* SWP - 27/2/99 - to enable input commands to not have to jump through
     * so many hoops I've changed to input support so the macros can fail
     * in such a way as to indicate that they have not handled this input,
     * this allows them to pick and choose.
     * The method chosen was to check the command variable status if it aborted, if 
     * defined to "0" then its not handled the input
     */
    if((ii=curbp->inputFunc) >= 0)
    {
        char *ss ;
        if(((cmdstatus = (execFunc(ii,f,n) == TRUE))) ||
           ((ss=getUsrLclCmdVar("status",&(cmdTable[ii]->varList))) == errorm) || meAtoi(ss))
            return cmdstatus ;
    }
    if(index >= 0)
        return (cmdstatus = (execFunc(index,f,n) == TRUE)) ;
    if(selhilight.flags)
    {
        selhilight.bp = NULL;
        selhilight.flags = 0;
    }
    if(c < 0x20 || c > 0xff)    /* If not an insertable char */
    {
        char outseq[40];	/* output buffer for keystroke sequence */
        cmdstr((unsigned short) c, &outseq[0]);	/* change to something printable */
        lastflag = 0;                       /* Fake last flags.     */
        cmdstatus = 0 ;
        /* don't complain about mouse_move* or mouse_time* as these are
         * special keys that are only added if they are bound and due
         * to their frequence they can be in the input buffer before they're
         * unbound leading to this error */
        c &= ~(ME_SHIFT|ME_CONTROL|ME_ALT) ;
        if(((c >= (ME_SPECIAL|SKEY_mouse_move)) && (c <= (ME_SPECIAL|SKEY_mouse_move_3))) ||
           ((c >= (ME_SPECIAL|SKEY_mouse_time)) && (c <= (ME_SPECIAL|SKEY_mouse_time_3))) )
            return TRUE ;
        return mlwrite(MWABORT,"[Key not bound \"%s\"]", outseq); /* complain */
    }
    
    if (n <= 0)            /* Fenceposts.          */
    {
        lastflag = 0;
        return (cmdstatus = ((n<0) ? FALSE : TRUE));
    }
    thisflag = 0;          /* For the future.      */
    
    if(bchange() != TRUE)               /* Check we can change the buffer */
        return (cmdstatus = FALSE) ;
    
    /*---    If a space was  typed, fill column is  defined, the argument is non-
       negative, wrap mode is enabled, and we are now past fill column, and
       we are not read-only, perform word wrap.  */
    
    if(meModeTest(curbp->b_mode,MDWRAP) && 
       ((c == ' ') || (strchr(filleos,c) != NULL)) &&
       (fillcol > 0) && (n >= 0) && (getccol() > fillcol) &&
       !meModeTest(curbp->b_mode,MDVIEW))
        wrapWord(FALSE, 1);
    
    /* insert the required number of chars */
    if(insertChar(c,n) != TRUE)
        return (cmdstatus = FALSE) ;
    
    if(meModeTest(curbp->b_mode,MDCMOD))
    {
        if((c == '}') || (c == '#'))
        {
            ii = curwp->w_doto ;
            curwp->w_doto = 0 ;
            if(gotoFrstNonWhite() == c)
            {
                curwp->w_doto = ii ;
                doCindent(&ii) ;
            }
            else
                curwp->w_doto = ii ;
        }
        else if((commentMargin >= 0) &&
                ((c == '*') || (c == '/')) && (curwp->w_doto > 2) &&
                (lgetc(curwp->w_dotp, curwp->w_doto-2) == '/') &&
                (lgetc(curwp->w_dotp, curwp->w_doto-3) != '/'))
        {
            ii = curwp->w_doto - 2 ;
            if((gotoFrstNonWhite() == 0) &&
               (curwp->w_doto=0,(gotoFrstNonWhite() != 0)) &&
               (curwp->w_doto < ii))
            {
                curwp->w_doto = ii ;
                if(ii > commentMargin)
                    ii = (tabsize-1) - ((ii-1)%tabsize) ;
                else
                    ii = commentMargin - ii ;
                linsert(ii,' ') ;
#if MEUNDO
                meUndoAddInsChars(ii) ;
#endif
            }
            else
                curwp->w_doto = ii ;
            curwp->w_doto += 2 ;
        }
    }
#if    CFENCE
    /* check for fence matching */
    if((c == '}') || (c == ')') || (c == ']'))
    {
        curwp->w_doto-- ;
        /* flag for delay move and only bell in cmode */
        gotoFence(TRUE,(meModeTest(curbp->b_mode,MDCMOD)) ? 3:2) ;
        curwp->w_doto++ ;
    }
#endif
    lastflag = thisflag;
    return (cmdstatus = TRUE) ;
}


/*
** Me info
** Creates a *info* buffer and populates it with info on the current state of things.
*/

void
addModesList(BUFFER *bp, register char *buf, char *name, meMODE mode,
             int res)
{
    register int ii, nlen, len, ll ;
    
    strcpy(buf,name) ;
    len = nlen = strlen(name) ;
    for (ii = 0; ii < MDNUMMODES; ii++)	/* add in the mode flags */
        if((meModeTest(mode,ii) != 0) == res)
        {
            ll = strlen(modeName[ii]);
            if(len+ll >= 78)
            {
                buf[len] = '\0' ;
                addLineToEob(bp,buf) ;
                len = nlen ;
                memset(buf,' ',len) ;
            }
            buf[len++] = ' ' ;
            strcpy(buf+len, modeName[ii]);
            len += ll ;
        }
    buf[len] = '\0' ;
    addLineToEob(bp,buf) ;
}    

void
addModesLists(BUFFER *bp, register char *buf, meMODE mode)
{
    addLineToEob(bp,"") ;
    addModesList(bp,buf,"  Modes on  :",mode,1) ;
    addModesList(bp,buf,"  Modes off :",mode,0) ;
    addLineToEob(bp,"") ;
}

char meCopyright[]="Copyright (C) 1988 - " meCENTURY meYEAR ", JASSPA (jasspa@geocities.com)" ;
int
meAbout(int f, int n)
{
    WINDOW *wp ;
    BUFFER *bp, *tbp ;
    long    numchars ;		/* # of chars in file */
    long    numlines ;		/* # of lines in file */
    long    predchars ;		/* # chars preceding point */
    long    predlines ;		/* # lines preceding point */
    char    buf[MAXBUF] ;
    int     ii ;
    
    if((wp = wpopup("*about*",(BFND_CREAT|BFND_CLEAR|WPOP_USESTR))) == NULL)
        return FALSE ;
    bp = wp->w_bufp ;
    
     /* definitions in evers.h */
    addLineToEob(bp,ENAME " " meVERSION " - Date " meDATE " - " meSYSTEM_NAME) ;
    addLineToEob(bp,"") ;
    addLineToEob(bp, "Global Status:") ;
    tbp = bheadp ;
    ii = 0 ;
    while(tbp != NULL)
    {
        ii++ ;
        tbp = tbp->b_bufp ;
    }
    sprintf(buf,"  # buffers : %d", ii) ;
    addLineToEob(bp,buf) ;
    addModesLists(bp,buf,globMode) ;
    addLineToEob(bp, "Current Buffer Status:") ;
    sprintf(buf,"  Buffer    : %s", curbp->b_bname) ;
    addLineToEob(bp,buf) ;
    sprintf(buf,"  File name : %s", (curbp->b_fname == NULL) ? 
            "":curbp->b_fname) ;
    addLineToEob(bp,buf) ;
    addLineToEob(bp,"") ;
    
    getBufferInfo(&numlines,&predlines,&numchars,&predchars) ;
    sprintf(buf,"  Lines     : Total %6ld, Current %6ld",numlines,predlines) ;
    addLineToEob(bp,buf) ;
    sprintf(buf,"  Characters: Total %6ld, Current %6ld",numchars,predchars) ;
    addLineToEob(bp,buf) ;
    
    addModesLists(bp,buf,curbp->b_mode) ;
    addLineToEob(bp,meCopyright) ;
    bp->b_dotp = lforw(bp->b_linep);
    bp->b_doto = 0 ;
    bp->line_no = 0 ;
    meModeClear(bp->b_mode,MDEDIT);     /* don't flag this as a change */
    meModeSet(bp->b_mode,MDVIEW);       /* put this buffer view mode */
    resetBufferWindows(bp) ;    
    return TRUE ;
}

/*
 * Quit command. If an argument, always quit. Otherwise confirm if a buffer
 * has been changed and not written out. Normally bound to "C-X C-C".
 */
int
exitEmacs(int f, int n)
{
    register int    s;
    
    s = 0 ;
    if((f != FALSE) ||                         /* Argument forces it.  */
       (((anycb() == FALSE) ||                 /* All buffers clean.   */
         ((s=mlyesno("Modified buffers exist. Leave anyway")) == TRUE))
#ifdef _IPIPES
        && ((ipipes == NULL) || 
#ifdef _CLIENTSERVER
            ((ipipes->pid == 0) && (ipipes->next == NULL)) || 
#endif
            ((s=mlyesno("Active processes. Leave anyway")) == TRUE))
#endif
       ))
    {
        BUFFER *bp, *nbp ;
        /* User says it's OK.   */
        /* go round removing all the active ipipes,
         * execute any dhooks set on active buffers 
         * and remove all auto-write files 
         */
#ifdef _IPIPES
        while(ipipes != NULL)
            ipipeRemove(ipipes) ;
#endif
        bp = bheadp ;
        while(bp != NULL)
        {
            nbp = bp->b_bufp ;
            autowriteremove(bp) ;
            if(!meModeTest(bp->b_mode,MDNACT) && (bp->dhook >= 0))
                execBufferFunc(bp,bp->dhook,0,1) ;     /* Execute the delete hook */
            bp = nbp ;
        }
        saveHistory(TRUE,0) ;
        TTend();
#ifdef _TCAP
#ifdef _XTERM
        if(meSystemCfg & meSYSTEM_CONSOLE)
#endif
            TCAPputc('\n');
#endif
#ifdef FREE_ALL_MEMORY
        {
            /* free of everything we can */
            extern void printFreeMemory(void) ;
            extern void osdFreeMemory(void) ;
            extern void TTfreeTranslateKey(void) ;
            extern unsigned char *ffbuf ;
            extern LINE  *mline ;
            extern char  *defHistFile ;
            extern unsigned long *colTable ;
            FRAMELINE    *flp;                   /* Frame store line pointer */
            meREGISTERS  *reg ;
            WINDOW       *wp ;
            meMACRO      *mac ;
            meVARIABLE   *nuv, *cuv ;
            KLIST        *thiskl ;
            KILL         *next, *kill ;
            meABREV      *abrev ;
            VVIDEO       *vvptr;
            int           ii, jj ;
            
            while(bheadp->b_bufp != NULL)
            {
                bheadp->b_bufp->intFlag |= BIFBLOW ;
                zotbuf(bheadp->b_bufp,1) ;
            }
            /* ensure the buffer we're left with is a simple *scratch* */
            bheadp->intFlag |= BIFBLOW ;
            zotbuf(bheadp,1) ;
            bclear(bheadp) ;
            meFree(bheadp->b_linep);             /* Release header line. */
            meNullFree(bheadp->b_fname) ;
            meNullFree(bheadp->b_bname) ;
            meNullFree(bheadp->modeLineStr) ;
            meFree(bheadp);                             /* Release buffer block */
            meNullFree(mlBinds) ;
            
            while (wheadp != NULL)
            {
                wp = wheadp ;
                wheadp = wp->w_wndp ;
                meFree(wp->model) ;
                meFree(wp->curLineOff) ;
                meFree(wp);
            }
            /* Destruct all virtual video blocks */
            meFree (vvideo.video);
            for (vvptr = vvideo.next; vvptr != NULL; vvptr = vvideo.next)
            {
                vvideo.next = vvptr->next;
                meFree (vvptr->video);
                meFree (vvptr);
            }
            meFree(hilCol) ;
            meFree(disLineBuff) ;
            meFree(mline) ;
            meFree(mlStore) ;
            meNullFree(defHistFile) ;
#ifdef _XTERM
            meNullFree(colTable) ;
#endif
            meNullFree(searchPath) ;
            meNullFree(homedir) ;
            meNullFree(curdir) ;
            
            for(ii=0 ; ii<numStrHist ; ii++)
                free(strHist[ii]) ;
            for(ii=0 ; ii<numBuffHist ; ii++)
                free(buffHist[ii]) ;
            for(ii=0 ; ii<numCommHist ; ii++)
                free(commHist[ii]) ;
            for(ii=0 ; ii<numFileHist ; ii++)
                free(fileHist[ii]) ;
            for(ii=0 ; ii<numSrchHist ; ii++)
                free(srchHist[ii]) ;
            free(strHist) ;
            
            while(klhead != NULL)
            {
                thiskl = klhead ;
                klhead = klhead->next ;
                kill = thiskl->kill ;
                while(kill != NULL)
                {
                    next = kill->next;
                    free(kill);
                    kill = next;
                }
                free(thiskl) ;
            }
            for(ii=0 ; ii<CK_MAX ; ii++)
            {
                cuv = cmdTable[ii]->varList.head ;
                while(cuv != NULL)
                {
                    nuv = cuv->next ;
                    meNullFree(cuv->value) ;
                    free(cuv) ;
                    cuv = nuv ;
                }
            }
            for( ; ii<cmdTableSize ; ii++)
            {
                mac = getMacro(ii) ;
                meFree(mac->name) ;
                meNullFree(mac->fname) ;
                freeLineLoop(mac->hlp,1) ;
                cuv = mac->varList.head ;
                while(cuv != NULL)
                {
                    nuv = cuv->next ;
                    meNullFree(cuv->value) ;
                    free(cuv) ;
                    cuv = nuv ;
                }
                free(mac) ;
            }
            if(cmdTable != __cmdTable)
                free(cmdTable) ;
            cuv = usrVarList.head ;
            while(cuv != NULL)
            {
                nuv = cuv->next ;
                meNullFree(cuv->value) ;
                free(cuv) ;
                cuv = nuv ;
            }
            meFree(meRegHead) ;
            meFree(ffbuf) ;
            meFree(progName) ;
            if(modeLineStr != orgModeLineStr)
                meNullFree(modeLineStr) ;
            meNullFree(flNextFileTemp) ;
            meNullFree(flNextLineTemp) ;
            meNullFree(commentCont) ;
            
            meNullFree(rcsFile) ;
            meNullFree(rcsCiStr) ;
            meNullFree(rcsCiFStr) ;
            meNullFree(rcsCoStr) ;
            meNullFree(rcsCoUStr) ;
            
            {
                extern meNAMESVAR buffNames ;
                extern meDIRLIST  fileNames ;
                extern meNAMESVAR commNames ;
                extern meNAMESVAR modeNames ;
                extern meNAMESVAR varbNames ;
                
                meNullFree(curDirList.path) ;
                freeFileList(curDirList.size,curDirList.list) ;
                meNullFree(fileNames.mask) ;
                meNullFree(fileNames.path) ;
                freeFileList(fileNames.size,fileNames.list) ;
                meNullFree(modeNames.mask) ;
                meNullFree(commNames.list) ;
                meNullFree(commNames.mask) ;
                meNullFree(buffNames.list) ;
                meNullFree(buffNames.mask) ;
                if(varbNames.list != NULL)
                    freeFileList(varbNames.size,varbNames.list) ;
                meNullFree(varbNames.mask) ;
            }
            
            if(noNextLine > 0)
            {
                for(ii=0 ; ii<noNextLine ; ii++)
                {
                    meFree(nextName[ii]) ;
                    for(jj=0 ; jj<nextLineCnt[ii] ; jj++)
                        meFree(nextLineStr[ii][jj]) ;
                    meNullFree(nextLineStr[ii]) ;
                }
                meFree(nextLineCnt) ;
                meFree(nextLineStr) ;
                meFree(nextName) ;
            }
            addFileHook(TRUE,0) ;
            
            deleteDict(1,6) ;
            addSpellRule(1,0) ;
            printFreeMemory() ;
            osdFreeMemory() ;
            TTfreeTranslateKey() ;            
#if HILIGHT
            if(noHilights > 0)
            {
                unsigned char hilno ;
                
                for(hilno=0 ; hilno < noHilights ; hilno++)
                {
                    if(hilights[hilno] != NULL)
                    {
                        hilights[hilno]->rtoken = NULL ;
                        freeToken(hilights[hilno]) ;
                    }
                }
                meFree(hilights) ;
            }
#endif
            meNullFree (fileIgnore) ;
            
            while(aheadp != NULL)
            {
                abrev = aheadp ;
                aheadp = abrev->next ;
                freeLineLoop(&(abrev->hlp),0) ;
                meFree(abrev->fname) ;
                meFree(abrev) ;
            }
            meFree(colSchmTbl) ;
#ifdef _TCAPFONT
            meFree(fntSchmTbl) ;
#endif
            for(flp=frameStore,ii=0; ii<TTmrow; ii++, flp++)
            {
                meFree(flp->text) ;
                meFree(flp->colour) ;
            }
            meFree(frameStore) ;
        }
#endif
        meExit(0);
    }
    mlerase(MWCLEXEC);
    
    return (s);
}


/*
 * Better quit command, query saves buffers and then queries the exit 
 * if modified buffers still exist
 */
int
saveExitEmacs(int f, int n)
{
    if((saveSomeBuffers(f,(n & 0x01)) == TRUE) &&
       (saveDict(f,2|(n & 0x01)) == TRUE) &&
       (saveRegistry(f,2|(n & 0x01)) == TRUE))
        return exitEmacs(f, n) ;            /* conditionally quit   */
    return FALSE ;
}

/*
 * Fancy quit command, as implemented by Norm. If the any buffer has
 * changed do a write on that buffer and exit emacs, otherwise simply exit.
 */
int
quickExit(int f, int n)
{
    return saveExitEmacs(1,0) ;
}

/*
 * Abort.
 * Beep the beeper. Kill off any keyboard macro, etc., that is in progress.
 * Sometimes called as a routine, to do general aborting of stuff.
 */
int
ctrlg(int f, int n)
{
    /* stop recording - execution will stop at this point so there's no 
     * point continuing, this provides an easy get out if the user forgets
     * the macro end record key */
    if(kbdmode == RECORD)
    {
        kbdmode = STOP ;
        addModeToWindows(WFMODE) ;  /* update ALL mode lines */
    }
    if(n)
        mlwrite(MWABORT,"[Aborted]");
    return ABORT ;
}

/*
** function to report a no mark set error */
int
noMarkSet(void)
{
    return mlwrite(MWABORT,"No mark set in this window");
}

/* tell the user that this command is illegal while we are in
   VIEW (read-only) mode                */

int
rdonly(void)
{
    return mlwrite(MWABORT,"[Command illegal in VIEW mode]");
}

/* void function, does nothing but return false if an argument of zero past,
 * else return true. Used to get rid of unwanted unbounded-key bells
 */
int
voidFunc(int f, int n)
{
    return ((n) ? TRUE:FALSE) ;
}

void
prefixFunc(void) /* dummy prefix function */
{
}
void
uniArgument(void)         /* dummy function for binding to universal-argument */
{
}

#ifdef _UNIX
#ifdef _BSD
extern void sigAlarm(void) ;
#else
extern void sigAlarm(int sig) ;
#endif

#ifdef _BSD
void
sigchild(void)
#else
void 
sigchild(int sig)
#endif
{
    int status ;
    
    if(noIpipes != 0)
    {
        IPIPEBUF *ipipe ;
        
        ipipe = ipipes ;
        while(ipipe != NULL)
        {
            if((ipipe->pid > 0) &&
               (waitpid(ipipe->pid,&status,WNOHANG) == ipipe->pid))
            {
                if(WIFEXITED(status))
                    ipipe->pid = -4 ;
                else if(WIFSIGNALED(status))
                    ipipe->pid = -3 ;
#ifdef WCOREDUMP
                else if(WCOREDUMP(status))
                    ipipe->pid = -2 ;
#endif
#if 0
                else
                {
                    if(WIFSTOPPED(status))
                        mlwrite(0,"Process %s has STOPPED",ipipe->buf->b_bname) ;
#ifdef _IRIX
                    else if(WIFCONTINUED(status))
                        mlwrite(0,"Process %s has CONTINUED",ipipe->buf->b_bname) ;
#endif
                    else
                        mlwrite(0,"Process %s sent an UNKNOWN SIGNAL",ipipe->buf->b_bname) ;
                }
#endif
            }
            ipipe = ipipe->next ;
        }
    }
    /* clear up any zoombie children */
    waitpid(-1,&status,WNOHANG) ;
}
#endif

#if (defined _UNIX) || (defined _WIN32)
int
meDie(void)
{
    register BUFFER *bp;    /* scanning pointer to buffers */
#ifdef _XTERM
    int      usedX = 0 ;
#endif
    /*
     * To get here we have received a signal and am about to die!
     * Warn the user.
     */
    /* we know about the signal and we're dealing with it */
    alarmState = 0 ;
#ifdef _XTERM
    /* As the X window will always go, always output to the console
     * or terminal, so if using X, do some grotty stuff.
     * This is really important if the X-term window is destroyed,
     * see waitForEvent in unixterm.c. This is because the windows 
     * already gone, so we can't use it!
     */
    if(!(meSystemCfg & meSYSTEM_CONSOLE))
    {
        usedX = 1 ;
        meSystemCfg |= meSYSTEM_CONSOLE ;
#ifdef _TCAP
        /* This one is used */
        CM = "" ;
#endif
    }
#endif
    /* Make a noisy BELL */
    meModeClear(globMode,MDQUIET) ;
    mlwrite(MWCURSOR|MWABORT,"*** Emergency quit ***");
#ifdef _TCAP
    TCAPputc('\n');
#endif
    
    bp = bheadp;
    while (bp != NULL) 
    {
        if(bufferNeedSaving(bp))
        {
            autowriteout(bp) ;
#ifdef _TCAP
            TCAPputc('\n');
#endif
        }
        bp = bp->b_bufp;            /* on to the next buffer */
    }
    saveHistory(1,0) ;
    saveRegistry(1,2) ;
#ifdef _IPIPES
    /* try to kill any child processes */
    while(ipipes != NULL)
        ipipeRemove(ipipes) ;
#endif
#ifdef _TCAP
    TCAPputc('\n');
#endif
#ifdef _XTERM
    /* Don't bother with the TTend if this was an xterm, to dangerous */
    if(!usedX)
#endif
        TTend();
    
    meExit(1) ;
}
#endif

void
autoSaveHandler(void)
{
    struct meTimeval   tp ;
    register BUFFER  *bp ;
    register unsigned long tim, next=0xffffffff ;

    gettimeofday(&tp,NULL) ;
    tim = ((tp.tv_sec-startTime)*1000) + (tp.tv_usec/1000) ;
    bp  = bheadp;
    while (bp != NULL) 
    {
        if(bp->autotime < next)
        {
            if(bp->autotime <= tim)
                autowriteout(bp) ;
            else 
                next = bp->autotime ;
        }
        bp = bp->b_bufp ;
    }
    if(next != 0xffffffff)
        timerSet(AUTOS_TIMER_ID,next,next-tim) ;
    else
        timerClearExpired(AUTOS_TIMER_ID) ;
}

void
callBackHandler(void)
{
    struct meTimeval   tp ;
    register meMACRO *mac ;
    register unsigned long tim, next=0xffffffff ;
    register int ii ;
        
    gettimeofday(&tp,NULL) ;
    tim = ((tp.tv_sec-startTime)*1000) + (tp.tv_usec/1000) ;
    
    /* Loop through all the macros executing them first */
    for(ii=CK_MAX ; ii<cmdTableSize ; ii++)
    {
        mac = getMacro(ii) ;
        if(mac->callback < tim)
        {
             mac->callback = 0xffffffff ;
            /* If the current buffer has an input handler, that macro will
             * receive the key "callback", this must be bound to this macro,
             * otherwise the input handler will not know what macro to call
             */
            if(curbp->inputFunc >= 0)
            {
                /* binary chop through the key table looking for the character code.  
                ** If found then return the index into the names table.
                */
                register KEYTAB  *ktp;			/* Keyboard character array */
                register short    low;			/* Lowest index in table. */
                register short    hi;			/* Hightest index in table. */
                register short    mid;			/* Mid value. */
                register int      status;		/* Status of comparison. */
                ktp = keytab ;
                hi  = keyTableSize;		/* Set hi water to end of table */
                low = 0;			/* Set low water to start of table */
                do
                {
                    mid = (low + hi) >> 1;		/* Get mid value. */
                    if ((status=(ME_SPECIAL|SKEY_callback)-ktp[mid].code) == 0)
                    {
                        /* Found - return index */
                        ktp[mid].index = ii ;
                        break ;
                    }
                    else if (status < 0)
                        hi = mid - 1;		/* Discard bottom half */
                    else
                        low = mid + 1;		/* Discard top half */
                } while (low <= hi);		/* Until converges */
            }
            execFuncHidden(ME_SPECIAL|SKEY_callback,ii,0) ;
        }
    }    
    /* Loop through all the macros to set the next */
    for(ii=CK_MAX ; ii<cmdTableSize ; ii++)
    {
        mac = getMacro(ii) ;
        if(mac->callback < next)
            next = mac->callback ;
    }
    if(next != 0xffffffff)
        timerSet(CALLB_TIMER_ID,next,next-tim) ;
    else
        timerClearExpired(CALLB_TIMER_ID) ;
}

#ifdef _UNIX
#ifdef _BSD
void
sigeat(void)
#else
void 
sigeat(int sig)
#endif
{
    /*
     * Trap nasty signals under UNIX. Write files and quit.
     * Can't just do that cos god knows what me is doing (broken links etc).
     * So can only flag request here and hope its seen.
     */
    alarmState |= 8 ;
}

#endif    /* _UNIX */

void
doOneKey(void)
{
    register int    c;
    register int    f;
    register int    n;
    register int    mflag;
    int     basec;              /* c stripped of meta character   */
    
    update(FALSE);                          /* Fix up the screen    */
    
    /*
     * If we are not playing or recording a macro. This is the ONLY place
     * where we may get an idle key back from the input stream. Set the kdb
     * mode to IDLE to indicate that we will accept an idle key. 
     * 
     * kdbmode is probably not the best variable to use. However it is global
     * and it is not currntly being used - hence it is available for abuse !!
     */
    if (kbdmode == STOP)
    {
        kbdmode = KBD_IDLE;             /* In an idle state  */
        c = getkeycmd(FALSE, 1, 0);     /* Get an idle key   */
    }
    else
        c = getkeycmd(FALSE, 1, 0);     /* Get a key */
        
    if (mlStatus & MLSTATUS_CLEAR)
        mlerase(MWCLEXEC) ;
    
    f = FALSE;
    n = 1;
    
    /* do ME_PREFIX1-# processing if needed */
    basec = c & ~ME_PREFIX1 ;        /* strip meta char off if there */
    if((c & ME_PREFIX1) && ((basec >= '0' && basec <= '9') || basec == '-'))
    {
        f = TRUE;
        if(basec == '-')
        {
            mflag = -1 ;
            n = 0;
        }
        else
        {
            mflag = 1 ;
            n = basec - '0' ;
        }
        while(((c=getkeycmd(TRUE,(n * mflag),0)) >= '0') && (c <= '9'))
            n = n * 10 + (c - '0');
        n *= mflag;    /* figure in the sign */
    }
    
    /* do ^U repeat argument processing */
    if(c == reptc) 
    {                           /* ^U, start argument   */
        f = TRUE;               /* In case not set */
        mflag = 1;              /* current minus flag */
        for(;;c = getkeycmd(f, n,0))
        {
            switch(c)
            {
            case 21 :                   /* \U (^U) */
                n *= 4;
                continue;    /* Loop for more */
            case '-':
                /* If '-', and argument is +ve, make arg -ve */
                mflag = -1;
                if(n < 0)
                    break;
                n *= -1;
                continue;    /* Loop for more */
            }
            /* Not ^U or first '-' .. drop out */
            break;
        }
        /* handle "^U n" to give "n" (GNU Emacs does this) */
        if ((c >= '0' && c <= '9'))
        {
            for(n=c-'0' ;;)
            {
                /*
                 * get the next key, if a digit, update the
                 * count note that we do not handle "-" here
                 */
                c = getkeycmd(TRUE,(mflag*n),0);
                if(c >= '0' && c <= '9')
                    n = n * 10 + (c - '0');
                else
                    break;
            }
            n *= mflag;    /* figure in the sign */
        }
    }
        
    if(f == TRUE)        /* Zap count from cmd line ? */
        mlerase(MWCLEXEC);

    execute(c, f, n) ;   /* Do it. */
}



void
mesetup(int argc, char *argv[])
{
    extern unsigned char *ffbuf ;
    BUFFER *bp;
    int     carg,rarg;          /* current arg to scan            */
    int     searchflag=FALSE;   /* Do we need to search at start? */
    int     binflag=0 ;         /* load next file as a binary file*/
    int     gline = 0 ;         /* what line? (-g option)         */
    int     noFiles=0 ;
    char   *file=NULL ;
#ifdef _UNIX
    int     sigcatch=1 ;        /* Dont catch signals             */
#endif
#ifdef _DOS
    int     dumpScreen=0 ;
#endif
#ifdef _CLIENTSERVER
    char   *clientMessage=NULL ;
    int     userClientServer=0 ;
#endif
    startTime = time(NULL) ;
    
#ifdef _UNIX
    /* Get the usr id and group id for mode-line and file permissions */
    meUmask = umask(0);
    umask(meUmask);
    meUmask = meUmask ^ 00666 ;
    meUid = getuid();
    meGid = getgid();
    /* Set the required alarms first so we always have them */ 
    /* setup alarm process for timers */
    {
        struct sigaction sa ;
        
        sigemptyset(&meSigHoldMask) ;
        sigaddset(&meSigHoldMask,SIGCHLD);
        sigaddset(&meSigHoldMask,SIGALRM);
        meSigRelease() ;

        sigemptyset(&sa.sa_mask) ;
        sa.sa_flags=0;
        sa.sa_handler=sigAlarm ;
        sigaction(SIGALRM,&sa,NULL) ;
        /* sigset(SIGALRM,sigAlarm);*/
        /* setup child process exit capture for ipipe */
        sa.sa_handler=sigchild ;
        sigaction(SIGCHLD,&sa,NULL) ;
        /* sigset(SIGCHLD,sigchild);*/
    }
#endif
    count_key_table() ;
    
    if(((meRegHead = meMalloc(sizeof(meREGISTERS))) == NULL) ||
       ((ffbuf = meMalloc(FIOBUFSIZ+1)) == NULL))
        exit(1) ;
    /* Make the head registers point back to themselves so that
     * accessing #p? gets #g? and not a core-dump
     */
    meRegHead->prev = meRegHead ;
    meRegHead->execstr = NULL ;
    meRegHead->varList = NULL ;
    meRegCurr = meRegHead ;
    /* Initialise the head as this is dumped in list-variables */
    for(carg = 0 ; carg<meNUMREG ; carg++)
        meRegHead->reg[carg][0] = '\0' ;
    
    /* initialize the editor and process the command line arguments */
    initHistory() ;           /* allocate history space */
    set_dirs();               /* setup directories */
    
    {
        /* setup the $progname make it an absolute path. */ 
        char buff[FILEBUF] ;
        if(executableLookup(argv[0],buff))
            progName = strdup(buff) ;
        else
#ifdef FREE_ALL_MEMORY
            /* stops problems on exit */
            progName = strdup(argv[0]) ;
#else
            progName = argv[0] ;
#endif
    }
    /* scan through the command line and get all global options */
    carg = rarg = 1 ;
    for( ; carg < argc; ++carg) 
    {
        /* if its a switch, process it */
        if (argv[carg][0] == '-') 
        {
            switch (argv[carg][1]) 
            {
            case 'c':
                HistNoFilesLoaded = -1 ;
                break ;
                
            case 'd':    /* -d debug */
                macbug = TRUE;
                break;
            
            case 'h':
                printf("usage     : me [options] [files]\n\n") ;
                printf("where options can be:-\n") ;
                printf("  @<file> : Setup using <file>[.emf], default is me.emf\n") ;
                printf("  -b      : Load next file as a binary file\n") ;
                printf("  -c      : Continuation mode (last edit, must have history setup)\n") ;
                printf("  -d      : Debug mode (for macro files)\n") ;
                printf("  -h      : For this help page\n") ;
#ifdef _DOS
                printf("  -i      : Insert the current screen into the *scratch* buffer\n") ;
#endif
                printf("  +<n> or\n") ;
                printf("  -l<n>   : Go to line <n> in the next given file\n") ;
#ifdef _CLIENTSERVER
                printf("  -m<msg> : Post message to MicroEmacs server\n") ;
#endif
#ifdef _WINCON
                printf("  -n      : For not MS window, uses console instead\n") ;
#endif
#ifdef _XTERM
                printf("  -n      : For not X window, uses termcap instead\n") ;
#endif
#ifdef _CLIENTSERVER
                printf("  -o      : One MicroEmacs, use ME server if available\n") ;
#endif
                printf("  -p      : Pipe stdin into *stdin*, when saved output to stdout\n") ;
                printf("  -r      : Read-only, all buffers will be in view mode\n") ;
                printf("  -s<s>   : Search for string \"s\" in the current buffer\n") ;
                printf("  -v<v=s> : Set variable v to string s\n") ;
#ifdef _UNIX
                printf("  -x      : Don't catch signals\n") ;
#endif
                meExit(0) ;
#ifdef _CLIENTSERVER
            case 'm':    /* -m send a client server message */
                userClientServer = 1 ;
                if(argv[carg][2] == '\0')
                    clientMessage = argv[++carg] ;
                else
                    clientMessage = argv[carg]+2 ;
                break;
#endif
#if (defined _XTERM) || (defined _WINCON)
            case 'n':
                meSystemCfg |= meSYSTEM_CONSOLE ;
                break ;
#endif
#ifdef _CLIENTSERVER
            case 'o':
                userClientServer=1 ;
                break ;
#endif
#ifdef _DOS
            case 'i':
                dumpScreen = 1 ;
                break ;
#endif
            case 'p':
                alarmState |= 0x80 ;
#if (defined _XTERM) || (defined _WINCON)
                meSystemCfg |= meSYSTEM_CONSOLE ;
#endif
                break ;
                
            case 'r':
                meModeSet(globMode,MDVIEW) ;
                break ;

            case 's':    /* -s for initial search string */
                searchflag = TRUE;
                if(argv[carg][2] == '\0')
                    addHistory(MLSEARCH,argv[++carg]) ;
                else
                    addHistory(MLSEARCH,argv[carg]+2) ;
                break;
                
            case 'v':
            {
                char *ss, *tt, cc ;
                if (argv[carg][2] == 0)
                    ss = argv[++carg] ;
                else
                    ss = argv[carg] + 2 ;
                if((tt = strchr(ss,'=')) != NULL)
                {
                    *tt = ' ' ;
                    tt = execstr ;
                    execstr = ss ;
                    cc = clexec ;
                    clexec = TRUE ;
                    if(setVariable(FALSE,1) != TRUE)		/* set a variable */
                    {
                        printf("Unable to setVariable [%s]\n",ss) ;
                        meExit(1) ;
                    }
                    execstr = tt ;
                    clexec = cc ;
                }
                break ;
            }
            case 'x':
#ifdef _UNIX
                sigcatch = 0 ;
                break;
#endif
            case 'l':    /* -l for initial goto line */
                if (argv[carg][2] == 0)
                {
                    char *ss = argv[carg++] ;
                    argv[rarg++] = argv[carg] ;
                    argv[rarg++] = ss ;
                    break ;
                }
            case 'b':    /* -b bin flag */
                /* don't want these options yet as they apply to the 
                 * next file on the command line
                 */
                argv[rarg++] = argv[carg] ;
                break ;
            
            default:
                fprintf(stderr,"%s: Unknown option %c. Option -h gives further help\n",argv[0],argv[carg][1]) ;
                meExit(1) ;
            }
        } 
        else if(argv[carg][0] == '@')
            file = argv[carg] + 1 ;
        else
            argv[rarg++] = argv[carg] ;
    }
#ifdef _CLIENTSERVER
    if(userClientServer && TTconnectClientServer())
    {
        BUFFER *bp ;
        
        file=NULL ;
        for(carg=rarg-1 ; carg > 0 ; carg--) 
        {
            /* if its a switch, process it */
            if(argv[carg][0] == '-') 
            {
                switch(argv[carg][1]) 
                {
                case 'l':    /* -l for initial goto line */
                    if (argv[carg][2] == 0)
                        gline = meAtoi(argv[--carg]);
                    else
                        gline = meAtoi((argv[carg])+2);
                    break;
                case 'b':    /* -b bin flag */
                    binflag = BFND_BINARY ;
                    break;
                }
            } 
            else if (argv[carg][0] == '+')
                gline = meAtoi((argv[carg])+1);
            else 
            {
                /* process a file name */
                if(file != NULL)
                    /* set up a buffer for this file */
                    noFiles += findFileList(file,(BFND_CREAT|BFND_MKNAM|binflag),gline) ;
                file = argv[carg] ;
                gline = 0 ;
                binflag = 0 ;
            }
        }
        if(file != NULL)
            /* set up a buffer for this file */
            noFiles += findFileList(file,(BFND_CREAT|BFND_MKNAM|binflag),gline) ;
        if(clientMessage != NULL)
            TTsendClientServer(clientMessage) ;
        bp = bheadp ;
        while(bp != NULL)
        {
            if(bp->b_fname != NULL)
            {
                char buff[MAXBUF+30] ;
                sprintf(buff,"C:ME:find-file \"%s\"\n",bp->b_fname) ;
                TTsendClientServer(buff) ;
                if(bp->line_no != 0)
                {
                    sprintf(buff,"C:ME:goto-line %d\n",bp->line_no) ;
                    TTsendClientServer(buff) ;
                }
            }
            bp = bp->b_bufp ;
        }
        meExit(0) ;
    }
    else if(userClientServer && clientMessage)
    {
        fprintf(stderr,"%s: Unable to connect to server\n",argv[0]) ;
        meExit(1) ;
    }
#endif
    
    vtinit();                 /* Displays.            */
    edinit(BmainN);           /* Buffers, windows.    */
    
#ifdef _DOS
    if(dumpScreen)
    {
        extern void TTdump(BUFFER *) ;
        TTdump(curbp) ;
        gotobob(FALSE,1) ;
        carg++ ;
    }
#endif
    if(alarmState & 0x80)
    {
        BUFFER *sbp ;
#ifdef _WIN32
        ffp = GetStdHandle(STD_INPUT_HANDLE) ;
#else
        ffp = stdin ;
#endif
        sbp = bfind("*stdin*",BFND_CREAT);
        readin(sbp,NULL) ;
    }
    
    mlerase(0);                /* Delete command line */
    bp = curbp ;
    /* run me.emf unless an @... arg was given in which case run that */
    startup(file) ;
    
    /* initalize colors to global defaults */
    meModeCopy(bp->b_mode,globMode) ;
    bp->b_scheme = globScheme;
#ifdef _CLIENTSERVER
    if((ipipes != NULL) && (ipipes->pid == 0))
        ipipes->bp->b_scheme = globScheme;
#endif
    
    file=NULL ;
    /* scan through the command line and get the files to edit */
    for(carg=rarg-1 ; carg > 0 ; carg--) 
    {
        /* if its a switch, process it */
        if(argv[carg][0] == '-') 
        {
            switch(argv[carg][1]) 
            {
            case 'l':    /* -l for initial goto line */
                if (argv[carg][2] == 0)
                    gline = meAtoi(argv[--carg]);
                else
                    gline = meAtoi((argv[carg])+2);
                break;
            case 'b':    /* -b bin flag */
                binflag = BFND_BINARY ;
                break;
            }
        } 
 		else if (argv[carg][0] == '+')
 			gline = meAtoi((argv[carg])+1);
        else 
        {
            /* process a file name */
            if(file != NULL)
                /* set up a buffer for this file */
                noFiles += findFileList(file,(BFND_CREAT|BFND_MKNAM|binflag),gline) ;
            file = argv[carg] ;
            gline = 0 ;
            binflag = 0 ;
        }
    }
    if(file != NULL)
        /* set up a buffer for this file */
        noFiles += findFileList(file,(BFND_CREAT|BFND_MKNAM|binflag),gline) ;
    if(!noFiles)
        noFiles = HistNoFilesLoaded ;

    if(noFiles > 0)
    {
        swbuffer(curwp,replacebuffer(bp)) ;
        if(noFiles > 1)
        {
            splitWindVert(TRUE,2) ;
            swbuffer(curwp,replacebuffer(bp)) ;
            prevwind(FALSE,1) ;
        }
        bp->histNo = -1 ;
    }
        
#ifdef _UNIX
    /*
     * Catch nasty signals when running under UNIX unless the -c
     * option appeared on the command line.
     */
    if(sigcatch)
    {
        struct sigaction sa ;
        
        sigemptyset(&sa.sa_mask) ;
        sa.sa_flags=0;
        sa.sa_handler=sigeat ;
        sigaction(SIGHUP,&sa,NULL) ;
        sigaction(SIGINT,&sa,NULL) ;
        sigaction(SIGQUIT,&sa,NULL) ;
        sigaction(SIGTERM,&sa,NULL) ;
        sigaction(SIGUSR1,&sa,NULL) ;
#if 0
        sigset(SIGHUP,  sigeat);
        sigset(SIGINT,  sigeat);
        sigset(SIGQUIT, sigeat);
        sigset(SIGTERM, sigeat);
        sigset(SIGUSR1, sigeat);
#endif
        /* These can't be trapped as caused by ME failure and will cause loop
           signal(SIGILL, sigeat);
           signal(SIGTRAP, sigeat);
           signal(SIGBUS, sigeat);
           signal(SIGSEGV, sigeat);
           signal(SIGSYS, sigeat);
           signal(SIGIOT, sigeat);
           signal(SIGEMT, sigeat);
         */
    }
#endif
    
    /* Deal with startup gotos and searches */
    if (searchflag)
        huntForw(FALSE,1) ;

    /* setup to process commands */
    lastflag = 0;                       /* Fake last flags.     */
    
    /* Set the screen for a redraw at this point cos its after the 
     * startup which can screw things up
     */
    sgarbf = TRUE;			 /* Erase-page clears */
    
    carg = decode_fncname("start-up",1) ;
    if(carg >= 0)
        execFunc(carg,FALSE,1) ;
}

#ifndef NDEBUG

/* _meAssert
 * The receiving end of the macro meAssert(<boolean>). This keeps microEMACS
 * up allowing the debugger to be attached. This is essential under Windows
 * since Microsofts assert() is the pits and invariably just exits the
 * program without invoking the debugger.
 */

void
_meAssert (char *file, int line)
{
    char buf [MAXBUF];                  /* String buffer */
    char cc;                            /* Character input */
    
    /* Put out the string */
    sprintf (buf,
             "!! ASSERT FAILED %s:%d !! - <S>ave. <Q>uit. <C>ontinue",
             file, line);
    mlwrite (MWABORT, buf);
    
    TTinflush ();                       /* Flush the input buffer */
    for (;;)
    {
        cc = TTgetc() ;                 /* Get character from keyboard */
        /* Try and perform an emergency save for the user - no guarantees
         * here I am afraid - we may be totally screwed at this point in
         * time. */
        if (cc == 'S')
        {
            register BUFFER *bp;        /* scanning pointer to buffers */
            
            bp = bheadp;
            while (bp != NULL)
            {
                if(bufferNeedSaving(bp))
                    autowriteout(bp) ;
                bp = bp->b_bufp;        /* on to the next buffer */
            }
            saveHistory(TRUE,0) ;
        }
        else if (cc == 'Q')
            meExit (1);                 /* Die the death */
        else if (cc == 'C')
            break;                      /* Let the sucker go !! */
    }
}
#endif

#ifndef _WIN32
void
main(int argc, char *argv[])
{
    mesetup(argc,argv) ;
    while(1)
    {
        doOneKey() ;
        if(TTbreakFlag)
        {
            TTinflush() ;
            if((selhilight.flags & (SELHIL_ACTIVE|SELHIL_KEEP)) == SELHIL_ACTIVE)
                selhilight.flags &= ~SELHIL_ACTIVE ;
            TTbreakFlag = 0 ;
        }
    }
}
#endif
