/*****************************************************************************
*
*	Title:		%M%
*
*	Synopsis:	next file/line routines.
*
******************************************************************************
*
*	Filename:		%P%
*
*	Author:			Steven Phillips
*
*	Creation Date:		01/06/94 12:37		<990706.1020>
*
*	Modification date:	%G% : %U%
*
*	Current rev:		%I%
*
*	Special Comments:
*
*	Contents Description:
*
*****************************************************************************
*
* Copyright (C) 1994 - 1999, JASSPA 
*
* The MicroEmacs Jasspa distribution can be copied and distributed freely for
* any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
* incorporated into commercial software with the expressed permission of
* JASSPA.
*  
****************************************************************************/

/*---	Include defintions */

#define	__NEXTC	1			/* Define the filename */

/*---	Include files */

#include "emain.h"
#include "esearch.h"

#if FLNEXT

int
flNextFind(char *str,char **curFilePtr, long *curLine)
{
    char    pat[MAXBUF] ;
    char   *n=NULL ;
    int     fileRpt=-1, lineRpt=-1, onRpt=0 ;
    int     soff, len ;

    soff = 0 ;
    while((soff < NPAT) && (*str != '\0'))
    {
        if(*str == '%')
        {
            str++ ;
            switch(*str)
            {
            case 'f':
                fileRpt = onRpt++ ;
                n = flNextFileTemp ;
                break ;
            case 'l':
                lineRpt = onRpt++ ;
                n = flNextLineTemp ;
                break ;
            }
        }
        if(n != NULL)
        {
            pat[soff++] = '\\' ;
            pat[soff++] = '(' ;
            strcpy(pat+soff,n) ;
            soff += strlen(n) ;
            pat[soff++] = '\\' ;
            pat[soff++] = ')' ;
            str++ ;
            n = NULL ;
        }
        else
            pat[soff++] = *str++ ;
    }
    if(soff >= NPAT)
        return mlwrite(MWABORT,"[Search string to long]") ;
    pat[soff] = '\0' ;

    if(iscanner(pat,1,ISCANNER_PTBEG|ISCANNER_MAGIC|ISCANNER_EXACT,NULL) != TRUE)
        return FALSE;

    /*
     * Found it, so fill in the slots.
     */

    soff = curwp->w_doto ;
    if(fileRpt >= 0)
    {
#ifdef _GNU_REGEX
        len = mereRegs.end[fileRpt+1] - mereRegs.start[fileRpt+1] ;
#else
        len = srchAutoRpt.end_pos[fileRpt] - srchAutoRpt.start_pos[fileRpt] ;
#endif
        meNullFree(*curFilePtr) ;
        if((*curFilePtr = (char *) meMalloc(len+1)) == NULL)
            return FALSE ;
#ifdef _GNU_REGEX
        strncpy(*curFilePtr,(char *) curwp->w_dotp->l_text+soff+mereRegs.start[fileRpt+1],len) ;
#else
        strncpy(*curFilePtr,(char *) curwp->w_dotp->l_text+soff+srchAutoRpt.start_pos[fileRpt],len) ;
#endif
        (*curFilePtr)[len] = '\0' ;
    }
    if(lineRpt >= 0)
#ifdef _GNU_REGEX
        *curLine = atol((char *) curwp->w_dotp->l_text+soff+mereRegs.start[lineRpt+1]) ;
#else
        *curLine = atol((char *) curwp->w_dotp->l_text+soff+srchAutoRpt.start_pos[lineRpt]) ;
#endif

    return TRUE ;
}

int
getNextLine(int f,int n)
{
    unsigned char noNextLines, ii, no ;
    char **nextLines ;
    char   *newName=NULL ;
    long    curLine=-1 ;
    BUFFER *bp=NULL, *tbp ;
    
    for(ii=0 ; ii<noNextLine ; ii++)
    {
        if((tbp = bfind(nextName[ii],0)) != NULL)
        {
            if(tbp == curbp)
            {
                bp = tbp ;
                no = ii ;
                break ;
            }
            if((bp == NULL) || (bp->b_nwnd < tbp->b_nwnd))
            {
                bp = tbp ;
                no = ii ;
            }
        }
    }
    if(bp == NULL)
        return mlwrite(MWABORT,"[No next buffer found]") ;
    if((noNextLines = nextLineCnt[no]) == 0)
        return mlwrite(MWABORT,"[No lines for next buffer %s]",bp->b_bname) ;
    nextLines = nextLineStr[no] ;
    
    wpopup(bp->b_bname,WPOP_MKCURR) ;

    if((curwp->w_dotp == bp->b_linep) ||
       ((curwp->w_dotp == lback(bp->b_linep)) &&
        (curwp->w_doto == llength(curwp->w_dotp))))
    {
        curwp->w_dotp = lforw(bp->b_linep) ;
        curwp->w_doto = 0 ;
        curwp->line_no = 0 ;
    }
    while((curwp->line_no)++,
          (curwp->w_dotp = lforw(curwp->w_dotp)) != bp->b_linep)
    {
        for(ii=0 ; ii<noNextLines ; ii++)
        {
            curwp->w_doto = 0 ;
            if(flNextFind(nextLines[ii],&newName,&curLine) == TRUE)
            {
                if(newName != NULL)
                {
                    char fname[FILEBUF] ;
                    
                    meNullFree(bp->nextFile) ;
                    if(bp->b_fname != NULL)
                    {
                        char tmp[FILEBUF] ;
                        getFilePath(bp->b_fname,tmp) ;
                        strcat(tmp,newName) ;
                        fileNameCorrect(tmp,fname,NULL) ;
                    }
                    else
                        fileNameCorrect(newName,fname,NULL) ;
                    bp->nextFile = strdup(fname) ;
                    meFree(newName) ;
                }
                curwp->w_doto = 0 ;
                curwp->w_flag |= WFMOVEL ;
                if(bp->nextFile == NULL)
                    return mlwrite(MWABORT,"[No File name]") ;
                if(wpopup(NULL,WPOP_MKCURR) == NULL)
                    return FALSE ;
                if(findSwapFileList(bp->nextFile,(BFND_CREAT|BFND_MKNAM),0) != TRUE)
                    return FALSE ;
                if(curLine < 0)
                    return mlwrite(MWABORT,"[No Line number]") ;
                /* if for some strange reason the file wasn't found, but the directory
                 * was and its read only, the findSwapFileList will succeed (new buffer)
                 * BUT the file name will be null! catch this and just use the buffer name */ 
                mlwrite(0,"File %s, line %d",
                        (curbp->b_fname == NULL) ? curbp->b_bname:curbp->b_fname,curLine) ;
#if NARROW
                return gotoAbsLine(curLine) ;
#else
                return gotoLine(curLine) ;
#endif
            }
        }
    }
    curwp->w_doto = 0 ;
    curwp->w_flag |= WFMOVEL ;
#if 0
    /* swp 24/12/97 - I don't think this is needed */
    addModeToWindows(WFMODE) ;  /* and update ALL mode lines */
#endif
    return mlwrite(MWABORT,"[No more lines found]") ;
}

int
addNextLine(int f, int n)
{
    char name[MAXBUF], line[MAXBUF] ;
    int  no, cnt ;
    
    if((mlreply("next name",0,0,name,MAXBUF) != TRUE) ||
       (mlreply("next line",0,0,line,MAXBUF) != TRUE))
        return FALSE ;
    for(no=0 ; no<noNextLine ; no++)
        if(!strcmp(nextName[no],name))
            break ;
    if(no == noNextLine)
    {
        noNextLine++ ;
        if(((nextName=meRealloc(nextName,noNextLine*sizeof(char *))) == NULL) ||
           ((nextName[no]=strdup(name)) == NULL) ||
           ((nextLineCnt=meRealloc(nextLineCnt,noNextLine*sizeof(unsigned char))) == NULL) ||
           ((nextLineStr=meRealloc(nextLineStr,noNextLine*sizeof(char **))) == NULL))
        {
            noNextLine = 0 ;
            return FALSE ;
        }
        nextLineCnt[no] = 0 ;
        nextLineStr[no] = NULL ;
    }
    cnt = nextLineCnt[no] ;
    if(line[0] == '\0')
    {
        while(cnt--)
            meFree(nextLineStr[no][cnt]) ;
        nextLineCnt[no] = 0 ;
    }
    else
    {
        if(((nextLineStr[no]=meRealloc(nextLineStr[no],(cnt+1)*sizeof(char *))) == NULL) ||
           ((nextLineStr[no][cnt]=strdup(line)) == NULL))
        {
            noNextLine = 0 ;
            return FALSE ;
        }
        nextLineCnt[no]++ ;
    }
    return TRUE ;
}

#endif

#if 0
int
flNextAddtoList(unsigned char *numList, char ***list, char *prompt)
{
    unsigned char i ;
    char     line[MAXBUF];

    if(mlreply(prompt,0,0,line,MAXBUF) == ABORT)
        return FALSE ;
    if(strlen(line) == 0)
    {
        if(*numList > 0)
            for(i=0 ; i<*numList ; i++)
                meFree((*list)[i]) ;
        *numList = 0 ;
    }
    else
    {
        if((*list = (char **) meRealloc(*list,
                                        (*numList+1)*sizeof(char *))) == NULL)
            return FALSE ;
        (*list)[(*numList)++] = strdup(line) ;
    }
    return TRUE ;
}
#endif

int
addFileHook(int f, int n)
{
    if(n == 0)
    {
        /* arg of 0 frees all hooks off */
        if(fileHookCount)
        {
            n = fileHookCount ;
            while(--n >= 0)
            {
                free(fileHookExt[n]) ;
                free(fileHookFunc[n]) ;
            }
            free(fileHookExt) ;
            free(fileHookFunc) ;
            free(fileHookArg) ;
            fileHookCount=0 ;
            fileHookExt=NULL ;
            fileHookFunc=NULL ;
            fileHookArg=NULL ;
        }
    }
    else
    {
        char buff[MAXBUF] ;
    
        if(((fileHookExt = meRealloc(fileHookExt,(fileHookCount+1)*sizeof(char *))) == NULL) ||
           ((fileHookFunc = meRealloc(fileHookFunc,(fileHookCount+1)*sizeof(char *))) == NULL) ||
           ((fileHookArg = meRealloc(fileHookArg,(fileHookCount+1)*sizeof(short))) == NULL))
        {
            fileHookCount=0 ;
            return FALSE ;
        }
        if((mlreply("File id",0,0,buff,MAXBUF) != TRUE) ||
           ((fileHookExt[fileHookCount] = strdup(buff)) == NULL) ||
           (mlreply("Hook func",0,0,buff,MAXBUF) != TRUE) ||
           ((fileHookFunc[fileHookCount] = strdup(buff)) == NULL))
            return FALSE ;
        if(f == FALSE)
            fileHookArg[fileHookCount] = 0 ;
        else
            fileHookArg[fileHookCount] = n ;
        fileHookCount++ ;
    }
    return TRUE ;
}


/*
 * Add the magic hook strings. Arguments as follows:-
 *
 * f=FALSE; n=1   : Magic String to search for from column 0
 * f=TRUE ; n=-ve : Emacs mode definition of form -*- mode -*-
 * f=TRUE ; n=+ve : Magic string to search for from column 'n'
 *
 * Note unlike the file hooks multiple strings are not present in an entry.
 * Since we are dealing with literal text it is quite legal for a magic
 * string to contain a <SPACE> character.
 *
 * e.g.
 *    ; Detect shell script files.
 *    add-magic-hook "#!/bin/sh" "fhook-shell"
 *
 *    ; Explicitly set the mode in a GNUEmacs type string.
 *    ; Syntax is "-*- mode -*-" or "-*- Mode: mode; xxxx -*-"
 *    ; To aviod conflict with GNU emacs the MicroEmacs mode string
 *    ; defined as -!- nroff -!- may be used.
 * -1 add-magic-hook "nroff"  "fhook-nroff"
 *
 *    ; Detect a file type foo with header syntax "V4.0a FOOFILE xxxx"
 * 6  add-magic-hook "FOOFILE" "fhook-foo"
 */
#if 0
int
addMagichook (int f, int n)
{
    unsigned char s = numMagicHook ;

    if((s=flNextAddtoList(&s, &magicHookString, "Magic String")) != TRUE)
        return s ;
    if (flNextAddtoList(&numMagicHook, &magicHookFunc, "Magic Hook Function") == TRUE)
    {
        if (f == FALSE)
            n = 0;
        magicHookStart = (short *) meRealloc (magicHookStart,
                                              sizeof (short) * numMagicHook);
        if (magicHookStart != NULL)
        {
            magicHookStart [numMagicHook-1] = n;
            return (TRUE);
        }
    }
    return (FALSE);
}
#endif

#if DORCS

int
rcsFilePresent(char *fname)
{
    char           tname[MAXBUF] ;
    register char *fn, *ld, *tn=tname ;

    if(rcsFile == NULL)
        return FALSE ;

    fn = fname ;
    if((ld = strrchr(fname,DIR_CHAR)) != NULL)
    {
        do
            *tn++ = *fn ;
        while(fn++ != ld) ;
    }
    ld = rcsFile ;
    while(*ld != '\0')
    {
        if((*ld == '%') &&
           (*++ld == 'f'))
        {
            strcpy(tn,fn) ;
            tn += strlen(fn) ;
        }
        else
            *tn++ = *ld ;
        ld++ ;
    }
    *tn = '\0' ;
    return !meTestExist(tname) ;
}

int
doRcsCommand(char *fname, register char *comStr)
{
    char  pat[MAXBUF] ;
    char  path[FILEBUF] ;
    char *bname ;
    register int   ii, len ;
    
    fileNameCorrect(fname,path,&bname) ;
    ii = 0 ;
    while(*comStr != '\0')
    {
        if((*comStr == '%') &&
           ((*++comStr == 'f') || (*comStr == 'm')))
        {
            if(*comStr++ == 'f')
            {
                len = strlen(bname) ;
                if((ii + len) >= NPAT)
                    break ;
                strcpy(pat+ii,bname) ;
                ii += len ;
            }
            else
            {
                if(mlreply("Enter message",0,0,pat+ii,MAXBUF-ii) != TRUE)
                    return FALSE ;
                ii = strlen(pat) ;
            }
        }
        else
        {
            if((ii + 1) >= NPAT)
                break ;
            pat[ii++] = *comStr++ ;
        }
    }
    if(*comStr != '\0')
        return mlwrite(MWABORT,"[command to long]") ;
    pat[ii] = '\0' ;
    *bname = '\0' ;
    /* must do a pipe not an ipipe as its sequential */
    ii = doPipeCommand(pat,path,BcommandN,LAUNCH_SILENT) ;

    return ii ;
}

int
rcsCiCoFile(int f, int n)
{
    register char *str ;
    register int   lineno, curcol, ss ;

    if(curbp->b_fname == NULL)
        return mlwrite(MWABORT,"No file name!") ;
    ss = rcsFilePresent(curbp->b_fname) ;
    if(meModeTest(curbp->b_mode,MDVIEW))	/* if read-only then unlock */
    {
        if(ss != TRUE)
        {
#ifdef _DOS
            curbp->stats.stmode &= ~0x01 ;
#endif
#ifdef _WIN32
            curbp->stats.stmode &= ~FILE_ATTRIBUTE_READONLY;/* Write permission for owner */
#endif
#ifdef _UNIX
            curbp->stats.stmode |= 00200 ;
#endif
            meModeClear(curbp->b_mode,MDVIEW) ;
            curwp->w_flag |= WFMODE ;
            return TRUE ;
        }
        if((str = rcsCoUStr) == NULL)
            return mlwrite(MWABORT,"[rcs cou command not set]") ;
    }
    else
    {
        if(ss == TRUE)
            str = rcsCiStr ;
        else
            str = rcsCiFStr ;
        if(str == NULL)
            return mlwrite(MWABORT,"[rcs ci or cif command not set]") ;
        if(meModeSet(curbp->b_mode,MDEDIT) &&
           ((ss=saveBuffer(TRUE,TRUE)) != TRUE))
            return ss ;
    }
    lineno = curwp->line_no ;
    curcol = curwp->w_doto ;
    if((doRcsCommand(curbp->b_fname,str) != TRUE) ||
       (readin(curbp,curbp->b_fname) != TRUE) ||
       (gotoLine(TRUE, lineno+1) != TRUE))
        return FALSE ;
    if(curcol > llength(curwp->w_dotp))
        curwp->w_doto = llength(curwp->w_dotp) ;
    else
        curwp->w_doto = curcol ;

    return TRUE ;
}

#endif
