/* -*- C -*- ****************************************************************
 *
 *  			Copyright 1996-1999 Jasspa
 *
 *
 *  System        : MicroEmacs Jasspa Distribution
 *  Object Name   : $RCSfile$
 *  Revision      : $Revision$
 *  Date          : $Date$
 *  Author        : $Author$
 *  Created By    : Steve Philips
 *  Created       : 1996
 *  Last Modified : <160699.0735>
 *
 *  Description
 *      Print drivers
 *
 *  Notes
 *      Jon Green took this apart in 1998 and put it back together
 *      It still needs more work !!
 *
 *  History
 *
 *  $Log$
 *
 ****************************************************************************
 *
 * Copyright (C) 1996-1999, JASSPA 
 *
 * The MicroEmacs Jasspa distribution can be copied and distributed freely for
 * any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
 * incorporated into commercial software with the expressed permission of
 * JASSPA.
 *  
 ****************************************************************************/

#include "emain.h"

#if (defined _UNIX) || (defined _DOS) || (defined _WIN32)
#include <sys/types.h>
#include <time.h>
#endif
#include "eprint.h"

#define TESTPRINT  0                    /* Enable printer testing */

char *printNames [] =
{
#define DEFPRINT(s,t,v) s,
#include "eprint.def"
    NULL
#undef DEFPRINT
};

char printTypes[] =
{
#define DEFPRINT(s,t,v) t,
#include "eprint.def"
    -1
#undef DEFPRINT
};

PRINTER printer;
static BUFFER *gbp;                     /* Source of print buffer */
static BUFFER *gpbp;                    /* Composition print buffer */

#define CHAR_FONT   (0*P2_NUMFONTS)
#define START_FONT  (1*P2_NUMFONTS)
#define END_FONT    (2*P2_NUMFONTS)
#define MAX_FONT    (3*P2_NUMFONTS)

static char *fontStrings [MAX_FONT];
static char *fontPaths[] = {"over", "start", "end"};

typedef struct
{
    int snum;                           /* Number of schemes in the system */
    char *curStab;                      /* Current scheme table ptr */
    int numStab;                        /* Current number in scheme table */
    unsigned char **stab;               /* Scheme table */
} PHILIGHT;

/* The hilighting characters */
static char hilchars[INUMFONTS+1] = {'b', 'i', 'u', 'r', 'n'};
static PHILIGHT pHilTab = {0, NULL};

#ifdef FREE_ALL_MEMORY
void printFreeMemory(void)
{
    if (pHilTab.snum)
    {
        int schmNo;                         /* Current scheme number */
        for(schmNo=0 ; schmNo<pHilTab.snum ; schmNo++)
            meNullFree(pHilTab.stab[schmNo]) ;
        free(pHilTab.stab) ;
    }
}
#endif

/*
 * printGetMask
 * Get the printer mask from the user
 */

static int
printGetMask (unsigned char *pmask)
{
    char buf [20];                      /* Prompt buffer */
    unsigned char mask;                 /* Scheme mask */
    char cc;                            /* Local character store */
    char *p;                            /* Local character pointer */

    if (mlreply ("Printer mask", 0, 0, buf, 20) == ABORT)
        return FALSE;
    for (p = buf, mask = 0; (cc = *p) != '\0'; p++)
    {
#if 0
        /* Treat digits as a hexadecimal mask */
        if ((cc >= '0') && (cc <= '9'))
            mask |= cc - '0';
        /* Treat A..F as a hexadecimal mask value */
        else if ((cc >= 'A') && (cc <= 'F'))
            mask |= (cc + 10) - 'A';
        /* Determine special charcters. We have:
         *    b = bold      (=1)
         *    i = italic    (=2)
         *    u = underline (=4)
         *    r = reverse   (=8)
         */
        else
            ;
#endif
        if (cc == 'b')
            mask |= 1;
        else if (cc == 'i')
            mask |= 2;
        else if (cc == 'u')
            mask |= 4;
        else if (cc == 'r')
            mask |= 8;
        else
            return mlwrite (MWABORT|MWPAUSE, "Illegal hilight print option %c", cc);
    }
    *pmask = mask;
    return TRUE;
}

/*
 * getTranslationLen
 * Compute the length of the string with characters translated
 */
static int
getTranslationLen (unsigned char *str, int n)
{
    int len = 0;                        /* Length of the string */
    int cc;                             /* Character pointer */
    char *p;

    while (((cc = *str++) != '\0') && (--n >= 0))
    {
        if ((p = printer.filter[cc]) == NULL)
            len++;
        else
            len += strlen (p);
    }
    return (len);
}

/*
 * doTranslation
 * Translate the characters in the string ready for the printer.
 */
static int
doTranslation (unsigned char *dest, unsigned char *str, int n)
{
    int len = 0;                        /* Length of the string */
    int cc;                             /* Character pointer */
    unsigned char *p;                   /* Character pointer */

    while (((cc = *str++) != '\0') && (--n >= 0))
    {
        /* Convert special characters i.e. box chars to an ASCII equivelent */
        if ((meSystemCfg & meSYSTEM_FONTFIX) && (cc < TTSPECCHARS))
            cc = ttSpeChars [cc];
        
        /* Process the character through the filter */
        if ((p = (unsigned char *) printer.filter[cc]) == NULL)
            dest [len++] = cc;
        else
        {

            while ((dest[len] = *p++) != '\0')
                len++;
        }
    }
    return len;
}

/*
 * printHilight
 * Pick up the colour scheme hilighting information.
 * We construct a 2-D dynamic table as follows, the primary index
 * is the colour scheme number, the secondary index is the hilight
 * colour. This addressing references a bit mask which identifies the
 * print code.
 *
 * This may appear to be a little cumbersome to set up; but it does
 * allow ALL hilighting schemes and ALL hilighting colours to be
 * represented.
 *
 * An alternative suggested was to only map the 1st 32 colours, however
 * as in most cases this usually bites in the end so in this case we
 * have not done that.
 */
int
printHilight (int f, int n)
{
    char buf [MAXBUF];                  /* Local character buffer */
    unsigned char mask;                 /* Scheme mask */
    unsigned char *up;                  /* Local pointer */
    int schmNo;                         /* Current scheme number */
    int noHilCols;                      /* number of hilight colours */
    int status;
    int ii;                             /* Local loop counter */

    /* Get the scheme number */
    if (mlreply ("Hilight scheme number", 0, 0, buf, 10) == ABORT)
        return FALSE;
    if (((schmNo = meAtoi(buf)) < 0) || (schmNo > 255))
        return mlwrite (MWABORT|MWPAUSE,"Invalid scheme number %d", schmNo);

    /* Add a new entry to the printer hilighter table */
    if (pHilTab.snum <= schmNo)
    {
        pHilTab.stab = realloc (pHilTab.stab, sizeof (char *) * (schmNo+1));
        if (pHilTab.stab == NULL)
        {
            /* Make safe */
            pHilTab.snum = 0;
            return mlwrite (MWABORT|MWPAUSE, "Cannot allocate printer hilight scheme table");
        }
        /* Fill the memory with zero's */
        memset (&pHilTab.stab[pHilTab.snum], 0,
                ((schmNo+1) - pHilTab.snum) * sizeof (char *));
        pHilTab.snum = schmNo+1;
    }

    /* If this is a reset operation then reset the table and quit. */
    if (f)
    {
        if (pHilTab.stab[schmNo] != NULL)
        {
            meFree (pHilTab.stab[schmNo]);
            pHilTab.stab[schmNo] = NULL;
        }
        return TRUE;
    }

    /* Get the colour mask information */
    if ((status = printGetMask (&mask)) != TRUE)
        return status;

    /* Point to the colour table and keep a track of the largest colour */
    if ((up = pHilTab.stab[schmNo]) == NULL)
        noHilCols = 0;
    else
        noHilCols = *up;

    /* Get a list of scheme numbers */
    for (;;)
    {
        /* Get the colour index out */
        if (((status = mlreply ("Color Number", 0, 0, buf, MAXBUF)) != TRUE) || (buf[0] == '\0'))
            break;
        if ((ii = meAtoi(buf)) < 0)
            break;
        if (ii > 255)
        {
            status = mlwrite (MWPAUSE|MWABORT, "Hilight color out of range %d", ii);
            break;
        }

        /* Insert into the table. */
        if (ii >= noHilCols)
        {
            if ((up = realloc (up, ii+2)) == NULL)
            {
                status = mlwrite (MWPAUSE|MWABORT, "Cannot allocate hilight table");
                break;
            }
            memset (&up[noHilCols+1], 0, (ii+1) - noHilCols);
            noHilCols = ii+1;
        }
        up[ii+1] = mask;
    }

    /* Save the table information back */
    if (up)
        *up = noHilCols;
    pHilTab.stab[schmNo] = up;
    return ((status == ABORT) ? ABORT:TRUE) ;
}

static int
printGetParams (void)
{
    REGHANDLE regPrnRoot;               /* Root of the print registry */
    REGHANDLE regHisRoot;               /* Root of the history registry */
    REGHANDLE regTmpRoot;               /* Temporary root */
    REGHANDLE regMajor;                 /* Root of the major device */
    REGHANDLE regMinor;                 /* Root of the minor device */
    REGHANDLE regData;                  /* Temporary data */
    char *ss ;
    int ii;
#if TESTPRINT 
    FILE *fp;
#endif
    mlwrite (0, "Printer reading the registry ... ");

    /* Get the name of the driver out of the registry history */
    if ((regHisRoot = regFind (NULL, "/history/printer")) == NULL)
        regHisRoot = regSet (NULL, "/history/printer", NULL);
    
    if (((regTmpRoot = regFind (regHisRoot, meSYSTEM_NAME)) == NULL) ||
        ((printer.pDeviceMajor = regGetString (regTmpRoot,NULL)) == NULL))
    {
#ifdef WIN32
        /* Under windows automatically create the driver */
        printer.pDeviceMajor = "windows";
        regTmpRoot = regSet (regHisRoot, meSYSTEM_NAME, printer.pDeviceMajor);
#else
        return mlwrite (MWABORT|MWPAUSE, "[Printer driver not setup]");
#endif
    }

    /* Get the page size out */
    if (((regHisRoot = regFind (regTmpRoot, printer.pDeviceMajor)) == NULL) ||
        ((printer.pDeviceMinor = regGetString (regHisRoot,NULL)) == NULL))
    {
#ifdef WIN32
        /* Automatic construct of printer minor in history registry if
         * it does not exist */
        if (strcmp (printer.pDeviceMajor, "windows") == 0)
        {
            printer.pDeviceMinor = "automatic";
            regHisRoot = regSet (regTmpRoot,printer.pDeviceMajor,printer.pDeviceMinor);
        }
        else
#endif
            return mlwrite (MWABORT|MWPAUSE, "[Printer page size not setup]");
    }
    
    /* Get the output destination */
    if (((regTmpRoot = regFind (regHisRoot, "dest")) == NULL) ||
        ((ss = regGetString (regTmpRoot,NULL)) == NULL) ||
        ((printer.pDestination=meAtoi(ss)) < 0) ||
        (printer.pDestination > PDEST_COMLINE))
        printer.pDestination = PDEST_BUFFER ;
#ifndef _WIN32
    /* Only windows support an internal device */
    if (printer.pDestination == PDEST_INTERNAL)
        /* Not supported - use buffer */
        printer.pDestination = PDEST_BUFFER ;
#else 
    if ((printer.pDestination == PDEST_INTERNAL) &&
        ((strcmp (printer.pDeviceMajor, "windows") != 0) ||
         (strcmp (printer.pDeviceMinor, "automatic") != 0)))
        printer.pDestination = PDEST_BUFFER ;
#endif

    /* Find the printer information from the registry */
    if ((regPrnRoot = regFind (NULL, "/printer")) == NULL)
    {
        char filename[FILEBUF] ;	        /* Filename */

        if ((!fileLookup("printer",".erf",meFL_CHECKDOT|meFL_USESRCHPATH,filename)) ||
            ((regPrnRoot = regRead ("printer", filename, 0)) == NULL))
            return mlwrite(MWABORT|MWWAIT,"[Cannot find printer file \"printer.erf\"]") ;
    }

    /* Find the description of the driver */
    if ((regMajor = vregFind (regPrnRoot, "devices/%s", printer.pDeviceMajor)) == NULL)
        return mlwrite (MWPAUSE|MWABORT, "[Cannot find printer driver description for %s]", printer.pDeviceMajor);
    if ((regMinor = vregFind (regMajor, "layout/%s", printer.pDeviceMinor)) == NULL)
        return mlwrite (MWPAUSE|MWABORT, "[Cannot find printer layout description for %s]", printer.pDeviceMinor);
    
#if TESTPRINT
    /*
     * Get all of the printer codes out of the registry.
     */
    fp = fopen ("print.cfg", "wb");

#endif
    /* Get all of the printer strings out. Simply link to the registry strings
     * we do not perfom any copy operation. We check in major first and then
     * in minor. If it exists in the minor then it over-writes the major.
     */
    for (ii = 0; ii < P2_NUMFONTS; ii++)
    {
        char subkey [INUMFONTS+1];
        int len;
        int jj;

        for (jj = 0, len = 0; jj < INUMFONTS; jj++)
        {
            if (ii & (1 << jj))
                subkey[len++] = hilchars[jj];
        }
        /* Special case of 'normal' text */
        if (len == 0)
            subkey [len++] = 'n';
        subkey[len] = '\0';

        /* Search for the font definitions */
        for (jj = 0; jj < 3; jj++)
        {
#if TESTPRINT
            fprintf (fp, "Searching for %s/%s\n",fontPaths[jj], subkey);
#endif
            /* Search in /printer/devices/<driver>/layout/<layout>/<section>/id */
            if ((regData = vregFind (regMinor, "%s/%s", fontPaths[jj], subkey)) == NULL)
                /* Search in /printer/devices/<driver>/<section>/id */
                regData = vregFind (regMajor, "%s/%s", fontPaths[jj], subkey);

            /* Get the font string out */
            fontStrings [ii+(jj*P2_NUMFONTS)] = (regData == NULL) ?
                      NULL : regGetString (regData, NULL) ;
#if TESTPRINT
            if (fontStrings [ii+(jj*P2_NUMFONTS)])
                fprintf (fp, "Font string [%d:%s] = %s\n",
                         ii+(jj*P2_NUMFONTS),
                         subkey,
                         fontStrings [ii+(jj*P2_NUMFONTS)]);
#endif
        }
    }

    for (ii = 0; ii < PRINT_MAX; ii++)
    {
        /* NOTE:- If any of these change print.emf must be up-dated also */
        /* Try /history/printer/<platform>/<driver>/<layout>/<id> */
        if ((regData = vregFind (regHisRoot, "%s/%s", printer.pDeviceMinor, printNames[ii])) != NULL)
            ;
        /* Try /printer/devices/<driver>/layout/<layout>/<id> */
        else if ((regData = regFind (regMinor, printNames[ii])) != NULL)
            ;
        /* Try /printer/devices/<driver>/<id> */
        else if ((regData = regFind (regMajor, printNames[ii])) != NULL)
            ;
        /* Try /printer/<platform>/<id> */
        else if ((regData = vregFind (regPrnRoot, "%s/%s", meSYSTEM_NAME, printNames[ii])) != NULL)
            ;
        /* Try /printer/<id> */
        else
            regData = regFind (regPrnRoot, printNames[ii]);

        /* Get the string or the integer out */
        if (regData == NULL)
        {
            if (printTypes[ii] & mePD_INT)
                printer.param[ii].l = 0;
            else
                printer.param[ii].p = NULL;
        }
        else if (printTypes[ii] & mePD_INT)
            printer.param[ii].l = regGetLong (regData, 0) ;
        else
            printer.param[ii].p = regGetString (regData, NULL);
#if TESTPRINT
        if (fp)
        {
            if (printTypes[ii] & mePD_INT)
                fprintf (fp, "Parameter %s = %ld\n", printNames [ii],
                         printer.param[ii].l);
            else
                fprintf (fp, "Parameter %s = %s\n", printNames [ii],
                         (printer.param[ii].p == NULL) ? "(nil)" : printer.param[ii].p);
        }
#endif
    }

    /*
     * Get the filter codes out of the registry
     */

    /* Allocate the filter table */
    if (printer.filter == NULL)
        printer.filter = meMalloc (sizeof (char *) * 256);
    memset(printer.filter,0, sizeof(char *) * 256) ;

    /* Search in /printer/devices/<driver>/layout/<layout>/filter */
    if ((regData = regFind (regMinor, "filter")) == NULL)
        /* Search in /printer/devices/<driver>/filter */
        regData = regFind (regMajor, "filter");
    if (regData != NULL)
    {
        char *nodeName;
        regData = regGetChild(regData) ;

        while (regData != NULL)
        {
            /* Get the node name out */
            nodeName = regGetName(regData) ;
            if (nodeName[0] != '\0')
            {
                if (nodeName[1] == '\0')
                    ii = nodeName[0];
                else
                    ii = meAtoi (nodeName);
                if ((ii >= 0) && (ii <= 255))
                    printer.filter [ii] = regGetString (regData, NULL);
            }
            regData = regGetNext (regData) ;
        }
    }
#if TESTPRINT
    if (fp)
        fclose (fp);
#endif
    mlwrite (0, "[Printer read the registry ...]");
    /* Add a few defaults */
    if (printer.param[mePI_COLS].l < 1)
        printer.param[mePI_COLS].l = 1;
    if (printer.param[mePI_ROWS].l < 1)
        printer.param[mePI_ROWS].l = 1;
    if (printer.param[mePS_ECONT].p == NULL)
        printer.param[mePS_ECONT].p = " ";
    /* Default the output states */
    if (printer.param[mePS_BUFFER].p == NULL)
        printer.param[mePS_BUFFER].p = "*printer*";
    return TRUE;
}

/*
 * printComputePageSetup
 * Compute the page setup for the printer
 */
int
printComputePageSetup (int f)
{
    int xtraWidth;
    int xtraDepth;
    /* Sort out the parameters */
    if (printer.param[mePI_COLS].l < 1)
        printer.param[mePI_COLS].l = 1;
    if (printer.param[mePI_ROWS].l < 1)
        printer.param[mePI_ROWS].l = 1;
    xtraWidth = printer.param[mePI_MLEFT].l;  /* Left margin */
    xtraWidth += (printer.param[mePI_COLS].l-1) * printer.param[mePI_WSEP].l;
    if (printer.pLineNumDigits > 0)
        xtraWidth += printer.param[mePI_COLS].l * (printer.pLineNumDigits);

    xtraDepth = (printer.param[mePI_ROWS].l-1);
    if (printer.param [mePI_FLAGS].l & PFLAG_ENBLFOOTER)
        xtraDepth += printer.param[mePI_FSEP].l + printer.param[mePI_FLEN].l;
    if (printer.param [mePI_FLAGS].l & PFLAG_ENBLHEADER)
        xtraDepth += printer.param[mePI_TSEP].l + printer.param[mePI_HLEN].l;

    /* Fill in some auto's if not defined. */
    if (printer.param[mePS_ECONT].p == NULL)
        printer.param[mePS_ECONT].p = " ";

    /* Automatically fill some of the entries. */
    if ((printer.param[mePI_PAGEX].l == 0) && (printer.param[mePI_PAGEY].l == 0))
    {
        /* The page size is undefined, generate the page size from the
         * paper size */
        printer.param[mePI_PAGEX].l = (printer.param[mePI_PAPERX].l - xtraWidth) /
                  printer.param[mePI_COLS].l;
        printer.param[mePI_PAGEY].l = (printer.param[mePI_PAPERY].l - xtraDepth) /
                  printer.param[mePI_ROWS].l;
    }
    else if ((printer.param[mePI_PAPERX].l == 0) && (printer.param[mePI_PAPERY].l == 0))
    {
        /* The paper size is undefined, determine the paper size */
        printer.param[mePI_PAPERX].l = printer.param[mePI_PAGEX].l * printer.param[mePI_COLS].l + xtraWidth;
        printer.param[mePI_PAPERY].l = printer.param[mePI_PAGEY].l * printer.param[mePI_ROWS].l + xtraDepth;
    }

    /* Check to ensure that the page request is legal. */
    if ((printer.param[mePI_PAPERX].l < (printer.param[mePI_PAGEX].l * printer.param[mePI_COLS].l + xtraWidth)) ||
        (printer.param[mePI_PAPERY].l < (printer.param[mePI_PAGEY].l * printer.param[mePI_ROWS].l + xtraDepth)))
        return mlwrite (MWABORT|MWPAUSE, "Invalid paper size %d x %d. Page requirements %d x %d",
                        printer.param[mePI_PAPERX].l,
                        printer.param[mePI_PAPERY].l,
                        printer.param[mePI_PAGEX].l * printer.param[mePI_COLS].l + xtraWidth,
                        (printer.param[mePI_PAGEY].l * printer.param[mePI_ROWS].l + xtraDepth));
    else if (!f)
        mlwrite (0, "Valid paper size %d x %d. Page requirements %d x %d of %d x %d",
                 printer.param[mePI_PAPERX].l,
                 printer.param[mePI_PAPERY].l,
                 printer.param[mePI_PAGEX].l,
                 printer.param[mePI_PAGEY].l,
                 printer.param[mePI_COLS],
                 printer.param[mePI_ROWS]);
    return TRUE;
}

/*
 * Initialise the driver ready for a print.
 * Recompute the print metrics ready for the page.
 */
static int
printInit (int f, int n)
{
#ifdef _WIN32
    extern int printSetup(void);
#endif
    /* Get the parameters out of the registry. */
    if (printGetParams () != TRUE)
        return FALSE;
    
    /* if the line numbers aren't required then set printer.pLineNumDigits to zero
     * so we only have to check that value */
    if (!(printer.param [mePI_FLAGS].l & PFLAG_ENBLLINENO))
        printer.pLineNumDigits = 0 ;

#ifdef _WIN32
    /* If this is the windows printer then invoke the windows setup. */
    if (strcmp (printer.pDeviceMajor, "windows") == 0)
    {
        if (printSetup() != TRUE)
            return FALSE;
    }
    else
#endif
        if (printComputePageSetup (f) != TRUE)
            return FALSE;

    mlwrite (0, "Checking printer parameters ...");

    /* Get the hilighting scheme out - if NULL try default (0) */
    if (((curbp->hiLight >= pHilTab.snum) ||
         ((pHilTab.curStab = (char *) pHilTab.stab[curbp->hiLight]) == NULL)) &&
        ((pHilTab.curStab = (char *) pHilTab.stab[0]) == NULL))
    {
        pHilTab.numStab = 0;
        pHilTab.curStab = NULL;         /* MUST set this to NULL here !! */
        mlwrite (MWPAUSE, "[Printer hilighting not defined for buffer]");
    }
    else
        pHilTab.numStab = *pHilTab.curStab++;

    mlwrite (0, "Setup printer parameters ...");
    return TRUE;
}

/*
 * addComposedLine
 * Add a composed line to the lpage line list
 */
static void
addComposedLine (LINE **head, LINE **tail, char *buf, int len)
{
    LINE *nline;                        /* New line */

    nline = lalloc (len);               /* Get the new line */
    if (len > 0)                        /* Copy in the data */
        memcpy (nline->l_text, buf, len);
    lforw (nline) = NULL;
    if ((lback (nline) = *tail) == NULL)
        *tail = (*head = nline);
    else
    {
        lforw (*tail) = nline;
        *tail = nline;
    }
}

/*
 * expandText
 * Expand control codes in the escape sequence
 */
static int
expandText (unsigned char *buf, char *strp, int level)
{
    unsigned char cc;
    unsigned char *bbuf = buf;
    unsigned char *ss;

    if (strp == NULL)
        return 0;

    while ((cc = *strp) != '\0')
    {
        strp++;
        /* Check out the escape characters */
        if (cc == '%')
        {
            switch (cc = *strp)
            {
            case 'c':
                if (((cc = *++strp) == 0xff) && (strp[1] == 0x01))
                {
                    strp++;
                    *buf++ = '\0';
                    }
                    else
                        *buf++ = cc;
                break;
            case 'h':
                buf += sprintf ((char *) buf, "%02d", printer.ptime->tm_hour);
                break;
            case 'm':
                buf += sprintf ((char *) buf, "%02d", printer.ptime->tm_min);
                break;
            case 's':
                buf += sprintf ((char *) buf, "%02d", printer.ptime->tm_sec);
                break;
            case 'D':
                buf += sprintf ((char *) buf, "%02d", printer.ptime->tm_mday);
                break;
            case 'M':
                buf += sprintf ((char *) buf, "%02d", printer.ptime->tm_mon+1);
                break;
            case 'Y':
                buf += sprintf ((char *) buf, "%02d", printer.ptime->tm_year);
                break;
            case 'f':
                if ((gpbp->b_bname[0] == '*') || (gpbp->b_fname == NULL))
                    ss = (unsigned char *) gpbp->b_bname;
                else
                    ss = (unsigned char *) gpbp->b_fname;
                goto cat_string;
            case 'b':
                ss = (unsigned char *) gbp->b_bname;
cat_string:
                if (ss != NULL)
                    buf += sprintf ((char *) buf, "%s", ss);
                break;
            case 'n':
                buf += sprintf ((char *) buf, "%d", printer.pPageNo);
                break;
            case '\0':
                break;
            default:
                *buf++ = cc;
                break;
            }
            if (*strp != '\0')
                strp++;
        }
        else if (cc == meNLCHAR)
        {
            if (level == 0)
                buf += expandText (buf, printer.param [mePS_EOL].p, 1);
            else
                *buf++ = meNLCHAR;
        }
        else if (cc == '\r')
        {
            if (level != 0)
                *buf++ = '\r';
        }
        else if (cc == '\t')
            *buf++ = ' ';
        else
            *buf++ = cc;
    }
    /* Return the delta length of the string */
    return buf - bbuf;
}
/*
 * expandTitleText
 * Expand the title text into the appropriate form
 */
static void
addFormatedLine (LINE **head, LINE **tail,
                 unsigned char *buf, int noLines, char *strp, int filter)
{
    unsigned char *bhead = buf;
    char *p = strp;

    while (--noLines >= 0)
    {
        int ii;
        unsigned char cc;
        unsigned char tbuf [MAXBUF];

        /* Get the new lines out */
        if (strp != NULL)
        {
            for (ii = 0; (cc = *p) != '\0'; /* NULL */)
            {
                p++;
                if (cc == meNLCHAR)
                    break;
                /* Filter the character if required. */
                if (filter == 0)
                    tbuf [ii++] = cc;
                else
                    ii += doTranslation (&tbuf[ii], &cc, 1);
            }
        }
        else
            ii = 0;
/*        tbuf [ii++] = meNLCHAR;*/
        tbuf [ii] = '\0';

        /* Add the start of line code */
        buf = bhead + expandText (bhead, printer.param [mePS_SOL].p, 1);
        /* Expand the header text and append to the composition buffer */
        buf = buf + expandText (buf, (char *) tbuf, 0);
        /* Add the end of line code */
        buf = buf + expandText (buf, printer.param [mePS_EOL].p, 1);
        addComposedLine (head, tail, (char *) bhead, buf - bhead);
    }
}

/*
 * Render the page into a line store list
 */
LINE *
composePage (int f)
{
    LINE *head = NULL;
    LINE *tail = NULL;
    unsigned char buf [1024*5];         /* Heafty line store !! */
    unsigned char *p;                   /* Pointer to the buffer */
    int xx;                             /* Page column iterator */
    int yy;                             /* Page row iterator */
    int ll;                             /* Lines on page iterator */
    int ii;
    LINE *lp;                           /* Pointer to the line */

    /* Handle the end of file case */
    if (f)
    {
        if (f < 0)
        {
            /* End of file. */
            if (printer.param [mePS_EOF].p != NULL)
            {
                p = buf + expandText (buf, printer.param [mePS_EOF].p, 1);
                addComposedLine (&head, &tail, (char *) buf, p - buf);
            }
            return head;
        }
        else
        {
            /* Start of file */
            if (printer.param [mePS_SOF].p != NULL)
            {
                p = buf + expandText (buf, printer.param [mePS_SOF].p, 1);
                addComposedLine (&head, &tail, (char *) buf, p - buf);
            }
            return head;
        }
    }

    /* Add the start of page marker */
    if (printer.param [mePS_SOP].p != NULL)
    {
        p = buf + expandText (buf, printer.param [mePS_SOP].p, 1);
        addComposedLine (&head, &tail, (char *) buf, p - buf);
    }

    /* Render the header and the separator between the header and the textual
     * page */
    addFormatedLine (&head, &tail, buf, printer.param [mePI_MTOP].l, NULL, 0);
    if (printer.param [mePI_FLAGS].l & PFLAG_ENBLHEADER)
    {
        addFormatedLine (&head, &tail, buf, printer.param [mePI_HLEN].l, printer.param [mePS_HEADER].p, 1);
        addFormatedLine (&head, &tail, buf, printer.param [mePI_TSEP].l, NULL, 0);
    }

    /* Iterate over the number of page rows */
    for (yy = 0; yy < printer.param [mePI_ROWS].l; yy++)
    {
        /* Iterate over the lines on the page */
        for (ll = 0; ll < printer.param [mePI_PAGEY].l; ll++)
        {
            /* Add the start of line code */
            p = buf + expandText (buf, printer.param [mePS_SOL].p, 1);

            /* Add the leading margin */
            for (ii = printer.param [mePI_MLEFT].l; --ii >= 0; *p++ = ' ')
                /* NULL */;

            /* Iterate over the number of page columns */
            for (xx = 0;;)
            {
                /* Go to the start of the line in the buffer */
                ii = ((yy * printer.param [mePI_COLS].l + xx) * printer.param [mePI_PAGEY].l) + ll;
                lp = lforw (gbp->b_linep);
                while (--ii >= 0)
                {
                    if ((lp = lforw (lp)) == gbp->b_linep)
                    {
                        lp = NULL;
                        break;
                    }
                }

                /* Process the line */
                if (lp != NULL)
                {
                    /* Add the continuation marker - we have already done the line numbers */
                    if (((lp->l_flag & LNCHNG) == 0) && (printer.pLineNumDigits > 0))
                    {
                        for (ii = printer.pLineNumDigits-2; --ii >= 0; *p++ = ' ')
                            /* NULL */;
                        p += expandText (p, printer.param [mePS_ECONT].p, 1);
                        *p++ = ' ' ;
                    }

                    /* Concatinate the composed line information */
                    {
                        unsigned char cc;
                        unsigned char *str;

                        str = lp->l_text;
                        while ((cc = *str++) != '\0')
                        {
                            if (cc == 255)
                            {
                                if ((cc=*str++) == 0x01)
                                    cc = 0;
                                else if (cc == 0x02)
                                    cc=*str++;
                            }
                            *p++ = cc;
                        }
                    }
                }
                else
                {
                    /* NULL line. Pad the line with spaces */
                    ii = printer.param [mePI_PAGEX].l;
                    if (printer.pLineNumDigits > 0)
                        ii += printer.pLineNumDigits ;
                    while (--ii >= 0)
                        *p++ = ' ';
                }

                /* End of the line reached ?? */
                if (++xx >= printer.param [mePI_COLS].l)
                    break;

                /* Add a new column */
                p += expandText (p, printer.param [mePS_VSEP].p, 1);
            }

            /* Add the end of line marker and push the buffer */
            if (printer.param [mePI_STRIP].l)
            {
                while ((p - buf) > 0)
                {
                    char cc;
                    if (((cc = *(p-1)) == ' ') || (cc == '\t'))
                        p--;
                    else
                        break;
                }
            }
            p += expandText (p, printer.param [mePS_EOL].p, 1);
            addComposedLine (&head, &tail, (char *) buf, p - buf);
        }

        /* Add the horizontal separator if required */
        if (yy+1 < printer.param [mePI_ROWS].l)
        {
            /* Add the start of line code */
            p = buf + expandText (buf, printer.param [mePS_SOL].p, 1);

            /* Add the leading margin */
            for (ii = printer.param [mePI_MLEFT].l; --ii >= 0; *p++ = ' ')
                /* NULL */;

            for (xx = printer.param [mePI_COLS].l; --xx >= 0; /* NULL */)
            {
                int nn;

                /* Work out the width - length of text plus line # width */
                nn = printer.param [mePI_PAGEX].l + printer.pLineNumDigits;
                while (--nn >= 0)
                    p += expandText (p, printer.param [mePS_HSEP].p, 1);

                /* If there is another column then add the vertical crossover
                 * separator. */
                if (xx > 0)
                    p += expandText (p, printer.param [mePS_XSEP].p, 1);
            }

            /* Add the end of line marker and push the buffer */
            p += expandText (p, printer.param [mePS_EOL].p, 1);
            addComposedLine (&head, &tail, (char *) buf, p - buf);
        }
    }

    /* Render the footer and the separator between the footer and the textual
     * page */
    if (printer.param [mePI_FLAGS].l & PFLAG_ENBLFOOTER)
    {
        addFormatedLine (&head, &tail, buf, printer.param [mePI_FSEP].l, NULL, 0);
        addFormatedLine (&head, &tail, buf, printer.param [mePI_FLEN].l, printer.param [mePS_FOOTER].p, 1);
    }
    
    /* Add the end of line marker and push the buffer */
    if (printer.param [mePS_EOP].p != NULL)
    {
        p = buf + expandText (buf, printer.param [mePS_EOP].p, 1);
        addComposedLine (&head, &tail, (char *) buf, p - buf);
    }
    return head;
}

/*
 * dumpToInternal
 * Write the line into the iternal line list
 */
static void
dumpToInternal (LINE **headp, LINE **tailp, LINE *lp)
{
    /* Check for NULL case - this should not happen */
    if (lp == NULL)
        return;

    /* Link to the head */
    if (*tailp == NULL)
        *headp = lp;
    else
    {
        lforw(*tailp) = lp;             /* Attach head to end of list */
        lback(lp) = *tailp;
    }

    /* Fix up the tail */
    while (lforw(lp) != NULL)
        lp = lforw(lp);
    *tailp = lp;
}

/*
 * dumpToFile
 * Write the composed lines into the file
 */
static void
dumpToFile (FILE *fp, LINE *lp)
{
    /* Dump the line to file */
    while (lp != NULL)
    {
        LINE *tlp;                      /* Temp line store */

        tlp = lp;
        lp = lforw (lp);

        fwrite (tlp->l_text, 1, (int)(tlp->l_used), fp);
#if TESTPRINT
        {
            FILE *fp2;
            if ((fp2 = fopen ("spool2.out","ab")) != NULL)
            {
                fwrite (tlp->l_text, 1, (int)(tlp->l_used), fp2);
                fclose (fp2);
            }
        }
#endif
        meFree (tlp);
    }
}

/*
 * dumpToBuffer
 * Write the composed lines into a buffer. We have to un-translate the
 * lines here. This is not particularly efficient since we have already
 * translated (nil) characters to '\0's, now we have to expand them again !!
 * However this is not really a problem, just a bit of a pain !!.
 */
static void
dumpToBuffer (BUFFER *bp, LINE *lp)
{
    int ii;
    char *p;

    while (lp != NULL)
    {
        LINE *tlp;                      /* Temp line store */
        char cc;
        int len;
        int jj, kk;

        /* Advance the line pointer */
        tlp = lp;
        lp = lforw (lp);

        /* Determine how many '\0's are in the string */
        jj = 0;
        while (jj != tlp->l_used)
        {
            LINE *nlp;
            char *q;
            
            for (kk = jj, ii = tlp->l_used - jj, len = 0, p = (char *) &tlp->l_text [jj]; --ii >= 0; /* NULL */)
            {
                jj++;
                if ((cc = *p++) == '\0')
                    len += 2;
                else if (cc == meNLCHAR)
                    break;
                else 
                    len++;
            }
            
            /* Construct a new line for the buffer and translate out any 0's */
            nlp = lalloc (len);
            for (p = (char *) &tlp->l_text[kk], q = (char *) nlp->l_text; --len >= 0; /* NULL */)
            {
                if ((cc = *p++) == '\0')
                {
                    *q++ = (char)0xff;
                    *q++ = (char)0x01;
                }
                else if (cc == meNLCHAR)
                    break;
                else
                    *q++ = cc;
            }
                
            /* Link the new line to the end of the buffer. */
            lforw(lback(bp->b_linep)) = nlp;
            lback(nlp) = lback(bp->b_linep);
            lback(bp->b_linep) = nlp;
            lforw(nlp) = bp->b_linep;
            bp->elineno += 1;
        }
        /* Free off the old line and assign the converted line. */
        meFree (tlp);
    }
}

void
printLinkLine (BUFFER *bp, LINE *nlp, unsigned short lno)
{
    bp->b_linep->l_bp->l_fp = nlp;
    nlp->l_bp = bp->b_linep->l_bp;
    bp->b_linep->l_bp = nlp;
    nlp->l_fp = bp->b_linep;
    if (lno)
        nlp->l_flag &= ~LNCHNG;
    else
        printer.pLineNo++;              /* Next line */
}

#define PRTSIZ 2048

#if HILIGHT

static void
getHilightInfo (unsigned char col, unsigned short *slen, unsigned char **sstr,
                unsigned short *elen, unsigned char **estr)
{
    /* If the scheme is not registered then there is no colour change
     * to perform */
    if (col >= pHilTab.numStab)
        col = 0;
    else
        col = pHilTab.curStab [col];    /* Get the colour out of table */

    /* Get the start code out of the print table */
    if ((*sstr = (unsigned char *) fontStrings [col + START_FONT]) == NULL)
    {
        *sstr = (unsigned char *) "";
        *slen = 0;
    }
    else
        *slen = strlen ((char *) *sstr);

    /* Get the end code out of the print table */
    if ((*estr = (unsigned char *) fontStrings [col + END_FONT]) == NULL)
    {
        *estr = (unsigned char *) "";
        *elen = 0;
    }
    else
        *elen = strlen ((char *) *estr);
}


int
printAddLine (BUFFER *bp, LINE *lp, VIDEO *vps)
#else
int
printAddLine (BUFFER *bp, LINE *lp)
#endif
{
    unsigned char  *sss, *ess;
    unsigned short noColChng, sln, eln;
    HILCOLOR       *colp;
    LINE           *nlp;
    unsigned short len, wid, ii, jj, kk, ll;

    /* Render the line to get the colour information */
#if HILIGHT
    if (vps[0].hilno)
    {
        noColChng = hilightLine (vps);
        colp=hilCol;
        wid = colp[noColChng-1].ccol;
    }
    else
#endif
    {
        wid = renderLine (lp->l_text,lp->l_used,0);
        noColChng = 0;
    }

    /* Commence rendering the line */
    len = 0;
    ii  = 0;
    for(;;)
    {
        /* Get the length of the text */
        ll = (unsigned short) printer.param[mePI_PAGEX].l;
        kk = wid-len ;
        if ((printer.param [mePI_FLAGS].l & PFLAG_ENBLTRUNCC))
        {
            if ((kk = wid-len) >= ll)
                kk = ll - 1 ;    /* Normalise to print length */
        }
        else if (kk > ll)
            kk = ll ;   /* Normalise to print length */

        /* Determine the length of all of the colour change strings. For
         * all of the colour changes to the end of the line, add the length
         * of storage of each to our line total (ll) */
        for (jj=0; jj<noColChng; jj++)
        {
            getHilightInfo ((unsigned char) (colp[jj].icol/HCTNUM),&sln,&sss,&eln,&ess);
            ll += sln+eln;
            if (colp[jj].ccol > len+kk)
                break;
        }

        /* Insert line numbers if required. We insert line numbers at the
         * initial expansion so that we do not need to remember where we
         * are later. However we do not add the start of line continuation
         * markers until later (i.e. a line that continues on the next line)
         * This basically saves some space and is easy to deal with later */
        if ((ii == 0) && (printer.pLineNumDigits > 0))
            ll += printer.pLineNumDigits ;
        ll += getTranslationLen (disLineBuff+len,kk) - kk;

        /* Construct a new composition line */
        if ((nlp=lalloc (ll)) == NULL)
            return (FALSE);
        printLinkLine (bp,nlp,ii++);

        /* Write the line numbers into the buffer if required */
        if ((nlp->l_flag & LNCHNG) && (printer.pLineNumDigits > 0))
        {
            char lnobuf [40];
            ll = printer.pLineNumDigits ;
            sprintf (lnobuf, "% 20d ", printer.pLineNo);
            memcpy (nlp->l_text, &lnobuf [21-printer.pLineNumDigits], ll);
        }
        else
            ll = 0;

        /* Copy the line into the buffer */
        if (noColChng == 0)
        {
            /* No colour changes - simply copy the line */
            ll += doTranslation (nlp->l_text+ll, disLineBuff+len, kk);
        }
        else
        {
            /* Propogate the colour change strings into the print buffer
             * whenever the colour changes. */
            for (jj = 0; jj < kk; /* NULL*/)
            {
                /* Get the colour change string */
                getHilightInfo ((unsigned char) (colp->icol/HCTNUM),&sln,&sss,&eln,&ess);
                /* Copy accross the hilight start text */
                strncpy ((char *) nlp->l_text+ll,(char *) sss,sln);
                ll += sln;
                if (colp->ccol > len+kk)
                {
                    /* Off the end of the line !! */
#if 0
                    strncpy ((char *) nlp->l_text+ll,(char *) disLineBuff+len+jj, kk-jj);
                    ll += kk-jj;
                    jj = kk;
#else
                    ll += doTranslation (nlp->l_text+ll,
                                         disLineBuff+len+jj,
                                         kk-jj);
                    jj = kk;
#endif
                }
                else
                {
                    /* All fits !! */
#if 0
                    strncpy ((char *) nlp->l_text+ll,
                             (char *) disLineBuff+len+jj,colp->ccol-len-jj);
                    ll += colp->ccol-len-jj;
                    jj = colp->ccol-len;
#else
                    ll += doTranslation (nlp->l_text+ll,
                                         disLineBuff+len+jj,
                                         colp->ccol-len-jj);
                    jj = colp->ccol-len;
#endif
                    colp++;
                }
                /* Terminate the string with the unhilight text */
                strncpy ((char *) nlp->l_text+ll,(char *) ess,eln);
                ll += eln;
            }
        }
        len += kk ;
        /* Handle the end of the line */
        if (len >= wid)
        {
            /* Pad the end of the line with spaces. */
            len = printer.param[mePI_PAGEX].l-kk ;
            memset(nlp->l_text+ll,' ',len);
            nlp->l_text[ll+len] = '\0';
            break ;
        }
        if ((printer.param [mePI_FLAGS].l & PFLAG_ENBLTRUNCC))
        /* Line continues - add the line continuation character */
            nlp->l_text[ll++] = '\\';
        nlp->l_text[ll] = '\0';
    }
    printer.pNoLines += ii;
    return (TRUE);
}

/*
 * printSection
 * Print a section of the buffer
 */

int
printSection (WINDOW *wp, long sLineNo, long numLines, LINE *sLine, LINE *eLine)
{
    time_t clock;                       /* Time in machine format. */
    BUFFER *bp;                         /* Composition buffer */
    WINDOW *dwp;                        /* Destination window */
    BUFFER *dbp = NULL;                 /* Destination buffer */
    FILE *fp = NULL;                    /* Output file stream */
    LINE *lp, *nlp;                     /* Local line pointers */
    LINE *clp;                          /* Composed page list */
    LINE *ihead = NULL;                 /* Internal destination header */
    LINE *itail = NULL;                 /* Internal destination trailer */
    int status = TRUE;                  /* Return status */
    char fname [MAXBUF];                /* File name buffer */
    int ii;                             /* Local loop counter. */

#if HILIGHT
    VIDEO          vps[2];
#endif
    /* Work out the line number range. */
    for (numLines += sLineNo, printer.pLineNumDigits = 1; numLines > 0; printer.pLineNumDigits++)
        numLines /= 10;

    /* Initialise the printer. */
    if (printInit (TRUE, 0) != TRUE)
        return ABORT;

    /* Get the clock out and initialise */
    clock = time (0);	                 /* Get system time */
    printer.ptime = localtime (&clock);  /* Get time frame */
    printer.pLineNo = sLineNo;
    printer.pPageNo = 0;
    printer.pNoLines = 0;

    /* Determine the destination */
    switch (printer.pDestination)
    {
    case PDEST_INTERNAL:
        break;
    case PDEST_BUFFER:
        if ((dwp = wpopup (printer.param [mePS_BUFFER].p,(BFND_CREAT|BFND_CLEAR|WPOP_USESTR))) == NULL)
            return mlwrite(MWABORT,"[Failed to create print buffer]") ;
        dbp = dwp->w_bufp;
        break;
    case PDEST_COMLINE:
    case PDEST_FILE:
        if ((printer.param [mePS_FILE].p == NULL) || (*printer.param [mePS_FILE].p == '\0'))
            mkTempName (fname, NULL, "prt");
        else
            strcpy (fname, printer.param [mePS_FILE].p);

        /* Endevour to open the file */
        if ((fp = fopen (fname, "wb")) == NULL)
            return mlwrite (MWABORT, "Unable to open printer spool file %s", fname);
        break;
    }
    
    /* Shuffle the isDisplayable test so it actually tests isPrintable */
    for(ii=0 ; ii<256 ; ii++)
        charMaskTbl1[ii] = (charMaskTbl1[ii] & ~(CHRMSK_DISPLAYABLE|CHRMSK_PRINTABLE)) |
              ((charMaskTbl1[ii] & CHRMSK_DISPLAYABLE) ? CHRMSK_PRINTABLE:0) |
              ((charMaskTbl1[ii] & CHRMSK_PRINTABLE) ? CHRMSK_DISPLAYABLE:0) ;
    /* Initialise the printer */
    if ((bp = createBuffer ("")) == NULL)
    {
        status = mlwrite (MWABORT,"Unable to setup Printer");
        goto quitEarly;
    }
    /* Set up the buffer modes correctly so that we save the codes correctly */
    meModeSet (bp->b_mode,MDAUTO);
    meModeClear (bp->b_mode,MDCRLF);
    meModeClear (bp->b_mode,MDCTRLZ);
    
#if HILIGHT
    /* Clear the 2nd entry just in case the hilighting is not defined.
     * The 2nd entry is used as a continuation when hilighting is active. */
    vps[1].wind = (vps[0].wind = wp);
    vps[1].hilno = (vps[0].hilno = wp->w_bufp->hiLight);
    vps[1].bracket = (vps[0].bracket = NULL);
    vps[1].flag = (vps[0].flag = 0);
#endif
    printer.pLinesPerPage = (printer.param[mePI_PAGEY].l *
                             printer.param[mePI_ROWS].l *
                             printer.param[mePI_COLS].l);
    /* Construct the start of file */
    if ((clp = composePage (1)) != NULL)
    {
        if (fp)                         /* File device */
            dumpToFile (fp, clp);
        else if (dbp)                   /* Buffer device */
            dumpToBuffer (dbp, clp);
        else                            /* Internal device */
            dumpToInternal (&ihead, &itail, clp);
    }

    do {
        printer.pPageNo++;              /* Next page */
        mlwrite (0, "Printing page %d ..", printer.pPageNo);

        /* Construcr a page worth of formatted data */
        while ((sLine != eLine) && (printer.pNoLines < printer.pLinesPerPage))
        {
#if HILIGHT
            vps[0].line = sLine;
            if (printAddLine (bp,sLine,vps) != TRUE)
#else
            if (printAddLine (bp,sLine) != TRUE)
#endif
            {
                status = mlwrite (MWABORT, "Internal error: Cannot add new prin line");
                goto quitEarly;
            }
            sLine = lforw (sLine);
#if HILIGHT
            /* Propogate the next line information back into the
             * current line */
            vps[0].flag = vps[1].flag;
            vps[0].hilno = vps[1].hilno;
            vps[0].bracket = vps[1].bracket;
#endif
        }

        /* Compose the formatted data into a page */
        gbp = bp;
        gpbp = wp->w_bufp;
        if ((clp = composePage (0)) == NULL)
        {
            status = mlwrite (MWABORT, "Internal error: Cannot compose printer page");
            goto quitEarly;
        }

        /* Output the formatted page to the approipriate destination */
        if (fp)                         /* File device */
            dumpToFile (fp, clp);
        else if (dbp)                   /* Buffer device */
            dumpToBuffer (dbp, clp);
        else                            /* Internal device */
            dumpToInternal (&ihead, &itail, clp);

        /* Delete the formatted page data */
        if(printer.pNoLines)
        {
            lp = lforw (bp->b_linep);
            for (ii = 0; ii < printer.pLinesPerPage; ii++)
            {
                nlp = lforw (lp);
                meFree (lp);
                lp = nlp;
                if (--printer.pNoLines == 0)
                    break;
            }
            /* Fix up the pointers */
            lforw (bp->b_linep) = lp;
            lback (lp) = bp->b_linep;
        }
    } while (sLine != eLine);

    /* Construct the end of file */
    if ((clp = composePage (-1)) != NULL)
    {
        if (fp)
            dumpToFile (fp, clp);
        else if (dbp)                   /* Buffer device */
            dumpToBuffer (dbp, clp);
        else                            /* Internal device */
            dumpToInternal (&ihead, &itail, clp);
    }

    /* Close the file descriptor */
    if (fp)
    {
        fclose (fp);
        fp = NULL;
    }
                  
    /* If this is a command line then print the file. */
    switch (printer.pDestination)
    {
    case PDEST_COMLINE:
        {
            char cmdLine [MAXBUF+20];
            char *p, *q;
            char cc;
            
            /* Expand the command line */
            p = printer.param [mePS_CMD].p;
            q = cmdLine;
            if ((p == NULL) || (*p == '\0'))
            {
                mlwrite (MWPAUSE|MWABORT, "No command line specified. Print file %s", fname);
                goto quitEarly;
            }
            
            while ((cc = *p++) != '\0')
            {
                if (cc == '%')
                {
                    /* %f - insert the filename */
                    if ((cc = *p) == 'f')
                    {
                        char *r = fname;
                        while ((*q = *r++) != '\0')
                            q++;
                        p++;                /* Skip the '%f' */
                    }
                    /* % at end of string, insert a % */
                    else if (cc == '\0')
                        *q++ = '%';
                    /* Otherwise insert the trailing character */
                    else
                    {
                        *q++ = cc;
                        p++;            /* Skip the character */
                    }
                }
                else
                    *q++ = cc;
            }
            *q = '\0';                  /* Terminate the string */
                    
            /* Call me's system to print the file. Me's version is
             * used cause its return value is correctly evaluated and
             * on unix a " </dev/null" safety arg is added to the cmdLine
             */
            if((status=doShellCommand(cmdLine)) == TRUE)
                status = (resultStr[0] == '0') ? TRUE:FALSE ;
            if(status != TRUE)
                mlwrite(MWABORT,"[Failed to print file %s]",fname);
            break;
        
        case PDEST_INTERNAL:
#ifdef _WIN32
            status = WinPrint ((curbp->b_fname != NULL) ? curbp->b_fname
                               : (curbp->b_bname != NULL) ? curbp->b_bname
                               :"UNKNOWN source",
                               ihead);
            ihead = NULL;               /* Reset - never returned */
            break;
#endif
        case PDEST_BUFFER:
            dbp->b_dotp = lforw (dbp->b_linep);
            dbp->b_doto = 0;
            dbp->line_no = 0;
            resetBufferWindows (dbp);
            break;
        }
    }
    
    /* Shut down the system */
quitEarly:
    /* Shut down the file */
    if (fp)
        fclose (fp);
    /* Kill off the buffer */
    if (bp)
        zotbuf (bp, TRUE);
    /* Free off internal list */
    while (ihead != NULL)
    {
        itail = ihead;
        ihead = lforw (ihead);
        meFree (itail);
    }
    /* Reshuffle the isDisplayable test back again */
    for(ii=0 ; ii<256 ; ii++)
        charMaskTbl1[ii] = (charMaskTbl1[ii] & ~(CHRMSK_DISPLAYABLE|CHRMSK_PRINTABLE)) |
              ((charMaskTbl1[ii] & CHRMSK_DISPLAYABLE) ? CHRMSK_PRINTABLE:0) |
              ((charMaskTbl1[ii] & CHRMSK_PRINTABLE) ? CHRMSK_DISPLAYABLE:0) ;

    /* Put out end of file indicator. */
    if (status == TRUE)
        mlwrite (0, "Printing - Done. %d page(s).", printer.pPageNo);
    return status;


}

int
printRegion (int f, int n)
{
    LINE *startLine, *endLine;
    long startLineNo;
    long numLines;

    if (curwp->w_markp == NULL)
        return noMarkSet ();

    if ((startLineNo=curwp->line_no-curwp->mlineno) == 0)
        return TRUE;
    numLines = abs (startLineNo);
    if (startLineNo < 0)
    {
        startLineNo = curwp->line_no;
        startLine = curwp->w_dotp;
        endLine = curwp->w_markp;
    }
    else
    {
        startLineNo = curwp->mlineno;
        startLine = curwp->w_markp;
        endLine = curwp->w_dotp;
    }
    return printSection (curwp, startLineNo, numLines, startLine, endLine);
}

int
printBuffer (int f, int n)
{
    return printSection (curwp, 0, curbp->elineno,
                         lforw (curbp->b_linep), curbp->b_linep);
}

