/*****************************************************************************
 *
 *	Title:		search.c
 *
 *	Synopsis:	Search routines.
 *
 ******************************************************************************
 *
 *	Filename:		search.c
 *
 *	Author:			Unknown
 *
 *	Creation Date:		07/05/85 08:19		
 *
 *	Modification date:	<990708.1304>
 *
 *	Current rev:		10.1
 *
 *	Special Comments:	
 *
 *	Contents Description:	
 *
 * The functions in this file implement commands that search in the forward
 * and backward directions.  There are no special characters in the search
 * strings.  Probably should have a regular expression search, or something
 * like that.
 *
 * Aug. 1986 John M. Gamble:
 *	Made forward and reverse search use the same scan routine.
 *
 *	Added a limited number of regular expressions - 'any',
 *	'character class', 'closure', 'beginning of line', and
 *	'end of line'.
 *
 *	Replacement metacharacters will have to wait for a re-write of
 *	the replaces function, and a new variation of ldelete().
 *
 *	For those curious as to my references, i made use of
 *	Kernighan & Plauger's "Software Tools."
 *	I deliberately did not look at any published grep or editor
 *	source (aside from this one) for inspiration.  I did make use of
 *	Allen Hollub's bitmap routines as published in Doctor Dobb's Journal,
 *	June, 1985 and modified them for the limited needs of character class
 *	matching.  Any inefficiences, bugs, stupid coding examples, etc.,
 *	are therefore my own responsibility.
 *
 * July 90. Jon Green (No references cos couldn't find fore-mentioned book !)
 *
 *	Corrected bug on 
 *	an 'undo' of a search replace function. Now insert the correct 
 *	text back into the source file, not the MAGIC pattern !!!. Required
 *	new ldelete() function, called mldelete. (see line.c).
 *
 *	Fixed bug on magic wild card '.' when used in conjunction with '*'.
 *	ie '.*'. The code always used to go to the LAST occurence of the 
 *	pattern on a foward search. This is not what is required. We want to go
 *	to the first occurence of the pattern. Tricky one that, have to
 *	look for the next literal in the MC string which we can search on.
 *	This is what we use to delimit on.
 *
 *	Added the repeat pattern, that is allow text to be imported from
 *	the search function into the replace function. ie consider the
 *	following :-
 *
 *	Old Text buffer : a += 1;
 *	                  a <<= 3;
 *
 *	Require :         a = a + 1;
 *	                  a = a << 3;
 *
 *	Can now do operation with search and replace as :=
 *
 *	Search : ^{.*} {.*}=
 *	Replace: $0 = $0 $1
 *
 *	Easy !! It always pissed me off that I couldn't do that. Wonder
 *	how many people will use it ?
 *
 * July 91	Jon Green
 *
 *	Added incrental replace facility on search and replace. Allows the same
 *	search pattern while varying the replace pattern.
 * 
 * June 99      Steven Phillips
 * 
 *      Moved over to GNU regex at last.
 *
 *****************************************************************************
 * 
 * Modifications to the original file by Jasspa. 
 *
 * Copyright (C) 1988 - 1999, JASSPA 
 *
 * The MicroEmacs Jasspa distribution can be copied and distributed freely for
 * any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
 * incorporated into commercial software with the expressed permission of
 * JASSPA.
 *  
 ****************************************************************************/

/*---	Include defintions */

#define	__SEARCHC			/* Define filename */

/*---	Include files */

#include "emain.h"
#include "esearch.h"

/*---	Local macro definitions */

#define notFound() (mlwrite(MWABORT|MWCLEXEC,notFoundStr))

/*---	External references */

/*---	Local type definitions */

/*---	Local variable definitions */

static char *notFoundStr = "[Not Found]";
static int  exactFlag = 1;              /* Use to cache EXACT mode state */
char srchPat [NPAT]="";                 /* Search pattern array */
char srchRPat[NPAT]="";                 /* Reversed pattern array */

/*
 * boundry -- Return information depending on whether we may search no
 *	further.  Beginning of file and end of file are the obvious
 *	cases, but we may want to add further optional boundry restrictions
 *	in future, a la VMS EDT.  At the moment, just return TRUE or
 *	FALSE depending on if a boundry is hit (ouch).
 */
#define boundry(curline,curoff,dir)                                          \
(((dir) == FORWARD) ?                                                        \
 (curline == curbp->b_linep) :                                               \
 ((curoff == 0) && (lback(curline) == curbp->b_linep)))


#if	MAGIC

static int magical = FALSE;

#ifdef _GNU_REGEX

struct re_pattern_buffer mereBuffer={0} ;
struct re_registers mereRegs={0} ;
char         mereLastPat[NPAT]="";  /* last pattern array - reset by set-char-mask */
static char  mereTranslate[256] ;
int          mereNumRegs=0 ;
static int   mereNumNewl=0 ;
static int   mereNewlStkSz=0 ;
static int  *mereNewlStk=NULL ;
static int   mereNewlBufSz=0 ;
static char *mereNewlBuf=NULL ;

#else
/*
 * The variable magical determines if there are actual metacharacters in the
 * string - if not, then we don't have to use the slower MAGIC mode search
 * functions.
 * 
 * The variable mclen holds the length of the matched string - used by the
 * replace functions.
 * 
 * The arrays srchMcPat and srchMcRPat hold the MC and reversed MC search structures.
 */
static int	mclen = 0;
MC              srchMcPat[NPAT];
MC              srchMcRPat[NPAT];

/* Auto repeat counters. */

A_RPT srchAutoRpt;	/* Auto repeat store */

#endif

static int      reportErrors = TRUE;

#endif

/*
 * nextch -- retrieve the next/previous character in the buffer,
 *	and advance/retreat the point.
 *	The order in which this is done is significant, and depends
 *	upon the direction of the search.  Forward searches look at
 *	the current character and move, reverse searches move and
 *	look at the character.
 */
#if 0
static int
nextch(LINE **pcurline, int *pcuroff, int dir)
{
    register LINE	*curline;
    register int	curoff;
    register int	c;
    
    curline = *pcurline;
    curoff = *pcuroff;

    if (dir == FORWARD)
    {
        if (curoff == llength(curline))		/* if at EOL */
        {
            if(curline == curbp->b_linep)
                c = '\0';			/* and return a <NL> */
            else
            {
                curline = lforw(curline);	/* skip to next line */
                *pcurline = curline;
                curoff = 0;
                c = meNLCHAR;			/* and return a <NL> */
            }
        }
        else
            c = lgetc(curline, curoff++);	/* get the char */
    }
    else			/* Reverse.*/
    {
        if (curoff == 0)
        {
            curline = lback(curline);
            if(curline == curbp->b_linep)
                c = '\0';			/* and return a <NL> */
            else
            {
                *pcurline = curline;
                curoff = llength(curline);
                c = meNLCHAR;
            }
        }
        else
            c = lgetc(curline, --curoff);
        
    }
    *pcuroff = curoff;
    return (c);
}
#else

#define nextch(cc,curline,curoff,dir)                                       \
    if (dir == FORWARD)                                                      \
    {                                                                        \
        if (curoff == llength(curline))		/* if at EOL */              \
        {                                                                    \
            if(curline == curbp->b_linep)                                    \
                cc = '\0';			/* END return a NULL */      \
            else                                                             \
            {                                                                \
                curline = lforw(curline) ;	/* skip to next line */      \
                curoff = 0;                                                  \
                cc = meNLCHAR;			/* and return a <NL> */      \
            }                                                                \
        }                                                                    \
        else                                                                 \
            cc = lgetc(curline, curoff++) ;	/* get the char */           \
    }                                                                        \
    else			/* Reverse.*/                                \
    {                                                                        \
        if (curoff == 0)                                                     \
        {                                                                    \
            if((curline = lback(curline)) == curbp->b_linep)                 \
                cc = '\0';                                                   \
            else                                                             \
            {                                                                \
                curoff = llength(curline);                                   \
                cc = meNLCHAR;                                               \
            }                                                                \
        }                                                                    \
        else                                                                 \
            cc = lgetc(curline, --curoff);                                   \
    }

#endif

#if	MAGIC

#ifdef _GNU_REGEX

int
mere_compile_pattern(char *apat)
{
    const char *err ;
    char cc, *ss ;
    int ii, ll ;
    
    if(!mereBuffer.allocated)
    {
        re_syntax_options = RE_CHAR_CLASSES ;
        mereBuffer.allocated = 8;
        mereBuffer.buffer = malloc(mereBuffer.allocated);
        mereBuffer.fastmap = malloc(256);
        mereBuffer.regs_allocated = REGS_UNALLOCATED;
    }
    else if(!strcmp(apat,mereLastPat))
        return TRUE ;

    /* set-up the translation table */
    if(exactFlag)
    {
        /* Map uppercase characters to corresponding lowercase ones.  */
        for (ii = 1; ii < 256 ; ii++)
            mereTranslate[ii] = toLower(ii) ;
        mereBuffer.translate = mereTranslate ;
    }
    else
        mereBuffer.translate = NULL ;
    
    mereNumNewl = 0 ; 
    ss = apat ;
    ll = 0 ;
    while((cc=*ss++) != '\0')
    {
        if(cc == meNLCHAR)
            mereNumNewl++ ;
        ll++ ;
    }
    if((err=re_compile_pattern (apat,ll,&mereBuffer)) != NULL)
    {
        if(reportErrors)
            mlwrite(MWABORT,"[Regex Error: %s]",err) ;
        return FALSE ;
    }
    if(mereNumNewl > mereNewlStkSz)
    {
        if((mereNewlStk = meRealloc(mereNewlStk,mereNumNewl*sizeof(int))) == NULL)
        {
            mereNewlStkSz = 0 ;
            return FALSE ;
        }
        mereNewlStkSz = mereNumNewl ;
    }
    strncpy(mereLastPat,apat,NPAT-1) ;
    mereLastPat[NPAT-1] = '\0' ;
    return TRUE ;
}

int
mere_scanner(int direct, int beg_or_end, int *n, SCANNERPOS *sp)
{
    register LINE *lp=curwp->w_dotp, *nlp ;
    register long  lnno=curwp->line_no, nlnno ;
    register int   ii=curwp->w_doto, jj=llength(lp), kk, count=*n ;
    
    /* Moved the TTbreakTest() to the start of the loop as some of the
     * main problems as caused by search string with no length, such as ^,
     * or [ ]*$ on a line with no end spaces. In these cases is the user
     * is trying to replace it then they couldn't interupt as this
     * immediately returned success, hence the move!
     */
    if(TTbreakTest(1))
    {
        ctrlg(FALSE,1) ;
        return FALSE ;
    }
    /* treat the first line as a special case */
    if(direct == FORWARD)
    {
        if(lp == curbp->b_linep)
            return FALSE ;
        ii = re_search_2(&mereBuffer,NULL,0,(char *) ltext(lp),jj,
                         ii,jj-ii,&mereRegs,jj) ;
    }
    else
    {
        ii = re_search_2(&mereBuffer,NULL,0,(char *) ltext(lp),jj,
                         ii-1,0-ii,&mereRegs,jj) ;
    }
    if(ii >= 0)
    {
        kk = mereRegs.end[0] ;
        nlp = lp ;
        nlnno = lnno ;
    }
    else if(mereNumNewl == 0)
    {
        if(direct == FORWARD)
        {
            /* On entry, if count=0 then no line limit, therefore test if (--count == 0) as count == -1 */ 
            while((--count != 0) && ((lp=lforw(lp)) != curbp->b_linep))
            {
                if(TTbreakTest(1))
                {
                    ctrlg(FALSE,1) ;
                    return FALSE ;
                }
                lnno++ ;
                ii = llength(lp) ;
                if((ii=re_search_2(&mereBuffer,NULL,0,(char *) ltext(lp),ii,0,ii,&mereRegs,ii)) >= 0)
                    break ;
            }
        }
        else
        {
            while((--count != 0) && ((lp=lback(lp)) != curbp->b_linep))
            {
                if(TTbreakTest(1))
                {
                    ctrlg(FALSE,1) ;
                    return FALSE ;
                }
                lnno-- ;
                ii = llength(lp) ;
                if((ii=re_search_2(&mereBuffer,NULL,0,(char *) ltext(lp),ii,ii-1,-ii,&mereRegs,ii)) >= 0)
                    break ;
            }
        }
        if(ii < 0)
            return FALSE ;
        kk = mereRegs.end[0] ;
        nlp = lp ;
        nlnno = lnno ;
    }
    else if(--count == 0)
        return FALSE ;
    else
    {
        /* treat the 1st-2nd line as a special case */
        ii = curwp->w_doto ;
        if(direct == FORWARD)
        {
            nlp = lforw(lp) ;
            kk = llength(nlp) ;
            lputc(lp,jj,meNLCHAR) ;
            ii = re_search_2(&mereBuffer,(char *) ltext(lp),jj+1,(char *) ltext(nlp),kk,
                             ii,jj+1+kk-ii,&mereRegs,jj+1+kk) ;
            lputc(lp,jj,0) ;
        }
        else
        {
            nlp = lp ;
            lp = lback(lp) ;
            if(lp == curbp->b_linep)
                return FALSE ;
            lnno-- ;
            kk = jj ;
            jj = llength(lp) ;
            lputc(lp,jj,meNLCHAR) ;
            ii = re_search_2(&mereBuffer,(char *) ltext(lp),jj+1,(char *) ltext(nlp),kk,
                             jj+1+ii-1,0-(jj+1+ii),&mereRegs,jj+1+kk) ;
            lputc(lp,jj,0) ;
        }
        if(ii < 0)
        {
            if(mereNumNewl == 1)
            {
                if(direct == FORWARD)
                {
                    /* On entry, if count=0 then no line limit, therefore test if (--count == 0) as count == -1 */ 
                    while((--count != 0) && ((lp=nlp) != curbp->b_linep))
                    {
                        if(TTbreakTest(1))
                        {
                            ctrlg(FALSE,1) ;
                            return FALSE ;
                        }
                        lnno++ ;
                        jj = kk ;
                        nlp = lforw(nlp) ;
                        kk = llength(nlp) ;
                        lputc(lp,jj,meNLCHAR) ;
                        ii = re_search_2(&mereBuffer,(char *) ltext(lp),jj+1,(char *) ltext(nlp),kk,
                                         0,jj+1+kk,&mereRegs,jj+1+kk) ;
                        lputc(lp,jj,0) ;
                        if(ii >= 0)
                            break ;
                    }
                }
                else
                {
                    /* On entry, if count=0 then no line limit, therefore test if (--count == 0) as count == -1 */ 
                    while((--count != 0) && ((nlp=lp),((lp=lback(lp)) != curbp->b_linep)))
                    {
                        if(TTbreakTest(1))
                        {
                            ctrlg(FALSE,1) ;
                            return FALSE ;
                        }
                        lnno-- ;
                        kk = jj ;
                        jj = llength(lp) ;
                        lputc(lp,jj,meNLCHAR) ;
                        ii = re_search_2(&mereBuffer,(char *) ltext(lp),jj+1,(char *) ltext(nlp),kk,
                                         jj+1+kk,0-(jj+1+kk),&mereRegs,jj+1+kk) ;
                        lputc(lp,jj,0) ;
                        if(ii >= 0)
                            break ;
                    }
                }
                
            }
            else
            {
                /* generic multi-line case */
                int noln=2, tnoln=mereNumNewl+1, offs ;
                
                if(direct != FORWARD)
                    offs = curwp->w_doto ;
                else
                    offs = llength(curwp->w_dotp)-curwp->w_doto ;
                while(--count != 0)
                {
                    if(TTbreakTest(1))
                    {
                        ctrlg(FALSE,1) ;
                        return FALSE ;
                    }
                    if(direct == FORWARD)
                    {
                        if(nlp == curbp->b_linep)
                            break ;
                        if(noln == tnoln)
                        {
                            lnno++ ;
                            lp = lforw(lp) ;
                        }
                    }
                    else
                    {
                        if((lp=lback(lp)) == curbp->b_linep)
                            break ;
                        lnno-- ;
                    }
                    if(noln == tnoln)
                        offs = 0 ;
                    else
                        noln++ ;
                    ii = noln ;
                    jj = ii ;
                    nlp = lp ;
                    while(--ii >= 0)
                    {
                        jj += llength(nlp) ;
                        nlp = lforw(nlp) ;
                    }
                    if(jj > mereNewlBufSz)
                    {
                        if((mereNewlBuf = meRealloc(mereNewlBuf,jj+128)) == NULL)
                        {
                            mereNewlBufSz = 0 ;
                            return FALSE ;
                        }
                        mereNewlBufSz = jj+128 ;
                    }
                    ii = noln ;
                    jj = 0 ;
                    nlp = lp ;
                    for(;;)
                    {
                        kk = llength(nlp) ;
                        memcpy(mereNewlBuf+jj,ltext(nlp),kk) ;
                        jj += kk ;
                        /* we don't want the last \n or move nlp to past the
                         * last line so drop out here */
                        if(--ii == 0)
                            break ;
                        mereNewlBuf[jj++] = meNLCHAR ;
                        nlp = lforw(nlp) ;
                    }
                    if(direct == FORWARD)
                        ii = re_search_2(&mereBuffer,NULL,0,mereNewlBuf,jj,
                                         offs,jj,&mereRegs,jj) ;
                    else
                        ii = re_search_2(&mereBuffer,NULL,0,mereNewlBuf,jj,
                                         jj-offs-1,0-(jj-offs),&mereRegs,jj) ;
                    if(ii >= 0)
                        break ;
                }
            }
            if(ii < 0)
                return FALSE ;
        }
        kk = mereRegs.end[0] ;
        while(ii > llength(lp))
        {
            jj = llength(lp)+1 ;
            ii -= jj ;
            kk -= jj ;
            lnno++ ;
            lp = lforw(lp) ;
        }
        nlp = lp ;
        nlnno = lnno ;
        while(kk > llength(nlp))
        {
            kk -= llength(nlp)+1 ;
            nlnno++ ;
            nlp = lforw(nlp) ;
        }
    }
    if(beg_or_end == PTEND)
    {
        curwp->w_dotp = nlp ;
        curwp->line_no = nlnno ;
        curwp->w_doto = kk ;
    }
    else
    {
        curwp->w_dotp = lp ;
        curwp->line_no = lnno ;
        curwp->w_doto = ii ;
    }
    if (sp != NULL)
    {
        sp->startline = lp ;
        sp->startline_no = lnno ;
        sp->startoff = ii ;
        sp->endline = nlp ;
        sp->endline_no = nlnno ;
        sp->endoffset = kk ;
    }
    setShowRegion(curbp,lnno,ii,nlnno,kk) ;
    curwp->w_flag |= WFMOVEL|WFSELHIL ;
    /* change the regs \( \) start and end points so the start is at zero */
    ii = mereRegs.start[0] ;
    jj = mereNumRegs = mereBuffer.re_nsub + 1 ;
    while(--jj >= 0)
    {
        if(mereRegs.start[jj] >= 0)
        {
            mereRegs.start[jj] -= ii ;
            mereRegs.end[jj]   -= ii ;
        }
    }
    return TRUE ;
}
#else
/*
 * open_arpt -- Set up for an open auto repeat search + replace character.
 */
 
static int
open_arpt (register MC *mc_ptr)
/* mc_ptr - Pointer to the Mata pattern */
{
    /*---	Check for the repeat count being too high. */

    if (srchAutoRpt.count >= 9)	/* Too high ?? */
        return (FALSE);		/* Yes - return error */
    
    /*---	Set auto repeat on and allocate the next auto pattern number */

    srchAutoRpt.do_repeat = TRUE;		/* Auto repeat on */
    mc_ptr->mc_type = SRCL;			/* Set up Meta-pattern type */
    mc_ptr->u.lchar = srchAutoRpt.count;	/* Allocate repeat No */
    srchAutoRpt.stack [srchAutoRpt.hi_water++] =
              srchAutoRpt.count++;		/* Stack close No */
    return (TRUE);				/* Valid exit */
}	/* End of "open_arpt" */

static int
close_arpt (register MC	*mc_ptr)
/* mc_ptr - Pointer to the Mata pattern */
{
    /*---	Check for the repeat stack being too low ie. No open. */
    
    if (srchAutoRpt.hi_water <= 0)	/* Too low ?? */
        return (FALSE);		/* Yes - return error */
    
    /*---	Set auto repeat on and decrement the stack. */
    
    srchAutoRpt.do_repeat = TRUE;		/* Auto repeat on */
    mc_ptr->mc_type = ERCL;			/* Set up Meta-pattern type */
    mc_ptr->u.lchar = 			/* Allocate repeat No */
              srchAutoRpt.stack [--srchAutoRpt.hi_water];	/* From the stack */
    return (TRUE);				/* Valid exit */
}	/* End of "open_arpt" */

static void
make_arpt(void)
{
    int	i;		/* Local loop counter */
    
    srchAutoRpt.do_repeat = FALSE;		/* No repeat yet. */
    srchAutoRpt.hi_water = 0;		/* Clear stack */
    srchAutoRpt.count = 0;			/* No repeats */
    
    for (i = 0; i < 9; i++)
        srchAutoRpt.end_pos [i] = srchAutoRpt.start_pos [i] = 0;

}	/* End of "make_arpt" */

#define check_arpt() ((srchAutoRpt.hi_water == 0) ? TRUE : FALSE)

static void
configure_arpt (int direct, int slen)
{
    register int	i;

    /* --- If we are scanning in the foward direction, then everything is ok. No
       configuration to perform. */
	
    if (direct == FORWARD ||		/* Righ way ?? */
        srchAutoRpt.do_repeat == FALSE)	/* No repeat strings ?? */
        return;				/* Yep !! - Exit no config */
    
    /* --- If we are scanning backwards, then translate the offsets in the repeat
       table to foward offsets. */
	
    for (i = srchAutoRpt.count; --i >= 0; /* Null */)
    {
        srchAutoRpt.start_pos [i] = slen - srchAutoRpt.start_pos [i] - 1;
        srchAutoRpt.end_pos [i] = slen - srchAutoRpt.end_pos [i] - 1;
    }
}


/*
 * freebits -- Free up any CCL bitmaps.
 */

static void
freebits(void)
{
    register MC	*mcptr;

    mcptr = &srchMcPat[0];

    while (mcptr->mc_type != MCNIL)
    {
        if (((mcptr->mc_type & MASKCL) == CCL) ||
            ((mcptr->mc_type & MASKCL) == NCCL))
            meFree(mcptr->u.cclmap);
        mcptr++;
    }
}


/*
 * clearbits -- Allocate and zero out a CCL bitmap.
 */


static	meBITMAP
clearbits(void)
{
    meBITMAP		cclstart, cclmap;
    register int	j;
    
    if ((cclmap = cclstart = (meBITMAP) meMalloc(meHIBYTE)) != NULL)
        for (j = 0; j < meHIBYTE; j++)
            *cclmap++ = 0;

    return (cclstart);
}

/*
 * setbit -- Set a bit (ON only) in the bitmap.
 */

#if 0

static void
setbit(int bc, meBITMAP cclmap)
{
    bc &= 0xff;
    *(cclmap + ((bc >> 3) & 0x00ff)) |= meBIT(bc & 7);
}
#else

#define setbit(bc,cclmap) \
(((bc) &= 0xff),(*((cclmap) + (((bc) >> 3) & 0x00ff)) |= meBIT((bc) & 7)))

#endif

/*
 * cclmake -- create the bitmap for the character class.
 *	ppatptr is left pointing to the end-of-character-class character,
 *	so that a loop may automatically increment with safety.
 */

static int
cclmake(unsigned char **ppatptr, MC *mcptr)
{
    meBITMAP		bmap;
    register unsigned char *patptr;
    register int	pchr, ochr;
    
    if((bmap = clearbits()) == NULL)
        return FALSE;

    mcptr->u.cclmap = bmap;
    patptr = *ppatptr;
    
    /*
     * Test the initial character(s) in ccl for special cases - negate ccl, or
     * an end ccl character as a first character. Anything else gets set in
     * the bitmap.
     */
    if (*++patptr == MC_NCCL)
    {
        patptr++;
        mcptr->mc_type = NCCL;
    }
    else
        mcptr->mc_type = CCL;
    
    if ((ochr = *patptr) == MC_ECCL)
        return ((reportErrors) ?
                mlwrite(MWABORT,"[No characters in character class]") :
                FALSE);
    
    /*
     * Handle numeric expressions in the escape sequence.
     */
    
    if (ochr == MC_ESC)
    {
        if ((ochr = *++patptr) == '0')
        {
            if (isHexDigit (patptr[1]) && isHexDigit (patptr[2]))
                ochr = ((hexToNum(patptr[1]) << 4) |
                        hexToNum(patptr[2])) & 0x00ff;
            else {
                meFree (bmap);
                return ((reportErrors) ?
                        mlwrite (MWABORT, "Numeric escape sequence expected") :
                        FALSE);
            }
            patptr += 2;
        }
    }
    
    setbit(ochr, bmap);
    patptr++;
    
    while (ochr != '\0' && (pchr = *patptr) != MC_ECCL)
    {
        switch (pchr)
        {
            /* Range character loses its meaning if it is the last character
             * in the class.
             */
        case MC_RCCL:
            if (*(patptr + 1) == MC_ECCL)
            {
                setbit(pchr, bmap);
                break;
            }
            
            if ((pchr = *++patptr) == MC_ESC)
            {
                if ((pchr = *++patptr) == '0') {
                    if (isHexDigit (patptr[1]) &&
                        isHexDigit (patptr[2]))
                        pchr = ((hexToNum(patptr[1]) << 4) |
                                hexToNum(patptr[2])) & 0x00ff;
                    else {
                        meFree (bmap);
                        return ((reportErrors) ?
                                mlwrite (MWABORT, "Numeric escape sequence expected") :
                                FALSE);
                    }
                    patptr += 2;
                }
            }
            if (pchr < ochr)
                do {
                    setbit(pchr, bmap);
                } while (++pchr <= ochr);
            else
                while (++ochr <= pchr)
                    setbit(ochr, bmap);
            break;

            /* Note: no break between case MC_ESC and the default.
             */
        case MC_ESC:
            if ((pchr = *++patptr) == '0') {
                if (isHexDigit (patptr[1]) &&
                    isHexDigit (patptr[2]))
                    pchr = ((hexToNum(patptr[1]) << 4) |
                            hexToNum(patptr[2])) & 0x00ff;
                else {
                    meFree (bmap);
                    return ((reportErrors) ?
                            mlwrite (MWABORT,
                                     "Numeric escape sequence expected") :
                            FALSE);
                }
                patptr += 2;
            }
        default:
            setbit(pchr, bmap);
            break;
        }
        patptr++;
        ochr = pchr;
    }
    
    *ppatptr = patptr;
    
    if (ochr == '\0')
    {
        meFree(bmap);
        return ((reportErrors) ?
                mlwrite(MWABORT,"Character class not ended") :
                FALSE);
    }
    return TRUE;
}


/*
 * mcstr -- Set up the 'magic' array.  The closure symbol is taken as
 *	a literal character when (1) it is the first character in the
 *	pattern, and (2) when preceded by a symbol that does not allow
 *	closure, such as a newline, beginning of line symbol, or another
 *	closure symbol.
 *
 *	Coding comment (jmg):  yes, i know i have gotos that are, strictly
 *	speaking, unnecessary.  But right now we are so cramped for
 *	code space that i will grab what i can in order to remain
 *	within the 64K limit.  C compilers actually do very little
 *	in the way of optimizing - they expect you to do that.
 * 
 *      JDG - Added 'ppat' argument to mcstr. This is a pointer to a pattern
 *      buffer. If specified as Non-NULL then the literal string is placed
 *      in this buffer. The resultant pattern may be used if the magical 
 *      status is FALSE, the string represents the literal search string
 *      with the magic literal characters removed (e.g. \0FE, \$ etc). 
 *      The generated ppat (processed pat !!) should be used by the caller
 *      in preference to 'pat' if scanner is invoked.
 *      Jon - 25/09/95
 */

static int
mcstr(unsigned char *pat, unsigned char *ppat)
{
    MC	*mcptr, *rtpcm;
    unsigned char *patptr;
    int c;
    int	mj;
    int	pchr;
    int	status = TRUE;
    int	does_closure = FALSE;
    
    /* If we had metacharacters in the MC array previously, free up any
     * bitmaps that may have been allocated. Initialise the auto repeater.
     */
    if (magical)
        freebits();
    make_arpt ();	/* Make auto repeat */

    magical = FALSE;
    mj = 0;
    mcptr = &srchMcPat[0];
    patptr = pat;
    
    while (((pchr = *patptr) != 0) && status)
    {
        switch (pchr)
        {
        case MC_CCL:
            status = cclmake(&patptr, mcptr);
            magical = TRUE;
            does_closure = TRUE;
            break;
        case MC_BOL:
            if (mj != 0)
                goto litcase;
            
            mcptr->mc_type = BOL;
            magical = TRUE;
            does_closure = FALSE;
            break;
        case MC_EOL:
            if (*(patptr + 1) != '\0')
                goto litcase;
            
            mcptr->mc_type = EOL;
            magical = TRUE;
            does_closure = FALSE;
            break;
        case MC_ANY:
            mcptr->mc_type = ANY;
            magical = TRUE;
            does_closure = TRUE;
            break;
        case MC_S_CLOSURE:
        case MC_O_CLOSURE:
        case MC_CLOSURE:
            /* Does the closure symbol mean closure here? If so, back up to
             * the previous element and indicate it is enclosed.
             */
            if (!does_closure)
                goto litcase;
            mj--;
            (--mcptr)->mc_type |= ((pchr == MC_CLOSURE) ? CLOSURE :
                                   ((pchr == MC_O_CLOSURE) ? O_CLOSURE :
                                   S_CLOSURE));
            magical = TRUE;
            does_closure = FALSE;
            break;
            /*
             * Note: no break between MC_ESC case and the default.
            */
        case MC_ESC:
            if ((c = *(patptr + 1)) != '\0')
            {
                if (c == MC_SRCL) {
                    if (open_arpt (mcptr) == FALSE) 
                        status = FALSE;
                    else
                    {
                        patptr++;
                        does_closure = FALSE;
                        magical = TRUE;
                    }
                } 
                else if (c == MC_ERCL)
                {
                    if (close_arpt (mcptr) == FALSE)
                        status = FALSE;
                    else
                    {
                        patptr++;
                        does_closure = FALSE;
                        magical = TRUE;
                    }
                }
                else if (c == '0')
                {
                    if (isHexDigit (patptr[2]) &&
                        isHexDigit (patptr[3]))
                    {
                        pchr = ((hexToNum(patptr[2]) << 4) |
                                hexToNum(patptr[3])) & 0x00ff;
                        patptr += 3;
                        goto litcase;
                    }
                    else
                        status = FALSE;
                }
                else
                {
                    pchr = c;
                    patptr++;
                    goto litcase;
                }
            }
            else
                status = FALSE;
            break;
            
        default:
litcase:
            mcptr->mc_type = LITCHAR;
            mcptr->u.lchar = pchr;
            does_closure = (pchr != meNLCHAR);
            /*
             * For the literal case. If we have not detected magic mode yet
             * and ppat is specified then build up the literal processed
             * pattern (ppat).
             */
            
            if ((ppat != NULL) && (magical == FALSE))
                *ppat++ = pchr;
            
            break;
        }		/* End of switch.*/
        mcptr++;
        patptr++;
        mj++;
    }		/* End of while.*/

    if (ppat != NULL)                   /* Terminate the processed pattern */
        *ppat = '\0';
    mcptr->mc_type = MCNIL;             /* Close off the meta-string. */

    
    /* Test for replace closures being the same. */
    
    if (check_arpt() == FALSE)
        status = FALSE;
    
    /* Set up the reverse array, if the status is good. Please note the
     * structure assignment - your compiler may not like that. If the status
     * is not good, nil out the meta-pattern. The only way the status would be
     * bad is from the cclmake() routine, and the bitmap for that member is
     * guarenteed to be freed. So we stomp a MCNIL value there, call
     * freebits() to free any other bitmaps, and set the zeroth array to
     * MCNIL.
     */
    if (status)
    {
        rtpcm = &srchMcRPat[0];
        while (--mj >= 0)
            *rtpcm++ = *--mcptr;
        rtpcm->mc_type = MCNIL;
    }
    else
    {
        (--mcptr)->mc_type = MCNIL;
        freebits();
        srchMcPat[0].mc_type = srchMcRPat[0].mc_type = MCNIL;
    }
    
    return(status);
}

#endif
#endif


#if 0
/* Local copy eq () for search */

static int
cEq (register int bc, register int pc)
{
    
    if (bc == pc)                       /* Quick test first - think positive */
        return (TRUE);                  /* Phew - gambit paid off */
    if (!exactFlag)                     /* Get exact out the way quick */
        return (FALSE);                 /* Failed. !! */
    
    /*
     * We are not in exact mode - if we are dealing with alphabetic then
     * convert to the same case and compare. 
     */
    
    bc = toLower(bc) ;
    pc = toLower(pc) ;
    return (bc == pc);
}

#define cNotEq(a,b) (!cEq(a,b))

#else

#define cNotEq(a,b)                                                          \
(((a) != (b)) &&                                                             \
 (!exactFlag || (toggleCase(a) != (b))))

#define cEq(a,b)                                                             \
(((a) == (b)) ||                                                             \
 (exactFlag && (toggleCase(a) == (b))))

#endif

#if	MAGIC
#ifndef _GNU_REGEX
#if 0
/*
 * biteq -- is the character in the bitmap?
 */
static	int
bitEq(int bc, meBITMAP cclmap)
{
    if (bc >= meHICHAR)
        return FALSE;
    
    return((*(cclmap + ((bc >> 3) & 0xff)) & meBIT(bc & 7)) ? TRUE: FALSE );
}
#else

#define bitEq(bc,cclmap)                                                     \
(((bc) >= meHICHAR) ? FALSE:                                                 \
 ((*((cclmap) + (((bc) >> 3) & 0xff)) & meBIT((bc) & 7)) ? TRUE : FALSE))

#endif

/*
 * mceq -- meta-character equality with a character. In Kernighan & Plauger's
 * Software Tools, this is the function omatch(), but i felt there were too
 * many functions with the 'match' name already.
 */

static int
mcEq(int bc, MC *mt)
{
    register int result;
    
    switch(mt->mc_type & MASKCL)
    {
    case LITCHAR:
        result = cEq(bc, mt->u.lchar);
        break;
        
    case ANY:
        result = (bc != meNLCHAR);
        break;
        
    case CCL:
        if (((result = bitEq(bc, mt->u.cclmap)) == 0) &&
            (exactFlag && (isAlpha(bc))))
            result = bitEq(toggleCase(bc), mt->u.cclmap);
        break;
        
    case NCCL:
        result = !bitEq(bc, mt->u.cclmap);
        if (exactFlag && (isAlpha(bc)))
            result &= !bitEq(toggleCase(bc), mt->u.cclmap);
        break;
        
    case SRCL:
    case ERCL:	
        result = TRUE;		/* Dummy processing */
        break;
    default:
        result = mlwrite(MWABORT,"mcEq: what is %d?", mt->mc_type);
        break;
        
    }	/* End of switch.*/

    return (result);
}

/*
 * amatch -- Search for a meta-pattern in either direction. Based on the
 * recursive routine amatch() (for "anchored match") in Kernighan & Plauger's
 * "Software Tools".
 */
static int
amatch(register MC *mcptr, int direct, LINE **pcwline, int *pcwoff)
/* mcptr   - string to scan for */
/* direct  - which way to go.*/
/* pcwline - current line during scan */
/* pcwoff  - position within current line */
{
    register int c;			/* character at current position */
    LINE *curline;			/* current line during scan */
    int curoff;				/* position within current line */
    
    /* Set up local scan pointers to ".", and get the current character. Then
     * loop around the pattern pointer until success or failure.
     */
    curline = *pcwline;
    curoff = *pcwoff;
    /* Can have braces before the start of line character. Check these first,
     * otherwise do inside following loop.
     * 
     * The beginning-of-line and end-of-line metacharacters do not compare
     * against characters, they compare against positions.
     * 
     * BOL is guaranteed to be at the start of the pattern for forward
     * searches, and at the end of the pattern for reverse searches. The
     * reverse is true for EOL. So, for a start, we check for them on entry.
     */
    
    do
        switch (mcptr->mc_type)
        {
        case SRCL :		/* Open ?? */
            srchAutoRpt.start_pos [mcptr++->u.lchar] = mclen;
            continue;
            
        case ERCL :		/* Close ?? */
            srchAutoRpt.end_pos [mcptr++->u.lchar] = mclen;
            continue;
            
        case BOL :		/* Beginning of line ?? */
            if((curoff != 0) || (curline == curbp->b_linep))
                return FALSE;
            mcptr++;
            continue;
        case EOL :		/* End of line ?? */
            if((curoff != llength(curline)) || (curline == curbp->b_linep))
                return FALSE;
            mcptr++;
            continue;
        default:		/* None of above */
            break;
            
        }	/* End of 'switch' */
    while (0);	/* End of 'do-while' */
    
    /*---	Match the pattern with the rest of the string. */
    
    while (mcptr->mc_type != MCNIL)
    {
        nextch(c,curline,curoff,direct)

        if (mcptr->mc_type & CLOSURE_MASK)
        {
            int mclenCount = 0;
            
            if (mcptr->mc_type & S_CLOSURE)
            {
                if (mcEq(c, mcptr))
                {
                    nextch(c,curline,curoff,direct)
                    mclen++;
                    mclenCount++;
                }
            }
            else
            {
                /*---	If One close, match the 1 first */
            
                if (mcptr->mc_type & O_CLOSURE)
                {
                    if (mcEq(c, mcptr))
                    {
                        nextch(c,curline,curoff,direct)
                        mclen++;
                    }
                    else
                        return (FALSE);
                }
                else
                {
                    /* Must be a zero closure. We increment here to allow 
                     * a zero occurence to occur and continue correctly.
                     * (I think - Jon 10/06/98 - Leave this in until established
                     * as working) */
                    mclenCount++;           /* Allow to close to zero !! */
                }
            
                /* Try to match as many characters as possible against the
                 * current meta-character. A newline never matches a closure.
                 */
                if ((~CLOSURE_MASK & mcptr->mc_type) == ANY)
                {
                    while (c != meNLCHAR)
                    {
                        nextch(c,curline,curoff,direct)
                        mclenCount++;
                        mclen++;
                    }
                }
                else
                {
                    while (c != meNLCHAR && mcEq(c, mcptr))
                    {				
                        nextch(c,curline,curoff,direct)
                        mclenCount++;
                        mclen++;
                    }
                }
            }
            /* We are now at the character that made us fail. Try to match the
             * rest of the pattern. Shrink the closure by one for each
             * failure. Since closure matches *zero* or more occurences of a
             * pattern, a match may start even if the previous loop matched no
             * characters.
             */
            
            mcptr++;
            for (;;)
            {
                int omclen=mclen ;
                nextch(c,curline,curoff,(direct ^ REVERSE))
                if (amatch(mcptr, direct, &curline, &curoff))
                    goto success;
                mclen = omclen - 1 ;
                if (--mclenCount <= 0)
                    return FALSE;
            }
        }
        else			/* Not closure.*/
        {
            /* The only way we'd get a BOL metacharacter at this point is at
             * the end of the reversed pattern. The only way we'd get an EOL
             * metacharacter here is at the end of a regular pattern. So if we
             * match one or the other, and are at the appropriate position, we
             * are guaranteed success (since the next pattern character has to
             * be MCNIL). Before we report success, however, we back up by one
             * character, so as to leave the cursor in the correct position.
             * For example, a search for ")$" will leave the cursor at the end
             * of the line, while a search for ")<NL>" will leave the cursor
             * at the beginning of the next line. This follows the notion that
             * the meta-character '$' (and likewise '^') match positions, not
             * characters.
             */
	
            switch (mcptr->mc_type)
            {
            case BOL:
                if(curoff == llength(curline))
                {
                    /* Must now more forward a char */
                    curline = lforw(curline) ;
                    curoff = 0 ;
                    goto success;
                }
                else
                    return FALSE;
            case EOL:
                /* Check c is a \n, not that the offset is 0 as this catches
                 * the b_linep problem
                 */
                if (c == meNLCHAR)
                {
                    /* Must move backwards a char */
                    curline = lback(curline) ;
                    curoff = llength(curline) ;
                    goto success;
                }
                else
                    return FALSE;
                
            case SRCL:
                
                /*---	Save character position */ 
                
                srchAutoRpt.start_pos [mcptr->u.lchar] = mclen;
                
                /*---	Back up character & reduce character count */
                nextch(c,curline,curoff,(direct ^ REVERSE))
                mclen--;	/* Reduce charater count */
                break;					
                
            case ERCL:
                
                /*---	Save charater position */
                
                srchAutoRpt.end_pos [mcptr->u.lchar] = mclen;
                
                /*---	Back up character & reduce character count */
                
                nextch(c,curline,curoff,(direct ^ REVERSE))
                mclen--;	/* Reduce charater count */
                break;					
                
                /* Neither BOL nor EOL, so go through
                 * the meta-character equal function.
                 */
            default:
                if (!mcEq(c, mcptr))
                    return FALSE;
            }
        }
        
        /* Increment the length counter and
         * advance the pattern pointer.
         */
        mclen++;
        mcptr++;
    }			/* End of mcptr loop.*/
    
    /* A SUCCESSFULL MATCH!!!
     * Reset the "." pointers.
     */
success:
    *pcwline = curline;
    *pcwoff  = curoff;
    
    return TRUE;
}


/*
 * mcscanner -- Search for a meta-pattern in either direction.
 */
static int	
mcscanner (MC *mcpatrn, int direct, int beg_or_end, int *count, SCANNERPOS *sp)
/* mcpatrn    - pointer into pattern */
/* direct     - which way to go.*/
/* beg_or_end - put point at beginning or end of pattern.*/
/* count      - Number of lines to process */
{
    register LINE *lastline ;		/* last line position during scan */
    register int   lastoff ;		/* position within last line */
    register long  lastlno ;
    LINE *curline ;			/* current line during scan */
    int   curoff ;			/* position within current line */
    
    /* If we are going in reverse, then the 'end' is actually the beginning of
     * the pattern. Toggle it.
     */
    beg_or_end ^= direct ;
    
    /* Setup local scan pointers to global ".". */
    curline = curwp->w_dotp ;
    curoff  = curwp->w_doto ;
    lastlno = curwp->line_no ;

    /* Quick check on boundary conditions - lets not waste time !!! */
    if (boundry(curline, curoff, direct))
        return (FALSE);
    
    /* Scan each character until we hit the head link record. */
    for (;;)
    {
        /* Moved the TTbreakTest() to the start of the loop as some of the
         * main problems as caused by search string with no length, such as ^,
         * or [ ]*$ on a line with no end spaces. In these cases is the user
         * is trying to replace it then they couldn't interupt as this
         * immediately returned success, hence the move!
         */
        if(TTbreakTest(1))
        {
            ctrlg(FALSE,1) ;
            return FALSE ;
        }
        /* Save the current position in case we need to restore it on a match,
         * and initialize mclen to zero in case we are doing a search for
         * replacement.
         */
        lastline = curline;
        lastoff = curoff;
        mclen = 0;
        
        if (amatch(mcpatrn, direct, &curline, &curoff))
        {
            register long  curlno ;
            /*
             * A SUCCESSFULL MATCH!!!
             * Configuture the repeat values and reset global "." pointers.
             * Fill in the scanner position structure if required.
             */
            curlno = lastlno ;
            if(lastline != curline)
            {
                register LINE *lp=lastline ;
                if(direct == FORWARD)
                {
                    do {
                        curlno++ ;
                        lp = lforw(lp) ;
                    } while(lp != curline) ;
                }
                else
                {
                    do {
                        curlno-- ;
                        lp = lback(lp) ;
                    } while(lp != curline) ;
                }
            }
            configure_arpt (direct, mclen);	
            
            if (sp != NULL)
            {
                sp->startline = lastline;
                sp->startoff = lastoff;
                sp->startline_no = lastlno ;
                
                sp->endline = curline;
                sp->endoffset = curoff;
                sp->endline_no = curlno ;
            }
            setShowRegion(curbp,lastlno,lastoff,curlno,curoff) ;
            if (beg_or_end == PTEND)	/* at end of string */
            {
                curwp->w_dotp = curline;
                curwp->w_doto = curoff;
                curwp->line_no = curlno ;
            }
            else
            {
                curwp->w_dotp = lastline;
                curwp->w_doto = lastoff;
                curwp->line_no = lastlno ;
            }
             /* flag that we have moved and got a region to hilight.*/
            curwp->w_flag |= WFMOVEL|WFSELHIL ;
            return TRUE;
        }
        
        /* Advance the cursor. */
        if (direct == FORWARD)
        {
            if((curoff == llength(curline)) ||
               (mcpatrn->mc_type == BOL))       /* if at EOL */
            {
                curline = lforw(curline) ;	/* skip to next line */
                if(curline == curbp->b_linep)
                    break ;
                curoff = 0;
                lastlno++ ;
            }
            else
                curoff++ ;
        }
        else			/* Reverse.*/
        {
            if((curoff == 0) || (mcpatrn->mc_type == EOL))       /* if at BOL */
            {
                curline = lback(curline) ;
                if(curline == curbp->b_linep)
                    break ;
                curoff = llength(curline);
                lastlno-- ;
            }
            else
                curoff-- ;
        }
        /* Quit when we have done our work !! */
        if ((curoff == 0) && (*count > 0))
            if (--(*count) == 0)
                break;
    }
    
    return FALSE;	/* We could not find a match.*/
}
#endif
#endif


/*
 * scanner -- Search for a pattern in either direction.
 */
int 
scanner(unsigned char *patrn, int direct, int beg_or_end, int *count,
        SCANNERPOS *sp)
{
    register unsigned char cc, aa, bb ;	/* character at current position */
    register unsigned char *patptr;	/* pointer into pattern */
    register unsigned char *curptr;	/* pointer into pattern */
    register LINE *curline;		/* current line during scan */
    register int curoff;		/* position within current line */
    register unsigned char *matchptr;	/* pointer into pattern */
    register LINE *matchline;		/* current line during matching */
    register int matchoff;		/* position in matching line */
    register long curlnno, matchlnno ;
    
    /* If we are going in reverse, then the 'end' is actually the beginning of
     * the pattern. Toggle it.
     */
    beg_or_end ^= direct;
    
    /* Setup local scan pointers to global ".". */
    curline = curwp->w_dotp;
    curlnno = curwp->line_no;
    if (direct == FORWARD)
        curptr = curline->l_text+curwp->w_doto ;
    else			/* Reverse.*/
        curoff = curwp->w_doto;
    
    /* Scan each character until we hit the head link record. */
/*    if (boundry(curline, curoff, direct))*/
/*        return FALSE ;*/
    
    aa = patrn[0] ;
    for (;;)
    {
        /* Get the character resolving newlines, and
         * test it against first char in pattern.
         */
        if (direct == FORWARD)
        {
            if((cc=*curptr++) == '\0')		/* if at EOL */
            {
                if(curline != curbp->b_linep)
                {
                    curline = lforw(curline) ;	/* skip to next line */
                    curlnno++ ;
                    curptr = curline->l_text ;
                    cc = meNLCHAR;		/* and return a <NL> */
                    if((*count > 0) && (--(*count) == 0) &&
                       ((patrn[0] != meNLCHAR) || (patrn[1] != '\0')))
                        /* must check that the pattern isn't "\n" */
                        break ;
                }
            }
        }
        else			/* Reverse.*/
        {
            if (curoff == 0)
            {
                if((curline = lback(curline)) == curbp->b_linep)
                    cc = '\0';
                else
                {
                    if((*count > 0) && (--(*count) == 0))
                        break ;
                    curlnno-- ;
                    curoff = llength(curline);
                    cc = meNLCHAR;
                }
            }
            else
                cc = lgetc(curline, --curoff);
        }
        if(cc == '\0')
            break ;
        
        if(cEq(cc,aa))	/* if we find it..*/
        {
            /* Setup match pointers.
            */
            matchline = curline ;
            matchlnno = curlnno ;
            matchptr = curptr ;
            matchoff = curoff ;
            patptr = &patrn[0];
            
            /* Scan through the pattern for a match.
            */
            while((bb = *++patptr) != '\0')
            {
                if (direct == FORWARD)
                {
                    if((cc=*matchptr++) == '\0')		/* if at EOL */
                    {
                        if(matchline != curbp->b_linep)
                        {
                            matchline = lforw(matchline) ;	/* skip to next line */
                            matchlnno++ ;
                            matchptr = matchline->l_text ;
                            cc = meNLCHAR;			/* and return a <NL> */
                        }
                    }
                }
                else			/* Reverse.*/
                {
                    if (matchoff == 0)
                    {
                        if((matchline = lback(matchline)) == curbp->b_linep)
                            cc = '\0';
                        else
                        {
                            matchlnno-- ;
                            matchoff = llength(matchline);
                            cc = meNLCHAR;
                        }
                    }
                    else
                        cc = lgetc(matchline, --matchoff);
                }
                if (cNotEq(cc,bb))
                    goto fail;
            }
            
            /*
             * A SUCCESSFULL MATCH!!!
             * reset the curwp "." pointers
             * Fill in the scanner position structure if required.
             */
            /* Save the current position in case we need to restore it on a
             * match.
             */
            if (direct == FORWARD)
            {
                if(curptr == curline->l_text)
                {
                    curline = lback(curline) ;
                    curlnno-- ;
                    curoff = llength(curline);
                }
                else
                    curoff = ((size_t) curptr) - ((size_t) curline->l_text) - 1 ;
                matchoff = ((size_t) matchptr) - ((size_t) matchline->l_text) ;
            }
            else			/* Reverse.*/
            {
                if(curoff == llength(curline))		/* if at EOL */
                {
                    curline = lforw(curline) ;	/* skip to next line */
                    curlnno++ ;
                    curoff = 0;
                }
                else
                    curoff++ ;	/* get the char */
            }
        
            
            if (sp != NULL)
            {
                sp->startline = curline;
                sp->startoff = curoff;
                sp->startline_no = curlnno ;
                sp->endline = matchline;
                sp->endoffset = matchoff;
                sp->endline_no = matchlnno ;
            }
            setShowRegion(curbp,curlnno,curoff,matchlnno,matchoff) ;
            if (beg_or_end == PTEND)	/* at end of string */
            {
                curwp->w_dotp = matchline ;
                curwp->w_doto = matchoff ;
                curwp->line_no = matchlnno ;
            }
            else		/* at beginning of string */
            {
                curwp->w_dotp = curline;
                curwp->w_doto = curoff;
                curwp->line_no = curlnno ;
            }
             /* Flag that we have moved and got a region to hilight.*/
            curwp->w_flag |= WFMOVEL|WFSELHIL ;
            return TRUE;
            
        }
fail:			/* continue to search */
        if(TTbreakTest(1))
        {
            ctrlg(FALSE,1) ;
            return FALSE ;
        }
        
/*        if (boundry(curline, curoff, direct))*/
/*            break;*/
        /* Quit when we have done our work !! */
    }
    
    return FALSE;	/* We could not find a match */
}

int
expandchar(int c, char *d, int expChr)
{
    register int  doff=1 ;

    if(isDisplayable(c))
    {
        if(expChr && ((c == '\\') || (c == '"')))
        {
            *d = '\\';
            d[doff++] = c;
        }
        else			/* any other character */
            *d = c;
    }
    else if(c < 0x20)		/* control character */
    {   /* 0 -> \C@ this may not be okay */
        *d = '\\';
        d[doff++] = 'C' ;
        d[doff++] = c ^ 0x40;
    }
    else        		/* its a nasty character */
    {
        *d = '\\';
        d[doff++] = '0';
        d[doff++] = hexdigits[c >> 4] ;
        d[doff++] = hexdigits[c&0x0f] ;
    }
    return doff ;
}

int
expandexp(int slen, unsigned char *s, int dlen, int doff, char *d, 
          int cpos, int *opos, int expChr)
{
    unsigned char cc ;
    int  ii ;
    
    if(slen < 0)
        slen = strlen((char *) s) ;
    for(ii=0 ; (doff<dlen-6) && (ii<slen) ; ii++)
    {
        if(cpos == ii)
            *opos = doff ;
        if((cc = *s++) == 255)
        {
            if((cc = *s++) == 2)
            {
                d[doff++] = '\\' ;
                d[doff++] = 's' ;
                cc = *s++ ;
                ii++ ;
            }
            else if(cc == 1)
                cc = 0 ;
            ii++ ;
        }
        doff += expandchar( ((int) cc) & 0xff, d+doff, expChr) ;
    }	/* End of 'while' */
    d[doff] = '\0' ;
    return doff ;
}

/*
 * rvstrcpy -- Reverse string copy.
 */
static void
rvstrcpy(register char *rvstr, register char *str)
{
    register int i;
    
    str += (i = strlen(str));
    
    while (i-- > 0)
        *rvstr++ = *--str;
    
    *rvstr = '\0';
}

/*
 * readpattern -- Read a pattern.  Stash it in apat.  If it is the
 *	search string, create the reverse pattern and the magic
 *	pattern, assuming we are in MAGIC mode (and defined that way).
 *	Apat is not updated if the user types in an empty line.  If
 *	the user typed an empty line, and there is no old pattern, it is
 *	an error.  Display the old pattern, in the style of Jeff Lomicka.
 *	There is some do-it-yourself control expansion.  Change to using
 *	<META> to delimit the end-of-pattern to allow <NL>s in the search
 *	string. 
 */
static int
readpattern(char *prompt, char *apat, char *ppat, char **rpat, int defnum)
{
    int	status;

    if (rpat != NULL)
        *rpat = apat;
    
    /* Read a pattern.  Either we get one,
     * or we just get the META charater, and use the previous pattern.
     * Then, if it's the search string, make a reversed pattern.
     * *Then*, make the meta-pattern, if we are defined that way.
     */
    if((status = mlreply(prompt,MLSEARCH,defnum,apat,NPAT)) == TRUE)
    {
#if	MAGIC
        /* Only make the meta-pattern if in magic mode,
         * since the pattern in question might have an
         * invalid meta combination.
         */
        if((magical=meModeTest(curbp->b_mode,MDMAGIC)) == 0)
        {
            /* Reverse string copy. Reset magic string */
            rvstrcpy(srchRPat, apat);
#ifndef _GNU_REGEX
            srchMcPat[0].mc_type = srchMcRPat[0].mc_type = MCNIL;
#endif
        }
        else
        {
#ifdef _GNU_REGEX
            status = mere_compile_pattern(apat) ;
#else
            status = mcstr((unsigned char *) apat,(unsigned char *) ppat);
            /* SWP 10/1/98 - For hunting backwards we must
             * reverse the search string
             */ 
            if((ppat == NULL) || magical)
                rvstrcpy(srchRPat, apat);
            else
            {
                rvstrcpy(srchRPat, ppat);
                if (rpat != NULL)
                    *rpat = ppat;
            }
#endif
        }
#else
        rvstrcpy(srchRPat, apat);
#endif
    }
    return status ;
}
/*
 * replaces -- Search for a string and replace it with another
 *	string.  Query might be enabled (according to kind).
 */
static int
replaces(int kind, int ff, int nn)
/* kind - Query enabled flag */
{
    /*---	Local defintions for the state machine */

#define	SL_EXIT		0	/* Exit replaces */
#define	SL_GETSEARCH	1	/* Get the search string */
#define	SL_GETREPLACE	2	/* Get the replace string */
#define	SL_INCREPLACE	3	/* Incremental replace */
#define	SL_FIRSTREPLACE	4	/* First replacement condition */
#define	SL_NEXTREPLACE	5	/* Next replacement condition */
#define	SL_QUERYINPUT	6	/* Get the query input from the user */
#define	SL_SUBSTITUTE	7	/* Substitute string */

/*---	Local variable defintions */

    static char	*dpat = NULL;
    static int 	 dpatlen = 0;

    register int i;		/* loop index */
    register int slength;	/* Length of search string */
    register int rlength;	/* length of replace string */
    char         ppat[NPAT];    /* Processed Search pattern	*/
    char         *wpat;         /* Working pattern */
    char	 rpat[NPAT];	/* replacement pattern		*/
    int	         numsub;	/* number of substitutions */
    int	         nummatch;	/* number of found matches */
    int	         status;	/* success flag on pattern inputs */
    int	         nlflag;	/* last char of search string a <NL>? */
    int	         nlrepl;	/* was a replace done on the last line? */
    int	         origoff;	/* and offset (for . query option) */
    long         origlno;       /* line no (for . query option) */
    LINE	*lastline = 0;	/* position of last replace and */
    int	         lastoff;       /* offset (for 'u' query option) */
    long         lastlno;       /* line no (for 'u' query option) */
    int	onemore =FALSE;		/* only do one more replace */
    int	ilength = 0;		/* Last insert length */
    int	state_mc;		/* State machine */
    
    int 	cc;		/* input char for query */
    char	tmpc;		/* temporary character */
    char	tpat[NPAT];	/* temporary to hold search pattern */
    
    /* Determine if we are  in the correct viewing  mode, and if the number
       of repetitions is correct. */

    if (ff)
    {
        if (nn < 0) {           /* Check for negative repetitions. */
            ff = -nn;           /* Force line counting */
            nn = -1;            /* Force to be 0 - do as many as can */
        }
        else
            ff = 0;
    }
    else
        nn = -1;
    exactFlag = (meModeTest(curbp->b_mode,MDEXACT) == 0);
    
    /*--- Execute a state machine  to get around  the replace mechanism.  This
      is  not  super  efficient  in  terms  of  speed,  but  is  good  for
      versitility.  Hence we should be able to add what we like at a later
      date.  */

    for (state_mc = SL_GETSEARCH; state_mc != SL_EXIT; /* NULL */)
    {
        switch (state_mc)
        {
            /*
             * Ask the user for the text of a pattern.
             */
        case SL_GETSEARCH:
            
            if ((status = readpattern
                 ((kind == FALSE ? "Replace" : "Query replace"), 
                  srchPat, ppat, &wpat, 1+lastReplace)) != TRUE)
		return (status);		/* Aborted out - Exit */
            
            slength = strlen (wpat);	/* Get length of search string */
            
/*            mlwrite (MWWAIT,"Search [%d:%s]", slength, wpat);*/
            
            /* Set up flags so we can make sure not to do a recursive
             * replace on the last line. */
            
            nlflag = (wpat[slength - 1] == meNLCHAR);
            nlrepl = FALSE;
            state_mc = SL_GETREPLACE;	/* Move to next state */
            break;
            
        case SL_GETREPLACE :	/* Ask for replacement string */
        case SL_INCREPLACE :	/* Ask for incremental replacement */
            
            strcpy(tpat,"Replace [") ;
            i = expandexp(-1,(unsigned char *) srchPat, NPAT-17, 9, tpat, -1, NULL, 0) ;
            strcpy(tpat+i,"] with");
            
            if ((status=mlreply(tpat,MLSEARCH,0,&rpat[0],NPAT)) != TRUE)
            {
                lastReplace = 0 ;
                return (status);
            }	/* End of 'if' */
            
            rlength = strlen(&rpat[0]);	/* Get length of replace string. */
            
            if (kind)	  	/* Build query replace question string.  */
            {
                tpat[i+6] = ' ' ;
                tpat[i+7] = '[' ;
                i = expandexp(-1,(unsigned char *) rpat, NPAT-i-13, i+8, tpat, -1, NULL, 0) ;
                strcpy(tpat+i,"] ? ") ;
            }            
            state_mc = (state_mc == SL_GETREPLACE) ?
                SL_FIRSTREPLACE : SL_NEXTREPLACE;
            lastReplace = 1 ;
            break;
            
        case SL_FIRSTREPLACE:		/* Do the first replacement */
            
            /* Save original '.' position, initialise  the number of matches 
               and substitutions, and scan through the file.  */
            
            origoff = curwp->w_doto;
            origlno = curwp->line_no;
            numsub = 0;
            nummatch = 0;
            
            /* Initialize last replaced pointers. */
            lastline = NULL;
            
            state_mc = SL_NEXTREPLACE;		/* Do the next replacement */
            break;	
            
        case  SL_NEXTREPLACE:		/* Next replcaement condition */
            
            if (!((nn < 0 || nn > nummatch) &&
                  (nlflag == FALSE || nlrepl == FALSE) &&
                  onemore == FALSE))
            {
                /*---	Exit condition has been met */
                
		state_mc = SL_EXIT;		/* State machine to exit */
		break;				/* Exit case */
            }	/* End of 'if' */
            
            /*--- Search  for  the  pattern.   If  we  search  with  a regular
              expression, also save the true length of matched string.  */
            
#if	MAGIC
#ifdef _GNU_REGEX
            if(magical)
            {
                if (!mere_scanner(FORWARD, PTBEG, &ff, NULL))
                {
                    state_mc = SL_EXIT;
                    break;
		}
            }
#else
            if(magical && meModeTest(curbp->b_mode,MDMAGIC))
            {
                if (!mcscanner(&srchMcPat[0], FORWARD, PTBEG, &ff, NULL))
                {
                    state_mc = SL_EXIT;
                    break;
		}
            }
#endif
            else
#endif
                if (!scanner((unsigned char *) wpat, FORWARD, PTBEG, &ff, NULL))
            {
                state_mc = SL_EXIT;
                break;		/* all done */
            }
            
            ++nummatch;	/* Increment # of matches */
            
            /* Check if we are on the last line. */
            
            nlrepl = (lforw(curwp->w_dotp) == curbp->b_linep);
            
            if (kind)
		state_mc = SL_QUERYINPUT;
            else
		state_mc = SL_SUBSTITUTE;
            break;
            
        case SL_QUERYINPUT:			/* Get input from user. */
            
#if MEUNDO
            undoContFlag++ ;
#endif
            cc = mlCharReply(tpat,mlCR_LOWER_CASE|mlCR_UPDATE_ON_USER|mlCR_CURSOR_IN_MAIN,"y nil!u.",
                            "(Y)es (N)o (I)nc (L)ast (!)Do rest (U)ndo (^G)Abort (.)Abort back ? ");
            
            switch (cc)			/* And respond appropriately. */
            {
            case -1:
                return ctrlg(FALSE,1);
                
            case 'y':			/* yes, substitute */
            case ' ':
		state_mc = SL_SUBSTITUTE;
		break;
                
            case 'n':			/* no, onword */
		WforwChar(curwp, 1);
		state_mc = SL_NEXTREPLACE;
		break;
                
            case 'i':			/* Incremental change */
		state_mc = SL_INCREPLACE;
		break;
							
            case 'l':			/* last one */
		onemore = TRUE;
		state_mc = SL_SUBSTITUTE;
		break;
                
            case '!':			/* yes/stop asking */
		kind = FALSE;
		state_mc = SL_SUBSTITUTE;
		break;

            case 'u':	/* undo last and re-prompt */
                
                if (lastline == NULL)
                {   /* Restore old position. */
                    TTbell();		/* There is nothing to undo. */
                    break;
		}
                
                /*---	Reset the last position */
                
                curwp->w_dotp  = lastline ; 
                curwp->w_doto  = lastoff ;
                curwp->line_no = lastlno ;
                curwp->w_flag |= WFMOVEL ;
		lastline = NULL;
                
		/* Delete the new string. */
                
		WbackChar(curwp, ilength);
                if (!ldelete(ilength,6))
                    return mlwrite(MWABORT,"[ERROR while deleting]");
                
		/* And put in the old one. */
                
                for (i = 0; i < slength; i++)
                {
                    tmpc = dpat[i];
                    status = ((tmpc == meNLCHAR) ? lnewline(): linsert(1, tmpc));
                    
                    /* Insertion error? */
                    
                    if (!status)
                        return mlwrite(MWABORT,"[Out of memory while inserting]");
                
                }	/* End of 'for' */
#if MEUNDO
                meUndoAddReplaceEnd(slength) ;
#endif
		/* Record one less substitution, backup, and reprompt. */
                
		--numsub;
                lastoff = curwp->w_doto ;
                lastlno = curwp->line_no ;
                WbackChar(curwp, slength);
                /* must research for this instance to setup the auto-repeat sections
                 * correctly */
                --nummatch;	/* decrement # of matches */
                state_mc = SL_NEXTREPLACE ;
		break;
		
            case '.':	/* abort! and return */
                /* restore old position */
                gotoLine(TRUE,origlno+1) ;
                curwp->w_doto  = origoff ;
                state_mc = SL_EXIT;		/* Exit state machine */
		break;
            }	/* end of switch */
            break;
            
            /*
             * Delete the sucker.
             */
        case SL_SUBSTITUTE:			/* Substitute the string */
            
            if(bchange() != TRUE)               /* Check we can change the buffer */
                return ABORT ;
#if	MAGIC
            /* only set the new regex slength here as if the user is doing a query replace
             * and does an undo, the slength must be correct for the last insertion */
#ifdef _GNU_REGEX
            slength = mereRegs.end[0] ;
#else
            slength = mclen;
#endif
#endif
            if (dpatlen < (slength+1))
            {
                if (dpat != 0)
                { 
                    meFree(dpat) ;
                    dpat = 0; 
                }
		dpatlen = slength+1;
		if ((dpat = (char *) meMalloc (dpatlen)) == NULL)
                    return (FALSE);
            }	/* End of 'if' */
	
            if(mldelete(slength,(unsigned char *)dpat) != 0)
                return mlwrite(MWABORT,"[ERROR while deleting]") ;
            ilength = 0 ;
#if	0
            /*
             * Special case substitution. Check for "\0" which means
             * substitute for nothing.
             */
            if (!magical || (rlength != 2) || (rpat[0] != MC_RRCL) || (rpat[1] != '0'))
#endif
            {
            
                /*
                 * And insert its replacement.
                 */
                
                for (i = 0 ; i < rlength; i++)
                {
                    switch (tmpc = rpat[i])
                    {
                    case MC_RRCL:		/* Replacement string */
#if	MAGIC
#ifdef _GNU_REGEX
                        if (!magical)	               /* Magic string ?? */
                            ;	                       /* No - drop out */
                        else if (i+1 >= rlength)       /* Too long ?? */
                            ;		               /* Yes - drop out */
                        /*
                         * Auto repeat number (\1 - \9) or all (\&) ??
                         */
                        else if ((((tmpc = rpat [i+1]) < '0') ||
                                  ((int)(tmpc-='0') >= mereNumRegs)) &&
                                 ((int)(tmpc-='&') != 0))
                            ;		        /* No - drop out */
                        else
                        {
                            /* Found auto repeat char replacement */
                            register int j,k;	/* Local loop counter */
                            
                            i++;		/* Skip next character */
                            /* if start offset is < 0, it was a ? auto repeat which was not found,
                               therefore replace str == "" */ 
                            if((j=mereRegs.start[tmpc]) >= 0)
                            {
                                for(k=mereRegs.end[(int) tmpc];
                                    ((j < k) && (status != FALSE)); ilength++)
                                {
                                    if ((tmpc=dpat[j++]) != meNLCHAR)
                                        status = linsert(1, tmpc);
                                    else
                                        status = lnewline ();
                                }
                            }
                            break;
                        }	/* End of 'if' */
#else
                        if (!magical)	               /* Magic string ?? */
                            ;	                       /* No - drop out */
                        else if (srchAutoRpt.count == 0)  /* Auto repeat ?? */
                            ;			       /* No - drop out */
                        else if (i+1 >= rlength)       /* Too long ?? */
                            ;		               /* Yes - drop out */
                        /*
                         * Auto repeat number ??
                         */
                        else if (((tmpc = rpat [i+1]) < '1') ||
                                 (tmpc > '9'))
                            ;		        /* No - drop out */
                        else if ((int)(tmpc -= '1') >= srchAutoRpt.count)
                            ;		/* Invalid counter - drop out */
                        else
                        {
                            /* Found auto repeat char replacement */
                            register int j,k;	/* Local loop counter */
                            
                            i++;		/* Skip next character */
                            for(j=srchAutoRpt.start_pos[(int)tmpc],k=srchAutoRpt.end_pos[(int) tmpc];
                                ((j < k) && (status != FALSE)); ilength++)
                            {
                                if ((tmpc=dpat[j++]) != meNLCHAR)
                                    status = linsert(1, tmpc);
                                else
                                    status = lnewline ();
                            }
                            break;
                        }	/* End of 'if' */
#endif
#endif	
                        /*---	Got this far so not a auto repeat
                           charater. Put out for normal character */
                        
                        if ((i+1) >= rlength)
                            tmpc = MC_RRCL;
                        else if (((i+3) <= rlength) &&
                                 (rpat [i+1] == '0') &&
                                 isHexDigit (rpat [i+2]) &&
                                 isHexDigit (rpat [i+3]))
                        {
                            tmpc = ((hexToNum (rpat [i+2]) << 4) |
                                    hexToNum (rpat [i+3])) & 0xff;
                            i+= 3;
                        }
                        else
                        {
                            tmpc = rpat [i+1];
                            i++;
                        }
                        /* Allow drop through */
                    default:
                        if (tmpc != meNLCHAR)
                            status = linsert(1, tmpc);
                        else
                            status = lnewline ();
                        ilength++;
                        break;
                    }	/* End of 'switch' */
                
                    /* Insertion error ? */
                    
                    if (!status) 
                        return mlwrite(MWABORT,"[Out of memory while inserting]"); 
                }	/* End of 'for' */
            }
#if MEUNDO
            meUndoAddReplace(dpat,ilength) ;
#endif
            
            if (kind) 
            {			/* Save position for undo. */
		lastline = curwp->w_dotp;
		lastoff = curwp->w_doto; 
		lastlno = curwp->line_no; 
            }
            numsub++;			/* increment # of substitutions */
            state_mc = SL_NEXTREPLACE;	/* Do next replacement */
            if (slength == 0)
            {
                WforwChar(curwp, 1);
                if ((ff > 0) && (curwp->w_doto == 0) && (--ff == 0))
                    state_mc = SL_EXIT;
            }
            break;
            
        default:
            return mlwrite(MWABORT,"[Bad state machine value %d]", state_mc);
        }	/* End of 'switch' on state_mc */
    }   /* End of 'for' */
    
    /*---	And report the results. */
    
    mlwrite(MWCLEXEC,"%d substitutions", numsub);
    return (TRUE);
    
}	/* End of 'replaces' */

/*
 * replaceStr -- Search and replace.
 */
int
replaceStr(int f, int n)
{
    return(replaces(FALSE, f, n));
}

/*
 * queryReplaceStr -- search and replace with query.
 */
int	
queryReplaceStr(int f, int n)
{
    return(replaces(TRUE, f, n));
}

/*
 * searchForw -- Search forward.  Get a search string from the user, and
 *	search, beginning at ".", for the string.  If found, reset the "."
 *	to be just after the match string, and (perhaps) repaint the display.
 */

int
searchForw(int f, int n)
{
    register int status;
    char         ppat[NPAT];            /* Processed Search pattern */
    char         *wpat;                 /* The pattern string */
    int          repCount;              /* Repeat count */

    /* Resolve the repeat count. */
    if (n <= 0) {
        n = -n;
        repCount = 1;
    }
    else {
        repCount = n;
        n = 0;
    }
    
    /* Ask the user for the text of a pattern.  If the
     * response is TRUE (responses other than FALSE are
     * possible), search for the pattern.
     */
    if ((status = readpattern("Search", srchPat, ppat, &wpat, 
                              1+lastReplace)) == TRUE)
    {
        lastReplace = 0 ;
        exactFlag = (meModeTest(curbp->b_mode,MDEXACT) == 0);
        do
        {
#if	MAGIC
#ifdef _GNU_REGEX
            if(magical)
                status = mere_scanner(FORWARD, PTEND, &n, NULL);
#else
            if(meModeTest(curbp->b_mode,MDMAGIC))
            {
                if (magical)
                    status = mcscanner(&srchMcPat[0], FORWARD, PTEND, &n, NULL);
                else 
                    status = scanner((unsigned char *) wpat, FORWARD, PTEND, &n, NULL);
            }
#endif
            else
#endif
                status = scanner((unsigned char *) wpat, FORWARD, PTEND, &n, NULL);
        } while ((--repCount > 0) && status);
        
        /* ...and complain if not there.
         */
        if(status == FALSE)
            notFound();
    }
    return(status);
}

/*
 * huntForw -- Search forward for a previously acquired search string,
 *	beginning at ".".  If found, reset the "." to be just after
 *	the match string, and (perhaps) repaint the display.
 */

int
huntForw(int f, int n)
{
    register int status;
    int          repCount;              /* Repeat count */

    /* Resolve the repeat count. */
    if (n <= 0) {
        n = -n;
        repCount = 1;
    }
    else {
        repCount = n;
        n = 0;
    }
    
    /* Make sure a pattern exists, or that we didn't switch
     * into MAGIC mode until after we entered the pattern.
     */
    if(srchPat[0] == '\0')
        return mlwrite(MWABORT,"[No pattern set]");
#if	MAGIC
#ifndef _GNU_REGEX
    if(meModeTest(curbp->b_mode,MDMAGIC) &&
       (srchMcPat[0].mc_type == MCNIL))
    {
        if (!mcstr((unsigned char *) srchPat, NULL))
            return FALSE;
    }
#endif
#endif
    
    /* Search for the pattern... */
    do
    {
#if	MAGIC
#ifdef _GNU_REGEX
        if(magical)
            status = mere_scanner(FORWARD, PTEND, &n, NULL);
#else
        if(magical && meModeTest(curbp->b_mode,MDMAGIC))
            status = mcscanner(&srchMcPat[0], FORWARD, PTEND, &n, NULL);
#endif
        else
#endif
            status = scanner((unsigned char *) srchPat, FORWARD, PTEND, &n, NULL);
    } while ((--repCount > 0) && status);
    
    /* ...and complain if not there. */
    if(status == FALSE)
        notFound() ;
    
    return(status);
}

/*
 * searchBack -- Reverse search.  Get a search string from the user, and
 *	search, starting at "." and proceeding toward the front of the buffer.
 *	If found "." is left pointing at the first character of the pattern
 *	(the last character that was matched).
 */
int
searchBack(int f, int n)
{
    register int status;
    char         ppat[NPAT];            /* Search pattern */
    char         *wpat;                 /* Returned search pattern */
    int          repCount;              /* Repeat count */
    
    /* Resolve the repeat count. */
    if (n <= 0) {
        n = -n;
        repCount = 1;
    }
    else {
        repCount = n;
        n = 0;
    }
    
    /* Ask the user for the text of a pattern.  If the
     * response is TRUE (responses other than FALSE are
     * possible), search for the pattern.
     */
    if ((status = readpattern("Reverse search", srchPat, ppat, &wpat, 
                              1+lastReplace)) == TRUE)
    {
        lastReplace = 0 ;
        exactFlag = (meModeTest(curbp->b_mode,MDEXACT) == 0);
        do
        {
#if	MAGIC
#ifdef _GNU_REGEX
            if(magical)
                status = mere_scanner(REVERSE, PTBEG, &n, NULL);
#else
            if(magical && meModeTest(curbp->b_mode,MDMAGIC))
                status = mcscanner(&srchMcRPat[0], REVERSE, PTBEG, &n, NULL);
#endif
            else
#endif
                status = scanner((unsigned char *) &srchRPat[0], REVERSE, PTBEG, &n, NULL);
        } while ((--repCount > 0) && status);
        
        /* ...and complain if not there.
         */
        if(status == FALSE)
            notFound();
    }
    return(status);
}

/*
 * huntBack -- Reverse search for a previously acquired search string,
 *	starting at "." and proceeding toward the front of the buffer.
 *	If found "." is left pointing at the first character of the pattern
 *	(the last character that was matched).
 */
int
huntBack(int f, int n)
{
    register int status;
    int          repCount;              /* Repeat count */
    
    /* Resolve null and negative arguments. */
    if (n <= 0) {
        n = -n;
        repCount = 1;
    }
    else {
        repCount = n;
        n = 0;
    }
    
    /* Make sure a pattern exists, or that we didn't switch
     * into MAGIC mode until after we entered the pattern.
     */
    if(srchPat[0] == '\0')
        return mlwrite(MWABORT,"[No pattern set]");
#if	MAGIC
#ifndef _GNU_REGEX
    if(meModeTest(curbp->b_mode,MDMAGIC) &&
       (srchMcRPat[0].mc_type == MCNIL))
    {
        if (!mcstr((unsigned char *) srchPat, NULL))
            return FALSE;
    }
#endif
#endif
    
    /* Go search for it...
     */
    do
    {
#if	MAGIC
#ifdef _GNU_REGEX
        if(magical)
            status = mere_scanner(REVERSE, PTBEG, &n, NULL);
#else
        if(magical && meModeTest(curbp->b_mode,MDMAGIC))
            status = mcscanner (&srchMcRPat[0], REVERSE, PTBEG, &n, NULL);
#endif
        else
#endif
            status = scanner ((unsigned char *) &srchRPat[0], REVERSE, PTBEG, &n, NULL);
    } while ((--repCount > 0) && status);
    
    /* ...and complain if not there. */
    if(status == FALSE)
        notFound();
    return(status);
}


int
iscanner (char *apat, int n, int flags, SCANNERPOS *sp)
{
    int direct;
    int beg_or_end;
    
    direct = (flags & ISCANNER_BACKW) ? REVERSE:FORWARD ;
    beg_or_end = (flags & ISCANNER_PTBEG) ? PTBEG : PTEND;
    exactFlag = ((flags & ISCANNER_EXACT) == 0);

#if	MAGIC
    /*
     * Only make the meta-pattern if in magic mode, since the pattern in
     * question might have an invalid meta combination.
     */
    if((magical = (flags & ISCANNER_MAGIC)) != 0)
    {
#ifdef _GNU_REGEX
        int status ;
        
        if (flags & ISCANNER_QUIET)
            reportErrors = FALSE;
        status = mere_compile_pattern(apat);
        if (flags & ISCANNER_QUIET)
            reportErrors = TRUE;        
        /*
         * Bail out here if we have a false status, we cannot construct a valid
         * search string yet it is not compleate.
         */
        if (status != TRUE)
            return ABORT ;
        
        return mere_scanner(direct,beg_or_end,&n,sp);
#else
        int status ;
        char ppat [NPAT];               /* Pre-processed pattern */
        
        if (flags & ISCANNER_QUIET)
            reportErrors = FALSE;
        status = mcstr((unsigned char *) apat,(unsigned char *) ppat);
        if (flags & ISCANNER_QUIET)
            reportErrors = TRUE;        
        /*
         * Bail out here if we have a false status, we cannot construct a valid
         * search string yet it is not compleate.
         */
        if (status != TRUE)
            return ABORT ;
        
        /* Reverse string copy, always want to do this incase a hunt-backwards is used next */
        rvstrcpy(srchRPat, ppat);
        
        if(magical)
            return mcscanner(((direct == FORWARD) ? &srchMcPat[0] : &srchMcRPat[0]),
                             direct, beg_or_end, &n, sp);

        /*
         * The string is not a magic pattern. 
         * Use the modified pattern as the source string (returned from mcstr())
         * since the literal characters have been decomposed.
         */
        return scanner ((unsigned char *) ((direct == FORWARD) ? ppat : srchRPat),
                        direct, beg_or_end, &n, sp);
#endif
    }
    
    /* Reverse string copy, always want to do this incase a hunt-backwards is used next */
    rvstrcpy(srchRPat, apat);
#ifndef _GNU_REGEX
    srchMcPat[0].mc_type = srchMcRPat[0].mc_type = MCNIL;
#endif
#else
    /* Reverse string copy, always want to do this incase a hunt-backwards is used next */
    rvstrcpy(srchRPat, apat);
#endif
    
    return scanner ((unsigned char *) ((direct == FORWARD) ? apat : srchRPat),
                          direct, beg_or_end, &n, sp);
}


