/*****************************************************************************
*
*	Title:		%M%
*
*	Synopsis:	Find tag positions
*
******************************************************************************
*
*	Filename:		%P%
*
*	Author:			Mike Rendell of ROOT Computers Ltd.
*
*	Creation Date:		10/05/91 08:27		<160699.0735>
*
*	Modification date:	%G% : %U%
*
*	Current rev:		%I%
*
*	Special Comments:	
*
*	Contents Description:	Find tags from a tag file. Only defined if
*				running some sort of UNIX.
*
*****************************************************************************
* 
* Modifications to the original file by Jasspa. 
* 
* Copyright (C) 1988 - 1999, JASSPA 
* The MicroEmacs Jasspa distribution can be copied and distributed freely for
* any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
* incorporated into commercial software with the expressed permission of
* JASSPA.
* 
****************************************************************************/


/*---	Include defintions */

#define	__TAGC			/* Define filename */

/*---	Include files */

#include "emain.h"
#include "esearch.h"

/*
 * Find the function in the tag file.  From Mike Rendell.
 * Return the line of the tags file, and let do_tag handle it from here.
 */

int
find_fun(register char *key, register char *file, register char *patt)
{
    int		   tmp;
    register FILE *fp;
    char	   function[MAXBUF], *baseName ;
    register long  pos, start, end, End;
    char          *tagFile, *tagTemp ;
    
    *patt++ = meNLCHAR;
    if(((tagFile = getUsrVar("tag-file")) == errorm) ||
       ((tagTemp = getUsrVar("tag-template")) == errorm))
        return mlwrite(MWABORT,"[tags not setup]");
    
    getFilePath(curbp->b_fname,file) ;
    baseName = file + strlen(file) ;
    strcpy(baseName,tagFile) ;
    
    if((fp = fopen(file, "r")) == (FILE *) NULL)
        return mlwrite(MWABORT|MWCLEXEC,"[no tags file : %s]",file);
    start = 0;
    fseek(fp, 0L, 2);
    end = End = ftell(fp);
    for(;;)
    {
        pos = (start+end)/2;
        fseek(fp, pos, 0);
        if(end || start)
            for( ; (fgetc(fp) != meNLCHAR) ; )
                ;
        
        /* Get line of info */
        fscanf(fp,tagTemp,function,baseName,patt) ;
        
        if (!(tmp = strcmp(key, function)))
        {		/* found */
            char cc, *p1, *p2 ;
            fclose(fp);
            p1 = p2 = patt ;
            while((cc=*p2++) != '\0')
            {
                if(cc == '\\')
                    cc = *p2++ ;
                *p1++ = cc ;
            }
            *(p1-2) = '\0' ;
            return TRUE ;
        }
        else if (tmp < 0)	/* back */
            end = pos;
        else			/* foward */
            start = pos;
        if (end - start < 9)
        {   /* minimium number of chars per line */
            if (!start && end)
                end = 0;
            else if (end == End)
            {
                char buf[1024];
                int i;

                fseek(fp, 1024 > End ? -End : -1024L, 1);
                i = fread(buf, 1, 1024, fp);
                if (i > 2)
                    i -= 2;
                while ( buf[i] != meNLCHAR && i)
                    i--;
                start -= 2*(1024 > End ? End - i : 1024 - i);
                End++;
            }
            else
                break;
        }
    }
    fclose(fp);
    mlwrite(MWABORT|MWCLEXEC,"[Can't find tag [%s] in tags file : %s]",
            key, tagFile);
    return ABORT ;
}

static	int
do_tag(register int nrcurb, register char tag[])
{
    char  file[FILEBUF];	/* File name */
    char  patt[MAXBUF];		/* Pattern */
    int   ii ;
    
    /*
     * Since every tag in the tag file starts with a '^', representing
     * a newline, we start patt off with a newline character and pass
     * the address of the second element (ie patt[1]) to find_fun().
     */

    if(find_fun(tag, file, patt) != TRUE)
        return FALSE ;
    
    if(nrcurb && (wpopup(NULL,WPOP_MKCURR) == NULL))
        return FALSE ;
    if(findSwapFileList(file,(BFND_CREAT|BFND_MKNAM),0) != TRUE)
        return mlwrite(MWABORT,"[Can't find %s]", file);
    gotobob(TRUE, 0);	/* can't return error */
    
    ii = iscanner(patt,0,ISCANNER_PTBEG|ISCANNER_EXACT,NULL) ;
    
    if(ii != TRUE)
        return mlwrite(MWABORT,"[Can't find %s]", patt) ;
    
    /*
     * Since we started "patt" off with a newline character, we are
     * now sitting one character off the "target", and so need to
     * advance our postion.
     */
    WforwChar(curwp, 1);
    
    return TRUE ;
}

/*ARGSUSED*/

int
findTag(int f, int n)
{
    char            tag [MAXBUF];
    register LINE  *startp;
    register short  starto, i;
    
    /* Determine if we are in a word. If not then get a word from the 
       user. */
    
    if((n & 0x02) || (inWord() == FALSE))
    {
	/*---	Get user word. */
        if(mlreply("Enter Tag", MLNOSPACE, 0, tag,MAXBUF) != TRUE)
            return ABORT ;
    }
    else
    {
	/*---	Save the current cursor postion. */
        
        startp = curwp->w_dotp;		
        starto = curwp->w_doto;
        
	/*---	Go to the start of the word. */

        while(WbackChar(curwp,1) != FALSE)	/* Back up */
            if(inWord() == FALSE)
            {
                WforwChar(curwp, 1);		/* Back into the word */
                break ;
            }
        
	/*---	Get the word required. */
        
        i = 0;
        while (inWord() != FALSE)
        {
            tag[i++] = lgetc(curwp->w_dotp, curwp->w_doto);
            if (i >= MAXBUF-1)
            {
                i = 0;	/* Too big !! */
                break;	/* Break out */
            }
            WforwChar(curwp, 1);
        }
        tag[i] = 0;			/* End of string */
        
	/*---	Restore old cursor position, find tag type and find the string */
        
        curwp->w_dotp = startp;
        curwp->w_doto = starto;

    }	/* End of 'if' */
    
    /*---	Call the tag find. */

    return do_tag((n & 0x01),tag);

}


