/*****************************************************************************
*
*	Title:		%M%
*
*	Synopsis:	Time stamp the files.
*
******************************************************************************
*
*	Filename:		%P%
*
*	Author:			Jon Green of Akebia Ltd.
*
*	Creation Date:		10/05/91 08:27		<160699.0828>
*
*	Modification date:	%G% : %U%
*
*	Current rev:		%I%
*
*	Special Comments:	
*
*	Contents Description:	Set auto file time stamping. Only defined if
*				running some sort of UNIX/DOS/WIN32.
*
*****************************************************************************
* 
* Copyright (C) 1991 - 1999, JASSPA 
* 
* The MicroEmacs Jasspa distribution can be copied and distributed freely for
* any non-commercial purposes. The MicroEmacs Jasspa Distribution can only be
* incorporated into commercial software with the expressed permission of
* JASSPA.
* 
****************************************************************************/

/*---	Include defintions */

#define	__TIMEC			/* Define filename */

/*---	Include files */

#include "emain.h"
#include "esearch.h"

#if (defined _UNIX) || (defined _DOS) || (defined _WIN32)
#include <sys/types.h>
#include <time.h>
#endif


#if	TIMSTMP


/*---	Local macro definitions */

/*---	External references */

/*---	Local type definitions */

/*---	Local variable definitions */

char	time_stamp[TSTMPLEN] = "<%Y%M%D.%h%m>";	/* Time stamp string */

#define TSNUMFIELD 6
static char TSFIELDS[]="YMDhms" ;

void
createTimeStampSrch(char *buf, unsigned char *pos)
{
    unsigned char curPos=0 ;
    char cc, *ss, *tt ;
    int ii ;
    memset(pos,0,TSNUMFIELD) ;
    
    ss = time_stamp ;
    while((cc=*ss++) != '\0')
    {
        if((cc == '%') && ((cc=*ss++) != '\0') &&
           ((tt = strchr(TSFIELDS,cc)) != NULL))
        {
            ii = tt-TSFIELDS ;
            pos[ii] = ++curPos ;
#ifdef _GNU_REGEX
            if(ii == 0)
            {
                /* for the year flags cater for an optional century (4 digits) */
                strcpy(buf,"\\(\\([0-9][0-9]\\)?[0-9][0-9]\\)") ;
                buf += 29 ;
                curPos++ ;
            }
            else
#endif
            {
                strcpy(buf,"\\([0-9][0-9]\\)") ;
                buf += 14 ;
            }
        }
        else
            *buf++ = cc ;
    }
    *buf = '\0' ;
}


int
set_timestamp(BUFFER *bp)
{
    WINDOW         *owp, win ;
    BUFFER         *obp ;
    unsigned char   pos[TSNUMFIELD];	/* pos of fields in search */
    char 	    patt[1024];    	/* Search pattern */
    int             ii, jj ;
    
    /*---	Determine if time stamping is to be performed. */
    
    if(!meModeTest(bp->b_mode,MDTIME) ||	/* Time stamping mode on ?? */
       (time_stamp[0] == '\0'))			/* No time stamp defined */
        return TRUE ;				/* No - exit */
    
    mlwrite(MWCURSOR,"[Time stamping File]");
    
    /*---	Save current position in buffer and go to the start of the buffer. */
    
    obp = curbp ;
    owp = curwp ;
    curwp = &win ;
    curbp = bp ;
    win.w_bufp = curbp ;
    win.w_dotp = lforw(curbp->b_linep) ;	/* Save position */
    win.w_doto = 0 ;
    win.w_flag = 0 ;
    win.line_no = 0 ;    
    
    /* create magic search string */
    createTimeStampSrch(patt,pos) ;
    
    /*---	Search for the time stamp string. */
    if(iscanner(patt,100,ISCANNER_PTBEG|ISCANNER_MAGIC|ISCANNER_EXACT,NULL) == TRUE)
    {
        unsigned char   values[TSNUMFIELD] ;
        time_t	        clock;		/* Time in machine format. */
        struct tm      *time_ptr;	/* Pointer to time frame. */
        unsigned short  soff ;

        /* Found it, so fill in the slots */
        
        clock = time(0);				/* Get system time */
        time_ptr = (struct tm *) localtime (&clock);	/* Get time frame */
        
        values[0] = time_ptr->tm_year ;
        values[1] = time_ptr->tm_mon+1 ;
        values[2] = time_ptr->tm_mday ;
        values[3] = time_ptr->tm_hour ;
        values[4] = time_ptr->tm_min ;
        values[5] = time_ptr->tm_sec ;
        
        soff = curwp->w_doto ;
        for(ii=0 ; ii<TSNUMFIELD ; ii++)
            if((jj=pos[ii]) != 0)
            {
#ifdef _GNU_REGEX
                if(!ii && ((mereRegs.end[jj]-mereRegs.start[jj]) == 4))
                {
                    curwp->w_dotp->l_text[soff+mereRegs.end[jj]-4] = '0' +  ((((int) values[ii])+1900)/1000) ;
                    curwp->w_dotp->l_text[soff+mereRegs.end[jj]-3] = '0' + (((((int) values[ii])+1900)/100)%10) ;
                }
                curwp->w_dotp->l_text[soff+mereRegs.end[jj]-2] = '0' + (values[ii]/10) ;
                curwp->w_dotp->l_text[soff+mereRegs.end[jj]-1] = '0' + (values[ii]%10) ;
#else
                curwp->w_dotp->l_text[soff+srchAutoRpt.start_pos[jj-1]]   = '0' + (values[ii]/10) ;
                curwp->w_dotp->l_text[soff+srchAutoRpt.start_pos[jj-1]+1] = '0' + (values[ii]%10) ;
#endif
            }
        curwp->w_dotp->l_flag |= LNCHNG ;
    }
    else
        ii = 0 ;
    curbp = obp ;
    curwp = owp ;
    if(ii)
        /* if found flag any window displaying it to update */
        addModeToBufferWindows(bp,WFMAIN) ;

    return TRUE ;
}

#endif  /* TIMSTMP */

