#____________________________________________________________________________
#
# Make file for MicroEmacs using the Microsoft MSCV 5.0 development kit.
# 
# Author      : Jon Green
# Created     :  020197.1002
# Last Edited : <990716.0749>
#
# History     : JG  020197 Original  Constructed from MS envoronment and
#				     UNIX makefile.
#		
# Details     : 
#               Build from the command line using nmake. 
#
#               nmake              - builds locally.
#		nmake clean        - cleans emacs directory
#		nmake spotless     - cleans some more !!
#		nmake install	   - builds and copies to c:\emacs.
#____________________________________________________________________________
# Uncomment the next line for no console support
SUPSYS	= windows
# Uncomment the next 2 lines for console support
#CONDEF	= -D_WINCON
#SUPSYS	= console
#
# Uncomment the next 2 lines for URL support
#URLDEF	= -D_URLSUPP
#URLLIB	= $(LIBRARY)\ws2_32.lib
#
#
# TOOLS Environment - change to location of your tools.
#
TOOLSDIR=       c:\progra~1\DevStudio\vc
INCLUDE=	$(TOOLSDIR)\include
LIBRARY=	$(TOOLSDIR)\lib

#
# Destination directory.
#
BINDIR	=	c:\emacs
#
# CTYPE - pick optimised or debugged.
CTYPE   = $(COPTIMISE)
#CTYPE  = $(CDEBUG)

#
# Standard Compilation tools and flags
#
CC	=	cl
RC	=	rc
RM	=	erase
LD	=	link
INSTALL	=	copy
MKDIR	=	mkdir

# Standard File Extensions
I	=	-I
O	=	.o
EXE	=	.exe

# Extended build rules.
.SUFFIXES: .c .o
.c.o:
	$(CC) $(CFLAGS) -c $< -Fo$*.o

#
# Standard include stuff
#
CDEBUG		=	-nologo -G5 -W3 -GX -Z7 -YX -Yd -Od
CPROFILE	=	-nologo -G5 -YX -GX
COPTIMISE	=	-nologo -G5 -YX -GX -O2 -DNDEBUG=1
CNONOPTIMISE	=	-nologo -G5 -YX -GX

LDFLAGS		=	/DEBUG

CINCLUDES	=	$(I). $(I)$(INCLUDE)
CFLAGS		=	$(CINCLUDES) $(CTYPE) -D_JASSPA_ME -D_WIN32 -DWIN32 $(CONDEF) $(URLDEF) -D_WINDOWS -D_MBCS
LDFLAGS		=	$(LDFLAGS) /NOLOGO /SUBSYSTEM:$(SUPSYS) /INCREMENTAL:no /MACHINE:IX86 /PDB:NONE

SYSLIBS         =	$(LIBRARY)\kernel32.lib	\
			$(URLLIB)		\
			$(LIBRARY)\libc.lib	\
			$(LIBRARY)\user32.lib	\
			$(LIBRARY)\gdi32.lib	\
			$(LIBRARY)\winmm.lib	\
                        $(LIBRARY)\wsock32.lib	\
                        $(LIBRARY)\winspool.lib	\
                        $(LIBRARY)\comdlg32.lib	\
                        $(LIBRARY)\advapi32.lib	\
                        $(LIBRARY)\shell32.lib	\
                        $(LIBRARY)\ole32.lib	\
                        $(LIBRARY)\oleaut32.lib	\
                        $(LIBRARY)\uuid.lib	\
                        $(LIBRARY)\msvcrt.lib	\
                        $(LIBRARY)\oldnames.lib
# Note that oldnames.lib is MSCV 2.0 NOT 4.0

#
# Define the files involved in the build.
#
HEADERS = ebind.h edef.h eextrn.h efunc.h esearch.h estruct.h evar.h \
	  evers.h eterm.h ebind.def efunc.def evar.def emain.h eprint.h
PLTHDR  = wintermr.h
STDSRC	= abbrev.c basic.c bind.c buffer.c crypt.c dirlist.c display.c \
	  eval.c exec.c file.c fileio.c hilight.c history.c input.c \
	  isearch.c key.c line.c macro.c main.c narrow.c next.c osd.c \
	  print.c registry.c random.c region.c search.c spawn.c spell.c \
	  tag.c termio.c time.c undo.c window.c word.c regex.c
PLTSRC  = winterm.c winprint.c

STDOBJ	= $(STDSRC:.c=.o)	
PLTOBJ  = $(PLTSRC:.c=.o)
OBJ	= $(STDOBJ) $(PLTOBJ)

#
# Standard build rules
#
all::	me32$(EXE)

me32$(EXE):$(OBJ) me.res
	$(LD) $(LDFLAGS) /out:$@ $(OBJ) me.res $(SYSLIBS)

$(STDOBJ): $(HEADERS)
$(PLTOBJ): $(HEADERS) $(PLTHDR)

me.res:
	$(RC) -v -i . -i $(INCLUDE) -i $(TOOLSDIR)\mfc\include -fo $@ $*.rc 
relink::
	$(RM)	me32$(EXE)
	$(MAKE) -f $(MAKEFILE) $(MFLAGS) me32$(EXE)

install:: me32$(EXE)
	$(INSTALL) me32$(EXE) $(BINDIR)
	@echo Install done.

spotless:: clean
	$(RM) *.??~
clean::
	- $(RM) *.pch
	- $(RM) *.o
	- $(RM) *.exe
	- $(RM) *.res
