#____________________________________________________________________________
#
# Make file for MicroEmacs using the Microsoft MSCV 2.0 development kit.
# 
# Author      : Jon Green
# Created     :  020197.1002
# Last Edited : <990617.0909>
#
# History     : JG  020197 Original  Constructed from MS envoronment and
#				     UNIX makefile.
# 
# Notes       : This is a win32 build for V2.0 of the MSVc compiler. 
#               For win32s then use win32s.mak. The only difference between
#               this file and the win32.mak is the resource file is 
#               slightly different.
#		
# Details     : 
#               Build from the command line using nmake. 
#
#               nmake              - builds locally.
#		nmake clean        - cleans emacs directory
#		nmake spotless     - cleans some more !!
#		nmake install	   - builds and copies to c:\emacs.
#____________________________________________________________________________
#
#
#
# TOOLS Environment - change to location of your tools.
#
TOOLSDIR=       c:\msvc20
INCLUDE=	$(TOOLSDIR)\include
LIBRARY=	$(TOOLSDIR)\lib
TOOLSBIN=	$(TOOLSDIR)\bin

#
# Destination directory.
#
BINDIR	=	c:\emacs

#
# Standard Compilation tools and flags
#
CC	=	$(TOOLSBIN)\cl
RC	=	$(TOOLSBIN)\rc
RM	=	erase
LD	=	$(TOOLSBIN)\link
INSTALL	=	copy
MKDIR	=	mkdir

# Standard File Extensions
I	=	-I
O	=	.o
EXE	=	.exe

# Extended build rules.
.SUFFIXES: .c .o
.c.o:
	$(CC) $(CFLAGS) -c $< -Fo$*.o

#
# Standard include stuff
#
CDEBUGFLAGS	=	-nologo -G5 -W3 -GX -Z7 -YX -Yd -Od
CPROFILEFLAGS	=	-nologo -G5 -YX -GX
COPTIMISEFLAGS	=	-nologo -G5 -YX -GX -O2 -DNDEBUG=1
CNONOPTIMISEFLAGS=	-nologo -G5 -YX -GX

LDDEBUGFLAGS	=	/DEBUG
LDOPTIMISEFLAGS	=	/DEBUG

CINCLUDES	=	$(I). $(I)$(INCLUDE)
#CDEFS		=	$(CDEBUGFLAGS)
CDEFS		=	$(COPTIMISEFLAGS)
CFLAGS		=	$(CINCLUDES) $(CDEFS) -D_JASSPA_ME -D_WIN32 -DWIN32 -D_WINDOWS -D_MBCS
#LDFLAGS	=	$(LDDEBUGFLAGS) /NOLOGO /SUBSYSTEM:windows /INCREMENTAL:no /MACHINE:I386 /PDB:NONE
LDFLAGS		=	$(LDOPTIMISEFLAGS) /NOLOGO /SUBSYSTEM:windows /INCREMENTAL:no /MACHINE:I386 /PDB:NONE

SYSLIBS         =	$(LIBRARY)\kernel32.lib	\
			$(LIBRARY)\wsock32.lib	\
			$(LIBRARY)\libc.lib	\
			$(LIBRARY)\user32.lib	\
			$(LIBRARY)\gdi32.lib	\
			$(LIBRARY)\winmm.lib	\
                        $(LIBRARY)\wsock32.lib	\
                        $(LIBRARY)\winspool.lib	\
                        $(LIBRARY)\comdlg32.lib	\
                        $(LIBRARY)\advapi32.lib	\
                        $(LIBRARY)\shell32.lib	\
                        $(LIBRARY)\ole32.lib	\
                        $(LIBRARY)\oleaut32.lib	\
                        $(LIBRARY)\uuid.lib	\
                        $(LIBRARY)\msvcrt.lib	\
                        $(LIBRARY)\oldnames.lib
                  
#
# Define the files involved in the build.
#
HEADERS = ebind.h edef.h eextrn.h efunc.h esearch.h estruct.h evar.h \
	  evers.h eterm.h ebind.def efunc.def evar.def emain.h eprint.h
PLTHDR  = wintermr.h
STDSRC	= abbrev.c basic.c bind.c buffer.c crypt.c dirlist.c display.c \
	  eval.c exec.c file.c fileio.c hilight.c history.c input.c \
	  isearch.c key.c line.c macro.c main.c narrow.c next.c osd.c \
	  print.c registry.c random.c region.c search.c spawn.c spell.c \
	  tag.c termio.c time.c undo.c window.c word.c regex.c
PLTSRC  = winterm.c winprint.c

STDOBJ	= $(STDSRC:.c=.o)	
PLTOBJ  = $(PLTSRC:.c=.o)
OBJ	= $(STDOBJ) $(PLTOBJ)

#
# Standard build rules
#
all::	me32$(EXE) 

#
# Build the executable 
#
me32$(EXE):$(OBJ) memsvc20.res 
	$(LD) $(LDFLAGS) /out:$@ $(OBJ) memsvc20.res $(SYSLIBS)

$(STDOBJ): $(HEADERS)
$(PLTOBJ): $(HEADERS) $(PLTHDR)

memsvc20.res:
	$(RC) -v -i . -i $(INCLUDE) -i $(TOOLSDIR)\mfc\include -fo $@ $*.rc 
relink::
	$(RM)	me32$(EXE)
	$(MAKE) -f $(MAKEFILE) $(MFLAGS) me32$(EXE)

install:: me32$(EXE)
	$(INSTALL) me32$(EXE) $(BINDIR)
	@echo Install done.

spotless:: clean
	- $(RM) *.??~
clean::
	- $(RM) *.pch
	- $(RM) *.o
	- $(RM) *.exe
	- $(RM) *.res
